class TaskDescriptionError(Exception):
    pass


class MessageDecodingError(Exception):
    pass


class RosbagSchemaError(Exception):
    pass


class RuntimeError(Exception):
    pass


class CertificationError(Exception):
    pass
