# from nltk.corpus import wordnet
# import nltk
# nltk.download('omw-1.4')
# nltk.download('wordnet')

# wnids = [x.strip() for x in open('imagenet21k_wordnet_ids.txt', 'r')]

# in21k_class_names = []
# for wnid in wnids:
# synset = wordnet.synset_from_pos_and_offset('n', int(wnid[1:]))
# synonyms = [x.name() for x in synset.lemmas()]
# in21k_class_names.append(synonyms[0])

# IN21K_CATEGORIES = repr(in21k_class_names) + "  # noqa"
# with open("/tmp/in21k_categories.py", "wt") as f:
# f.write(f"IN21K_CATEGORIES = {IN21K_CATEGORIES}")

IN21K_CATEGORIES = [
    "organism",
    "benthos",
    "heterotroph",
    "cell",
    "person",
    "animal",
    "plant",
    "food",
    "artifact",
    "hop",
    "check-in",
    "dressage",
    "curvet",
    "piaffe",
    "funambulism",
    "rock_climbing",
    "contact_sport",
    "outdoor_sport",
    "gymnastics",
    "acrobatics",
    "track_and_field",
    "track",
    "jumping",
    "broad_jump",
    "high_jump",
    "Fosbury_flop",
    "skiing",
    "cross-country_skiing",
    "ski_jumping",
    "water_sport",
    "swimming",
    "bathe",
    "dip",
    "dive",
    "floating",
    "dead-man's_float",
    "belly_flop",
    "cliff_diving",
    "flip",
    "gainer",
    "half_gainer",
    "jackknife",
    "swan_dive",
    "skin_diving",
    "scuba_diving",
    "snorkeling",
    "surfing",
    "water-skiing",
    "rowing",
    "sculling",
    "boxing",
    "professional_boxing",
    "in-fighting",
    "fight",
    "rope-a-dope",
    "spar",
    "archery",
    "sledding",
    "tobogganing",
    "luging",
    "bobsledding",
    "wrestling",
    "Greco-Roman_wrestling",
    "professional_wrestling",
    "sumo",
    "skating",
    "ice_skating",
    "figure_skating",
    "rollerblading",
    "roller_skating",
    "skateboarding",
    "speed_skating",
    "racing",
    "auto_racing",
    "boat_racing",
    "hydroplane_racing",
    "camel_racing",
    "greyhound_racing",
    "horse_racing",
    "riding",
    "equestrian_sport",
    "pony-trekking",
    "showjumping",
    "cross-country_riding",
    "cycling",
    "bicycling",
    "motorcycling",
    "dune_cycling",
    "blood_sport",
    "bullfighting",
    "cockfighting",
    "hunt",
    "battue",
    "beagling",
    "coursing",
    "deer_hunting",
    "ducking",
    "fox_hunting",
    "pigsticking",
    "fishing",
    "angling",
    "fly-fishing",
    "troll",
    "casting",
    "bait_casting",
    "fly_casting",
    "overcast",
    "surf_casting",
    "day_game",
    "athletic_game",
    "ice_hockey",
    "tetherball",
    "water_polo",
    "outdoor_game",
    "golf",
    "professional_golf",
    "round_of_golf",
    "medal_play",
    "match_play",
    "miniature_golf",
    "croquet",
    "quoits",
    "shuffleboard",
    "field_game",
    "field_hockey",
    "shinny",
    "football",
    "American_football",
    "professional_football",
    "touch_football",
    "hurling",
    "rugby",
    "ball_game",
    "baseball",
    "ball",
    "professional_baseball",
    "hardball",
    "perfect_game",
    "no-hit_game",
    "one-hitter",
    "two-hitter",
    "three-hitter",
    "four-hitter",
    "five-hitter",
    "softball",
    "rounders",
    "stickball",
    "cricket",
    "lacrosse",
    "polo",
    "pushball",
    "soccer",
    "court_game",
    "handball",
    "racquetball",
    "fives",
    "squash",
    "volleyball",
    "jai_alai",
    "badminton",
    "battledore",
    "basketball",
    "professional_basketball",
    "deck_tennis",
    "netball",
    "tennis",
    "professional_tennis",
    "singles",
    "singles",
    "doubles",
    "doubles",
    "royal_tennis",
    "pallone",
    "sport",
    "clasp",
    "judo",
    "team_sport",
    "Last_Supper",
    "Seder",
    "camping",
    "pest",
    "critter",
    "creepy-crawly",
    "darter",
    "peeper",
    "homeotherm",
    "poikilotherm",
    "range_animal",
    "scavenger",
    "bottom-feeder",
    "bottom-feeder",
    "work_animal",
    "beast_of_burden",
    "draft_animal",
    "pack_animal",
    "domestic_animal",
    "feeder",
    "feeder",
    "stocker",
    "hatchling",
    "head",
    "migrator",
    "molter",
    "pet",
    "stayer",
    "stunt",
    "marine_animal",
    "by-catch",
    "female",
    "hen",
    "male",
    "adult",
    "young",
    "orphan",
    "young_mammal",
    "baby",
    "pup",
    "wolf_pup",
    "puppy",
    "cub",
    "lion_cub",
    "bear_cub",
    "tiger_cub",
    "kit",
    "suckling",
    "sire",
    "dam",
    "thoroughbred",
    "giant",
    "mutant",
    "carnivore",
    "herbivore",
    "insectivore",
    "acrodont",
    "pleurodont",
    "microorganism",
    "monohybrid",
    "arbovirus",
    "adenovirus",
    "arenavirus",
    "Marburg_virus",
    "Arenaviridae",
    "vesiculovirus",
    "Reoviridae",
    "variola_major",
    "viroid",
    "coliphage",
    "paramyxovirus",
    "poliovirus",
    "herpes",
    "herpes_simplex_1",
    "herpes_zoster",
    "herpes_varicella_zoster",
    "cytomegalovirus",
    "varicella_zoster_virus",
    "polyoma",
    "lyssavirus",
    "reovirus",
    "rotavirus",
    "moneran",
    "archaebacteria",
    "bacteroid",
    "Bacillus_anthracis",
    "Yersinia_pestis",
    "Brucella",
    "spirillum",
    "botulinus",
    "clostridium_perfringens",
    "cyanobacteria",
    "trichodesmium",
    "nitric_bacteria",
    "spirillum",
    "Francisella",
    "gonococcus",
    "Corynebacterium_diphtheriae",
    "enteric_bacteria",
    "klebsiella",
    "Salmonella_typhimurium",
    "typhoid_bacillus",
    "nitrate_bacterium",
    "nitrite_bacterium",
    "actinomycete",
    "streptomyces",
    "Streptomyces_erythreus",
    "Streptomyces_griseus",
    "tubercle_bacillus",
    "pus-forming_bacteria",
    "streptobacillus",
    "myxobacteria",
    "staphylococcus",
    "diplococcus",
    "pneumococcus",
    "streptococcus",
    "spirochete",
    "planktonic_algae",
    "zooplankton",
    "parasite",
    "endoparasite",
    "ectoparasite",
    "pathogen",
    "commensal",
    "myrmecophile",
    "protoctist",
    "protozoan",
    "sarcodinian",
    "heliozoan",
    "endameba",
    "ameba",
    "globigerina",
    "testacean",
    "arcella",
    "difflugia",
    "ciliate",
    "paramecium",
    "stentor",
    "alga",
    "arame",
    "seagrass",
    "golden_algae",
    "yellow-green_algae",
    "brown_algae",
    "kelp",
    "fucoid",
    "fucoid",
    "fucus",
    "bladderwrack",
    "green_algae",
    "pond_scum",
    "chlorella",
    "stonewort",
    "desmid",
    "sea_moss",
    "eukaryote",
    "prokaryote",
    "zooid",
    "Leishmania",
    "zoomastigote",
    "polymastigote",
    "costia",
    "giardia",
    "cryptomonad",
    "sporozoan",
    "sporozoite",
    "trophozoite",
    "merozoite",
    "coccidium",
    "gregarine",
    "plasmodium",
    "leucocytozoan",
    "microsporidian",
    "Ostariophysi",
    "cypriniform_fish",
    "loach",
    "cyprinid",
    "carp",
    "domestic_carp",
    "leather_carp",
    "mirror_carp",
    "European_bream",
    "tench",
    "dace",
    "chub",
    "shiner",
    "common_shiner",
    "roach",
    "rudd",
    "minnow",
    "gudgeon",
    "goldfish",
    "crucian_carp",
    "electric_eel",
    "catostomid",
    "buffalo_fish",
    "black_buffalo",
    "hog_sucker",
    "redhorse",
    "cyprinodont",
    "killifish",
    "mummichog",
    "striped_killifish",
    "rivulus",
    "flagfish",
    "swordtail",
    "guppy",
    "topminnow",
    "mosquitofish",
    "platy",
    "mollie",
    "squirrelfish",
    "reef_squirrelfish",
    "deepwater_squirrelfish",
    "Holocentrus_ascensionis",
    "soldierfish",
    "anomalops",
    "flashlight_fish",
    "John_Dory",
    "boarfish",
    "boarfish",
    "cornetfish",
    "stickleback",
    "three-spined_stickleback",
    "ten-spined_stickleback",
    "pipefish",
    "dwarf_pipefish",
    "deepwater_pipefish",
    "seahorse",
    "snipefish",
    "shrimpfish",
    "trumpetfish",
    "pellicle",
    "embryo",
    "fetus",
    "abortus",
    "spawn",
    "blastula",
    "blastocyst",
    "gastrula",
    "morula",
    "yolk",
    "chordate",
    "cephalochordate",
    "lancelet",
    "tunicate",
    "ascidian",
    "sea_squirt",
    "salp",
    "doliolum",
    "larvacean",
    "appendicularia",
    "ascidian_tadpole",
    "vertebrate",
    "Amniota",
    "amniote",
    "aquatic_vertebrate",
    "jawless_vertebrate",
    "ostracoderm",
    "heterostracan",
    "anaspid",
    "conodont",
    "cyclostome",
    "lamprey",
    "sea_lamprey",
    "hagfish",
    "Myxine_glutinosa",
    "eptatretus",
    "gnathostome",
    "placoderm",
    "cartilaginous_fish",
    "holocephalan",
    "chimaera",
    "rabbitfish",
    "elasmobranch",
    "shark",
    "cow_shark",
    "mackerel_shark",
    "porbeagle",
    "mako",
    "shortfin_mako",
    "longfin_mako",
    "bonito_shark",
    "great_white_shark",
    "basking_shark",
    "thresher",
    "carpet_shark",
    "nurse_shark",
    "sand_tiger",
    "whale_shark",
    "requiem_shark",
    "bull_shark",
    "sandbar_shark",
    "blacktip_shark",
    "whitetip_shark",
    "dusky_shark",
    "lemon_shark",
    "blue_shark",
    "tiger_shark",
    "soupfin_shark",
    "dogfish",
    "smooth_dogfish",
    "smoothhound",
    "American_smooth_dogfish",
    "Florida_smoothhound",
    "whitetip_shark",
    "spiny_dogfish",
    "Atlantic_spiny_dogfish",
    "Pacific_spiny_dogfish",
    "hammerhead",
    "smooth_hammerhead",
    "smalleye_hammerhead",
    "shovelhead",
    "angel_shark",
    "ray",
    "electric_ray",
    "sawfish",
    "smalltooth_sawfish",
    "guitarfish",
    "stingray",
    "roughtail_stingray",
    "butterfly_ray",
    "eagle_ray",
    "spotted_eagle_ray",
    "cownose_ray",
    "manta",
    "Atlantic_manta",
    "devil_ray",
    "skate",
    "grey_skate",
    "little_skate",
    "thorny_skate",
    "barndoor_skate",
    "bird",
    "dickeybird",
    "fledgling",
    "nestling",
    "cock",
    "gamecock",
    "hen",
    "nester",
    "night_bird",
    "night_raven",
    "bird_of_passage",
    "archaeopteryx",
    "archaeornis",
    "ratite",
    "carinate",
    "ostrich",
    "cassowary",
    "emu",
    "kiwi",
    "rhea",
    "rhea",
    "elephant_bird",
    "moa",
    "passerine",
    "nonpasserine_bird",
    "oscine",
    "songbird",
    "honey_eater",
    "accentor",
    "hedge_sparrow",
    "lark",
    "skylark",
    "wagtail",
    "pipit",
    "meadow_pipit",
    "finch",
    "chaffinch",
    "brambling",
    "goldfinch",
    "linnet",
    "siskin",
    "red_siskin",
    "redpoll",
    "redpoll",
    "New_World_goldfinch",
    "pine_siskin",
    "house_finch",
    "purple_finch",
    "canary",
    "common_canary",
    "serin",
    "crossbill",
    "bullfinch",
    "junco",
    "dark-eyed_junco",
    "New_World_sparrow",
    "vesper_sparrow",
    "white-throated_sparrow",
    "white-crowned_sparrow",
    "chipping_sparrow",
    "field_sparrow",
    "tree_sparrow",
    "song_sparrow",
    "swamp_sparrow",
    "bunting",
    "indigo_bunting",
    "ortolan",
    "reed_bunting",
    "yellowhammer",
    "yellow-breasted_bunting",
    "snow_bunting",
    "honeycreeper",
    "banana_quit",
    "sparrow",
    "English_sparrow",
    "tree_sparrow",
    "grosbeak",
    "evening_grosbeak",
    "hawfinch",
    "pine_grosbeak",
    "cardinal",
    "pyrrhuloxia",
    "towhee",
    "chewink",
    "green-tailed_towhee",
    "weaver",
    "baya",
    "whydah",
    "Java_sparrow",
    "avadavat",
    "grassfinch",
    "zebra_finch",
    "honeycreeper",
    "lyrebird",
    "scrubbird",
    "broadbill",
    "tyrannid",
    "New_World_flycatcher",
    "kingbird",
    "Arkansas_kingbird",
    "Cassin's_kingbird",
    "eastern_kingbird",
    "grey_kingbird",
    "pewee",
    "western_wood_pewee",
    "phoebe",
    "vermillion_flycatcher",
    "cotinga",
    "cock_of_the_rock",
    "cock_of_the_rock",
    "manakin",
    "bellbird",
    "umbrella_bird",
    "ovenbird",
    "antbird",
    "ant_thrush",
    "ant_shrike",
    "spotted_antbird",
    "woodhewer",
    "pitta",
    "scissortail",
    "Old_World_flycatcher",
    "spotted_flycatcher",
    "thickhead",
    "thrush",
    "missel_thrush",
    "song_thrush",
    "fieldfare",
    "redwing",
    "blackbird",
    "ring_ouzel",
    "robin",
    "clay-colored_robin",
    "hermit_thrush",
    "veery",
    "wood_thrush",
    "nightingale",
    "thrush_nightingale",
    "bulbul",
    "Old_World_chat",
    "stonechat",
    "whinchat",
    "solitaire",
    "redstart",
    "wheatear",
    "bluebird",
    "robin",
    "bluethroat",
    "warbler",
    "gnatcatcher",
    "kinglet",
    "goldcrest",
    "gold-crowned_kinglet",
    "ruby-crowned_kinglet",
    "Old_World_warbler",
    "blackcap",
    "greater_whitethroat",
    "lesser_whitethroat",
    "wood_warbler",
    "sedge_warbler",
    "wren_warbler",
    "tailorbird",
    "babbler",
    "New_World_warbler",
    "parula_warbler",
    "Wilson's_warbler",
    "flycatching_warbler",
    "American_redstart",
    "Cape_May_warbler",
    "yellow_warbler",
    "Blackburn",
    "Audubon's_warbler",
    "myrtle_warbler",
    "blackpoll",
    "New_World_chat",
    "yellow-breasted_chat",
    "ovenbird",
    "water_thrush",
    "yellowthroat",
    "common_yellowthroat",
    "riflebird",
    "New_World_oriole",
    "northern_oriole",
    "Baltimore_oriole",
    "Bullock's_oriole",
    "orchard_oriole",
    "meadowlark",
    "eastern_meadowlark",
    "western_meadowlark",
    "cacique",
    "bobolink",
    "New_World_blackbird",
    "grackle",
    "purple_grackle",
    "rusty_blackbird",
    "cowbird",
    "red-winged_blackbird",
    "Old_World_oriole",
    "golden_oriole",
    "fig-bird",
    "starling",
    "common_starling",
    "rose-colored_starling",
    "myna",
    "crested_myna",
    "hill_myna",
    "corvine_bird",
    "crow",
    "American_crow",
    "raven",
    "rook",
    "jackdaw",
    "chough",
    "jay",
    "Old_World_jay",
    "common_European_jay",
    "New_World_jay",
    "blue_jay",
    "Canada_jay",
    "Rocky_Mountain_jay",
    "nutcracker",
    "common_nutcracker",
    "Clark's_nutcracker",
    "magpie",
    "European_magpie",
    "American_magpie",
    "Australian_magpie",
    "butcherbird",
    "currawong",
    "piping_crow",
    "wren",
    "winter_wren",
    "house_wren",
    "marsh_wren",
    "long-billed_marsh_wren",
    "sedge_wren",
    "rock_wren",
    "Carolina_wren",
    "cactus_wren",
    "mockingbird",
    "blue_mockingbird",
    "catbird",
    "thrasher",
    "brown_thrasher",
    "New_Zealand_wren",
    "rock_wren",
    "rifleman_bird",
    "creeper",
    "brown_creeper",
    "European_creeper",
    "wall_creeper",
    "European_nuthatch",
    "red-breasted_nuthatch",
    "white-breasted_nuthatch",
    "titmouse",
    "chickadee",
    "black-capped_chickadee",
    "tufted_titmouse",
    "Carolina_chickadee",
    "blue_tit",
    "bushtit",
    "wren-tit",
    "verdin",
    "fairy_bluebird",
    "swallow",
    "barn_swallow",
    "cliff_swallow",
    "tree_swallow",
    "white-bellied_swallow",
    "martin",
    "house_martin",
    "bank_martin",
    "purple_martin",
    "wood_swallow",
    "tanager",
    "scarlet_tanager",
    "western_tanager",
    "summer_tanager",
    "hepatic_tanager",
    "shrike",
    "butcherbird",
    "European_shrike",
    "northern_shrike",
    "white-rumped_shrike",
    "loggerhead_shrike",
    "migrant_shrike",
    "bush_shrike",
    "black-fronted_bush_shrike",
    "bowerbird",
    "satin_bowerbird",
    "great_bowerbird",
    "water_ouzel",
    "European_water_ouzel",
    "American_water_ouzel",
    "vireo",
    "red-eyed_vireo",
    "solitary_vireo",
    "blue-headed_vireo",
    "waxwing",
    "cedar_waxwing",
    "Bohemian_waxwing",
    "bird_of_prey",
    "Accipitriformes",
    "hawk",
    "eyas",
    "tiercel",
    "goshawk",
    "sparrow_hawk",
    "Cooper's_hawk",
    "chicken_hawk",
    "buteonine",
    "redtail",
    "rough-legged_hawk",
    "red-shouldered_hawk",
    "buzzard",
    "honey_buzzard",
    "kite",
    "black_kite",
    "swallow-tailed_kite",
    "white-tailed_kite",
    "harrier",
    "marsh_harrier",
    "Montagu's_harrier",
    "marsh_hawk",
    "harrier_eagle",
    "falcon",
    "peregrine",
    "falcon-gentle",
    "gyrfalcon",
    "kestrel",
    "sparrow_hawk",
    "pigeon_hawk",
    "hobby",
    "caracara",
    "Audubon's_caracara",
    "carancha",
    "eagle",
    "young_bird",
    "eaglet",
    "harpy",
    "golden_eagle",
    "tawny_eagle",
    "bald_eagle",
    "sea_eagle",
    "Kamchatkan_sea_eagle",
    "ern",
    "fishing_eagle",
    "osprey",
    "vulture",
    "Aegypiidae",
    "Old_World_vulture",
    "griffon_vulture",
    "bearded_vulture",
    "Egyptian_vulture",
    "black_vulture",
    "secretary_bird",
    "New_World_vulture",
    "buzzard",
    "condor",
    "Andean_condor",
    "California_condor",
    "black_vulture",
    "king_vulture",
    "owl",
    "owlet",
    "little_owl",
    "horned_owl",
    "great_horned_owl",
    "great_grey_owl",
    "tawny_owl",
    "barred_owl",
    "screech_owl",
    "screech_owl",
    "scops_owl",
    "spotted_owl",
    "Old_World_scops_owl",
    "Oriental_scops_owl",
    "hoot_owl",
    "hawk_owl",
    "long-eared_owl",
    "laughing_owl",
    "barn_owl",
    "amphibian",
    "Ichyostega",
    "urodele",
    "salamander",
    "European_fire_salamander",
    "spotted_salamander",
    "alpine_salamander",
    "newt",
    "common_newt",
    "red_eft",
    "Pacific_newt",
    "rough-skinned_newt",
    "California_newt",
    "eft",
    "ambystomid",
    "mole_salamander",
    "spotted_salamander",
    "tiger_salamander",
    "axolotl",
    "waterdog",
    "hellbender",
    "giant_salamander",
    "olm",
    "mud_puppy",
    "dicamptodon",
    "Pacific_giant_salamander",
    "olympic_salamander",
    "lungless_salamander",
    "eastern_red-backed_salamander",
    "western_red-backed_salamander",
    "dusky_salamander",
    "climbing_salamander",
    "arboreal_salamander",
    "slender_salamander",
    "web-toed_salamander",
    "Shasta_salamander",
    "limestone_salamander",
    "amphiuma",
    "siren",
    "frog",
    "true_frog",
    "wood-frog",
    "leopard_frog",
    "bullfrog",
    "green_frog",
    "cascades_frog",
    "goliath_frog",
    "pickerel_frog",
    "tarahumara_frog",
    "grass_frog",
    "leptodactylid_frog",
    "robber_frog",
    "barking_frog",
    "crapaud",
    "tree_frog",
    "tailed_frog",
    "Liopelma_hamiltoni",
    "true_toad",
    "bufo",
    "agua",
    "European_toad",
    "natterjack",
    "American_toad",
    "Eurasian_green_toad",
    "American_green_toad",
    "Yosemite_toad",
    "Texas_toad",
    "southwestern_toad",
    "western_toad",
    "obstetrical_toad",
    "midwife_toad",
    "fire-bellied_toad",
    "spadefoot",
    "western_spadefoot",
    "southern_spadefoot",
    "plains_spadefoot",
    "tree_toad",
    "spring_peeper",
    "Pacific_tree_toad",
    "canyon_treefrog",
    "chameleon_tree_frog",
    "cricket_frog",
    "northern_cricket_frog",
    "eastern_cricket_frog",
    "chorus_frog",
    "lowland_burrowing_treefrog",
    "western_narrow-mouthed_toad",
    "eastern_narrow-mouthed_toad",
    "sheep_frog",
    "tongueless_frog",
    "Surinam_toad",
    "African_clawed_frog",
    "South_American_poison_toad",
    "caecilian",
    "reptile",
    "anapsid",
    "diapsid",
    "Diapsida",
    "chelonian",
    "turtle",
    "sea_turtle",
    "green_turtle",
    "loggerhead",
    "ridley",
    "Atlantic_ridley",
    "Pacific_ridley",
    "hawksbill_turtle",
    "leatherback_turtle",
    "snapping_turtle",
    "common_snapping_turtle",
    "alligator_snapping_turtle",
    "mud_turtle",
    "musk_turtle",
    "terrapin",
    "diamondback_terrapin",
    "red-bellied_terrapin",
    "slider",
    "cooter",
    "box_turtle",
    "Western_box_turtle",
    "painted_turtle",
    "tortoise",
    "European_tortoise",
    "giant_tortoise",
    "gopher_tortoise",
    "desert_tortoise",
    "Texas_tortoise",
    "soft-shelled_turtle",
    "spiny_softshell",
    "smooth_softshell",
    "tuatara",
    "saurian",
    "lizard",
    "gecko",
    "flying_gecko",
    "banded_gecko",
    "iguanid",
    "common_iguana",
    "marine_iguana",
    "desert_iguana",
    "chuckwalla",
    "zebra-tailed_lizard",
    "fringe-toed_lizard",
    "earless_lizard",
    "collared_lizard",
    "leopard_lizard",
    "spiny_lizard",
    "fence_lizard",
    "western_fence_lizard",
    "eastern_fence_lizard",
    "sagebrush_lizard",
    "side-blotched_lizard",
    "tree_lizard",
    "horned_lizard",
    "Texas_horned_lizard",
    "basilisk",
    "American_chameleon",
    "worm_lizard",
    "night_lizard",
    "skink",
    "western_skink",
    "mountain_skink",
    "teiid_lizard",
    "whiptail",
    "racerunner",
    "plateau_striped_whiptail",
    "Chihuahuan_spotted_whiptail",
    "western_whiptail",
    "checkered_whiptail",
    "teju",
    "caiman_lizard",
    "agamid",
    "agama",
    "frilled_lizard",
    "moloch",
    "mountain_devil",
    "anguid_lizard",
    "alligator_lizard",
    "blindworm",
    "glass_lizard",
    "legless_lizard",
    "Lanthanotus_borneensis",
    "venomous_lizard",
    "Gila_monster",
    "beaded_lizard",
    "lacertid_lizard",
    "sand_lizard",
    "green_lizard",
    "chameleon",
    "African_chameleon",
    "horned_chameleon",
    "monitor",
    "African_monitor",
    "Komodo_dragon",
    "crocodilian_reptile",
    "crocodile",
    "African_crocodile",
    "Asian_crocodile",
    "Morlett's_crocodile",
    "false_gavial",
    "alligator",
    "American_alligator",
    "Chinese_alligator",
    "caiman",
    "spectacled_caiman",
    "gavial",
    "armored_dinosaur",
    "stegosaur",
    "ankylosaur",
    "Edmontonia",
    "bone-headed_dinosaur",
    "pachycephalosaur",
    "ceratopsian",
    "protoceratops",
    "triceratops",
    "styracosaur",
    "psittacosaur",
    "ornithopod",
    "hadrosaur",
    "trachodon",
    "saurischian",
    "sauropod",
    "apatosaur",
    "barosaur",
    "diplodocus",
    "argentinosaur",
    "theropod",
    "ceratosaur",
    "coelophysis",
    "tyrannosaur",
    "allosaur",
    "ornithomimid",
    "maniraptor",
    "oviraptorid",
    "velociraptor",
    "deinonychus",
    "utahraptor",
    "synapsid",
    "dicynodont",
    "pelycosaur",
    "dimetrodon",
    "pterosaur",
    "pterodactyl",
    "ichthyosaur",
    "ichthyosaurus",
    "stenopterygius",
    "plesiosaur",
    "nothosaur",
    "snake",
    "colubrid_snake",
    "hoop_snake",
    "thunder_snake",
    "ringneck_snake",
    "hognose_snake",
    "leaf-nosed_snake",
    "green_snake",
    "smooth_green_snake",
    "rough_green_snake",
    "green_snake",
    "racer",
    "blacksnake",
    "blue_racer",
    "horseshoe_whipsnake",
    "whip-snake",
    "coachwhip",
    "California_whipsnake",
    "Sonoran_whipsnake",
    "rat_snake",
    "corn_snake",
    "black_rat_snake",
    "chicken_snake",
    "Indian_rat_snake",
    "glossy_snake",
    "bull_snake",
    "gopher_snake",
    "pine_snake",
    "king_snake",
    "common_kingsnake",
    "milk_snake",
    "garter_snake",
    "common_garter_snake",
    "ribbon_snake",
    "Western_ribbon_snake",
    "lined_snake",
    "ground_snake",
    "eastern_ground_snake",
    "water_snake",
    "common_water_snake",
    "water_moccasin",
    "grass_snake",
    "viperine_grass_snake",
    "red-bellied_snake",
    "sand_snake",
    "banded_sand_snake",
    "black-headed_snake",
    "vine_snake",
    "lyre_snake",
    "Sonoran_lyre_snake",
    "night_snake",
    "blind_snake",
    "western_blind_snake",
    "indigo_snake",
    "eastern_indigo_snake",
    "constrictor",
    "boa",
    "boa_constrictor",
    "rubber_boa",
    "rosy_boa",
    "anaconda",
    "python",
    "carpet_snake",
    "reticulated_python",
    "Indian_python",
    "rock_python",
    "amethystine_python",
    "elapid",
    "coral_snake",
    "eastern_coral_snake",
    "western_coral_snake",
    "coral_snake",
    "African_coral_snake",
    "Australian_coral_snake",
    "copperhead",
    "cobra",
    "Indian_cobra",
    "asp",
    "black-necked_cobra",
    "hamadryad",
    "ringhals",
    "mamba",
    "black_mamba",
    "green_mamba",
    "death_adder",
    "tiger_snake",
    "Australian_blacksnake",
    "krait",
    "banded_krait",
    "taipan",
    "sea_snake",
    "viper",
    "adder",
    "asp",
    "puff_adder",
    "gaboon_viper",
    "horned_viper",
    "pit_viper",
    "copperhead",
    "water_moccasin",
    "rattlesnake",
    "diamondback",
    "timber_rattlesnake",
    "canebrake_rattlesnake",
    "prairie_rattlesnake",
    "sidewinder",
    "Western_diamondback",
    "rock_rattlesnake",
    "tiger_rattlesnake",
    "Mojave_rattlesnake",
    "speckled_rattlesnake",
    "massasauga",
    "ground_rattler",
    "fer-de-lance",
    "carcase",
    "carrion",
    "arthropod",
    "trilobite",
    "arachnid",
    "harvestman",
    "scorpion",
    "false_scorpion",
    "book_scorpion",
    "whip-scorpion",
    "vinegarroon",
    "spider",
    "orb-weaving_spider",
    "black_and_gold_garden_spider",
    "barn_spider",
    "garden_spider",
    "comb-footed_spider",
    "black_widow",
    "tarantula",
    "wolf_spider",
    "European_wolf_spider",
    "trap-door_spider",
    "acarine",
    "tick",
    "hard_tick",
    "Ixodes_dammini",
    "Ixodes_neotomae",
    "Ixodes_pacificus",
    "Ixodes_scapularis",
    "sheep-tick",
    "Ixodes_persulcatus",
    "Ixodes_dentatus",
    "Ixodes_spinipalpis",
    "wood_tick",
    "soft_tick",
    "mite",
    "web-spinning_mite",
    "acarid",
    "trombidiid",
    "trombiculid",
    "harvest_mite",
    "acarus",
    "itch_mite",
    "rust_mite",
    "spider_mite",
    "red_spider",
    "myriapod",
    "garden_centipede",
    "tardigrade",
    "centipede",
    "house_centipede",
    "millipede",
    "sea_spider",
    "Merostomata",
    "horseshoe_crab",
    "Asian_horseshoe_crab",
    "eurypterid",
    "tongue_worm",
    "gallinaceous_bird",
    "domestic_fowl",
    "Dorking",
    "Plymouth_Rock",
    "Cornish",
    "Rock_Cornish",
    "game_fowl",
    "cochin",
    "jungle_fowl",
    "jungle_cock",
    "jungle_hen",
    "red_jungle_fowl",
    "chicken",
    "bantam",
    "chick",
    "cock",
    "cockerel",
    "capon",
    "hen",
    "cackler",
    "brood_hen",
    "mother_hen",
    "layer",
    "pullet",
    "spring_chicken",
    "Rhode_Island_red",
    "Dominique",
    "Orpington",
    "turkey",
    "turkey_cock",
    "ocellated_turkey",
    "grouse",
    "black_grouse",
    "European_black_grouse",
    "Asian_black_grouse",
    "blackcock",
    "greyhen",
    "ptarmigan",
    "red_grouse",
    "moorhen",
    "capercaillie",
    "spruce_grouse",
    "sage_grouse",
    "ruffed_grouse",
    "sharp-tailed_grouse",
    "prairie_chicken",
    "greater_prairie_chicken",
    "lesser_prairie_chicken",
    "heath_hen",
    "guan",
    "curassow",
    "piping_guan",
    "chachalaca",
    "Texas_chachalaca",
    "megapode",
    "mallee_fowl",
    "mallee_hen",
    "brush_turkey",
    "maleo",
    "phasianid",
    "pheasant",
    "ring-necked_pheasant",
    "afropavo",
    "argus",
    "golden_pheasant",
    "bobwhite",
    "northern_bobwhite",
    "Old_World_quail",
    "migratory_quail",
    "monal",
    "peafowl",
    "peachick",
    "peacock",
    "peahen",
    "blue_peafowl",
    "green_peafowl",
    "quail",
    "California_quail",
    "tragopan",
    "partridge",
    "Hungarian_partridge",
    "red-legged_partridge",
    "Greek_partridge",
    "mountain_quail",
    "guinea_fowl",
    "guinea_hen",
    "hoatzin",
    "tinamou",
    "columbiform_bird",
    "dodo",
    "pigeon",
    "pouter_pigeon",
    "dove",
    "rock_dove",
    "band-tailed_pigeon",
    "wood_pigeon",
    "turtledove",
    "Streptopelia_turtur",
    "ringdove",
    "Australian_turtledove",
    "mourning_dove",
    "domestic_pigeon",
    "squab",
    "fairy_swallow",
    "roller",
    "homing_pigeon",
    "carrier_pigeon",
    "passenger_pigeon",
    "sandgrouse",
    "painted_sandgrouse",
    "pin-tailed_sandgrouse",
    "pallas's_sandgrouse",
    "parrot",
    "popinjay",
    "poll",
    "African_grey",
    "amazon",
    "macaw",
    "kea",
    "cockatoo",
    "sulphur-crested_cockatoo",
    "pink_cockatoo",
    "cockateel",
    "lovebird",
    "lory",
    "lorikeet",
    "varied_Lorikeet",
    "rainbow_lorikeet",
    "parakeet",
    "Carolina_parakeet",
    "budgerigar",
    "ring-necked_parakeet",
    "cuculiform_bird",
    "cuckoo",
    "European_cuckoo",
    "black-billed_cuckoo",
    "roadrunner",
    "ani",
    "coucal",
    "crow_pheasant",
    "touraco",
    "coraciiform_bird",
    "roller",
    "European_roller",
    "ground_roller",
    "kingfisher",
    "Eurasian_kingfisher",
    "belted_kingfisher",
    "kookaburra",
    "bee_eater",
    "hornbill",
    "hoopoe",
    "Euopean_hoopoe",
    "wood_hoopoe",
    "motmot",
    "tody",
    "apodiform_bird",
    "swift",
    "European_swift",
    "chimney_swift",
    "swiftlet",
    "tree_swift",
    "hummingbird",
    "Archilochus_colubris",
    "thornbill",
    "goatsucker",
    "European_goatsucker",
    "chuck-will's-widow",
    "whippoorwill",
    "poorwill",
    "frogmouth",
    "oilbird",
    "piciform_bird",
    "woodpecker",
    "green_woodpecker",
    "downy_woodpecker",
    "flicker",
    "yellow-shafted_flicker",
    "gilded_flicker",
    "red-shafted_flicker",
    "ivorybill",
    "redheaded_woodpecker",
    "sapsucker",
    "yellow-bellied_sapsucker",
    "red-breasted_sapsucker",
    "wryneck",
    "piculet",
    "barbet",
    "puffbird",
    "honey_guide",
    "jacamar",
    "toucan",
    "toucanet",
    "trogon",
    "quetzal",
    "resplendent_quetzel",
    "aquatic_bird",
    "waterfowl",
    "anseriform_bird",
    "duck",
    "drake",
    "quack-quack",
    "duckling",
    "diving_duck",
    "dabbling_duck",
    "mallard",
    "black_duck",
    "teal",
    "greenwing",
    "bluewing",
    "garganey",
    "widgeon",
    "American_widgeon",
    "shoveler",
    "pintail",
    "sheldrake",
    "shelduck",
    "ruddy_duck",
    "bufflehead",
    "goldeneye",
    "Barrow's_goldeneye",
    "canvasback",
    "pochard",
    "redhead",
    "scaup",
    "greater_scaup",
    "lesser_scaup",
    "wild_duck",
    "wood_duck",
    "wood_drake",
    "mandarin_duck",
    "muscovy_duck",
    "sea_duck",
    "eider",
    "scoter",
    "common_scoter",
    "old_squaw",
    "merganser",
    "goosander",
    "American_merganser",
    "red-breasted_merganser",
    "smew",
    "hooded_merganser",
    "goose",
    "gosling",
    "gander",
    "Chinese_goose",
    "greylag",
    "blue_goose",
    "snow_goose",
    "brant",
    "common_brant_goose",
    "honker",
    "barnacle_goose",
    "coscoroba",
    "swan",
    "cob",
    "pen",
    "cygnet",
    "mute_swan",
    "whooper",
    "tundra_swan",
    "whistling_swan",
    "Bewick's_swan",
    "trumpeter",
    "black_swan",
    "screamer",
    "horned_screamer",
    "crested_screamer",
    "chaja",
    "mammal",
    "female_mammal",
    "tusker",
    "prototherian",
    "monotreme",
    "echidna",
    "echidna",
    "platypus",
    "marsupial",
    "opossum",
    "common_opossum",
    "crab-eating_opossum",
    "opossum_rat",
    "bandicoot",
    "rabbit-eared_bandicoot",
    "kangaroo",
    "giant_kangaroo",
    "wallaby",
    "common_wallaby",
    "hare_wallaby",
    "nail-tailed_wallaby",
    "rock_wallaby",
    "pademelon",
    "tree_wallaby",
    "musk_kangaroo",
    "rat_kangaroo",
    "potoroo",
    "bettong",
    "jerboa_kangaroo",
    "phalanger",
    "cuscus",
    "brush-tailed_phalanger",
    "flying_phalanger",
    "koala",
    "wombat",
    "dasyurid_marsupial",
    "dasyure",
    "eastern_dasyure",
    "native_cat",
    "thylacine",
    "Tasmanian_devil",
    "pouched_mouse",
    "numbat",
    "pouched_mole",
    "placental",
    "livestock",
    "bull",
    "cow",
    "calf",
    "calf",
    "yearling",
    "buck",
    "doe",
    "insectivore",
    "mole",
    "starnose_mole",
    "brewer's_mole",
    "golden_mole",
    "shrew_mole",
    "Asiatic_shrew_mole",
    "American_shrew_mole",
    "shrew",
    "common_shrew",
    "masked_shrew",
    "short-tailed_shrew",
    "water_shrew",
    "American_water_shrew",
    "European_water_shrew",
    "Mediterranean_water_shrew",
    "least_shrew",
    "hedgehog",
    "tenrec",
    "tailless_tenrec",
    "otter_shrew",
    "eiderdown",
    "aftershaft",
    "sickle_feather",
    "contour_feather",
    "bastard_wing",
    "saddle_hackle",
    "encolure",
    "hair",
    "squama",
    "scute",
    "sclerite",
    "plastron",
    "scallop_shell",
    "oyster_shell",
    "theca",
    "invertebrate",
    "sponge",
    "choanocyte",
    "glass_sponge",
    "Venus's_flower_basket",
    "metazoan",
    "coelenterate",
    "planula",
    "polyp",
    "medusa",
    "jellyfish",
    "scyphozoan",
    "Chrysaora_quinquecirrha",
    "hydrozoan",
    "hydra",
    "siphonophore",
    "nanomia",
    "Portuguese_man-of-war",
    "praya",
    "apolemia",
    "anthozoan",
    "sea_anemone",
    "actinia",
    "sea_pen",
    "coral",
    "gorgonian",
    "sea_feather",
    "sea_fan",
    "red_coral",
    "stony_coral",
    "brain_coral",
    "staghorn_coral",
    "mushroom_coral",
    "ctenophore",
    "beroe",
    "platyctenean",
    "sea_gooseberry",
    "Venus's_girdle",
    "worm",
    "helminth",
    "woodworm",
    "woodborer",
    "acanthocephalan",
    "arrowworm",
    "bladder_worm",
    "flatworm",
    "planarian",
    "fluke",
    "cercaria",
    "liver_fluke",
    "Fasciolopsis_buski",
    "schistosome",
    "tapeworm",
    "echinococcus",
    "taenia",
    "ribbon_worm",
    "beard_worm",
    "rotifer",
    "nematode",
    "common_roundworm",
    "chicken_roundworm",
    "pinworm",
    "eelworm",
    "vinegar_eel",
    "trichina",
    "hookworm",
    "filaria",
    "Guinea_worm",
    "annelid",
    "archiannelid",
    "oligochaete",
    "earthworm",
    "polychaete",
    "lugworm",
    "sea_mouse",
    "bloodworm",
    "leech",
    "medicinal_leech",
    "horseleech",
    "mollusk",
    "scaphopod",
    "tooth_shell",
    "gastropod",
    "abalone",
    "ormer",
    "scorpion_shell",
    "conch",
    "giant_conch",
    "snail",
    "edible_snail",
    "garden_snail",
    "brown_snail",
    "Helix_hortensis",
    "slug",
    "seasnail",
    "neritid",
    "nerita",
    "bleeding_tooth",
    "neritina",
    "whelk",
    "moon_shell",
    "periwinkle",
    "limpet",
    "common_limpet",
    "keyhole_limpet",
    "river_limpet",
    "sea_slug",
    "sea_hare",
    "Hermissenda_crassicornis",
    "bubble_shell",
    "physa",
    "cowrie",
    "money_cowrie",
    "tiger_cowrie",
    "solenogaster",
    "chiton",
    "bivalve",
    "spat",
    "clam",
    "seashell",
    "soft-shell_clam",
    "quahog",
    "littleneck",
    "cherrystone",
    "geoduck",
    "razor_clam",
    "giant_clam",
    "cockle",
    "edible_cockle",
    "oyster",
    "Japanese_oyster",
    "Virginia_oyster",
    "pearl_oyster",
    "saddle_oyster",
    "window_oyster",
    "ark_shell",
    "blood_clam",
    "mussel",
    "marine_mussel",
    "edible_mussel",
    "freshwater_mussel",
    "pearly-shelled_mussel",
    "thin-shelled_mussel",
    "zebra_mussel",
    "scallop",
    "bay_scallop",
    "sea_scallop",
    "shipworm",
    "teredo",
    "piddock",
    "cephalopod",
    "chambered_nautilus",
    "octopod",
    "octopus",
    "paper_nautilus",
    "decapod",
    "squid",
    "loligo",
    "ommastrephes",
    "architeuthis",
    "cuttlefish",
    "spirula",
    "crustacean",
    "malacostracan_crustacean",
    "decapod_crustacean",
    "brachyuran",
    "crab",
    "stone_crab",
    "hard-shell_crab",
    "soft-shell_crab",
    "Dungeness_crab",
    "rock_crab",
    "Jonah_crab",
    "swimming_crab",
    "English_lady_crab",
    "American_lady_crab",
    "blue_crab",
    "fiddler_crab",
    "pea_crab",
    "king_crab",
    "spider_crab",
    "European_spider_crab",
    "giant_crab",
    "lobster",
    "true_lobster",
    "American_lobster",
    "European_lobster",
    "Cape_lobster",
    "Norway_lobster",
    "spiny_lobster",
    "crayfish",
    "Old_World_crayfish",
    "American_crayfish",
    "hermit_crab",
    "shrimp",
    "snapping_shrimp",
    "prawn",
    "long-clawed_prawn",
    "tropical_prawn",
    "krill",
    "Euphausia_pacifica",
    "opossum_shrimp",
    "stomatopod",
    "mantis_shrimp",
    "squilla",
    "isopod",
    "woodlouse",
    "pill_bug",
    "sow_bug",
    "sea_louse",
    "amphipod",
    "skeleton_shrimp",
    "whale_louse",
    "daphnia",
    "fairy_shrimp",
    "brine_shrimp",
    "tadpole_shrimp",
    "copepod",
    "cyclops",
    "seed_shrimp",
    "barnacle",
    "acorn_barnacle",
    "goose_barnacle",
    "onychophoran",
    "wading_bird",
    "stork",
    "white_stork",
    "black_stork",
    "adjutant_bird",
    "marabou",
    "openbill",
    "jabiru",
    "saddlebill",
    "policeman_bird",
    "wood_ibis",
    "shoebill",
    "ibis",
    "wood_ibis",
    "sacred_ibis",
    "spoonbill",
    "common_spoonbill",
    "roseate_spoonbill",
    "flamingo",
    "heron",
    "great_blue_heron",
    "great_white_heron",
    "egret",
    "little_blue_heron",
    "snowy_egret",
    "little_egret",
    "great_white_heron",
    "American_egret",
    "cattle_egret",
    "night_heron",
    "black-crowned_night_heron",
    "yellow-crowned_night_heron",
    "boatbill",
    "bittern",
    "American_bittern",
    "European_bittern",
    "least_bittern",
    "crane",
    "whooping_crane",
    "courlan",
    "limpkin",
    "crested_cariama",
    "chunga",
    "rail",
    "weka",
    "crake",
    "corncrake",
    "spotted_crake",
    "gallinule",
    "Florida_gallinule",
    "moorhen",
    "purple_gallinule",
    "European_gallinule",
    "American_gallinule",
    "notornis",
    "coot",
    "American_coot",
    "Old_World_coot",
    "bustard",
    "great_bustard",
    "plain_turkey",
    "button_quail",
    "striped_button_quail",
    "plain_wanderer",
    "trumpeter",
    "Brazilian_trumpeter",
    "seabird",
    "shorebird",
    "plover",
    "piping_plover",
    "killdeer",
    "dotterel",
    "golden_plover",
    "lapwing",
    "turnstone",
    "ruddy_turnstone",
    "black_turnstone",
    "sandpiper",
    "surfbird",
    "European_sandpiper",
    "spotted_sandpiper",
    "least_sandpiper",
    "red-backed_sandpiper",
    "greenshank",
    "redshank",
    "yellowlegs",
    "greater_yellowlegs",
    "lesser_yellowlegs",
    "pectoral_sandpiper",
    "knot",
    "curlew_sandpiper",
    "sanderling",
    "upland_sandpiper",
    "ruff",
    "reeve",
    "tattler",
    "Polynesian_tattler",
    "willet",
    "woodcock",
    "Eurasian_woodcock",
    "American_woodcock",
    "snipe",
    "whole_snipe",
    "Wilson's_snipe",
    "great_snipe",
    "jacksnipe",
    "dowitcher",
    "greyback",
    "red-breasted_snipe",
    "curlew",
    "European_curlew",
    "Eskimo_curlew",
    "godwit",
    "Hudsonian_godwit",
    "stilt",
    "black-necked_stilt",
    "black-winged_stilt",
    "white-headed_stilt",
    "kaki",
    "stilt",
    "banded_stilt",
    "avocet",
    "oystercatcher",
    "phalarope",
    "red_phalarope",
    "northern_phalarope",
    "Wilson's_phalarope",
    "pratincole",
    "courser",
    "cream-colored_courser",
    "crocodile_bird",
    "stone_curlew",
    "coastal_diving_bird",
    "larid",
    "gull",
    "mew",
    "black-backed_gull",
    "herring_gull",
    "laughing_gull",
    "ivory_gull",
    "kittiwake",
    "tern",
    "sea_swallow",
    "skimmer",
    "jaeger",
    "parasitic_jaeger",
    "skua",
    "great_skua",
    "auk",
    "auklet",
    "razorbill",
    "little_auk",
    "guillemot",
    "black_guillemot",
    "pigeon_guillemot",
    "murre",
    "common_murre",
    "thick-billed_murre",
    "puffin",
    "Atlantic_puffin",
    "horned_puffin",
    "tufted_puffin",
    "gaviiform_seabird",
    "loon",
    "podicipitiform_seabird",
    "grebe",
    "great_crested_grebe",
    "red-necked_grebe",
    "black-necked_grebe",
    "dabchick",
    "pied-billed_grebe",
    "pelecaniform_seabird",
    "pelican",
    "white_pelican",
    "Old_world_white_pelican",
    "frigate_bird",
    "gannet",
    "solan",
    "booby",
    "cormorant",
    "snakebird",
    "water_turkey",
    "tropic_bird",
    "sphenisciform_seabird",
    "penguin",
    "Adelie",
    "king_penguin",
    "emperor_penguin",
    "jackass_penguin",
    "rock_hopper",
    "pelagic_bird",
    "procellariiform_seabird",
    "albatross",
    "wandering_albatross",
    "black-footed_albatross",
    "petrel",
    "white-chinned_petrel",
    "giant_petrel",
    "fulmar",
    "shearwater",
    "Manx_shearwater",
    "storm_petrel",
    "stormy_petrel",
    "Mother_Carey's_chicken",
    "diving_petrel",
    "aquatic_mammal",
    "cetacean",
    "whale",
    "baleen_whale",
    "right_whale",
    "bowhead",
    "rorqual",
    "blue_whale",
    "finback",
    "sei_whale",
    "lesser_rorqual",
    "humpback",
    "grey_whale",
    "toothed_whale",
    "sperm_whale",
    "pygmy_sperm_whale",
    "dwarf_sperm_whale",
    "beaked_whale",
    "bottle-nosed_whale",
    "dolphin",
    "common_dolphin",
    "bottlenose_dolphin",
    "Atlantic_bottlenose_dolphin",
    "Pacific_bottlenose_dolphin",
    "porpoise",
    "harbor_porpoise",
    "vaquita",
    "grampus",
    "killer_whale",
    "pilot_whale",
    "river_dolphin",
    "narwhal",
    "white_whale",
    "sea_cow",
    "manatee",
    "dugong",
    "Steller's_sea_cow",
    "carnivore",
    "omnivore",
    "pinniped_mammal",
    "seal",
    "crabeater_seal",
    "eared_seal",
    "fur_seal",
    "guadalupe_fur_seal",
    "fur_seal",
    "Alaska_fur_seal",
    "sea_lion",
    "South_American_sea_lion",
    "California_sea_lion",
    "Australian_sea_lion",
    "Steller_sea_lion",
    "earless_seal",
    "harbor_seal",
    "harp_seal",
    "elephant_seal",
    "bearded_seal",
    "hooded_seal",
    "walrus",
    "Atlantic_walrus",
    "Pacific_walrus",
    "Fissipedia",
    "fissiped_mammal",
    "aardvark",
    "canine",
    "bitch",
    "brood_bitch",
    "dog",
    "pooch",
    "cur",
    "feist",
    "pariah_dog",
    "lapdog",
    "toy_dog",
    "Chihuahua",
    "Japanese_spaniel",
    "Maltese_dog",
    "Pekinese",
    "Shih-Tzu",
    "toy_spaniel",
    "English_toy_spaniel",
    "Blenheim_spaniel",
    "King_Charles_spaniel",
    "papillon",
    "toy_terrier",
    "hunting_dog",
    "courser",
    "Rhodesian_ridgeback",
    "hound",
    "Afghan_hound",
    "basset",
    "beagle",
    "bloodhound",
    "bluetick",
    "boarhound",
    "coonhound",
    "coondog",
    "black-and-tan_coonhound",
    "dachshund",
    "sausage_dog",
    "foxhound",
    "American_foxhound",
    "Walker_hound",
    "English_foxhound",
    "harrier",
    "Plott_hound",
    "redbone",
    "wolfhound",
    "borzoi",
    "Irish_wolfhound",
    "greyhound",
    "Italian_greyhound",
    "whippet",
    "Ibizan_hound",
    "Norwegian_elkhound",
    "otterhound",
    "Saluki",
    "Scottish_deerhound",
    "staghound",
    "Weimaraner",
    "terrier",
    "bullterrier",
    "Staffordshire_bullterrier",
    "American_Staffordshire_terrier",
    "Bedlington_terrier",
    "Border_terrier",
    "Kerry_blue_terrier",
    "Irish_terrier",
    "Norfolk_terrier",
    "Norwich_terrier",
    "Yorkshire_terrier",
    "rat_terrier",
    "Manchester_terrier",
    "toy_Manchester",
    "fox_terrier",
    "smooth-haired_fox_terrier",
    "wire-haired_fox_terrier",
    "wirehair",
    "Lakeland_terrier",
    "Welsh_terrier",
    "Sealyham_terrier",
    "Airedale",
    "cairn",
    "Australian_terrier",
    "Dandie_Dinmont",
    "Boston_bull",
    "schnauzer",
    "miniature_schnauzer",
    "giant_schnauzer",
    "standard_schnauzer",
    "Scotch_terrier",
    "Tibetan_terrier",
    "silky_terrier",
    "Skye_terrier",
    "Clydesdale_terrier",
    "soft-coated_wheaten_terrier",
    "West_Highland_white_terrier",
    "Lhasa",
    "sporting_dog",
    "bird_dog",
    "water_dog",
    "retriever",
    "flat-coated_retriever",
    "curly-coated_retriever",
    "golden_retriever",
    "Labrador_retriever",
    "Chesapeake_Bay_retriever",
    "pointer",
    "German_short-haired_pointer",
    "setter",
    "vizsla",
    "English_setter",
    "Irish_setter",
    "Gordon_setter",
    "spaniel",
    "Brittany_spaniel",
    "clumber",
    "field_spaniel",
    "springer_spaniel",
    "English_springer",
    "Welsh_springer_spaniel",
    "cocker_spaniel",
    "Sussex_spaniel",
    "water_spaniel",
    "American_water_spaniel",
    "Irish_water_spaniel",
    "griffon",
    "working_dog",
    "watchdog",
    "kuvasz",
    "attack_dog",
    "housedog",
    "schipperke",
    "shepherd_dog",
    "Belgian_sheepdog",
    "groenendael",
    "malinois",
    "briard",
    "kelpie",
    "komondor",
    "Old_English_sheepdog",
    "Shetland_sheepdog",
    "collie",
    "Border_collie",
    "Bouvier_des_Flandres",
    "Rottweiler",
    "German_shepherd",
    "police_dog",
    "pinscher",
    "Doberman",
    "miniature_pinscher",
    "Sennenhunde",
    "Greater_Swiss_Mountain_dog",
    "Bernese_mountain_dog",
    "Appenzeller",
    "EntleBucher",
    "boxer",
    "mastiff",
    "bull_mastiff",
    "Tibetan_mastiff",
    "bulldog",
    "French_bulldog",
    "Great_Dane",
    "guide_dog",
    "Seeing_Eye_dog",
    "hearing_dog",
    "Saint_Bernard",
    "seizure-alert_dog",
    "sled_dog",
    "Eskimo_dog",
    "malamute",
    "Siberian_husky",
    "dalmatian",
    "liver-spotted_dalmatian",
    "affenpinscher",
    "basenji",
    "pug",
    "Leonberg",
    "Newfoundland",
    "Great_Pyrenees",
    "spitz",
    "Samoyed",
    "Pomeranian",
    "chow",
    "keeshond",
    "griffon",
    "Brabancon_griffon",
    "corgi",
    "Pembroke",
    "Cardigan",
    "poodle",
    "toy_poodle",
    "miniature_poodle",
    "standard_poodle",
    "large_poodle",
    "Mexican_hairless",
    "wolf",
    "timber_wolf",
    "white_wolf",
    "red_wolf",
    "coyote",
    "coydog",
    "jackal",
    "wild_dog",
    "dingo",
    "dhole",
    "crab-eating_dog",
    "raccoon_dog",
    "African_hunting_dog",
    "hyena",
    "striped_hyena",
    "brown_hyena",
    "spotted_hyena",
    "aardwolf",
    "fox",
    "vixen",
    "Reynard",
    "red_fox",
    "black_fox",
    "silver_fox",
    "red_fox",
    "kit_fox",
    "kit_fox",
    "Arctic_fox",
    "blue_fox",
    "grey_fox",
    "feline",
    "cat",
    "domestic_cat",
    "kitty",
    "mouser",
    "alley_cat",
    "stray",
    "tom",
    "gib",
    "tabby",
    "kitten",
    "tabby",
    "tiger_cat",
    "tortoiseshell",
    "Persian_cat",
    "Angora",
    "Siamese_cat",
    "blue_point_Siamese",
    "Burmese_cat",
    "Egyptian_cat",
    "Maltese",
    "Abyssinian",
    "Manx",
    "wildcat",
    "sand_cat",
    "European_wildcat",
    "cougar",
    "ocelot",
    "jaguarundi",
    "kaffir_cat",
    "jungle_cat",
    "serval",
    "leopard_cat",
    "margay",
    "manul",
    "lynx",
    "common_lynx",
    "Canada_lynx",
    "bobcat",
    "spotted_lynx",
    "caracal",
    "big_cat",
    "leopard",
    "leopardess",
    "panther",
    "snow_leopard",
    "jaguar",
    "lion",
    "lioness",
    "lionet",
    "tiger",
    "Bengal_tiger",
    "tigress",
    "liger",
    "tiglon",
    "cheetah",
    "saber-toothed_tiger",
    "Smiledon_californicus",
    "bear",
    "brown_bear",
    "bruin",
    "Syrian_bear",
    "grizzly",
    "Alaskan_brown_bear",
    "American_black_bear",
    "cinnamon_bear",
    "Asiatic_black_bear",
    "ice_bear",
    "sloth_bear",
    "viverrine",
    "civet",
    "large_civet",
    "small_civet",
    "binturong",
    "Cryptoprocta",
    "fossa",
    "fanaloka",
    "genet",
    "banded_palm_civet",
    "mongoose",
    "Indian_mongoose",
    "ichneumon",
    "palm_cat",
    "meerkat",
    "slender-tailed_meerkat",
    "suricate",
    "bat",
    "fruit_bat",
    "flying_fox",
    "Pteropus_capestratus",
    "Pteropus_hypomelanus",
    "harpy",
    "Cynopterus_sphinx",
    "carnivorous_bat",
    "mouse-eared_bat",
    "leafnose_bat",
    "macrotus",
    "spearnose_bat",
    "Phyllostomus_hastatus",
    "hognose_bat",
    "horseshoe_bat",
    "horseshoe_bat",
    "orange_bat",
    "false_vampire",
    "big-eared_bat",
    "vespertilian_bat",
    "frosted_bat",
    "red_bat",
    "brown_bat",
    "little_brown_bat",
    "cave_myotis",
    "big_brown_bat",
    "serotine",
    "pallid_bat",
    "pipistrelle",
    "eastern_pipistrel",
    "jackass_bat",
    "long-eared_bat",
    "western_big-eared_bat",
    "freetail",
    "guano_bat",
    "pocketed_bat",
    "mastiff_bat",
    "vampire_bat",
    "Desmodus_rotundus",
    "hairy-legged_vampire_bat",
    "predator",
    "prey",
    "game",
    "big_game",
    "game_bird",
    "fossorial_mammal",
    "tetrapod",
    "quadruped",
    "hexapod",
    "biped",
    "insect",
    "social_insect",
    "holometabola",
    "defoliator",
    "pollinator",
    "gallfly",
    "scorpion_fly",
    "hanging_fly",
    "collembolan",
    "beetle",
    "tiger_beetle",
    "ladybug",
    "two-spotted_ladybug",
    "Mexican_bean_beetle",
    "Hippodamia_convergens",
    "vedalia",
    "ground_beetle",
    "bombardier_beetle",
    "calosoma",
    "searcher",
    "firefly",
    "glowworm",
    "long-horned_beetle",
    "sawyer",
    "pine_sawyer",
    "leaf_beetle",
    "flea_beetle",
    "Colorado_potato_beetle",
    "carpet_beetle",
    "buffalo_carpet_beetle",
    "black_carpet_beetle",
    "clerid_beetle",
    "bee_beetle",
    "lamellicorn_beetle",
    "scarabaeid_beetle",
    "dung_beetle",
    "scarab",
    "tumblebug",
    "dorbeetle",
    "June_beetle",
    "green_June_beetle",
    "Japanese_beetle",
    "Oriental_beetle",
    "rhinoceros_beetle",
    "melolonthid_beetle",
    "cockchafer",
    "rose_chafer",
    "rose_chafer",
    "stag_beetle",
    "elaterid_beetle",
    "click_beetle",
    "firefly",
    "wireworm",
    "water_beetle",
    "whirligig_beetle",
    "deathwatch_beetle",
    "weevil",
    "snout_beetle",
    "boll_weevil",
    "blister_beetle",
    "oil_beetle",
    "Spanish_fly",
    "Dutch-elm_beetle",
    "bark_beetle",
    "spruce_bark_beetle",
    "rove_beetle",
    "darkling_beetle",
    "mealworm",
    "flour_beetle",
    "seed_beetle",
    "pea_weevil",
    "bean_weevil",
    "rice_weevil",
    "Asian_longhorned_beetle",
    "web_spinner",
    "louse",
    "common_louse",
    "head_louse",
    "body_louse",
    "crab_louse",
    "bird_louse",
    "flea",
    "Pulex_irritans",
    "dog_flea",
    "cat_flea",
    "chigoe",
    "sticktight",
    "dipterous_insect",
    "gall_midge",
    "Hessian_fly",
    "fly",
    "housefly",
    "tsetse_fly",
    "blowfly",
    "bluebottle",
    "greenbottle",
    "flesh_fly",
    "tachina_fly",
    "gadfly",
    "botfly",
    "human_botfly",
    "sheep_botfly",
    "warble_fly",
    "horsefly",
    "bee_fly",
    "robber_fly",
    "fruit_fly",
    "apple_maggot",
    "Mediterranean_fruit_fly",
    "drosophila",
    "vinegar_fly",
    "leaf_miner",
    "louse_fly",
    "horse_tick",
    "sheep_ked",
    "horn_fly",
    "mosquito",
    "wiggler",
    "gnat",
    "yellow-fever_mosquito",
    "Asian_tiger_mosquito",
    "anopheline",
    "malarial_mosquito",
    "common_mosquito",
    "Culex_quinquefasciatus",
    "gnat",
    "punkie",
    "midge",
    "fungus_gnat",
    "psychodid",
    "sand_fly",
    "fungus_gnat",
    "armyworm",
    "crane_fly",
    "blackfly",
    "hymenopterous_insect",
    "bee",
    "drone",
    "queen_bee",
    "worker",
    "soldier",
    "worker_bee",
    "honeybee",
    "Africanized_bee",
    "black_bee",
    "Carniolan_bee",
    "Italian_bee",
    "carpenter_bee",
    "bumblebee",
    "cuckoo-bumblebee",
    "andrena",
    "Nomia_melanderi",
    "leaf-cutting_bee",
    "mason_bee",
    "potter_bee",
    "wasp",
    "vespid",
    "paper_wasp",
    "hornet",
    "giant_hornet",
    "common_wasp",
    "bald-faced_hornet",
    "yellow_jacket",
    "Polistes_annularis",
    "mason_wasp",
    "potter_wasp",
    "Mutillidae",
    "velvet_ant",
    "sphecoid_wasp",
    "mason_wasp",
    "digger_wasp",
    "cicada_killer",
    "mud_dauber",
    "gall_wasp",
    "chalcid_fly",
    "strawworm",
    "chalcis_fly",
    "ichneumon_fly",
    "sawfly",
    "birch_leaf_miner",
    "ant",
    "pharaoh_ant",
    "little_black_ant",
    "army_ant",
    "carpenter_ant",
    "fire_ant",
    "wood_ant",
    "slave_ant",
    "Formica_fusca",
    "slave-making_ant",
    "sanguinary_ant",
    "bulldog_ant",
    "Amazon_ant",
    "termite",
    "dry-wood_termite",
    "Reticulitermes_lucifugus",
    "Mastotermes_darwiniensis",
    "Mastotermes_electrodominicus",
    "powder-post_termite",
    "orthopterous_insect",
    "grasshopper",
    "short-horned_grasshopper",
    "locust",
    "migratory_locust",
    "migratory_grasshopper",
    "long-horned_grasshopper",
    "katydid",
    "mormon_cricket",
    "sand_cricket",
    "cricket",
    "mole_cricket",
    "European_house_cricket",
    "field_cricket",
    "tree_cricket",
    "snowy_tree_cricket",
    "phasmid",
    "walking_stick",
    "diapheromera",
    "walking_leaf",
    "cockroach",
    "oriental_cockroach",
    "American_cockroach",
    "Australian_cockroach",
    "German_cockroach",
    "giant_cockroach",
    "mantis",
    "praying_mantis",
    "bug",
    "hemipterous_insect",
    "leaf_bug",
    "mirid_bug",
    "four-lined_plant_bug",
    "lygus_bug",
    "tarnished_plant_bug",
    "lace_bug",
    "lygaeid",
    "chinch_bug",
    "coreid_bug",
    "squash_bug",
    "leaf-footed_bug",
    "bedbug",
    "backswimmer",
    "true_bug",
    "heteropterous_insect",
    "water_bug",
    "giant_water_bug",
    "water_scorpion",
    "water_boatman",
    "water_strider",
    "common_pond-skater",
    "assassin_bug",
    "conenose",
    "wheel_bug",
    "firebug",
    "cotton_stainer",
    "homopterous_insect",
    "whitefly",
    "citrus_whitefly",
    "greenhouse_whitefly",
    "sweet-potato_whitefly",
    "superbug",
    "cotton_strain",
    "coccid_insect",
    "scale_insect",
    "soft_scale",
    "brown_soft_scale",
    "armored_scale",
    "San_Jose_scale",
    "cochineal_insect",
    "mealybug",
    "citrophilous_mealybug",
    "Comstock_mealybug",
    "citrus_mealybug",
    "plant_louse",
    "aphid",
    "apple_aphid",
    "blackfly",
    "greenfly",
    "green_peach_aphid",
    "ant_cow",
    "woolly_aphid",
    "woolly_apple_aphid",
    "woolly_alder_aphid",
    "adelgid",
    "balsam_woolly_aphid",
    "spruce_gall_aphid",
    "woolly_adelgid",
    "jumping_plant_louse",
    "cicada",
    "dog-day_cicada",
    "seventeen-year_locust",
    "spittle_insect",
    "froghopper",
    "meadow_spittlebug",
    "pine_spittlebug",
    "Saratoga_spittlebug",
    "leafhopper",
    "plant_hopper",
    "treehopper",
    "lantern_fly",
    "psocopterous_insect",
    "psocid",
    "bark-louse",
    "booklouse",
    "common_booklouse",
    "ephemerid",
    "mayfly",
    "stonefly",
    "neuropteron",
    "ant_lion",
    "doodlebug",
    "lacewing",
    "aphid_lion",
    "green_lacewing",
    "brown_lacewing",
    "dobson",
    "hellgrammiate",
    "fish_fly",
    "alderfly",
    "snakefly",
    "mantispid",
    "odonate",
    "dragonfly",
    "damselfly",
    "trichopterous_insect",
    "caddis_fly",
    "caseworm",
    "caddisworm",
    "thysanuran_insect",
    "bristletail",
    "silverfish",
    "firebrat",
    "jumping_bristletail",
    "thysanopter",
    "thrips",
    "tobacco_thrips",
    "onion_thrips",
    "earwig",
    "common_European_earwig",
    "lepidopterous_insect",
    "butterfly",
    "nymphalid",
    "mourning_cloak",
    "tortoiseshell",
    "painted_beauty",
    "admiral",
    "red_admiral",
    "white_admiral",
    "banded_purple",
    "red-spotted_purple",
    "viceroy",
    "anglewing",
    "ringlet",
    "comma",
    "fritillary",
    "silverspot",
    "emperor_butterfly",
    "purple_emperor",
    "peacock",
    "danaid",
    "monarch",
    "pierid",
    "cabbage_butterfly",
    "small_white",
    "large_white",
    "southern_cabbage_butterfly",
    "sulphur_butterfly",
    "lycaenid",
    "blue",
    "copper",
    "American_copper",
    "hairstreak",
    "Strymon_melinus",
    "moth",
    "moth_miller",
    "tortricid",
    "leaf_roller",
    "tea_tortrix",
    "orange_tortrix",
    "codling_moth",
    "lymantriid",
    "tussock_caterpillar",
    "gypsy_moth",
    "browntail",
    "gold-tail_moth",
    "geometrid",
    "Paleacrita_vernata",
    "Alsophila_pometaria",
    "cankerworm",
    "spring_cankerworm",
    "fall_cankerworm",
    "measuring_worm",
    "pyralid",
    "bee_moth",
    "corn_borer",
    "Mediterranean_flour_moth",
    "tobacco_moth",
    "almond_moth",
    "raisin_moth",
    "tineoid",
    "tineid",
    "clothes_moth",
    "casemaking_clothes_moth",
    "webbing_clothes_moth",
    "carpet_moth",
    "gelechiid",
    "grain_moth",
    "angoumois_moth",
    "potato_moth",
    "potato_tuberworm",
    "noctuid_moth",
    "cutworm",
    "underwing",
    "red_underwing",
    "antler_moth",
    "heliothis_moth",
    "army_cutworm",
    "armyworm",
    "armyworm",
    "Spodoptera_exigua",
    "beet_armyworm",
    "Spodoptera_frugiperda",
    "fall_armyworm",
    "hawkmoth",
    "Manduca_sexta",
    "tobacco_hornworm",
    "Manduca_quinquemaculata",
    "tomato_hornworm",
    "death's-head_moth",
    "bombycid",
    "domestic_silkworm_moth",
    "silkworm",
    "saturniid",
    "emperor",
    "imperial_moth",
    "giant_silkworm_moth",
    "silkworm",
    "luna_moth",
    "cecropia",
    "cynthia_moth",
    "ailanthus_silkworm",
    "io_moth",
    "polyphemus_moth",
    "pernyi_moth",
    "tussah",
    "atlas_moth",
    "arctiid",
    "tiger_moth",
    "cinnabar",
    "lasiocampid",
    "eggar",
    "tent-caterpillar_moth",
    "tent_caterpillar",
    "tent-caterpillar_moth",
    "forest_tent_caterpillar",
    "lappet",
    "lappet_caterpillar",
    "webworm",
    "webworm_moth",
    "Hyphantria_cunea",
    "fall_webworm",
    "garden_webworm",
    "instar",
    "caterpillar",
    "corn_borer",
    "bollworm",
    "pink_bollworm",
    "corn_earworm",
    "cabbageworm",
    "woolly_bear",
    "woolly_bear_moth",
    "larva",
    "nymph",
    "leptocephalus",
    "grub",
    "maggot",
    "leatherjacket",
    "pupa",
    "chrysalis",
    "imago",
    "queen",
    "phoronid",
    "bryozoan",
    "brachiopod",
    "peanut_worm",
    "echinoderm",
    "starfish",
    "brittle_star",
    "basket_star",
    "Astrophyton_muricatum",
    "sea_urchin",
    "edible_sea_urchin",
    "sand_dollar",
    "heart_urchin",
    "crinoid",
    "sea_lily",
    "feather_star",
    "sea_cucumber",
    "trepang",
    "Duplicidentata",
    "lagomorph",
    "leporid",
    "rabbit",
    "rabbit_ears",
    "lapin",
    "bunny",
    "European_rabbit",
    "wood_rabbit",
    "eastern_cottontail",
    "swamp_rabbit",
    "marsh_hare",
    "hare",
    "leveret",
    "European_hare",
    "jackrabbit",
    "white-tailed_jackrabbit",
    "blacktail_jackrabbit",
    "polar_hare",
    "snowshoe_hare",
    "Belgian_hare",
    "Angora",
    "pika",
    "little_chief_hare",
    "collared_pika",
    "rodent",
    "mouse",
    "rat",
    "pocket_rat",
    "murine",
    "house_mouse",
    "harvest_mouse",
    "field_mouse",
    "nude_mouse",
    "European_wood_mouse",
    "brown_rat",
    "wharf_rat",
    "sewer_rat",
    "black_rat",
    "bandicoot_rat",
    "jerboa_rat",
    "kangaroo_mouse",
    "water_rat",
    "beaver_rat",
    "New_World_mouse",
    "American_harvest_mouse",
    "wood_mouse",
    "white-footed_mouse",
    "deer_mouse",
    "cactus_mouse",
    "cotton_mouse",
    "pygmy_mouse",
    "grasshopper_mouse",
    "muskrat",
    "round-tailed_muskrat",
    "cotton_rat",
    "wood_rat",
    "dusky-footed_wood_rat",
    "vole",
    "packrat",
    "dusky-footed_woodrat",
    "eastern_woodrat",
    "rice_rat",
    "pine_vole",
    "meadow_vole",
    "water_vole",
    "prairie_vole",
    "water_vole",
    "red-backed_mouse",
    "phenacomys",
    "hamster",
    "Eurasian_hamster",
    "golden_hamster",
    "gerbil",
    "jird",
    "tamarisk_gerbil",
    "sand_rat",
    "lemming",
    "European_lemming",
    "brown_lemming",
    "grey_lemming",
    "pied_lemming",
    "Hudson_bay_collared_lemming",
    "southern_bog_lemming",
    "northern_bog_lemming",
    "porcupine",
    "Old_World_porcupine",
    "brush-tailed_porcupine",
    "long-tailed_porcupine",
    "New_World_porcupine",
    "Canada_porcupine",
    "pocket_mouse",
    "silky_pocket_mouse",
    "plains_pocket_mouse",
    "hispid_pocket_mouse",
    "Mexican_pocket_mouse",
    "kangaroo_rat",
    "Ord_kangaroo_rat",
    "kangaroo_mouse",
    "jumping_mouse",
    "meadow_jumping_mouse",
    "jerboa",
    "typical_jerboa",
    "Jaculus_jaculus",
    "dormouse",
    "loir",
    "hazel_mouse",
    "lerot",
    "gopher",
    "plains_pocket_gopher",
    "southeastern_pocket_gopher",
    "valley_pocket_gopher",
    "northern_pocket_gopher",
    "squirrel",
    "tree_squirrel",
    "eastern_grey_squirrel",
    "western_grey_squirrel",
    "fox_squirrel",
    "black_squirrel",
    "red_squirrel",
    "American_red_squirrel",
    "chickeree",
    "antelope_squirrel",
    "ground_squirrel",
    "mantled_ground_squirrel",
    "suslik",
    "flickertail",
    "rock_squirrel",
    "Arctic_ground_squirrel",
    "prairie_dog",
    "blacktail_prairie_dog",
    "whitetail_prairie_dog",
    "eastern_chipmunk",
    "chipmunk",
    "baronduki",
    "American_flying_squirrel",
    "southern_flying_squirrel",
    "northern_flying_squirrel",
    "marmot",
    "groundhog",
    "hoary_marmot",
    "yellowbelly_marmot",
    "Asiatic_flying_squirrel",
    "beaver",
    "Old_World_beaver",
    "New_World_beaver",
    "mountain_beaver",
    "cavy",
    "guinea_pig",
    "aperea",
    "mara",
    "capybara",
    "agouti",
    "paca",
    "mountain_paca",
    "coypu",
    "chinchilla",
    "mountain_chinchilla",
    "viscacha",
    "abrocome",
    "mole_rat",
    "mole_rat",
    "sand_rat",
    "naked_mole_rat",
    "queen",
    "Damaraland_mole_rat",
    "Ungulata",
    "ungulate",
    "unguiculate",
    "dinoceras",
    "hyrax",
    "rock_hyrax",
    "odd-toed_ungulate",
    "equine",
    "horse",
    "roan",
    "stablemate",
    "gee-gee",
    "eohippus",
    "foal",
    "filly",
    "colt",
    "male_horse",
    "ridgeling",
    "stallion",
    "stud",
    "gelding",
    "mare",
    "broodmare",
    "saddle_horse",
    "remount",
    "palfrey",
    "warhorse",
    "cavalry_horse",
    "charger",
    "steed",
    "prancer",
    "hack",
    "cow_pony",
    "quarter_horse",
    "Morgan",
    "Tennessee_walker",
    "American_saddle_horse",
    "Appaloosa",
    "Arabian",
    "Lippizan",
    "pony",
    "polo_pony",
    "mustang",
    "bronco",
    "bucking_bronco",
    "buckskin",
    "crowbait",
    "dun",
    "grey",
    "wild_horse",
    "tarpan",
    "Przewalski's_horse",
    "cayuse",
    "hack",
    "hack",
    "plow_horse",
    "pony",
    "Shetland_pony",
    "Welsh_pony",
    "Exmoor",
    "racehorse",
    "thoroughbred",
    "steeplechaser",
    "racer",
    "finisher",
    "pony",
    "yearling",
    "dark_horse",
    "mudder",
    "nonstarter",
    "stalking-horse",
    "harness_horse",
    "cob",
    "hackney",
    "workhorse",
    "draft_horse",
    "packhorse",
    "carthorse",
    "Clydesdale",
    "Percheron",
    "farm_horse",
    "shire",
    "pole_horse",
    "post_horse",
    "coach_horse",
    "pacer",
    "pacer",
    "trotting_horse",
    "pole_horse",
    "stepper",
    "chestnut",
    "liver_chestnut",
    "bay",
    "sorrel",
    "palomino",
    "pinto",
    "ass",
    "domestic_ass",
    "burro",
    "moke",
    "jack",
    "jennet",
    "mule",
    "hinny",
    "wild_ass",
    "African_wild_ass",
    "kiang",
    "onager",
    "chigetai",
    "zebra",
    "common_zebra",
    "mountain_zebra",
    "grevy's_zebra",
    "quagga",
    "rhinoceros",
    "Indian_rhinoceros",
    "woolly_rhinoceros",
    "white_rhinoceros",
    "black_rhinoceros",
    "tapir",
    "New_World_tapir",
    "Malayan_tapir",
    "even-toed_ungulate",
    "swine",
    "hog",
    "piglet",
    "sucking_pig",
    "porker",
    "boar",
    "sow",
    "razorback",
    "wild_boar",
    "babirusa",
    "warthog",
    "peccary",
    "collared_peccary",
    "white-lipped_peccary",
    "hippopotamus",
    "ruminant",
    "bovid",
    "bovine",
    "ox",
    "cattle",
    "ox",
    "stirk",
    "bullock",
    "bull",
    "cow",
    "heifer",
    "bullock",
    "dogie",
    "maverick",
    "beef",
    "longhorn",
    "Brahman",
    "zebu",
    "aurochs",
    "yak",
    "banteng",
    "Welsh",
    "red_poll",
    "Santa_Gertrudis",
    "Aberdeen_Angus",
    "Africander",
    "dairy_cattle",
    "Ayrshire",
    "Brown_Swiss",
    "Charolais",
    "Jersey",
    "Devon",
    "grade",
    "Durham",
    "milking_shorthorn",
    "Galloway",
    "Friesian",
    "Guernsey",
    "Hereford",
    "cattalo",
    "Old_World_buffalo",
    "water_buffalo",
    "Indian_buffalo",
    "carabao",
    "anoa",
    "tamarau",
    "Cape_buffalo",
    "Asian_wild_ox",
    "gaur",
    "gayal",
    "bison",
    "American_bison",
    "wisent",
    "musk_ox",
    "sheep",
    "ewe",
    "ram",
    "wether",
    "lamb",
    "lambkin",
    "baa-lamb",
    "hog",
    "teg",
    "Persian_lamb",
    "black_sheep",
    "domestic_sheep",
    "Cotswold",
    "Hampshire",
    "Lincoln",
    "Exmoor",
    "Cheviot",
    "broadtail",
    "longwool",
    "merino",
    "Rambouillet",
    "wild_sheep",
    "argali",
    "Marco_Polo_sheep",
    "urial",
    "Dall_sheep",
    "mountain_sheep",
    "bighorn",
    "mouflon",
    "aoudad",
    "goat",
    "kid",
    "billy",
    "nanny",
    "domestic_goat",
    "Cashmere_goat",
    "Angora",
    "wild_goat",
    "bezoar_goat",
    "markhor",
    "ibex",
    "goat_antelope",
    "mountain_goat",
    "goral",
    "serow",
    "chamois",
    "takin",
    "antelope",
    "blackbuck",
    "gerenuk",
    "addax",
    "gnu",
    "dik-dik",
    "hartebeest",
    "sassaby",
    "impala",
    "gazelle",
    "Thomson's_gazelle",
    "Gazella_subgutturosa",
    "springbok",
    "bongo",
    "kudu",
    "greater_kudu",
    "lesser_kudu",
    "harnessed_antelope",
    "nyala",
    "mountain_nyala",
    "bushbuck",
    "nilgai",
    "sable_antelope",
    "saiga",
    "steenbok",
    "eland",
    "common_eland",
    "giant_eland",
    "kob",
    "lechwe",
    "waterbuck",
    "puku",
    "oryx",
    "gemsbok",
    "forest_goat",
    "pronghorn",
    "deer",
    "stag",
    "royal",
    "pricket",
    "fawn",
    "red_deer",
    "hart",
    "hind",
    "brocket",
    "sambar",
    "wapiti",
    "Japanese_deer",
    "Virginia_deer",
    "mule_deer",
    "black-tailed_deer",
    "elk",
    "fallow_deer",
    "roe_deer",
    "roebuck",
    "caribou",
    "woodland_caribou",
    "barren_ground_caribou",
    "brocket",
    "muntjac",
    "musk_deer",
    "pere_david's_deer",
    "chevrotain",
    "kanchil",
    "napu",
    "water_chevrotain",
    "camel",
    "Arabian_camel",
    "Bactrian_camel",
    "llama",
    "domestic_llama",
    "guanaco",
    "alpaca",
    "vicuna",
    "giraffe",
    "okapi",
    "musteline_mammal",
    "weasel",
    "ermine",
    "stoat",
    "New_World_least_weasel",
    "Old_World_least_weasel",
    "longtail_weasel",
    "mink",
    "American_mink",
    "polecat",
    "ferret",
    "black-footed_ferret",
    "muishond",
    "snake_muishond",
    "striped_muishond",
    "otter",
    "river_otter",
    "Eurasian_otter",
    "sea_otter",
    "skunk",
    "striped_skunk",
    "hooded_skunk",
    "hog-nosed_skunk",
    "spotted_skunk",
    "badger",
    "American_badger",
    "Eurasian_badger",
    "ratel",
    "ferret_badger",
    "hog_badger",
    "wolverine",
    "glutton",
    "grison",
    "marten",
    "pine_marten",
    "sable",
    "American_marten",
    "stone_marten",
    "fisher",
    "yellow-throated_marten",
    "tayra",
    "fictional_animal",
    "pachyderm",
    "edentate",
    "armadillo",
    "peba",
    "apar",
    "tatouay",
    "peludo",
    "giant_armadillo",
    "pichiciago",
    "sloth",
    "three-toed_sloth",
    "two-toed_sloth",
    "two-toed_sloth",
    "megatherian",
    "mylodontid",
    "anteater",
    "ant_bear",
    "silky_anteater",
    "tamandua",
    "pangolin",
    "coronet",
    "scapular",
    "tadpole",
    "primate",
    "simian",
    "ape",
    "anthropoid",
    "anthropoid_ape",
    "hominoid",
    "hominid",
    "homo",
    "world",
    "Homo_erectus",
    "Pithecanthropus",
    "Java_man",
    "Peking_man",
    "Sinanthropus",
    "Homo_soloensis",
    "Javanthropus",
    "Homo_habilis",
    "Homo_sapiens",
    "Neandertal_man",
    "Cro-magnon",
    "Homo_sapiens_sapiens",
    "australopithecine",
    "Australopithecus_afarensis",
    "Australopithecus_africanus",
    "Australopithecus_boisei",
    "Zinjanthropus",
    "Australopithecus_robustus",
    "Paranthropus",
    "Sivapithecus",
    "rudapithecus",
    "proconsul",
    "Aegyptopithecus",
    "great_ape",
    "orangutan",
    "gorilla",
    "western_lowland_gorilla",
    "eastern_lowland_gorilla",
    "mountain_gorilla",
    "silverback",
    "chimpanzee",
    "western_chimpanzee",
    "eastern_chimpanzee",
    "central_chimpanzee",
    "pygmy_chimpanzee",
    "lesser_ape",
    "gibbon",
    "siamang",
    "monkey",
    "Old_World_monkey",
    "guenon",
    "talapoin",
    "grivet",
    "vervet",
    "green_monkey",
    "mangabey",
    "patas",
    "baboon",
    "chacma",
    "mandrill",
    "drill",
    "macaque",
    "rhesus",
    "bonnet_macaque",
    "Barbary_ape",
    "crab-eating_macaque",
    "langur",
    "entellus",
    "colobus",
    "guereza",
    "proboscis_monkey",
    "New_World_monkey",
    "marmoset",
    "true_marmoset",
    "pygmy_marmoset",
    "tamarin",
    "silky_tamarin",
    "pinche",
    "capuchin",
    "douroucouli",
    "howler_monkey",
    "saki",
    "uakari",
    "titi",
    "spider_monkey",
    "squirrel_monkey",
    "woolly_monkey",
    "tree_shrew",
    "prosimian",
    "lemur",
    "Madagascar_cat",
    "aye-aye",
    "slender_loris",
    "slow_loris",
    "potto",
    "angwantibo",
    "galago",
    "indri",
    "woolly_indris",
    "tarsier",
    "Tarsius_syrichta",
    "Tarsius_glis",
    "flying_lemur",
    "Cynocephalus_variegatus",
    "proboscidean",
    "elephant",
    "rogue_elephant",
    "Indian_elephant",
    "African_elephant",
    "mammoth",
    "woolly_mammoth",
    "columbian_mammoth",
    "imperial_mammoth",
    "mastodon",
    "plantigrade_mammal",
    "digitigrade_mammal",
    "procyonid",
    "raccoon",
    "common_raccoon",
    "crab-eating_raccoon",
    "bassarisk",
    "kinkajou",
    "coati",
    "lesser_panda",
    "giant_panda",
    "twitterer",
    "fish",
    "fingerling",
    "game_fish",
    "food_fish",
    "rough_fish",
    "groundfish",
    "young_fish",
    "parr",
    "mouthbreeder",
    "spawner",
    "barracouta",
    "crossopterygian",
    "coelacanth",
    "lungfish",
    "ceratodus",
    "catfish",
    "silurid",
    "European_catfish",
    "electric_catfish",
    "bullhead",
    "horned_pout",
    "brown_bullhead",
    "channel_catfish",
    "blue_catfish",
    "flathead_catfish",
    "armored_catfish",
    "sea_catfish",
    "gadoid",
    "cod",
    "codling",
    "Atlantic_cod",
    "Pacific_cod",
    "whiting",
    "burbot",
    "haddock",
    "pollack",
    "hake",
    "silver_hake",
    "ling",
    "cusk",
    "grenadier",
    "eel",
    "elver",
    "common_eel",
    "tuna",
    "moray",
    "conger",
    "teleost_fish",
    "beaked_salmon",
    "clupeid_fish",
    "whitebait",
    "brit",
    "shad",
    "common_American_shad",
    "river_shad",
    "allice_shad",
    "alewife",
    "menhaden",
    "herring",
    "Atlantic_herring",
    "Pacific_herring",
    "sardine",
    "sild",
    "brisling",
    "pilchard",
    "Pacific_sardine",
    "anchovy",
    "mediterranean_anchovy",
    "salmonid",
    "salmon",
    "parr",
    "blackfish",
    "redfish",
    "Atlantic_salmon",
    "landlocked_salmon",
    "sockeye",
    "chinook",
    "coho",
    "trout",
    "brown_trout",
    "rainbow_trout",
    "sea_trout",
    "lake_trout",
    "brook_trout",
    "char",
    "Arctic_char",
    "whitefish",
    "lake_whitefish",
    "cisco",
    "round_whitefish",
    "smelt",
    "sparling",
    "capelin",
    "tarpon",
    "ladyfish",
    "bonefish",
    "argentine",
    "lanternfish",
    "lizardfish",
    "lancetfish",
    "opah",
    "New_World_opah",
    "ribbonfish",
    "dealfish",
    "oarfish",
    "batfish",
    "goosefish",
    "toadfish",
    "oyster_fish",
    "frogfish",
    "sargassum_fish",
    "needlefish",
    "timucu",
    "flying_fish",
    "monoplane_flying_fish",
    "halfbeak",
    "saury",
    "spiny-finned_fish",
    "lingcod",
    "percoid_fish",
    "perch",
    "climbing_perch",
    "perch",
    "yellow_perch",
    "European_perch",
    "pike-perch",
    "walleye",
    "blue_pike",
    "snail_darter",
    "cusk-eel",
    "brotula",
    "pearlfish",
    "robalo",
    "snook",
    "pike",
    "northern_pike",
    "muskellunge",
    "pickerel",
    "chain_pickerel",
    "redfin_pickerel",
    "sunfish",
    "crappie",
    "black_crappie",
    "white_crappie",
    "freshwater_bream",
    "pumpkinseed",
    "bluegill",
    "spotted_sunfish",
    "freshwater_bass",
    "rock_bass",
    "black_bass",
    "Kentucky_black_bass",
    "smallmouth",
    "largemouth",
    "bass",
    "serranid_fish",
    "white_perch",
    "yellow_bass",
    "blackmouth_bass",
    "rock_sea_bass",
    "striped_bass",
    "stone_bass",
    "grouper",
    "hind",
    "rock_hind",
    "creole-fish",
    "jewfish",
    "soapfish",
    "surfperch",
    "rainbow_seaperch",
    "bigeye",
    "catalufa",
    "cardinalfish",
    "flame_fish",
    "tilefish",
    "bluefish",
    "cobia",
    "remora",
    "sharksucker",
    "whale_sucker",
    "carangid_fish",
    "jack",
    "crevalle_jack",
    "yellow_jack",
    "runner",
    "rainbow_runner",
    "leatherjacket",
    "threadfish",
    "moonfish",
    "lookdown",
    "amberjack",
    "yellowtail",
    "kingfish",
    "pompano",
    "Florida_pompano",
    "permit",
    "scad",
    "horse_mackerel",
    "horse_mackerel",
    "bigeye_scad",
    "mackerel_scad",
    "round_scad",
    "dolphinfish",
    "Coryphaena_hippurus",
    "Coryphaena_equisetis",
    "pomfret",
    "characin",
    "tetra",
    "cardinal_tetra",
    "piranha",
    "cichlid",
    "bolti",
    "snapper",
    "red_snapper",
    "grey_snapper",
    "mutton_snapper",
    "schoolmaster",
    "yellowtail",
    "grunt",
    "margate",
    "Spanish_grunt",
    "tomtate",
    "cottonwick",
    "sailor's-choice",
    "porkfish",
    "pompon",
    "pigfish",
    "sparid",
    "sea_bream",
    "porgy",
    "red_porgy",
    "European_sea_bream",
    "Atlantic_sea_bream",
    "sheepshead",
    "pinfish",
    "sheepshead_porgy",
    "snapper",
    "black_bream",
    "scup",
    "scup",
    "sciaenid_fish",
    "striped_drum",
    "jackknife-fish",
    "silver_perch",
    "red_drum",
    "mulloway",
    "maigre",
    "croaker",
    "Atlantic_croaker",
    "yellowfin_croaker",
    "whiting",
    "kingfish",
    "king_whiting",
    "northern_whiting",
    "corbina",
    "white_croaker",
    "white_croaker",
    "sea_trout",
    "weakfish",
    "spotted_weakfish",
    "mullet",
    "goatfish",
    "red_goatfish",
    "yellow_goatfish",
    "mullet",
    "striped_mullet",
    "white_mullet",
    "liza",
    "silversides",
    "jacksmelt",
    "barracuda",
    "great_barracuda",
    "sweeper",
    "sea_chub",
    "Bermuda_chub",
    "spadefish",
    "butterfly_fish",
    "chaetodon",
    "angelfish",
    "rock_beauty",
    "damselfish",
    "beaugregory",
    "anemone_fish",
    "clown_anemone_fish",
    "sergeant_major",
    "wrasse",
    "pigfish",
    "hogfish",
    "slippery_dick",
    "puddingwife",
    "bluehead",
    "pearly_razorfish",
    "tautog",
    "cunner",
    "parrotfish",
    "threadfin",
    "jawfish",
    "stargazer",
    "sand_stargazer",
    "blenny",
    "shanny",
    "Molly_Miller",
    "clinid",
    "pikeblenny",
    "bluethroat_pikeblenny",
    "gunnel",
    "rock_gunnel",
    "eelblenny",
    "wrymouth",
    "wolffish",
    "viviparous_eelpout",
    "ocean_pout",
    "sand_lance",
    "dragonet",
    "goby",
    "mudskipper",
    "sleeper",
    "flathead",
    "archerfish",
    "surgeonfish",
    "gempylid",
    "snake_mackerel",
    "escolar",
    "oilfish",
    "cutlassfish",
    "scombroid",
    "mackerel",
    "common_mackerel",
    "Spanish_mackerel",
    "chub_mackerel",
    "wahoo",
    "Spanish_mackerel",
    "king_mackerel",
    "Scomberomorus_maculatus",
    "cero",
    "sierra",
    "tuna",
    "albacore",
    "bluefin",
    "yellowfin",
    "bonito",
    "skipjack",
    "Chile_bonito",
    "skipjack",
    "bonito",
    "swordfish",
    "sailfish",
    "Atlantic_sailfish",
    "billfish",
    "marlin",
    "blue_marlin",
    "black_marlin",
    "striped_marlin",
    "white_marlin",
    "spearfish",
    "louvar",
    "dollarfish",
    "palometa",
    "harvestfish",
    "driftfish",
    "barrelfish",
    "clingfish",
    "tripletail",
    "Atlantic_tripletail",
    "Pacific_tripletail",
    "mojarra",
    "yellowfin_mojarra",
    "silver_jenny",
    "whiting",
    "ganoid",
    "bowfin",
    "paddlefish",
    "Chinese_paddlefish",
    "sturgeon",
    "Pacific_sturgeon",
    "beluga",
    "gar",
    "scorpaenoid",
    "scorpaenid",
    "scorpionfish",
    "plumed_scorpionfish",
    "lionfish",
    "stonefish",
    "rockfish",
    "copper_rockfish",
    "vermillion_rockfish",
    "red_rockfish",
    "rosefish",
    "bullhead",
    "miller's-thumb",
    "sea_raven",
    "lumpfish",
    "lumpsucker",
    "pogge",
    "greenling",
    "kelp_greenling",
    "painted_greenling",
    "flathead",
    "gurnard",
    "tub_gurnard",
    "sea_robin",
    "northern_sea_robin",
    "flying_gurnard",
    "plectognath",
    "triggerfish",
    "queen_triggerfish",
    "filefish",
    "leatherjacket",
    "boxfish",
    "cowfish",
    "puffer",
    "spiny_puffer",
    "porcupinefish",
    "balloonfish",
    "burrfish",
    "ocean_sunfish",
    "sharptail_mola",
    "flatfish",
    "flounder",
    "righteye_flounder",
    "plaice",
    "European_flatfish",
    "yellowtail_flounder",
    "winter_flounder",
    "lemon_sole",
    "American_plaice",
    "halibut",
    "Atlantic_halibut",
    "Pacific_halibut",
    "lefteye_flounder",
    "southern_flounder",
    "summer_flounder",
    "whiff",
    "horned_whiff",
    "sand_dab",
    "windowpane",
    "brill",
    "turbot",
    "tonguefish",
    "sole",
    "European_sole",
    "English_sole",
    "hogchoker",
    "aba",
    "abacus",
    "abandoned_ship",
    "A_battery",
    "abattoir",
    "abaya",
    "Abbe_condenser",
    "abbey",
    "abbey",
    "abbey",
    "Abney_level",
    "abrader",
    "abrading_stone",
    "abutment",
    "abutment_arch",
    "academic_costume",
    "academic_gown",
    "accelerator",
    "accelerator",
    "accelerator",
    "accelerometer",
    "accessory",
    "accommodating_lens_implant",
    "accommodation",
    "accordion",
    "acetate_disk",
    "acetate_rayon",
    "achromatic_lens",
    "acoustic_delay_line",
    "acoustic_device",
    "acoustic_guitar",
    "acoustic_modem",
    "acropolis",
    "acrylic",
    "acrylic",
    "actinometer",
    "action",
    "active_matrix_screen",
    "actuator",
    "adapter",
    "adder",
    "adding_machine",
    "addressing_machine",
    "adhesive_bandage",
    "adit",
    "adjoining_room",
    "adjustable_wrench",
    "adobe",
    "adz",
    "aeolian_harp",
    "aerator",
    "aerial_torpedo",
    "aerosol",
    "Aertex",
    "afghan",
    "Afro-wig",
    "afterburner",
    "after-shave",
    "agateware",
    "agglomerator",
    "aglet",
    "aglet",
    "agora",
    "aigrette",
    "aileron",
    "air_bag",
    "airbrake",
    "airbrush",
    "airbus",
    "air_compressor",
    "air_conditioner",
    "aircraft",
    "aircraft_carrier",
    "aircraft_engine",
    "air_cushion",
    "airdock",
    "airfield",
    "air_filter",
    "airfoil",
    "airframe",
    "air_gun",
    "air_hammer",
    "air_horn",
    "airing_cupboard",
    "airliner",
    "airmailer",
    "airplane",
    "airplane_propeller",
    "airport",
    "air_pump",
    "air_search_radar",
    "airship",
    "air_terminal",
    "air-to-air_missile",
    "air-to-ground_missile",
    "aisle",
    "Aladdin's_lamp",
    "alarm",
    "alarm_clock",
    "alb",
    "alcazar",
    "alcohol_thermometer",
    "alehouse",
    "alembic",
    "algometer",
    "alidade",
    "alidade",
    "A-line",
    "Allen_screw",
    "Allen_wrench",
    "alligator_wrench",
    "alms_dish",
    "alpaca",
    "alpenstock",
    "altar",
    "altar",
    "altarpiece",
    "altazimuth",
    "alternator",
    "altimeter",
    "Amati",
    "ambulance",
    "amen_corner",
    "American_organ",
    "ammeter",
    "ammonia_clock",
    "ammunition",
    "amphibian",
    "amphibian",
    "amphitheater",
    "amphitheater",
    "amphora",
    "amplifier",
    "ampulla",
    "amusement_arcade",
    "analog_clock",
    "analog_computer",
    "analog_watch",
    "analytical_balance",
    "analyzer",
    "anamorphosis",
    "anastigmat",
    "anchor",
    "anchor_chain",
    "anchor_light",
    "AND_circuit",
    "andiron",
    "android",
    "anechoic_chamber",
    "anemometer",
    "aneroid_barometer",
    "angiocardiogram",
    "angioscope",
    "angle_bracket",
    "angledozer",
    "ankle_brace",
    "anklet",
    "anklet",
    "ankus",
    "anode",
    "anode",
    "answering_machine",
    "antenna",
    "anteroom",
    "antiaircraft",
    "antiballistic_missile",
    "antifouling_paint",
    "anti-G_suit",
    "antimacassar",
    "antiperspirant",
    "anti-submarine_rocket",
    "anvil",
    "ao_dai",
    "apadana",
    "apartment",
    "apartment_building",
    "aperture",
    "aperture",
    "apiary",
    "apparatus",
    "apparel",
    "applecart",
    "appliance",
    "appliance",
    "applicator",
    "appointment",
    "apron",
    "apron_string",
    "apse",
    "aqualung",
    "aquaplane",
    "aquarium",
    "arabesque",
    "arbor",
    "arcade",
    "arch",
    "architecture",
    "architrave",
    "arch_support",
    "arc_lamp",
    "arctic",
    "area",
    "areaway",
    "argyle",
    "ark",
    "arm",
    "armament",
    "armature",
    "armband",
    "armchair",
    "armet",
    "arm_guard",
    "armhole",
    "armilla",
    "armlet",
    "armoire",
    "armor",
    "armored_car",
    "armored_car",
    "armored_personnel_carrier",
    "armored_vehicle",
    "armor_plate",
    "armory",
    "armrest",
    "arquebus",
    "array",
    "array",
    "arrester",
    "arrow",
    "arsenal",
    "arterial_road",
    "arthrogram",
    "arthroscope",
    "artificial_heart",
    "artificial_horizon",
    "artificial_joint",
    "artificial_kidney",
    "artificial_skin",
    "artillery",
    "artillery_shell",
    "artist's_loft",
    "art_school",
    "ascot",
    "ashcan",
    "ash-pan",
    "ashtray",
    "aspergill",
    "aspersorium",
    "aspirator",
    "aspirin_powder",
    "assault_gun",
    "assault_rifle",
    "assegai",
    "assembly",
    "assembly",
    "assembly_hall",
    "assembly_plant",
    "astatic_coils",
    "astatic_galvanometer",
    "astrodome",
    "astrolabe",
    "astronomical_telescope",
    "astronomy_satellite",
    "athenaeum",
    "athletic_sock",
    "athletic_supporter",
    "atlas",
    "atmometer",
    "atom_bomb",
    "atomic_clock",
    "atomic_pile",
    "atomizer",
    "atrium",
    "attache_case",
    "attachment",
    "attack_submarine",
    "attenuator",
    "attic",
    "attic_fan",
    "attire",
    "audio_amplifier",
    "audiocassette",
    "audio_CD",
    "audiometer",
    "audio_system",
    "audiotape",
    "audiotape",
    "audiovisual",
    "auditorium",
    "auger",
    "autobahn",
    "autoclave",
    "autofocus",
    "autogiro",
    "autoinjector",
    "autoloader",
    "automat",
    "automat",
    "automatic_choke",
    "automatic_firearm",
    "automatic_pistol",
    "automatic_rifle",
    "automatic_transmission",
    "automation",
    "automaton",
    "automobile_engine",
    "automobile_factory",
    "automobile_horn",
    "autopilot",
    "autoradiograph",
    "autostrada",
    "auxiliary_boiler",
    "auxiliary_engine",
    "auxiliary_pump",
    "auxiliary_research_submarine",
    "auxiliary_storage",
    "aviary",
    "awl",
    "awning",
    "ax",
    "ax_handle",
    "ax_head",
    "axis",
    "axle",
    "axle_bar",
    "axletree",
    "babushka",
    "baby_bed",
    "baby_buggy",
    "baby_grand",
    "baby_powder",
    "baby_shoe",
    "back",
    "back",
    "backbench",
    "backboard",
    "backboard",
    "backbone",
    "back_brace",
    "backgammon_board",
    "background",
    "backhoe",
    "backlighting",
    "backpack",
    "backpacking_tent",
    "backplate",
    "back_porch",
    "backsaw",
    "backscratcher",
    "backseat",
    "backspace_key",
    "backstairs",
    "backstay",
    "backstop",
    "backsword",
    "backup_system",
    "badminton_court",
    "badminton_equipment",
    "badminton_racket",
    "bag",
    "bag",
    "bag",
    "baggage",
    "baggage",
    "baggage_car",
    "baggage_claim",
    "bagpipe",
    "bailey",
    "bailey",
    "Bailey_bridge",
    "bain-marie",
    "bait",
    "baize",
    "bakery",
    "balaclava",
    "balalaika",
    "balance",
    "balance_beam",
    "balance_wheel",
    "balbriggan",
    "balcony",
    "balcony",
    "baldachin",
    "baldric",
    "bale",
    "baling_wire",
    "ball",
    "ball",
    "ball_and_chain",
    "ball-and-socket_joint",
    "ballast",
    "ball_bearing",
    "ball_cartridge",
    "ballcock",
    "balldress",
    "ballet_skirt",
    "ball_gown",
    "ballistic_galvanometer",
    "ballistic_missile",
    "ballistic_pendulum",
    "ballistocardiograph",
    "balloon",
    "balloon_bomb",
    "balloon_sail",
    "ballot_box",
    "ballpark",
    "ball-peen_hammer",
    "ballpoint",
    "ballroom",
    "ball_valve",
    "balsa_raft",
    "baluster",
    "banana_boat",
    "band",
    "bandage",
    "Band_Aid",
    "bandanna",
    "bandbox",
    "banderilla",
    "bandoleer",
    "bandoneon",
    "bandsaw",
    "bandwagon",
    "bangalore_torpedo",
    "bangle",
    "banjo",
    "banner",
    "bannister",
    "banquette",
    "banyan",
    "baptismal_font",
    "bar",
    "bar",
    "barbecue",
    "barbed_wire",
    "barbell",
    "barber_chair",
    "barbershop",
    "barbette_carriage",
    "barbican",
    "bar_bit",
    "bareboat",
    "barge",
    "barge_pole",
    "baritone",
    "bark",
    "bar_magnet",
    "bar_mask",
    "barn",
    "barndoor",
    "barn_door",
    "barnyard",
    "barograph",
    "barometer",
    "barong",
    "barouche",
    "bar_printer",
    "barrack",
    "barrage_balloon",
    "barrel",
    "barrel",
    "barrelhouse",
    "barrel_knot",
    "barrel_organ",
    "barrel_vault",
    "barrette",
    "barricade",
    "barrier",
    "barroom",
    "barrow",
    "bascule",
    "base",
    "base",
    "baseball",
    "baseball_bat",
    "baseball_cap",
    "baseball_equipment",
    "baseball_glove",
    "basement",
    "basement",
    "basic_point_defense_missile_system",
    "basilica",
    "basilica",
    "basilisk",
    "basin",
    "basinet",
    "basket",
    "basket",
    "basketball",
    "basketball_court",
    "basketball_equipment",
    "basket_weave",
    "bass",
    "bass_clarinet",
    "bass_drum",
    "basset_horn",
    "bass_fiddle",
    "bass_guitar",
    "bass_horn",
    "bassinet",
    "bassinet",
    "bassoon",
    "baster",
    "bastinado",
    "bastion",
    "bastion",
    "bat",
    "bath",
    "bath_chair",
    "bathhouse",
    "bathhouse",
    "bathing_cap",
    "bath_oil",
    "bathrobe",
    "bathroom",
    "bath_salts",
    "bath_towel",
    "bathtub",
    "bathyscaphe",
    "bathysphere",
    "batik",
    "batiste",
    "baton",
    "baton",
    "baton",
    "baton",
    "battering_ram",
    "batter's_box",
    "battery",
    "battery",
    "batting_cage",
    "batting_glove",
    "batting_helmet",
    "battle-ax",
    "battle_cruiser",
    "battle_dress",
    "battlement",
    "battleship",
    "battle_sight",
    "bay",
    "bay",
    "bayonet",
    "bay_rum",
    "bay_window",
    "bazaar",
    "bazaar",
    "bazooka",
    "B_battery",
    "BB_gun",
    "beach_house",
    "beach_towel",
    "beach_wagon",
    "beachwear",
    "beacon",
    "beading_plane",
    "beaker",
    "beaker",
    "beam",
    "beam_balance",
    "beanbag",
    "beanie",
    "bearing",
    "bearing_rein",
    "bearing_wall",
    "bearskin",
    "beater",
    "beating-reed_instrument",
    "beaver",
    "beaver",
    "Beckman_thermometer",
    "bed",
    "bed",
    "bed_and_breakfast",
    "bedclothes",
    "Bedford_cord",
    "bed_jacket",
    "bedpan",
    "bedpost",
    "bedroll",
    "bedroom",
    "bedroom_furniture",
    "bedsitting_room",
    "bedspread",
    "bedspring",
    "bedstead",
    "beefcake",
    "beehive",
    "beeper",
    "beer_barrel",
    "beer_bottle",
    "beer_can",
    "beer_garden",
    "beer_glass",
    "beer_hall",
    "beer_mat",
    "beer_mug",
    "belaying_pin",
    "belfry",
    "bell",
    "bell_arch",
    "bellarmine",
    "bellbottom_trousers",
    "bell_cote",
    "bell_foundry",
    "bell_gable",
    "bell_jar",
    "bellows",
    "bellpull",
    "bell_push",
    "bell_seat",
    "bell_tent",
    "bell_tower",
    "bellyband",
    "belt",
    "belt",
    "belt_buckle",
    "belting",
    "bench",
    "bench_clamp",
    "bench_hook",
    "bench_lathe",
    "bench_press",
    "bender",
    "beret",
    "berlin",
    "Bermuda_shorts",
    "berth",
    "besom",
    "Bessemer_converter",
    "bethel",
    "betting_shop",
    "bevatron",
    "bevel",
    "bevel_gear",
    "B-flat_clarinet",
    "bib",
    "bib-and-tucker",
    "bicorn",
    "bicycle",
    "bicycle-built-for-two",
    "bicycle_chain",
    "bicycle_clip",
    "bicycle_pump",
    "bicycle_rack",
    "bicycle_seat",
    "bicycle_wheel",
    "bidet",
    "bier",
    "bier",
    "bi-fold_door",
    "bifocals",
    "Big_Blue",
    "big_board",
    "bight",
    "bikini",
    "bikini_pants",
    "bilge",
    "bilge_keel",
    "bilge_pump",
    "bilge_well",
    "bill",
    "bill",
    "billboard",
    "billiard_ball",
    "billiard_room",
    "bin",
    "binder",
    "binder",
    "bindery",
    "binding",
    "bin_liner",
    "binnacle",
    "binoculars",
    "binocular_microscope",
    "biochip",
    "biohazard_suit",
    "bioscope",
    "biplane",
    "birch",
    "birchbark_canoe",
    "birdbath",
    "birdcage",
    "birdcall",
    "bird_feeder",
    "birdhouse",
    "bird_shot",
    "biretta",
    "bishop",
    "bistro",
    "bit",
    "bit",
    "bite_plate",
    "bitewing",
    "bitumastic",
    "black",
    "black",
    "blackboard",
    "blackboard_eraser",
    "black_box",
    "blackface",
    "blackjack",
    "black_tie",
    "blackwash",
    "bladder",
    "blade",
    "blade",
    "blade",
    "blank",
    "blanket",
    "blast_furnace",
    "blasting_cap",
    "blazer",
    "blender",
    "blimp",
    "blind",
    "blind_curve",
    "blindfold",
    "bling",
    "blinker",
    "blister_pack",
    "block",
    "blockade",
    "blockade-runner",
    "block_and_tackle",
    "blockbuster",
    "blockhouse",
    "block_plane",
    "bloodmobile",
    "bloomers",
    "blouse",
    "blower",
    "blowtorch",
    "blucher",
    "bludgeon",
    "blue",
    "blue_chip",
    "blunderbuss",
    "blunt_file",
    "boarding",
    "boarding_house",
    "boardroom",
    "boards",
    "boat",
    "boater",
    "boat_hook",
    "boathouse",
    "boatswain's_chair",
    "boat_train",
    "boatyard",
    "bobbin",
    "bobby_pin",
    "bobsled",
    "bobsled",
    "bocce_ball",
    "bodega",
    "bodice",
    "bodkin",
    "bodkin",
    "bodkin",
    "body",
    "body_armor",
    "body_lotion",
    "body_stocking",
    "body_plethysmograph",
    "body_pad",
    "bodywork",
    "Bofors_gun",
    "bogy",
    "boiler",
    "boiling_water_reactor",
    "bolero",
    "bollard",
    "bolo",
    "bolo_tie",
    "bolt",
    "bolt",
    "bolt",
    "bolt_cutter",
    "bomb",
    "bombazine",
    "bomb_calorimeter",
    "bomber",
    "bomber_jacket",
    "bomblet",
    "bomb_rack",
    "bombshell",
    "bomb_shelter",
    "bone-ash_cup",
    "bone_china",
    "bones",
    "boneshaker",
    "bongo",
    "bonnet",
    "book",
    "book_bag",
    "bookbindery",
    "bookcase",
    "bookend",
    "bookmark",
    "bookmobile",
    "bookshelf",
    "bookshop",
    "boom",
    "boom",
    "boomerang",
    "booster",
    "booster",
    "boot",
    "boot",
    "boot_camp",
    "bootee",
    "booth",
    "booth",
    "booth",
    "boothose",
    "bootjack",
    "bootlace",
    "bootleg",
    "bootstrap",
    "bore_bit",
    "boron_chamber",
    "borstal",
    "bosom",
    "Boston_rocker",
    "bota",
    "bottle",
    "bottle",
    "bottle_bank",
    "bottlebrush",
    "bottlecap",
    "bottle_opener",
    "bottling_plant",
    "bottom",
    "boucle",
    "boudoir",
    "boulle",
    "bouncing_betty",
    "bouquet",
    "boutique",
    "boutonniere",
    "bow",
    "bow",
    "bow",
    "bow_and_arrow",
    "bowed_stringed_instrument",
    "Bowie_knife",
    "bowl",
    "bowl",
    "bowl",
    "bowler_hat",
    "bowline",
    "bowling_alley",
    "bowling_ball",
    "bowling_equipment",
    "bowling_pin",
    "bowling_shoe",
    "bowsprit",
    "bowstring",
    "bow_tie",
    "box",
    "box",
    "box",
    "box_beam",
    "box_camera",
    "boxcar",
    "box_coat",
    "boxing_equipment",
    "boxing_glove",
    "box_office",
    "box_spring",
    "box_wrench",
    "brace",
    "brace",
    "brace",
    "brace",
    "brace_and_bit",
    "bracelet",
    "bracer",
    "brace_wrench",
    "bracket",
    "bradawl",
    "brake",
    "brake",
    "brake_band",
    "brake_cylinder",
    "brake_disk",
    "brake_drum",
    "brake_lining",
    "brake_pad",
    "brake_pedal",
    "brake_shoe",
    "brake_system",
    "brass",
    "brass",
    "brass",
    "brassard",
    "brasserie",
    "brassie",
    "brassiere",
    "brass_knucks",
    "brattice",
    "brazier",
    "breadbasket",
    "bread-bin",
    "bread_knife",
    "breakable",
    "breakfast_area",
    "breakfast_table",
    "breakwater",
    "breast_drill",
    "breast_implant",
    "breastplate",
    "breast_pocket",
    "breathalyzer",
    "breechblock",
    "breechcloth",
    "breeches",
    "breeches_buoy",
    "breechloader",
    "breeder_reactor",
    "Bren",
    "brewpub",
    "brick",
    "brickkiln",
    "bricklayer's_hammer",
    "brick_trowel",
    "brickwork",
    "bridal_gown",
    "bridge",
    "bridge",
    "bridle",
    "bridle_path",
    "bridoon",
    "briefcase",
    "briefcase_bomb",
    "briefcase_computer",
    "briefs",
    "brig",
    "brig",
    "brigandine",
    "brigantine",
    "brilliantine",
    "brilliant_pebble",
    "brim",
    "bristle_brush",
    "britches",
    "broad_arrow",
    "broadax",
    "brochette",
    "broadcaster",
    "broadcloth",
    "broadcloth",
    "broad_hatchet",
    "broadloom",
    "broadside",
    "broadsword",
    "brocade",
    "brogan",
    "broiler",
    "broken_arch",
    "bronchoscope",
    "broom",
    "broom_closet",
    "broomstick",
    "brougham",
    "Browning_automatic_rifle",
    "Browning_machine_gun",
    "brownstone",
    "brunch_coat",
    "brush",
    "Brussels_carpet",
    "Brussels_lace",
    "bubble",
    "bubble_chamber",
    "bubble_jet_printer",
    "buckboard",
    "bucket",
    "bucket_seat",
    "bucket_shop",
    "buckle",
    "buckram",
    "bucksaw",
    "buckskins",
    "buff",
    "buffer",
    "buffer",
    "buffet",
    "buffing_wheel",
    "buggy",
    "bugle",
    "building",
    "building_complex",
    "bulldog_clip",
    "bulldog_wrench",
    "bulldozer",
    "bullet",
    "bulletproof_vest",
    "bullet_train",
    "bullhorn",
    "bullion",
    "bullnose",
    "bullpen",
    "bullpen",
    "bullring",
    "bulwark",
    "bumboat",
    "bumper",
    "bumper",
    "bumper_car",
    "bumper_guard",
    "bumper_jack",
    "bundle",
    "bung",
    "bungalow",
    "bungee",
    "bunghole",
    "bunk",
    "bunk",
    "bunk_bed",
    "bunker",
    "bunker",
    "bunker",
    "bunsen_burner",
    "bunting",
    "bur",
    "Burberry",
    "burette",
    "burglar_alarm",
    "burial_chamber",
    "burial_garment",
    "burial_mound",
    "burin",
    "burqa",
    "burlap",
    "burn_bag",
    "burner",
    "burnous",
    "burp_gun",
    "burr",
    "bus",
    "bushel_basket",
    "bushing",
    "bush_jacket",
    "business_suit",
    "buskin",
    "bustier",
    "bustle",
    "butcher_knife",
    "butcher_shop",
    "butter_dish",
    "butterfly_valve",
    "butter_knife",
    "butt_hinge",
    "butt_joint",
    "button",
    "buttonhook",
    "buttress",
    "butt_shaft",
    "butt_weld",
    "buzz_bomb",
    "buzzer",
    "BVD",
    "bypass_condenser",
    "byway",
    "cab",
    "cab",
    "cab",
    "cabana",
    "cabaret",
    "caber",
    "cabin",
    "cabin",
    "cabin_car",
    "cabin_class",
    "cabin_cruiser",
    "cabinet",
    "cabinet",
    "cabinet",
    "cabinetwork",
    "cabin_liner",
    "cable",
    "cable",
    "cable_car",
    "cache",
    "caddy",
    "caesium_clock",
    "cafe",
    "cafeteria",
    "cafeteria_tray",
    "caff",
    "caftan",
    "caftan",
    "cage",
    "cage",
    "cagoule",
    "caisson",
    "calash",
    "calceus",
    "calcimine",
    "calculator",
    "caldron",
    "calico",
    "caliper",
    "call-board",
    "call_center",
    "caller_ID",
    "calliope",
    "calorimeter",
    "calpac",
    "camail",
    "camber_arch",
    "cambric",
    "camcorder",
    "camel's_hair",
    "camera",
    "camera_lens",
    "camera_lucida",
    "camera_obscura",
    "camera_tripod",
    "camise",
    "camisole",
    "camisole",
    "camlet",
    "camouflage",
    "camouflage",
    "camp",
    "camp",
    "camp",
    "campaign_hat",
    "campanile",
    "camp_chair",
    "camper",
    "camper_trailer",
    "campstool",
    "camshaft",
    "can",
    "canal",
    "canal_boat",
    "candelabrum",
    "candid_camera",
    "candle",
    "candlepin",
    "candlesnuffer",
    "candlestick",
    "candlewick",
    "candy_thermometer",
    "cane",
    "cane",
    "cangue",
    "canister",
    "cannery",
    "cannikin",
    "cannikin",
    "cannon",
    "cannon",
    "cannon",
    "cannon",
    "cannonball",
    "canoe",
    "can_opener",
    "canopic_jar",
    "canopy",
    "canopy",
    "canopy",
    "canteen",
    "canteen",
    "canteen",
    "canteen",
    "canteen",
    "cant_hook",
    "cantilever",
    "cantilever_bridge",
    "cantle",
    "Canton_crepe",
    "canvas",
    "canvas",
    "canvas_tent",
    "cap",
    "cap",
    "cap",
    "capacitor",
    "caparison",
    "cape",
    "capital_ship",
    "capitol",
    "cap_opener",
    "capote",
    "capote",
    "cap_screw",
    "capstan",
    "capstone",
    "capsule",
    "captain's_chair",
    "car",
    "car",
    "car",
    "carabiner",
    "carafe",
    "caravansary",
    "car_battery",
    "carbine",
    "car_bomb",
    "carbon_arc_lamp",
    "carboy",
    "carburetor",
    "car_carrier",
    "cardcase",
    "cardiac_monitor",
    "cardigan",
    "card_index",
    "cardiograph",
    "cardioid_microphone",
    "car_door",
    "cardroom",
    "card_table",
    "card_table",
    "car-ferry",
    "cargo_area",
    "cargo_container",
    "cargo_door",
    "cargo_hatch",
    "cargo_helicopter",
    "cargo_liner",
    "cargo_ship",
    "carillon",
    "car_mirror",
    "caroche",
    "carousel",
    "carpenter's_hammer",
    "carpenter's_kit",
    "carpenter's_level",
    "carpenter's_mallet",
    "carpenter's_rule",
    "carpenter's_square",
    "carpetbag",
    "carpet_beater",
    "carpet_loom",
    "carpet_pad",
    "carpet_sweeper",
    "carpet_tack",
    "carport",
    "carrack",
    "carrel",
    "carriage",
    "carriage",
    "carriage_bolt",
    "carriageway",
    "carriage_wrench",
    "carrick_bend",
    "carrier",
    "carryall",
    "carrycot",
    "car_seat",
    "cart",
    "car_tire",
    "carton",
    "cartouche",
    "car_train",
    "cartridge",
    "cartridge",
    "cartridge_belt",
    "cartridge_extractor",
    "cartridge_fuse",
    "cartridge_holder",
    "cartwheel",
    "carving_fork",
    "carving_knife",
    "car_wheel",
    "caryatid",
    "cascade_liquefier",
    "cascade_transformer",
    "case",
    "case",
    "case",
    "casein_paint",
    "case_knife",
    "case_knife",
    "casement",
    "casement_window",
    "casern",
    "case_shot",
    "cash_bar",
    "cashbox",
    "cash_machine",
    "cashmere",
    "cash_register",
    "casing",
    "casino",
    "casket",
    "casque",
    "casquet",
    "Cassegrainian_telescope",
    "casserole",
    "cassette",
    "cassette_deck",
    "cassette_player",
    "cassette_recorder",
    "cassette_tape",
    "cassock",
    "cast",
    "caster",
    "caster",
    "castle",
    "castle",
    "catacomb",
    "catafalque",
    "catalytic_converter",
    "catalytic_cracker",
    "catamaran",
    "catapult",
    "catapult",
    "catboat",
    "cat_box",
    "catch",
    "catchall",
    "catcher's_mask",
    "catchment",
    "Caterpillar",
    "cathedra",
    "cathedral",
    "cathedral",
    "catheter",
    "cathode",
    "cathode-ray_tube",
    "cat-o'-nine-tails",
    "cat's-paw",
    "catsup_bottle",
    "cattle_car",
    "cattle_guard",
    "cattleship",
    "cautery",
    "cavalier_hat",
    "cavalry_sword",
    "cavetto",
    "cavity_wall",
    "C_battery",
    "C-clamp",
    "CD_drive",
    "CD_player",
    "CD-R",
    "CD-ROM",
    "CD-ROM_drive",
    "cedar_chest",
    "ceiling",
    "celesta",
    "cell",
    "cell",
    "cellar",
    "cellblock",
    "cello",
    "cellophane",
    "cellular_telephone",
    "cellulose_tape",
    "cenotaph",
    "censer",
    "center",
    "center_punch",
    "Centigrade_thermometer",
    "central_processing_unit",
    "centrifugal_pump",
    "centrifuge",
    "ceramic",
    "ceramic_ware",
    "cereal_bowl",
    "cereal_box",
    "cerecloth",
    "cesspool",
    "chachka",
    "chador",
    "chafing_dish",
    "chain",
    "chain",
    "chainlink_fence",
    "chain_mail",
    "chain_printer",
    "chain_saw",
    "chain_store",
    "chain_tongs",
    "chain_wrench",
    "chair",
    "chair",
    "chair_of_state",
    "chairlift",
    "chaise",
    "chaise_longue",
    "chalet",
    "chalice",
    "chalk",
    "challis",
    "chamberpot",
    "chambray",
    "chamfer_bit",
    "chamfer_plane",
    "chamois_cloth",
    "chancel",
    "chancellery",
    "chancery",
    "chandelier",
    "chandlery",
    "chanfron",
    "chanter",
    "chantry",
    "chap",
    "chapel",
    "chapterhouse",
    "chapterhouse",
    "character_printer",
    "charcuterie",
    "charge-exchange_accelerator",
    "charger",
    "chariot",
    "chariot",
    "charnel_house",
    "chassis",
    "chassis",
    "chasuble",
    "chateau",
    "chatelaine",
    "checker",
    "checkout",
    "cheekpiece",
    "cheeseboard",
    "cheesecloth",
    "cheese_cutter",
    "cheese_press",
    "chemical_bomb",
    "chemical_plant",
    "chemical_reactor",
    "chemise",
    "chemise",
    "chenille",
    "chessman",
    "chest",
    "chesterfield",
    "chest_of_drawers",
    "chest_protector",
    "cheval-de-frise",
    "cheval_glass",
    "chicane",
    "chicken_coop",
    "chicken_wire",
    "chicken_yard",
    "chiffon",
    "chiffonier",
    "child's_room",
    "chime",
    "chimney_breast",
    "chimney_corner",
    "china",
    "china_cabinet",
    "chinchilla",
    "Chinese_lantern",
    "Chinese_puzzle",
    "chinning_bar",
    "chino",
    "chino",
    "chin_rest",
    "chin_strap",
    "chintz",
    "chip",
    "chip",
    "chisel",
    "chlamys",
    "choir",
    "choir_loft",
    "choke",
    "choke",
    "chokey",
    "choo-choo",
    "chopine",
    "chordophone",
    "Christmas_stocking",
    "chronograph",
    "chronometer",
    "chronoscope",
    "chuck",
    "chuck_wagon",
    "chukka",
    "church",
    "church_bell",
    "church_hat",
    "church_key",
    "church_tower",
    "churidars",
    "churn",
    "ciderpress",
    "cigar_band",
    "cigar_box",
    "cigar_cutter",
    "cigarette_butt",
    "cigarette_case",
    "cigarette_holder",
    "cigar_lighter",
    "cinch",
    "cinema",
    "cinquefoil",
    "circle",
    "circlet",
    "circuit",
    "circuit_board",
    "circuit_breaker",
    "circuitry",
    "circular_plane",
    "circular_saw",
    "circus_tent",
    "cistern",
    "cistern",
    "cittern",
    "city_hall",
    "cityscape",
    "city_university",
    "civies",
    "civilian_clothing",
    "clack_valve",
    "clamp",
    "clamshell",
    "clapper",
    "clapperboard",
    "clarence",
    "clarinet",
    "Clark_cell",
    "clasp",
    "clasp_knife",
    "classroom",
    "clavichord",
    "clavier",
    "clay_pigeon",
    "claymore_mine",
    "claymore",
    "cleaners",
    "cleaning_implement",
    "cleaning_pad",
    "clean_room",
    "clearway",
    "cleat",
    "cleat",
    "cleats",
    "cleaver",
    "clerestory",
    "clevis",
    "clews",
    "cliff_dwelling",
    "climbing_frame",
    "clinch",
    "clinch",
    "clincher",
    "clinic",
    "clinical_thermometer",
    "clinker",
    "clinometer",
    "clip",
    "clip_lead",
    "clip-on",
    "clipper",
    "clipper",
    "clipper",
    "cloak",
    "cloak",
    "cloakroom",
    "cloche",
    "cloche",
    "clock",
    "clock_pendulum",
    "clock_radio",
    "clock_tower",
    "clockwork",
    "clog",
    "cloisonne",
    "cloister",
    "closed_circuit",
    "closed-circuit_television",
    "closed_loop",
    "closet",
    "closeup_lens",
    "cloth_cap",
    "cloth_covering",
    "clothesbrush",
    "clothes_closet",
    "clothes_dryer",
    "clothes_hamper",
    "clotheshorse",
    "clothespin",
    "clothes_tree",
    "clothing",
    "clothing_store",
    "clout_nail",
    "clove_hitch",
    "club_car",
    "clubroom",
    "cluster_bomb",
    "clutch",
    "clutch",
    "clutch_bag",
    "coach",
    "coach_house",
    "coal_car",
    "coal_chute",
    "coal_house",
    "coal_shovel",
    "coaming",
    "coaster_brake",
    "coat",
    "coat_button",
    "coat_closet",
    "coatdress",
    "coatee",
    "coat_hanger",
    "coating",
    "coating",
    "coat_of_paint",
    "coatrack",
    "coattail",
    "coaxial_cable",
    "cobweb",
    "cobweb",
    "Cockcroft_and_Walton_accelerator",
    "cocked_hat",
    "cockhorse",
    "cockleshell",
    "cockpit",
    "cockpit",
    "cockpit",
    "cockscomb",
    "cocktail_dress",
    "cocktail_lounge",
    "cocktail_shaker",
    "cocotte",
    "codpiece",
    "coelostat",
    "coffee_can",
    "coffee_cup",
    "coffee_filter",
    "coffee_maker",
    "coffee_mill",
    "coffee_mug",
    "coffeepot",
    "coffee_stall",
    "coffee_table",
    "coffee_urn",
    "coffer",
    "Coffey_still",
    "coffin",
    "cog",
    "coif",
    "coil",
    "coil",
    "coil",
    "coil_spring",
    "coin_box",
    "colander",
    "cold_cathode",
    "cold_chisel",
    "cold_cream",
    "cold_frame",
    "collar",
    "collar",
    "college",
    "collet",
    "collider",
    "colliery",
    "collimator",
    "collimator",
    "cologne",
    "colonnade",
    "colonoscope",
    "colorimeter",
    "colors",
    "color_television",
    "color_tube",
    "color_wash",
    "Colt",
    "colter",
    "columbarium",
    "columbarium",
    "column",
    "column",
    "comb",
    "comb",
    "comber",
    "combination_lock",
    "combination_plane",
    "combine",
    "comforter",
    "command_module",
    "commissary",
    "commissary",
    "commodity",
    "common_ax",
    "common_room",
    "communications_satellite",
    "communication_system",
    "community_center",
    "commutator",
    "commuter",
    "compact",
    "compact",
    "compact_disk",
    "compact-disk_burner",
    "companionway",
    "compartment",
    "compartment",
    "compass",
    "compass",
    "compass_card",
    "compass_saw",
    "compound",
    "compound_lens",
    "compound_lever",
    "compound_microscope",
    "compress",
    "compression_bandage",
    "compressor",
    "computer",
    "computer_circuit",
    "computerized_axial_tomography_scanner",
    "computer_keyboard",
    "computer_monitor",
    "computer_network",
    "computer_screen",
    "computer_store",
    "computer_system",
    "concentration_camp",
    "concert_grand",
    "concert_hall",
    "concertina",
    "concertina",
    "concrete_mixer",
    "condensation_pump",
    "condenser",
    "condenser",
    "condenser",
    "condenser_microphone",
    "condominium",
    "condominium",
    "conductor",
    "cone_clutch",
    "confectionery",
    "conference_center",
    "conference_room",
    "conference_table",
    "confessional",
    "conformal_projection",
    "congress_boot",
    "conic_projection",
    "connecting_rod",
    "connecting_room",
    "connection",
    "conning_tower",
    "conning_tower",
    "conservatory",
    "conservatory",
    "console",
    "console",
    "console_table",
    "consulate",
    "contact",
    "contact",
    "container",
    "container_ship",
    "containment",
    "contrabassoon",
    "control",
    "control_center",
    "control_circuit",
    "control_key",
    "control_panel",
    "control_rod",
    "control_room",
    "control_system",
    "control_tower",
    "convector",
    "convenience_store",
    "convent",
    "conventicle",
    "converging_lens",
    "converter",
    "convertible",
    "convertible",
    "conveyance",
    "conveyer_belt",
    "cooker",
    "cookfire",
    "cookhouse",
    "cookie_cutter",
    "cookie_jar",
    "cookie_sheet",
    "cooking_utensil",
    "cookstove",
    "coolant_system",
    "cooler",
    "cooling_system",
    "cooling_system",
    "cooling_tower",
    "coonskin_cap",
    "cope",
    "coping_saw",
    "copperware",
    "copyholder",
    "coquille",
    "coracle",
    "corbel",
    "corbel_arch",
    "corbel_step",
    "corbie_gable",
    "cord",
    "cord",
    "cordage",
    "cords",
    "core",
    "core_bit",
    "core_drill",
    "corer",
    "cork",
    "corker",
    "corkscrew",
    "corncrib",
    "corner",
    "corner",
    "corner_post",
    "cornet",
    "cornice",
    "cornice",
    "cornice",
    "correctional_institution",
    "corrugated_fastener",
    "corselet",
    "corset",
    "cosmetic",
    "cosmotron",
    "costume",
    "costume",
    "costume",
    "costume",
    "cosy",
    "cot",
    "cottage_tent",
    "cotter",
    "cotter_pin",
    "cotton",
    "cotton_flannel",
    "cotton_mill",
    "couch",
    "couch",
    "couchette",
    "coude_telescope",
    "counter",
    "counter",
    "counter",
    "counterbore",
    "counter_tube",
    "country_house",
    "country_store",
    "coupe",
    "coupling",
    "court",
    "court",
    "court",
    "court",
    "Courtelle",
    "courthouse",
    "courthouse",
    "coverall",
    "covered_bridge",
    "covered_couch",
    "covered_wagon",
    "covering",
    "coverlet",
    "cover_plate",
    "cowbarn",
    "cowbell",
    "cowboy_boot",
    "cowboy_hat",
    "cowhide",
    "cowl",
    "cow_pen",
    "CPU_board",
    "crackle",
    "cradle",
    "craft",
    "cramp",
    "crampon",
    "crampon",
    "crane",
    "craniometer",
    "crank",
    "crankcase",
    "crankshaft",
    "crash_barrier",
    "crash_helmet",
    "crate",
    "cravat",
    "crayon",
    "crazy_quilt",
    "cream",
    "cream_pitcher",
    "creche",
    "creche",
    "credenza",
    "creel",
    "crematory",
    "crematory",
    "crepe",
    "crepe_de_Chine",
    "crescent_wrench",
    "cretonne",
    "crib",
    "crib",
    "cricket_ball",
    "cricket_bat",
    "cricket_equipment",
    "cringle",
    "crinoline",
    "crinoline",
    "crochet_needle",
    "crock",
    "Crock_Pot",
    "crook",
    "Crookes_radiometer",
    "Crookes_tube",
    "croquet_ball",
    "croquet_equipment",
    "croquet_mallet",
    "cross",
    "crossbar",
    "crossbar",
    "crossbar",
    "crossbench",
    "cross_bit",
    "crossbow",
    "crosscut_saw",
    "crossjack",
    "crosspiece",
    "crotchet",
    "croupier's_rake",
    "crowbar",
    "crown",
    "crown",
    "crown_jewels",
    "crown_lens",
    "crow's_nest",
    "crucible",
    "crucifix",
    "cruet",
    "cruet-stand",
    "cruise_control",
    "cruise_missile",
    "cruiser",
    "cruiser",
    "cruise_ship",
    "crupper",
    "cruse",
    "crusher",
    "crutch",
    "cryometer",
    "cryoscope",
    "cryostat",
    "crypt",
    "crystal",
    "crystal_detector",
    "crystal_microphone",
    "crystal_oscillator",
    "crystal_set",
    "cubitiere",
    "cucking_stool",
    "cuckoo_clock",
    "cuddy",
    "cudgel",
    "cue",
    "cue_ball",
    "cuff",
    "cuirass",
    "cuisse",
    "cul",
    "culdoscope",
    "cullis",
    "culotte",
    "cultivator",
    "culverin",
    "culverin",
    "culvert",
    "cup",
    "cupboard",
    "cup_hook",
    "cupola",
    "cupola",
    "curb",
    "curb_roof",
    "curbstone",
    "curette",
    "curler",
    "curling_iron",
    "currycomb",
    "cursor",
    "curtain",
    "customhouse",
    "cutaway",
    "cutlas",
    "cutoff",
    "cutout",
    "cutter",
    "cutter",
    "cutting_implement",
    "cutting_room",
    "cutty_stool",
    "cutwork",
    "cybercafe",
    "cyclopean_masonry",
    "cyclostyle",
    "cyclotron",
    "cylinder",
    "cylinder",
    "cylinder_lock",
    "cymbal",
    "dacha",
    "Dacron",
    "dado",
    "dado_plane",
    "dagger",
    "dairy",
    "dais",
    "daisy_print_wheel",
    "daisywheel_printer",
    "dam",
    "damask",
    "dampener",
    "damper",
    "damper_block",
    "dark_lantern",
    "darkroom",
    "darning_needle",
    "dart",
    "dart",
    "dashboard",
    "dashiki",
    "dash-pot",
    "data_converter",
    "data_input_device",
    "data_multiplexer",
    "data_system",
    "davenport",
    "davenport",
    "davit",
    "daybed",
    "daybook",
    "day_nursery",
    "day_school",
    "dead_axle",
    "deadeye",
    "deadhead",
    "deanery",
    "deathbed",
    "death_camp",
    "death_house",
    "death_knell",
    "death_seat",
    "deck",
    "deck",
    "deck_chair",
    "deck-house",
    "deckle",
    "deckle_edge",
    "declinometer",
    "decoder",
    "decolletage",
    "decoupage",
    "dedicated_file_server",
    "deep-freeze",
    "deerstalker",
    "defense_system",
    "defensive_structure",
    "defibrillator",
    "defilade",
    "deflector",
    "delayed_action",
    "delay_line",
    "delft",
    "delicatessen",
    "delivery_truck",
    "delta_wing",
    "demijohn",
    "demitasse",
    "den",
    "denim",
    "densimeter",
    "densitometer",
    "dental_appliance",
    "dental_floss",
    "dental_implant",
    "dentist's_drill",
    "denture",
    "deodorant",
    "department_store",
    "departure_lounge",
    "depilatory",
    "depressor",
    "depth_finder",
    "depth_gauge",
    "derrick",
    "derrick",
    "derringer",
    "desk",
    "desk_phone",
    "desktop_computer",
    "dessert_spoon",
    "destroyer",
    "destroyer_escort",
    "detached_house",
    "detector",
    "detector",
    "detention_home",
    "detonating_fuse",
    "detonator",
    "developer",
    "device",
    "Dewar_flask",
    "dhoti",
    "dhow",
    "dial",
    "dial",
    "dial",
    "dialog_box",
    "dial_telephone",
    "dialyzer",
    "diamante",
    "diaper",
    "diaper",
    "diaphone",
    "diaphragm",
    "diaphragm",
    "diathermy_machine",
    "dibble",
    "dice_cup",
    "dicer",
    "dickey",
    "dickey",
    "Dictaphone",
    "die",
    "diesel",
    "diesel-electric_locomotive",
    "diesel-hydraulic_locomotive",
    "diesel_locomotive",
    "diestock",
    "differential_analyzer",
    "differential_gear",
    "diffuser",
    "diffuser",
    "digester",
    "diggings",
    "digital-analog_converter",
    "digital_audiotape",
    "digital_camera",
    "digital_clock",
    "digital_computer",
    "digital_display",
    "digital_subscriber_line",
    "digital_voltmeter",
    "digital_watch",
    "digitizer",
    "dilator",
    "dildo",
    "dimity",
    "dimmer",
    "diner",
    "dinette",
    "dinghy",
    "dining_area",
    "dining_car",
    "dining-hall",
    "dining_room",
    "dining-room_furniture",
    "dining-room_table",
    "dining_table",
    "dinner_bell",
    "dinner_dress",
    "dinner_jacket",
    "dinner_napkin",
    "dinner_pail",
    "dinner_table",
    "dinner_theater",
    "diode",
    "diode",
    "dip",
    "diplomatic_building",
    "dipole",
    "dipper",
    "dipstick",
    "DIP_switch",
    "directional_antenna",
    "directional_microphone",
    "direction_finder",
    "dirk",
    "dirndl",
    "dirndl",
    "dirty_bomb",
    "discharge_lamp",
    "discharge_pipe",
    "disco",
    "discount_house",
    "discus",
    "disguise",
    "dish",
    "dish",
    "dishpan",
    "dish_rack",
    "dishrag",
    "dishtowel",
    "dishwasher",
    "disk",
    "disk_brake",
    "disk_clutch",
    "disk_controller",
    "disk_drive",
    "diskette",
    "disk_harrow",
    "dispatch_case",
    "dispensary",
    "dispenser",
    "display",
    "display_adapter",
    "display_panel",
    "display_window",
    "disposal",
    "disrupting_explosive",
    "distaff",
    "distillery",
    "distributor",
    "distributor_cam",
    "distributor_cap",
    "distributor_housing",
    "distributor_point",
    "ditch",
    "ditch_spade",
    "ditty_bag",
    "divan",
    "divan",
    "dive_bomber",
    "diverging_lens",
    "divided_highway",
    "divider",
    "diving_bell",
    "divining_rod",
    "diving_suit",
    "dixie",
    "Dixie_cup",
    "dock",
    "doeskin",
    "dogcart",
    "doggie_bag",
    "dogsled",
    "dog_wrench",
    "doily",
    "doll",
    "dollhouse",
    "dolly",
    "dolman",
    "dolman",
    "dolman_sleeve",
    "dolmen",
    "dome",
    "dome",
    "domino",
    "dongle",
    "donkey_jacket",
    "door",
    "door",
    "door",
    "doorbell",
    "doorframe",
    "doorjamb",
    "doorlock",
    "doormat",
    "doornail",
    "doorplate",
    "doorsill",
    "doorstop",
    "Doppler_radar",
    "dormer",
    "dormer_window",
    "dormitory",
    "dormitory",
    "dosemeter",
    "dossal",
    "dot_matrix_printer",
    "double_bed",
    "double-bitted_ax",
    "double_boiler",
    "double-breasted_jacket",
    "double-breasted_suit",
    "double_door",
    "double_glazing",
    "double-hung_window",
    "double_knit",
    "doubler",
    "double_reed",
    "double-reed_instrument",
    "doublet",
    "doubletree",
    "douche",
    "dovecote",
    "Dover's_powder",
    "dovetail",
    "dovetail_plane",
    "dowel",
    "downstage",
    "drafting_instrument",
    "drafting_table",
    "Dragunov",
    "drainage_ditch",
    "drainage_system",
    "drain_basket",
    "drainplug",
    "drape",
    "drapery",
    "drawbar",
    "drawbridge",
    "drawer",
    "drawers",
    "drawing_chalk",
    "drawing_room",
    "drawing_room",
    "drawknife",
    "drawstring_bag",
    "dray",
    "dreadnought",
    "dredge",
    "dredger",
    "dredging_bucket",
    "dress",
    "dress_blues",
    "dresser",
    "dress_hat",
    "dressing",
    "dressing_case",
    "dressing_gown",
    "dressing_room",
    "dressing_sack",
    "dressing_table",
    "dress_rack",
    "dress_shirt",
    "dress_suit",
    "dress_uniform",
    "drift_net",
    "drill",
    "electric_drill",
    "drilling_platform",
    "drill_press",
    "drill_rig",
    "drinking_fountain",
    "drinking_vessel",
    "drip_loop",
    "drip_mat",
    "drip_pan",
    "dripping_pan",
    "drip_pot",
    "drive",
    "drive",
    "drive_line",
    "driver",
    "driveshaft",
    "driveway",
    "driving_iron",
    "driving_wheel",
    "drogue",
    "drogue_parachute",
    "drone",
    "drone",
    "drop_arch",
    "drop_cloth",
    "drop_curtain",
    "drop_forge",
    "drop-leaf_table",
    "dropper",
    "droshky",
    "drove",
    "drugget",
    "drugstore",
    "drum",
    "drum",
    "drum_brake",
    "drumhead",
    "drum_printer",
    "drum_sander",
    "drumstick",
    "dry_battery",
    "dry-bulb_thermometer",
    "dry_cell",
    "dry_dock",
    "dryer",
    "dry_fly",
    "dry_kiln",
    "dry_masonry",
    "dry_point",
    "dry_wall",
    "dual_scan_display",
    "duck",
    "duckboard",
    "duckpin",
    "dudeen",
    "duffel",
    "duffel_bag",
    "duffel_coat",
    "dugout",
    "dugout_canoe",
    "dulciana",
    "dulcimer",
    "dulcimer",
    "dumbbell",
    "dumb_bomb",
    "dumbwaiter",
    "dumdum",
    "dumpcart",
    "Dumpster",
    "dump_truck",
    "Dumpy_level",
    "dunce_cap",
    "dune_buggy",
    "dungeon",
    "duplex_apartment",
    "duplex_house",
    "duplicator",
    "dust_bag",
    "dustcloth",
    "dust_cover",
    "dust_cover",
    "dustmop",
    "dustpan",
    "Dutch_oven",
    "Dutch_oven",
    "dwelling",
    "dye-works",
    "dynamo",
    "dynamometer",
    "Eames_chair",
    "earflap",
    "early_warning_radar",
    "early_warning_system",
    "earmuff",
    "earphone",
    "earplug",
    "earplug",
    "earthenware",
    "earthwork",
    "easel",
    "easy_chair",
    "eaves",
    "ecclesiastical_attire",
    "echinus",
    "echocardiograph",
    "edger",
    "edge_tool",
    "efficiency_apartment",
    "egg-and-dart",
    "eggbeater",
    "egg_timer",
    "eiderdown",
    "eight_ball",
    "ejection_seat",
    "elastic",
    "elastic_bandage",
    "Elastoplast",
    "elbow",
    "elbow_pad",
    "electric",
    "electrical_cable",
    "electrical_contact",
    "electrical_converter",
    "electrical_device",
    "electrical_system",
    "electric_bell",
    "electric_blanket",
    "electric_chair",
    "electric_clock",
    "electric-discharge_lamp",
    "electric_fan",
    "electric_frying_pan",
    "electric_furnace",
    "electric_guitar",
    "electric_hammer",
    "electric_heater",
    "electric_lamp",
    "electric_locomotive",
    "electric_meter",
    "electric_mixer",
    "electric_motor",
    "electric_organ",
    "electric_range",
    "electric_refrigerator",
    "electric_toothbrush",
    "electric_typewriter",
    "electro-acoustic_transducer",
    "electrode",
    "electrodynamometer",
    "electroencephalograph",
    "electrograph",
    "electrolytic",
    "electrolytic_cell",
    "electromagnet",
    "electrometer",
    "electromyograph",
    "electron_accelerator",
    "electron_gun",
    "electronic_balance",
    "electronic_converter",
    "electronic_device",
    "electronic_equipment",
    "electronic_fetal_monitor",
    "electronic_instrument",
    "electronic_voltmeter",
    "electron_microscope",
    "electron_multiplier",
    "electrophorus",
    "electroscope",
    "electrostatic_generator",
    "electrostatic_printer",
    "elevator",
    "elevator",
    "elevator_shaft",
    "embankment",
    "embassy",
    "embellishment",
    "emergency_room",
    "emesis_basin",
    "emitter",
    "empty",
    "emulsion",
    "enamel",
    "enamel",
    "enamelware",
    "encaustic",
    "encephalogram",
    "enclosure",
    "endoscope",
    "energizer",
    "engine",
    "engine",
    "engineering",
    "enginery",
    "English_horn",
    "English_saddle",
    "enlarger",
    "ensemble",
    "ensign",
    "entablature",
    "entertainment_center",
    "entrenching_tool",
    "entrenchment",
    "envelope",
    "envelope",
    "envelope",
    "eolith",
    "epauliere",
    "epee",
    "epergne",
    "epicyclic_train",
    "epidiascope",
    "epilating_wax",
    "equalizer",
    "equatorial",
    "equipment",
    "erasable_programmable_read-only_memory",
    "eraser",
    "erecting_prism",
    "erection",
    "Erlenmeyer_flask",
    "escape_hatch",
    "escapement",
    "escape_wheel",
    "escarpment",
    "escutcheon",
    "esophagoscope",
    "espadrille",
    "espalier",
    "espresso_maker",
    "espresso_shop",
    "establishment",
    "estaminet",
    "estradiol_patch",
    "etagere",
    "etamine",
    "etching",
    "ethernet",
    "ethernet_cable",
    "Eton_jacket",
    "etui",
    "eudiometer",
    "euphonium",
    "evaporative_cooler",
    "evening_bag",
    "exercise_bike",
    "exercise_device",
    "exhaust",
    "exhaust_fan",
    "exhaust_valve",
    "exhibition_hall",
    "Exocet",
    "expansion_bit",
    "expansion_bolt",
    "explosive_detection_system",
    "explosive_device",
    "explosive_trace_detection",
    "express",
    "extension",
    "extension_cord",
    "external-combustion_engine",
    "external_drive",
    "extractor",
    "eyebrow_pencil",
    "eyecup",
    "eyeliner",
    "eyepatch",
    "eyepiece",
    "eyeshadow",
    "fabric",
    "facade",
    "face_guard",
    "face_mask",
    "faceplate",
    "face_powder",
    "face_veil",
    "facing",
    "facing",
    "facing",
    "facsimile",
    "factory",
    "factory_ship",
    "fagot",
    "fagot_stitch",
    "Fahrenheit_thermometer",
    "faience",
    "faille",
    "fairlead",
    "fairy_light",
    "falchion",
    "fallboard",
    "fallout_shelter",
    "false_face",
    "false_teeth",
    "family_room",
    "fan",
    "fan_belt",
    "fan_blade",
    "fancy_dress",
    "fanion",
    "fanlight",
    "fanjet",
    "fanjet",
    "fanny_pack",
    "fan_tracery",
    "fan_vaulting",
    "farm_building",
    "farmer's_market",
    "farmhouse",
    "farm_machine",
    "farmplace",
    "farmyard",
    "farthingale",
    "fastener",
    "fast_reactor",
    "fat_farm",
    "fatigues",
    "faucet",
    "fauld",
    "fauteuil",
    "feather_boa",
    "featheredge",
    "fedora",
    "feedback_circuit",
    "feedlot",
    "fell",
    "felloe",
    "felt",
    "felt-tip_pen",
    "felucca",
    "fence",
    "fencing_mask",
    "fencing_sword",
    "fender",
    "fender",
    "Ferris_wheel",
    "ferrule",
    "ferry",
    "ferule",
    "festoon",
    "fetoscope",
    "fetter",
    "fez",
    "fiber",
    "fiber_optic_cable",
    "fiberscope",
    "fichu",
    "fiddlestick",
    "field_artillery",
    "field_coil",
    "field-effect_transistor",
    "field-emission_microscope",
    "field_glass",
    "field_hockey_ball",
    "field_hospital",
    "field_house",
    "field_lens",
    "field_magnet",
    "field-sequential_color_television",
    "field_tent",
    "fieldwork",
    "fife",
    "fifth_wheel",
    "fighter",
    "fighting_chair",
    "fig_leaf",
    "figure_eight",
    "figure_loom",
    "figure_skate",
    "filament",
    "filature",
    "file",
    "file",
    "file_folder",
    "file_server",
    "filigree",
    "filling",
    "film",
    "film",
    "film_advance",
    "filter",
    "filter",
    "finder",
    "finery",
    "fine-tooth_comb",
    "finger",
    "fingerboard",
    "finger_bowl",
    "finger_paint",
    "finger-painting",
    "finger_plate",
    "fingerstall",
    "finish_coat",
    "finish_coat",
    "finisher",
    "fin_keel",
    "fipple",
    "fipple_flute",
    "fire",
    "fire_alarm",
    "firearm",
    "fire_bell",
    "fireboat",
    "firebox",
    "firebrick",
    "fire_control_radar",
    "fire_control_system",
    "fire_engine",
    "fire_extinguisher",
    "fire_iron",
    "fireman's_ax",
    "fireplace",
    "fire_screen",
    "fire_tongs",
    "fire_tower",
    "firewall",
    "firing_chamber",
    "firing_pin",
    "firkin",
    "firmer_chisel",
    "first-aid_kit",
    "first-aid_station",
    "first_base",
    "first_class",
    "fishbowl",
    "fisherman's_bend",
    "fisherman's_knot",
    "fisherman's_lure",
    "fishhook",
    "fishing_boat",
    "fishing_gear",
    "fishing_rod",
    "fish_joint",
    "fish_knife",
    "fishnet",
    "fish_slice",
    "fitment",
    "fixative",
    "fixer-upper",
    "flag",
    "flageolet",
    "flagon",
    "flagpole",
    "flagship",
    "flail",
    "flambeau",
    "flamethrower",
    "flange",
    "flannel",
    "flannel",
    "flannelette",
    "flap",
    "flash",
    "flash",
    "flash_camera",
    "flasher",
    "flashlight",
    "flashlight_battery",
    "flash_memory",
    "flask",
    "flat_arch",
    "flatbed",
    "flatbed_press",
    "flat_bench",
    "flatcar",
    "flat_file",
    "flatlet",
    "flat_panel_display",
    "flats",
    "flat_tip_screwdriver",
    "fleece",
    "fleet_ballistic_missile_submarine",
    "fleur-de-lis",
    "flight_simulator",
    "flintlock",
    "flintlock",
    "flip-flop",
    "flipper",
    "float",
    "floating_dock",
    "floatplane",
    "flood",
    "floor",
    "floor",
    "floor",
    "floorboard",
    "floor_cover",
    "floor_joist",
    "floor_lamp",
    "flophouse",
    "florist",
    "floss",
    "flotsam",
    "flour_bin",
    "flour_mill",
    "flowerbed",
    "flugelhorn",
    "fluid_drive",
    "fluid_flywheel",
    "flume",
    "fluorescent_lamp",
    "fluoroscope",
    "flush_toilet",
    "flute",
    "flute",
    "flux_applicator",
    "fluxmeter",
    "fly",
    "flying_boat",
    "flying_buttress",
    "flying_carpet",
    "flying_jib",
    "fly_rod",
    "fly_tent",
    "flytrap",
    "flywheel",
    "fob",
    "foghorn",
    "foglamp",
    "foil",
    "fold",
    "folder",
    "folding_chair",
    "folding_door",
    "folding_saw",
    "food_court",
    "food_processor",
    "food_hamper",
    "foot",
    "footage",
    "football",
    "football_helmet",
    "football_stadium",
    "footbath",
    "foot_brake",
    "footbridge",
    "foothold",
    "footlocker",
    "foot_rule",
    "footstool",
    "footwear",
    "footwear",
    "forceps",
    "force_pump",
    "fore-and-after",
    "fore-and-aft_sail",
    "forecastle",
    "forecourt",
    "foredeck",
    "fore_edge",
    "foreground",
    "foremast",
    "fore_plane",
    "foresail",
    "forestay",
    "foretop",
    "fore-topmast",
    "fore-topsail",
    "forge",
    "fork",
    "forklift",
    "formalwear",
    "Formica",
    "fortification",
    "fortress",
    "forty-five",
    "Foucault_pendulum",
    "foulard",
    "foul-weather_gear",
    "foundation_garment",
    "foundry",
    "fountain",
    "fountain_pen",
    "four-in-hand",
    "four-poster",
    "four-pounder",
    "four-stroke_engine",
    "four-wheel_drive",
    "four-wheel_drive",
    "four-wheeler",
    "fowling_piece",
    "foxhole",
    "fragmentation_bomb",
    "frail",
    "fraise",
    "frame",
    "frame",
    "frame_buffer",
    "framework",
    "Francis_turbine",
    "franking_machine",
    "free_house",
    "free-reed",
    "free-reed_instrument",
    "freewheel",
    "freight_car",
    "freight_elevator",
    "freight_liner",
    "freight_train",
    "French_door",
    "French_horn",
    "French_polish",
    "French_roof",
    "French_window",
    "Fresnel_lens",
    "fret",
    "friary",
    "friction_clutch",
    "frieze",
    "frieze",
    "frigate",
    "frigate",
    "frill",
    "Frisbee",
    "frock",
    "frock_coat",
    "frontlet",
    "front_porch",
    "front_projector",
    "fruit_machine",
    "frying_pan",
    "fuel_filter",
    "fuel_gauge",
    "fuel_injection",
    "fuel_system",
    "full-dress_uniform",
    "full_metal_jacket",
    "full_skirt",
    "fumigator",
    "funeral_home",
    "funnel",
    "funny_wagon",
    "fur",
    "fur_coat",
    "fur_hat",
    "furnace",
    "furnace_lining",
    "furnace_room",
    "furnishing",
    "furnishing",
    "furniture",
    "fur-piece",
    "furrow",
    "fuse",
    "fusee_drive",
    "fuselage",
    "fusil",
    "fustian",
    "futon",
    "gabardine",
    "gable",
    "gable_roof",
    "gadgetry",
    "gaff",
    "gaff",
    "gaff",
    "gaffsail",
    "gaff_topsail",
    "gag",
    "gaiter",
    "gaiter",
    "Galilean_telescope",
    "galleon",
    "gallery",
    "gallery",
    "galley",
    "galley",
    "galley",
    "gallows",
    "gallows_tree",
    "galvanometer",
    "gambling_house",
    "gambrel",
    "game",
    "gamebag",
    "game_equipment",
    "gaming_table",
    "gamp",
    "gangplank",
    "gangsaw",
    "gangway",
    "gantlet",
    "gantry",
    "garage",
    "garage",
    "Garand_rifle",
    "garbage",
    "garbage_truck",
    "garboard",
    "garden",
    "garden",
    "garden_rake",
    "garden_spade",
    "garden_tool",
    "garden_trowel",
    "gargoyle",
    "garibaldi",
    "garlic_press",
    "garment",
    "garment_bag",
    "garrison_cap",
    "garrote",
    "garter",
    "garter_belt",
    "garter_stitch",
    "gas_guzzler",
    "gas_shell",
    "gas_bracket",
    "gas_burner",
    "gas-cooled_reactor",
    "gas-discharge_tube",
    "gas_engine",
    "gas_fixture",
    "gas_furnace",
    "gas_gun",
    "gas_heater",
    "gas_holder",
    "gasket",
    "gas_lamp",
    "gas_maser",
    "gasmask",
    "gas_meter",
    "gasoline_engine",
    "gasoline_gauge",
    "gas_oven",
    "gas_oven",
    "gas_pump",
    "gas_range",
    "gas_ring",
    "gas_tank",
    "gas_thermometer",
    "gastroscope",
    "gas_turbine",
    "gas-turbine_ship",
    "gat",
    "gate",
    "gatehouse",
    "gateleg_table",
    "gatepost",
    "gathered_skirt",
    "Gatling_gun",
    "gauge",
    "gauntlet",
    "gauntlet",
    "gauze",
    "gauze",
    "gavel",
    "gazebo",
    "gear",
    "gear",
    "gear",
    "gearbox",
    "gearing",
    "gearset",
    "gearshift",
    "Geiger_counter",
    "Geiger_tube",
    "gene_chip",
    "general-purpose_bomb",
    "generator",
    "generator",
    "generator",
    "Geneva_gown",
    "geodesic_dome",
    "georgette",
    "gharry",
    "ghat",
    "ghetto_blaster",
    "gift_shop",
    "gift_wrapping",
    "gig",
    "gig",
    "gig",
    "gig",
    "gildhall",
    "gill_net",
    "gilt",
    "gimbal",
    "gingham",
    "girandole",
    "girder",
    "girdle",
    "glass",
    "glass",
    "glass_cutter",
    "glasses_case",
    "glebe_house",
    "Glengarry",
    "glider",
    "Global_Positioning_System",
    "glockenspiel",
    "glory_hole",
    "glove",
    "glove_compartment",
    "glow_lamp",
    "glow_tube",
    "glyptic_art",
    "glyptics",
    "gnomon",
    "goal",
    "goalmouth",
    "goalpost",
    "goblet",
    "godown",
    "goggles",
    "go-kart",
    "gold_plate",
    "golf_bag",
    "golf_ball",
    "golfcart",
    "golf_club",
    "golf-club_head",
    "golf_equipment",
    "golf_glove",
    "golliwog",
    "gondola",
    "gong",
    "goniometer",
    "Gordian_knot",
    "gorget",
    "gossamer",
    "Gothic_arch",
    "gouache",
    "gouge",
    "gourd",
    "government_building",
    "government_office",
    "gown",
    "gown",
    "gown",
    "grab",
    "grab_bag",
    "grab_bar",
    "grace_cup",
    "grade_separation",
    "graduated_cylinder",
    "graffito",
    "gramophone",
    "granary",
    "grandfather_clock",
    "grand_piano",
    "graniteware",
    "granny_knot",
    "grape_arbor",
    "grapnel",
    "grapnel",
    "grass_skirt",
    "grate",
    "grate",
    "grater",
    "graver",
    "gravestone",
    "gravimeter",
    "gravure",
    "gravy_boat",
    "grey",
    "grease-gun",
    "greasepaint",
    "greasy_spoon",
    "greatcoat",
    "great_hall",
    "greave",
    "greengrocery",
    "greenhouse",
    "grenade",
    "grid",
    "griddle",
    "grill",
    "grille",
    "grillroom",
    "grinder",
    "grinding_wheel",
    "grindstone",
    "gripsack",
    "gristmill",
    "grocery_bag",
    "grocery_store",
    "grogram",
    "groined_vault",
    "groover",
    "grosgrain",
    "gros_point",
    "ground",
    "ground_bait",
    "ground_control",
    "ground_floor",
    "groundsheet",
    "G-string",
    "guard",
    "guard_boat",
    "guardroom",
    "guardroom",
    "guard_ship",
    "guard's_van",
    "gueridon",
    "Guarnerius",
    "guesthouse",
    "guestroom",
    "guidance_system",
    "guided_missile",
    "guided_missile_cruiser",
    "guided_missile_frigate",
    "guildhall",
    "guilloche",
    "guillotine",
    "guimpe",
    "guimpe",
    "guitar",
    "guitar_pick",
    "gulag",
    "gun",
    "gunboat",
    "gun_carriage",
    "gun_case",
    "gun_emplacement",
    "gun_enclosure",
    "gunlock",
    "gunnery",
    "gunnysack",
    "gun_pendulum",
    "gun_room",
    "gunsight",
    "gun_trigger",
    "gurney",
    "gusher",
    "gusset",
    "gusset",
    "guy",
    "gymnastic_apparatus",
    "gym_shoe",
    "gym_suit",
    "gymslip",
    "gypsy_cab",
    "gyrocompass",
    "gyroscope",
    "gyrostabilizer",
    "habergeon",
    "habit",
    "habit",
    "hacienda",
    "hacksaw",
    "haft",
    "hairbrush",
    "haircloth",
    "hairdressing",
    "hairnet",
    "hairpiece",
    "hairpin",
    "hair_shirt",
    "hair_slide",
    "hair_spray",
    "hairspring",
    "hair_trigger",
    "halberd",
    "half_binding",
    "half_hatchet",
    "half_hitch",
    "half_track",
    "hall",
    "hall",
    "hall",
    "Hall_of_Fame",
    "hall_of_residence",
    "hallstand",
    "halter",
    "halter",
    "hame",
    "hammer",
    "hammer",
    "hammer",
    "hammerhead",
    "hammock",
    "hamper",
    "hand",
    "handball",
    "handbarrow",
    "handbell",
    "hand_blower",
    "handbow",
    "hand_brake",
    "hand_calculator",
    "handcar",
    "handcart",
    "hand_cream",
    "handcuff",
    "hand_drill",
    "hand_glass",
    "hand_glass",
    "hand_grenade",
    "hand-held_computer",
    "handhold",
    "handkerchief",
    "handlebar",
    "handloom",
    "hand_lotion",
    "hand_luggage",
    "hand-me-down",
    "hand_mower",
    "hand_pump",
    "handrest",
    "handsaw",
    "handset",
    "hand_shovel",
    "handspike",
    "handstamp",
    "hand_throttle",
    "hand_tool",
    "hand_towel",
    "hand_truck",
    "handwear",
    "handwheel",
    "handwheel",
    "hangar_queen",
    "hanger",
    "hang_glider",
    "hangman's_rope",
    "hank",
    "hansom",
    "harbor",
    "hard_disc",
    "hard_hat",
    "hardtop",
    "hardware",
    "hardware_store",
    "harmonica",
    "harmonium",
    "harness",
    "harness",
    "harp",
    "harp",
    "harpoon",
    "harpoon_gun",
    "harpoon_log",
    "harpsichord",
    "Harris_Tweed",
    "harrow",
    "harvester",
    "hash_house",
    "hasp",
    "hat",
    "hatbox",
    "hatch",
    "hatchback",
    "hatchback",
    "hatchel",
    "hatchet",
    "hatpin",
    "hauberk",
    "Hawaiian_guitar",
    "hawse",
    "hawser",
    "hawser_bend",
    "hay_bale",
    "hayfork",
    "hayloft",
    "haymaker",
    "hayrack",
    "hayrack",
    "hazard",
    "head",
    "head",
    "head",
    "headboard",
    "head_covering",
    "headdress",
    "header",
    "header",
    "header",
    "header",
    "headfast",
    "head_gasket",
    "head_gate",
    "headgear",
    "headlight",
    "headpiece",
    "headpin",
    "headquarters",
    "headrace",
    "headrest",
    "headsail",
    "headscarf",
    "headset",
    "head_shop",
    "headstall",
    "headstock",
    "health_spa",
    "hearing_aid",
    "hearing_aid",
    "hearse",
    "hearth",
    "hearthrug",
    "heart-lung_machine",
    "heat_engine",
    "heater",
    "heat_exchanger",
    "heating_pad",
    "heat_lamp",
    "heat_pump",
    "heat-seeking_missile",
    "heat_shield",
    "heat_sink",
    "heaume",
    "heaver",
    "heavier-than-air_craft",
    "heckelphone",
    "hectograph",
    "hedge",
    "hedge_trimmer",
    "helicon",
    "helicopter",
    "heliograph",
    "heliometer",
    "helm",
    "helmet",
    "helmet",
    "hematocrit",
    "hemming-stitch",
    "hemostat",
    "hemstitch",
    "henroost",
    "heraldry",
    "hermitage",
    "herringbone",
    "herringbone",
    "Herschelian_telescope",
    "Hessian_boot",
    "heterodyne_receiver",
    "hibachi",
    "hideaway",
    "hi-fi",
    "high_altar",
    "high-angle_gun",
    "highball_glass",
    "highboard",
    "highboy",
    "highchair",
    "high_gear",
    "high-hat_cymbal",
    "highlighter",
    "highlighter",
    "high-pass_filter",
    "high-rise",
    "high_table",
    "high-warp_loom",
    "hijab",
    "hinge",
    "hinging_post",
    "hip_boot",
    "hipflask",
    "hip_pad",
    "hip_pocket",
    "hippodrome",
    "hip_roof",
    "hitch",
    "hitch",
    "hitching_post",
    "hitchrack",
    "hob",
    "hobble_skirt",
    "hockey_skate",
    "hockey_stick",
    "hod",
    "hodoscope",
    "hoe",
    "hoe_handle",
    "hogshead",
    "hoist",
    "hold",
    "holder",
    "holding_cell",
    "holding_device",
    "holding_pen",
    "hollowware",
    "holster",
    "holster",
    "holy_of_holies",
    "home",
    "home_appliance",
    "home_computer",
    "home_plate",
    "home_room",
    "homespun",
    "homestead",
    "home_theater",
    "homing_torpedo",
    "hone",
    "honeycomb",
    "hood",
    "hood",
    "hood",
    "hood",
    "hood",
    "hood_latch",
    "hook",
    "hook",
    "hook",
    "hookah",
    "hook_and_eye",
    "hookup",
    "hookup",
    "hook_wrench",
    "hoopskirt",
    "hoosegow",
    "Hoover",
    "hope_chest",
    "hopper",
    "hopsacking",
    "horizontal_bar",
    "horizontal_stabilizer",
    "horizontal_tail",
    "horn",
    "horn",
    "horn",
    "horn_button",
    "hornpipe",
    "horse",
    "horsebox",
    "horsecar",
    "horse_cart",
    "horsecloth",
    "horse-drawn_vehicle",
    "horsehair",
    "horsehair_wig",
    "horseless_carriage",
    "horse_pistol",
    "horseshoe",
    "horseshoe",
    "horse-trail",
    "horsewhip",
    "hose",
    "hosiery",
    "hospice",
    "hospital",
    "hospital_bed",
    "hospital_room",
    "hospital_ship",
    "hospital_train",
    "hostel",
    "hostel",
    "hot-air_balloon",
    "hotel",
    "hotel-casino",
    "hotel-casino",
    "hotel_room",
    "hot_line",
    "hot_pants",
    "hot_plate",
    "hot_rod",
    "hot_spot",
    "hot_tub",
    "hot-water_bottle",
    "houndstooth_check",
    "hourglass",
    "hour_hand",
    "house",
    "house",
    "houseboat",
    "houselights",
    "house_of_cards",
    "house_of_correction",
    "house_paint",
    "housetop",
    "housing",
    "hovel",
    "hovercraft",
    "howdah",
    "huarache",
    "hub-and-spoke",
    "hubcap",
    "huck",
    "hug-me-tight",
    "hula-hoop",
    "hulk",
    "hull",
    "humeral_veil",
    "Humvee",
    "hunter",
    "hunting_knife",
    "hurdle",
    "hurricane_deck",
    "hurricane_lamp",
    "hut",
    "hutch",
    "hutment",
    "hydraulic_brake",
    "hydraulic_press",
    "hydraulic_pump",
    "hydraulic_system",
    "hydraulic_transmission",
    "hydroelectric_turbine",
    "hydrofoil",
    "hydrofoil",
    "hydrogen_bomb",
    "hydrometer",
    "hygrodeik",
    "hygrometer",
    "hygroscope",
    "hyperbaric_chamber",
    "hypercoaster",
    "hypermarket",
    "hypodermic_needle",
    "hypodermic_syringe",
    "hypsometer",
    "hysterosalpingogram",
    "I-beam",
    "ice_ax",
    "iceboat",
    "icebreaker",
    "iced-tea_spoon",
    "ice_hockey_rink",
    "ice_machine",
    "ice_maker",
    "ice_pack",
    "icepick",
    "ice_rink",
    "ice_skate",
    "ice_tongs",
    "icetray",
    "iconoscope",
    "Identikit",
    "idle_pulley",
    "igloo",
    "ignition_coil",
    "ignition_key",
    "ignition_switch",
    "imaret",
    "immovable_bandage",
    "impact_printer",
    "impeller",
    "implant",
    "implement",
    "impression",
    "imprint",
    "improvised_explosive_device",
    "impulse_turbine",
    "in-basket",
    "incendiary_bomb",
    "incinerator",
    "inclined_plane",
    "inclinometer",
    "inclinometer",
    "incrustation",
    "incubator",
    "index_register",
    "Indiaman",
    "Indian_club",
    "indicator",
    "induction_coil",
    "inductor",
    "industrial_watercourse",
    "inertial_guidance_system",
    "inflater",
    "inhaler",
    "injector",
    "ink_bottle",
    "ink_eraser",
    "ink-jet_printer",
    "inkle",
    "inkstand",
    "inkwell",
    "inlay",
    "inside_caliper",
    "insole",
    "instep",
    "instillator",
    "institution",
    "instrument",
    "instrument_of_punishment",
    "instrument_of_torture",
    "intaglio",
    "intake_valve",
    "integrated_circuit",
    "integrator",
    "Intelnet",
    "interceptor",
    "interchange",
    "intercommunication_system",
    "intercontinental_ballistic_missile",
    "interface",
    "interferometer",
    "interior_door",
    "internal-combustion_engine",
    "internal_drive",
    "internet",
    "interphone",
    "interrupter",
    "intersection",
    "interstice",
    "intraocular_lens",
    "intravenous_pyelogram",
    "inverter",
    "ion_engine",
    "ionization_chamber",
    "iPod",
    "video_iPod",
    "iron",
    "iron",
    "iron",
    "irons",
    "ironclad",
    "iron_foundry",
    "iron_horse",
    "ironing",
    "iron_lung",
    "ironmongery",
    "ironworks",
    "irrigation_ditch",
    "izar",
    "jabot",
    "jack",
    "jack",
    "jack",
    "jack",
    "jacket",
    "jacket",
    "jacket",
    "jack-in-the-box",
    "jack-o'-lantern",
    "jack_plane",
    "Jacob's_ladder",
    "jaconet",
    "Jacquard_loom",
    "jacquard",
    "jag",
    "jail",
    "jalousie",
    "jamb",
    "jammer",
    "jampot",
    "japan",
    "jar",
    "Jarvik_heart",
    "jaunting_car",
    "javelin",
    "jaw",
    "Jaws_of_Life",
    "jean",
    "jeep",
    "jellaba",
    "jerkin",
    "jeroboam",
    "jersey",
    "jersey",
    "jet",
    "jet_bridge",
    "jet_engine",
    "jetliner",
    "jeweler's_glass",
    "jewelled_headdress",
    "jew's_harp",
    "jib",
    "jibboom",
    "jig",
    "jig",
    "jiggermast",
    "jigsaw",
    "jigsaw_puzzle",
    "jinrikisha",
    "jobcentre",
    "jodhpurs",
    "jodhpur",
    "joinery",
    "joint",
    "Joint_Direct_Attack_Munition",
    "jointer",
    "joist",
    "jolly_boat",
    "jorum",
    "joss_house",
    "journal_bearing",
    "journal_box",
    "joystick",
    "jungle_gym",
    "junk",
    "jug",
    "jukebox",
    "jumbojet",
    "jumper",
    "jumper",
    "jumper",
    "jumper",
    "jumper_cable",
    "jump_seat",
    "jump_suit",
    "jump_suit",
    "junction",
    "junction",
    "junction_barrier",
    "junk_shop",
    "jury_box",
    "jury_mast",
    "kachina",
    "kaffiyeh",
    "kalansuwa",
    "Kalashnikov",
    "kameez",
    "kanzu",
    "katharometer",
    "kayak",
    "kazoo",
    "keel",
    "keelboat",
    "keelson",
    "keep",
    "keg",
    "kennel",
    "kepi",
    "keratoscope",
    "kerchief",
    "ketch",
    "kettle",
    "kettle",
    "key",
    "key",
    "keyboard",
    "keyboard_buffer",
    "keyboard_instrument",
    "keyhole",
    "keyhole_saw",
    "khadi",
    "khaki",
    "khakis",
    "khimar",
    "khukuri",
    "kick_pleat",
    "kicksorter",
    "kickstand",
    "kick_starter",
    "kid_glove",
    "kiln",
    "kilt",
    "kimono",
    "kinescope",
    "Kinetoscope",
    "king",
    "king",
    "kingbolt",
    "king_post",
    "Kipp's_apparatus",
    "kirk",
    "kirpan",
    "kirtle",
    "kirtle",
    "kit",
    "kit",
    "kitbag",
    "kitchen",
    "kitchen_appliance",
    "kitchenette",
    "kitchen_table",
    "kitchen_utensil",
    "kitchenware",
    "kite_balloon",
    "klaxon",
    "klieg_light",
    "klystron",
    "knee_brace",
    "knee-high",
    "knee_pad",
    "knee_piece",
    "knife",
    "knife",
    "knife_blade",
    "knight",
    "knit",
    "knitting_machine",
    "knitting_needle",
    "knitwear",
    "knob",
    "knob",
    "knobble",
    "knobkerrie",
    "knocker",
    "knot",
    "knuckle_joint",
    "kohl",
    "koto",
    "kraal",
    "kremlin",
    "kris",
    "krummhorn",
    "Kundt's_tube",
    "Kurdistan",
    "kurta",
    "kylix",
    "kymograph",
    "lab_bench",
    "lab_coat",
    "lace",
    "lacquer",
    "lacquerware",
    "lacrosse_ball",
    "ladder-back",
    "ladder-back",
    "ladder_truck",
    "ladies'_room",
    "ladle",
    "lady_chapel",
    "lagerphone",
    "lag_screw",
    "lake_dwelling",
    "lally",
    "lamasery",
    "lambrequin",
    "lame",
    "laminar_flow_clean_room",
    "laminate",
    "lamination",
    "lamp",
    "lamp",
    "lamp_house",
    "lamppost",
    "lampshade",
    "lanai",
    "lancet_arch",
    "lancet_window",
    "landau",
    "lander",
    "landing_craft",
    "landing_flap",
    "landing_gear",
    "landing_net",
    "landing_skid",
    "land_line",
    "land_mine",
    "land_office",
    "lanolin",
    "lantern",
    "lanyard",
    "lap",
    "laparoscope",
    "lapboard",
    "lapel",
    "lap_joint",
    "laptop",
    "laryngoscope",
    "laser",
    "laser-guided_bomb",
    "laser_printer",
    "lash",
    "lashing",
    "lasso",
    "latch",
    "latch",
    "latchet",
    "latchkey",
    "lateen",
    "latex_paint",
    "lath",
    "lathe",
    "latrine",
    "lattice",
    "launch",
    "launcher",
    "laundry",
    "laundry_cart",
    "laundry_truck",
    "lavalava",
    "lavaliere",
    "laver",
    "lawn_chair",
    "lawn_furniture",
    "lawn_mower",
    "layette",
    "lead-acid_battery",
    "lead-in",
    "leading_rein",
    "lead_pencil",
    "leaf_spring",
    "lean-to",
    "lean-to_tent",
    "leash",
    "leatherette",
    "leather_strip",
    "Leclanche_cell",
    "lectern",
    "lecture_room",
    "lederhosen",
    "ledger_board",
    "leg",
    "leg",
    "legging",
    "Leiden_jar",
    "leisure_wear",
    "lens",
    "lens",
    "lens_cap",
    "lens_implant",
    "leotard",
    "letter_case",
    "letter_opener",
    "levee",
    "level",
    "lever",
    "lever",
    "lever",
    "lever_lock",
    "Levi's",
    "Liberty_ship",
    "library",
    "library",
    "lid",
    "Liebig_condenser",
    "lie_detector",
    "lifeboat",
    "life_buoy",
    "life_jacket",
    "life_office",
    "life_preserver",
    "life-support_system",
    "life-support_system",
    "lifting_device",
    "lift_pump",
    "ligament",
    "ligature",
    "light",
    "light_arm",
    "light_bulb",
    "light_circuit",
    "light-emitting_diode",
    "lighter",
    "lighter-than-air_craft",
    "light_filter",
    "lighting",
    "light_machine_gun",
    "light_meter",
    "light_microscope",
    "lightning_rod",
    "light_pen",
    "lightship",
    "Lilo",
    "limber",
    "limekiln",
    "limiter",
    "limousine",
    "linear_accelerator",
    "linen",
    "line_printer",
    "liner",
    "liner",
    "lingerie",
    "lining",
    "link",
    "linkage",
    "Link_trainer",
    "linocut",
    "linoleum_knife",
    "Linotype",
    "linsey-woolsey",
    "linstock",
    "lion-jaw_forceps",
    "lip-gloss",
    "lipstick",
    "liqueur_glass",
    "liquid_crystal_display",
    "liquid_metal_reactor",
    "lisle",
    "lister",
    "litterbin",
    "little_theater",
    "live_axle",
    "living_quarters",
    "living_room",
    "load",
    "Loafer",
    "loaner",
    "lobe",
    "lobster_pot",
    "local",
    "local_area_network",
    "local_oscillator",
    "Lochaber_ax",
    "lock",
    "lock",
    "lock",
    "lock",
    "lockage",
    "locker",
    "locker_room",
    "locket",
    "lock-gate",
    "locking_pliers",
    "lockring",
    "lockstitch",
    "lockup",
    "locomotive",
    "lodge",
    "lodge",
    "lodge",
    "lodging_house",
    "loft",
    "loft",
    "loft",
    "log_cabin",
    "loggia",
    "longbow",
    "long_iron",
    "long_johns",
    "long_sleeve",
    "long_tom",
    "long_trousers",
    "long_underwear",
    "looking_glass",
    "lookout",
    "loom",
    "loop_knot",
    "lorgnette",
    "Lorraine_cross",
    "lorry",
    "lota",
    "lotion",
    "loudspeaker",
    "lounge",
    "lounger",
    "lounging_jacket",
    "lounging_pajama",
    "loungewear",
    "loupe",
    "louvered_window",
    "love_knot",
    "love_seat",
    "loving_cup",
    "lowboy",
    "low-pass_filter",
    "low-warp-loom",
    "LP",
    "L-plate",
    "lubber's_hole",
    "lubricating_system",
    "luff",
    "lug",
    "luge",
    "Luger",
    "luggage_carrier",
    "luggage_compartment",
    "luggage_rack",
    "lugger",
    "lugsail",
    "lug_wrench",
    "lumberjack",
    "lumbermill",
    "lunar_excursion_module",
    "lunchroom",
    "lunette",
    "lungi",
    "lunula",
    "lusterware",
    "lute",
    "luxury_liner",
    "lyceum",
    "lychgate",
    "lyre",
    "machete",
    "machicolation",
    "machine",
    "machine",
    "machine_bolt",
    "machine_gun",
    "machinery",
    "machine_screw",
    "machine_tool",
    "machinist's_vise",
    "machmeter",
    "mackinaw",
    "mackinaw",
    "mackinaw",
    "mackintosh",
    "macrame",
    "madras",
    "Mae_West",
    "magazine_rack",
    "magic_lantern",
    "magnet",
    "magnetic_bottle",
    "magnetic_compass",
    "magnetic_core_memory",
    "magnetic_disk",
    "magnetic_head",
    "magnetic_mine",
    "magnetic_needle",
    "magnetic_recorder",
    "magnetic_stripe",
    "magnetic_tape",
    "magneto",
    "magnetometer",
    "magnetron",
    "magnifier",
    "magnum",
    "magnus_hitch",
    "mail",
    "mailbag",
    "mailbag",
    "mailboat",
    "mailbox",
    "mail_car",
    "maildrop",
    "mailer",
    "maillot",
    "maillot",
    "mailsorter",
    "mail_train",
    "mainframe",
    "mainmast",
    "main_rotor",
    "mainsail",
    "mainspring",
    "main-topmast",
    "main-topsail",
    "main_yard",
    "maisonette",
    "majolica",
    "makeup",
    "Maksutov_telescope",
    "malacca",
    "mallet",
    "mallet",
    "mallet",
    "mammogram",
    "mandola",
    "mandolin",
    "manger",
    "mangle",
    "manhole",
    "manhole_cover",
    "man-of-war",
    "manometer",
    "manor",
    "manor_hall",
    "MANPAD",
    "mansard",
    "manse",
    "mansion",
    "mantel",
    "mantelet",
    "mantilla",
    "Mao_jacket",
    "map",
    "maquiladora",
    "maraca",
    "marble",
    "marching_order",
    "marimba",
    "marina",
    "marker",
    "marketplace",
    "marlinespike",
    "marocain",
    "marquee",
    "marquetry",
    "marriage_bed",
    "martello_tower",
    "martingale",
    "mascara",
    "maser",
    "masher",
    "mashie",
    "mashie_niblick",
    "masjid",
    "mask",
    "mask",
    "Masonite",
    "Mason_jar",
    "masonry",
    "mason's_level",
    "massage_parlor",
    "massage_parlor",
    "mass_spectrograph",
    "mass_spectrometer",
    "mast",
    "mast",
    "mastaba",
    "master_bedroom",
    "masterpiece",
    "mat",
    "mat",
    "match",
    "match",
    "matchboard",
    "matchbook",
    "matchbox",
    "matchlock",
    "match_plane",
    "matchstick",
    "material",
    "materiel",
    "maternity_hospital",
    "maternity_ward",
    "matrix",
    "Matthew_Walker",
    "matting",
    "mattock",
    "mattress_cover",
    "maul",
    "maulstick",
    "Mauser",
    "mausoleum",
    "maxi",
    "Maxim_gun",
    "maximum_and_minimum_thermometer",
    "maypole",
    "maze",
    "mazer",
    "means",
    "measure",
    "measuring_cup",
    "measuring_instrument",
    "measuring_stick",
    "meat_counter",
    "meat_grinder",
    "meat_hook",
    "meat_house",
    "meat_safe",
    "meat_thermometer",
    "mechanical_device",
    "mechanical_piano",
    "mechanical_system",
    "mechanism",
    "medical_building",
    "medical_instrument",
    "medicine_ball",
    "medicine_chest",
    "MEDLINE",
    "megalith",
    "megaphone",
    "memorial",
    "memory",
    "memory_chip",
    "memory_device",
    "menagerie",
    "mending",
    "menhir",
    "menorah",
    "Menorah",
    "man's_clothing",
    "men's_room",
    "mercantile_establishment",
    "mercury_barometer",
    "mercury_cell",
    "mercury_thermometer",
    "mercury-vapor_lamp",
    "mercy_seat",
    "merlon",
    "mess",
    "mess_jacket",
    "mess_kit",
    "messuage",
    "metal_detector",
    "metallic",
    "metal_screw",
    "metal_wood",
    "meteorological_balloon",
    "meter",
    "meterstick",
    "metronome",
    "mezzanine",
    "mezzanine",
    "microbalance",
    "microbrewery",
    "microfiche",
    "microfilm",
    "micrometer",
    "microphone",
    "microprocessor",
    "microscope",
    "microtome",
    "microwave",
    "microwave_diathermy_machine",
    "microwave_linear_accelerator",
    "middy",
    "midiron",
    "mihrab",
    "mihrab",
    "military_hospital",
    "military_quarters",
    "military_uniform",
    "military_vehicle",
    "milk_bar",
    "milk_can",
    "milk_float",
    "milking_machine",
    "milking_stool",
    "milk_wagon",
    "mill",
    "milldam",
    "miller",
    "milliammeter",
    "millinery",
    "millinery",
    "milling",
    "millivoltmeter",
    "millstone",
    "millstone",
    "millwheel",
    "mimeograph",
    "minaret",
    "mincer",
    "mine",
    "mine_detector",
    "minelayer",
    "mineshaft",
    "minibar",
    "minibike",
    "minibus",
    "minicar",
    "minicomputer",
    "ministry",
    "miniskirt",
    "minisub",
    "minivan",
    "miniver",
    "mink",
    "minster",
    "mint",
    "minute_hand",
    "Minuteman",
    "mirror",
    "missile",
    "missile_defense_system",
    "miter_box",
    "miter_joint",
    "mitten",
    "mixer",
    "mixer",
    "mixing_bowl",
    "mixing_faucet",
    "mizzen",
    "mizzenmast",
    "mobcap",
    "mobile_home",
    "moccasin",
    "mock-up",
    "mod_con",
    "Model_T",
    "modem",
    "modillion",
    "module",
    "module",
    "mohair",
    "moire",
    "mold",
    "moldboard",
    "moldboard_plow",
    "moleskin",
    "Molotov_cocktail",
    "monastery",
    "monastic_habit",
    "moneybag",
    "money_belt",
    "monitor",
    "monitor",
    "monitor",
    "monkey-wrench",
    "monk's_cloth",
    "monochrome",
    "monocle",
    "monofocal_lens_implant",
    "monoplane",
    "monotype",
    "monstrance",
    "mooring_tower",
    "Moorish_arch",
    "moped",
    "mop_handle",
    "moquette",
    "morgue",
    "morion",
    "morning_dress",
    "morning_dress",
    "morning_room",
    "Morris_chair",
    "mortar",
    "mortar",
    "mortarboard",
    "mortise_joint",
    "mosaic",
    "mosque",
    "mosquito_net",
    "motel",
    "motel_room",
    "Mother_Hubbard",
    "motion-picture_camera",
    "motion-picture_film",
    "motley",
    "motley",
    "motor",
    "motorboat",
    "motorcycle",
    "motor_hotel",
    "motorized_wheelchair",
    "motor_scooter",
    "motor_vehicle",
    "mound",
    "mound",
    "mount",
    "mountain_bike",
    "mountain_tent",
    "mouse",
    "mouse_button",
    "mousetrap",
    "mousse",
    "mouthpiece",
    "mouthpiece",
    "mouthpiece",
    "movement",
    "movie_projector",
    "moving-coil_galvanometer",
    "moving_van",
    "mud_brick",
    "mudguard",
    "mudhif",
    "muff",
    "muffle",
    "muffler",
    "mufti",
    "mug",
    "mulch",
    "mule",
    "multichannel_recorder",
    "multiengine_airplane",
    "multiplex",
    "multiplexer",
    "multiprocessor",
    "multistage_rocket",
    "munition",
    "Murphy_bed",
    "musette",
    "musette_pipe",
    "museum",
    "mushroom_anchor",
    "musical_instrument",
    "music_box",
    "music_hall",
    "music_school",
    "music_stand",
    "music_stool",
    "musket",
    "musket_ball",
    "muslin",
    "mustache_cup",
    "mustard_plaster",
    "mute",
    "muzzle_loader",
    "muzzle",
    "myelogram",
    "nacelle",
    "nail",
    "nailbrush",
    "nailfile",
    "nailhead",
    "nailhead",
    "nail_polish",
    "nainsook",
    "Napier's_bones",
    "nard",
    "narrowbody_aircraft",
    "narrow_wale",
    "narthex",
    "narthex",
    "nasotracheal_tube",
    "national_monument",
    "nautilus",
    "navigational_system",
    "naval_equipment",
    "naval_gun",
    "naval_missile",
    "naval_radar",
    "naval_tactical_data_system",
    "naval_weaponry",
    "nave",
    "navigational_instrument",
    "nebuchadnezzar",
    "neckband",
    "neck_brace",
    "neckcloth",
    "neckerchief",
    "necklace",
    "necklet",
    "neckline",
    "neckpiece",
    "necktie",
    "neckwear",
    "needle",
    "needle",
    "needlenose_pliers",
    "needlework",
    "negative",
    "negative_magnetic_pole",
    "negative_pole",
    "negligee",
    "neolith",
    "neon_lamp",
    "nephoscope",
    "nest",
    "nest_egg",
    "net",
    "net",
    "net",
    "net",
    "network",
    "network",
    "neutron_bomb",
    "newel",
    "newel_post",
    "newspaper",
    "newsroom",
    "newsroom",
    "newsstand",
    "Newtonian_telescope",
    "nib",
    "niblick",
    "nicad",
    "nickel-iron_battery",
    "Nicol_prism",
    "night_bell",
    "nightcap",
    "nightgown",
    "night_latch",
    "night-light",
    "nightshirt",
    "nightwear",
    "ninepin",
    "ninepin_ball",
    "ninon",
    "nipple",
    "nipple_shield",
    "niqab",
    "Nissen_hut",
    "nogging",
    "noisemaker",
    "nonsmoker",
    "non-volatile_storage",
    "Norfolk_jacket",
    "noria",
    "nosebag",
    "noseband",
    "nose_flute",
    "nosewheel",
    "notebook",
    "nuclear-powered_ship",
    "nuclear_reactor",
    "nuclear_rocket",
    "nuclear_weapon",
    "nude",
    "numdah",
    "nun's_habit",
    "nursery",
    "nut_and_bolt",
    "nutcracker",
    "nylon",
    "nylons",
    "oar",
    "oast",
    "oast_house",
    "obelisk",
    "object_ball",
    "objective",
    "oblique_bandage",
    "oboe",
    "oboe_da_caccia",
    "oboe_d'amore",
    "observation_dome",
    "observatory",
    "obstacle",
    "obturator",
    "ocarina",
    "octant",
    "odd-leg_caliper",
    "odometer",
    "oeil_de_boeuf",
    "office",
    "office_building",
    "office_furniture",
    "officer's_mess",
    "off-line_equipment",
    "ogee",
    "ogee_arch",
    "ohmmeter",
    "oil",
    "oilcan",
    "oilcloth",
    "oil_filter",
    "oil_heater",
    "oil_lamp",
    "oil_paint",
    "oil_pump",
    "oil_refinery",
    "oilskin",
    "oil_slick",
    "oilstone",
    "oil_tanker",
    "old_school_tie",
    "olive_drab",
    "olive_drab",
    "Olympian_Zeus",
    "omelet_pan",
    "omnidirectional_antenna",
    "omnirange",
    "onion_dome",
    "open-air_market",
    "open_circuit",
    "open-end_wrench",
    "opener",
    "open-hearth_furnace",
    "openside_plane",
    "open_sight",
    "openwork",
    "opera",
    "opera_cloak",
    "operating_microscope",
    "operating_room",
    "operating_table",
    "ophthalmoscope",
    "optical_device",
    "optical_disk",
    "optical_instrument",
    "optical_pyrometer",
    "optical_telescope",
    "orchestra_pit",
    "ordinary",
    "organ",
    "organdy",
    "organic_light-emitting_diode",
    "organ_loft",
    "organ_pipe",
    "organza",
    "oriel",
    "oriflamme",
    "O_ring",
    "Orlon",
    "orlop_deck",
    "orphanage",
    "orphrey",
    "orrery",
    "orthicon",
    "orthochromatic_film",
    "orthopter",
    "orthoscope",
    "oscillograph",
    "oscilloscope",
    "ossuary",
    "otoscope",
    "ottoman",
    "oubliette",
    "out-basket",
    "outboard_motor",
    "outboard_motorboat",
    "outbuilding",
    "outerwear",
    "outfall",
    "outfit",
    "outfitter",
    "outhouse",
    "output_device",
    "outrigger",
    "outrigger_canoe",
    "outside_caliper",
    "outside_mirror",
    "outwork",
    "oven",
    "oven_thermometer",
    "overall",
    "overall",
    "overcoat",
    "overdrive",
    "overgarment",
    "overhand_knot",
    "overhang",
    "overhead_projector",
    "overmantel",
    "overnighter",
    "overpass",
    "override",
    "overshoe",
    "overskirt",
    "oxbow",
    "Oxbridge",
    "oxcart",
    "oxeye",
    "oxford",
    "oximeter",
    "oxyacetylene_torch",
    "oxygen_mask",
    "oyster_bar",
    "oyster_bed",
    "pace_car",
    "pacemaker",
    "pack",
    "pack",
    "pack",
    "package",
    "package_store",
    "packaging",
    "packet",
    "packing_box",
    "packinghouse",
    "packinghouse",
    "packing_needle",
    "packsaddle",
    "paddle",
    "paddle",
    "paddle",
    "paddle_box",
    "paddle_steamer",
    "paddlewheel",
    "paddock",
    "padlock",
    "page_printer",
    "paint",
    "paintball",
    "paintball_gun",
    "paintbox",
    "paintbrush",
    "paisley",
    "pajama",
    "pajama",
    "palace",
    "palace",
    "palace",
    "palanquin",
    "paleolith",
    "palestra",
    "palette",
    "palette_knife",
    "palisade",
    "pallet",
    "pallette",
    "pallium",
    "pallium",
    "pan",
    "pan",
    "pancake_turner",
    "panchromatic_film",
    "panda_car",
    "paneling",
    "panhandle",
    "panic_button",
    "pannier",
    "pannier",
    "pannikin",
    "panopticon",
    "panopticon",
    "panpipe",
    "pantaloon",
    "pantechnicon",
    "pantheon",
    "pantheon",
    "pantie",
    "panting",
    "pant_leg",
    "pantograph",
    "pantry",
    "pants_suit",
    "panty_girdle",
    "pantyhose",
    "panzer",
    "paper_chain",
    "paper_clip",
    "paper_cutter",
    "paper_fastener",
    "paper_feed",
    "paper_mill",
    "paper_towel",
    "parabolic_mirror",
    "parabolic_reflector",
    "parachute",
    "parallel_bars",
    "parallel_circuit",
    "parallel_interface",
    "parang",
    "parapet",
    "parapet",
    "parasail",
    "parasol",
    "parer",
    "parfait_glass",
    "pargeting",
    "pari-mutuel_machine",
    "parka",
    "park_bench",
    "parking_meter",
    "parlor",
    "parquet",
    "parquetry",
    "parsonage",
    "Parsons_table",
    "partial_denture",
    "particle_detector",
    "partition",
    "parts_bin",
    "party_line",
    "party_wall",
    "parvis",
    "passenger_car",
    "passenger_ship",
    "passenger_train",
    "passenger_van",
    "passe-partout",
    "passive_matrix_display",
    "passkey",
    "pass-through",
    "pastry_cart",
    "patch",
    "patchcord",
    "patchouli",
    "patch_pocket",
    "patchwork",
    "patent_log",
    "paternoster",
    "patina",
    "patio",
    "patisserie",
    "patka",
    "patrol_boat",
    "patty-pan",
    "pave",
    "pavilion",
    "pavior",
    "pavis",
    "pawn",
    "pawnbroker's_shop",
    "pay-phone",
    "PC_board",
    "peach_orchard",
    "pea_jacket",
    "peavey",
    "pectoral",
    "pedal",
    "pedal_pusher",
    "pedestal",
    "pedestal_table",
    "pedestrian_crossing",
    "pedicab",
    "pediment",
    "pedometer",
    "peeler",
    "peep_sight",
    "peg",
    "peg",
    "peg",
    "peg",
    "pegboard",
    "Pelham",
    "pelican_crossing",
    "pelisse",
    "pelvimeter",
    "pen",
    "penal_colony",
    "penal_institution",
    "penalty_box",
    "pen-and-ink",
    "pencil",
    "pencil",
    "pencil_box",
    "pencil_sharpener",
    "pendant_earring",
    "pendulum",
    "pendulum_clock",
    "pendulum_watch",
    "penetration_bomb",
    "penile_implant",
    "penitentiary",
    "penknife",
    "penlight",
    "pennant",
    "pennywhistle",
    "penthouse",
    "pentode",
    "peplos",
    "peplum",
    "pepper_mill",
    "pepper_shaker",
    "pepper_spray",
    "percale",
    "percolator",
    "percussion_cap",
    "percussion_instrument",
    "perforation",
    "perfume",
    "perfumery",
    "perfumery",
    "perfumery",
    "peripheral",
    "periscope",
    "peristyle",
    "periwig",
    "permanent_press",
    "perpetual_motion_machine",
    "personal_computer",
    "personal_digital_assistant",
    "personnel_carrier",
    "pestle",
    "pestle",
    "petcock",
    "Petri_dish",
    "petrolatum_gauze",
    "pet_shop",
    "petticoat",
    "pew",
    "phial",
    "Phillips_screw",
    "Phillips_screwdriver",
    "phonograph_needle",
    "phonograph_record",
    "photocathode",
    "photocoagulator",
    "photocopier",
    "photographic_equipment",
    "photographic_paper",
    "photometer",
    "photomicrograph",
    "Photostat",
    "photostat",
    "physical_pendulum",
    "piano",
    "piano_action",
    "piano_keyboard",
    "piano_wire",
    "piccolo",
    "pick",
    "pick",
    "pick",
    "pickelhaube",
    "picket_boat",
    "picket_fence",
    "picket_ship",
    "pickle_barrel",
    "pickup",
    "picture",
    "picture_frame",
    "picture_hat",
    "picture_rail",
    "picture_window",
    "piece_of_cloth",
    "pied-a-terre",
    "pier",
    "pier",
    "pier_arch",
    "pier_glass",
    "pier_table",
    "pieta",
    "piezometer",
    "pig_bed",
    "piggery",
    "piggy_bank",
    "pilaster",
    "pile",
    "pile_driver",
    "pill_bottle",
    "pillbox",
    "pillion",
    "pillory",
    "pillow",
    "pillow_block",
    "pillow_lace",
    "pillow_sham",
    "pilot_bit",
    "pilot_boat",
    "pilot_burner",
    "pilot_cloth",
    "pilot_engine",
    "pilothouse",
    "pilot_light",
    "pin",
    "pin",
    "pin",
    "pinata",
    "pinball_machine",
    "pince-nez",
    "pincer",
    "pinch_bar",
    "pincurl_clip",
    "pinfold",
    "ping-pong_ball",
    "pinhead",
    "pinion",
    "pinnacle",
    "pinprick",
    "pinstripe",
    "pinstripe",
    "pinstripe",
    "pintle",
    "pinwheel",
    "pinwheel",
    "tabor_pipe",
    "pipe",
    "pipe_bomb",
    "pipe_cleaner",
    "pipe_cutter",
    "pipefitting",
    "pipet",
    "pipe_vise",
    "pipe_wrench",
    "pique",
    "pirate",
    "piste",
    "pistol",
    "pistol_grip",
    "piston",
    "piston_ring",
    "piston_rod",
    "pit",
    "pitcher",
    "pitchfork",
    "pitching_wedge",
    "pitch_pipe",
    "pith_hat",
    "piton",
    "Pitot-static_tube",
    "Pitot_tube",
    "pitsaw",
    "pivot",
    "pivoting_window",
    "pizzeria",
    "place_of_business",
    "place_of_worship",
    "placket",
    "planchet",
    "plane",
    "plane",
    "plane_seat",
    "planetarium",
    "planetarium",
    "planetarium",
    "planetary_gear",
    "plank-bed",
    "planking",
    "planner",
    "plant",
    "planter",
    "plaster",
    "plasterboard",
    "plastering_trowel",
    "plastic_bag",
    "plastic_bomb",
    "plastic_laminate",
    "plastic_wrap",
    "plastron",
    "plastron",
    "plastron",
    "plate",
    "plate",
    "plate",
    "platen",
    "platen",
    "plate_rack",
    "plate_rail",
    "platform",
    "platform",
    "platform",
    "platform_bed",
    "platform_rocker",
    "plating",
    "platter",
    "playback",
    "playbox",
    "playground",
    "playpen",
    "playsuit",
    "plaza",
    "pleat",
    "plenum",
    "plethysmograph",
    "pleximeter",
    "plexor",
    "pliers",
    "plimsoll",
    "plotter",
    "plow",
    "plug",
    "plug",
    "plug_fuse",
    "plughole",
    "plumb_bob",
    "plumb_level",
    "plunger",
    "plus_fours",
    "plush",
    "plywood",
    "pneumatic_drill",
    "p-n_junction",
    "p-n-p_transistor",
    "poacher",
    "pocket",
    "pocket_battleship",
    "pocketcomb",
    "pocket_flap",
    "pocket-handkerchief",
    "pocketknife",
    "pocket_watch",
    "pod",
    "pogo_stick",
    "point-and-shoot_camera",
    "pointed_arch",
    "pointing_trowel",
    "point_lace",
    "poker",
    "polarimeter",
    "Polaroid",
    "Polaroid_camera",
    "pole",
    "pole",
    "poleax",
    "poleax",
    "police_boat",
    "police_van",
    "polling_booth",
    "polo_ball",
    "polo_mallet",
    "polonaise",
    "polo_shirt",
    "polyester",
    "polygraph",
    "pomade",
    "pommel_horse",
    "poncho",
    "pongee",
    "poniard",
    "pontifical",
    "pontoon",
    "pontoon_bridge",
    "pony_cart",
    "pool_ball",
    "poolroom",
    "pool_table",
    "poop_deck",
    "poor_box",
    "poorhouse",
    "pop_bottle",
    "popgun",
    "poplin",
    "popper",
    "poppet",
    "pop_tent",
    "porcelain",
    "porch",
    "porkpie",
    "porringer",
    "portable",
    "portable_computer",
    "portable_circular_saw",
    "portcullis",
    "porte-cochere",
    "porte-cochere",
    "portfolio",
    "porthole",
    "portico",
    "portiere",
    "portmanteau",
    "portrait_camera",
    "portrait_lens",
    "positive_pole",
    "positive_pole",
    "positron_emission_tomography_scanner",
    "post",
    "postage_meter",
    "post_and_lintel",
    "post_chaise",
    "postern",
    "post_exchange",
    "posthole_digger",
    "post_horn",
    "posthouse",
    "pot",
    "pot",
    "potbelly",
    "Potemkin_village",
    "potential_divider",
    "potentiometer",
    "potentiometer",
    "potpourri",
    "potsherd",
    "potter's_wheel",
    "pottery",
    "pottle",
    "potty_seat",
    "pouch",
    "poultice",
    "pound",
    "pound_net",
    "powder",
    "powder_and_shot",
    "powdered_mustard",
    "powder_horn",
    "powder_keg",
    "power_brake",
    "power_cord",
    "power_drill",
    "power_line",
    "power_loom",
    "power_mower",
    "power_pack",
    "power_saw",
    "power_shovel",
    "power_steering",
    "power_takeoff",
    "power_tool",
    "praetorium",
    "prayer_rug",
    "prayer_shawl",
    "precipitator",
    "prefab",
    "presbytery",
    "presence_chamber",
    "press",
    "press",
    "press",
    "press_box",
    "press_gallery",
    "press_of_sail",
    "pressure_cabin",
    "pressure_cooker",
    "pressure_dome",
    "pressure_gauge",
    "pressurized_water_reactor",
    "pressure_suit",
    "pricket",
    "prie-dieu",
    "primary_coil",
    "Primus_stove",
    "Prince_Albert",
    "print",
    "print_buffer",
    "printed_circuit",
    "printer",
    "printer",
    "printer_cable",
    "priory",
    "prison",
    "prison_camp",
    "privateer",
    "private_line",
    "privet_hedge",
    "probe",
    "proctoscope",
    "prod",
    "production_line",
    "projectile",
    "projector",
    "projector",
    "prolonge",
    "prolonge_knot",
    "prompter",
    "prong",
    "propeller",
    "propeller_plane",
    "propjet",
    "proportional_counter_tube",
    "propulsion_system",
    "proscenium",
    "proscenium_arch",
    "prosthesis",
    "protective_covering",
    "protective_garment",
    "proton_accelerator",
    "protractor",
    "pruner",
    "pruning_knife",
    "pruning_saw",
    "pruning_shears",
    "psaltery",
    "psychrometer",
    "PT_boat",
    "public_address_system",
    "public_house",
    "public_toilet",
    "public_transport",
    "public_works",
    "puck",
    "pull",
    "pullback",
    "pull_chain",
    "pulley",
    "pull-off",
    "Pullman",
    "pullover",
    "pull-through",
    "pulse_counter",
    "pulse_generator",
    "pulse_timing_circuit",
    "pump",
    "pump",
    "pump_action",
    "pump_house",
    "pump_room",
    "pump-type_pliers",
    "pump_well",
    "punch",
    "punchboard",
    "punch_bowl",
    "punching_bag",
    "punch_pliers",
    "punch_press",
    "punnet",
    "punt",
    "pup_tent",
    "purdah",
    "purifier",
    "purl",
    "purse",
    "push-bike",
    "push_broom",
    "push_button",
    "push-button_radio",
    "pusher",
    "put-put",
    "puttee",
    "putter",
    "putty_knife",
    "puzzle",
    "pylon",
    "pylon",
    "pyramidal_tent",
    "pyrograph",
    "pyrometer",
    "pyrometric_cone",
    "pyrostat",
    "pyx",
    "pyx",
    "pyxis",
    "quad",
    "quadrant",
    "quadraphony",
    "quartering",
    "quarterstaff",
    "quartz_battery",
    "quartz_lamp",
    "queen",
    "queen",
    "queen_post",
    "quern",
    "quill",
    "quilt",
    "quilted_bedspread",
    "quilting",
    "quipu",
    "quirk_molding",
    "quirt",
    "quiver",
    "quoin",
    "quoit",
    "QWERTY_keyboard",
    "rabbet",
    "rabbet_joint",
    "rabbit_ears",
    "rabbit_hutch",
    "raceabout",
    "racer",
    "raceway",
    "racing_boat",
    "racing_gig",
    "racing_skiff",
    "rack",
    "rack",
    "rack",
    "rack_and_pinion",
    "racket",
    "racquetball",
    "radar",
    "radial",
    "radial_engine",
    "radiation_pyrometer",
    "radiator",
    "radiator",
    "radiator_cap",
    "radiator_hose",
    "radio",
    "radio_antenna",
    "radio_chassis",
    "radio_compass",
    "radiogram",
    "radio_interferometer",
    "radio_link",
    "radiometer",
    "radiomicrometer",
    "radio-phonograph",
    "radio_receiver",
    "radiotelegraph",
    "radiotelephone",
    "radio_telescope",
    "radiotherapy_equipment",
    "radio_transmitter",
    "radome",
    "raft",
    "rafter",
    "raft_foundation",
    "rag",
    "ragbag",
    "raglan",
    "raglan_sleeve",
    "rail",
    "rail_fence",
    "railhead",
    "railing",
    "railing",
    "railroad_bed",
    "railroad_tunnel",
    "rain_barrel",
    "raincoat",
    "rain_gauge",
    "rain_stick",
    "rake",
    "rake_handle",
    "RAM_disk",
    "ramekin",
    "ramjet",
    "rammer",
    "ramp",
    "rampant_arch",
    "rampart",
    "ramrod",
    "ramrod",
    "ranch",
    "ranch_house",
    "random-access_memory",
    "rangefinder",
    "range_hood",
    "range_pole",
    "rapier",
    "rariora",
    "rasp",
    "ratchet",
    "ratchet_wheel",
    "rathskeller",
    "ratline",
    "rat-tail_file",
    "rattan",
    "rattrap",
    "rayon",
    "razor",
    "razorblade",
    "reaction-propulsion_engine",
    "reaction_turbine",
    "reactor",
    "reading_lamp",
    "reading_room",
    "read-only_memory",
    "read-only_memory_chip",
    "readout",
    "read/write_head",
    "ready-to-wear",
    "real_storage",
    "reamer",
    "reamer",
    "rearview_mirror",
    "Reaumur_thermometer",
    "rebozo",
    "receiver",
    "receptacle",
    "reception_desk",
    "reception_room",
    "recess",
    "reciprocating_engine",
    "recliner",
    "reconnaissance_plane",
    "reconnaissance_vehicle",
    "record_changer",
    "recorder",
    "recording",
    "recording_system",
    "record_player",
    "record_sleeve",
    "recovery_room",
    "recreational_vehicle",
    "recreation_room",
    "recycling_bin",
    "recycling_plant",
    "redbrick_university",
    "red_carpet",
    "redoubt",
    "redoubt",
    "reduction_gear",
    "reed_pipe",
    "reed_stop",
    "reef_knot",
    "reel",
    "reel",
    "refectory",
    "refectory_table",
    "refinery",
    "reflecting_telescope",
    "reflectometer",
    "reflector",
    "reflex_camera",
    "reflux_condenser",
    "reformatory",
    "reformer",
    "refracting_telescope",
    "refractometer",
    "refrigeration_system",
    "refrigerator",
    "refrigerator_car",
    "refuge",
    "regalia",
    "regimentals",
    "regulator",
    "rein",
    "relay",
    "release",
    "religious_residence",
    "reliquary",
    "remote_control",
    "remote_terminal",
    "removable_disk",
    "rendering",
    "rep",
    "repair_shop",
    "repeater",
    "repeating_firearm",
    "repository",
    "reproducer",
    "rerebrace",
    "rescue_equipment",
    "research_center",
    "reseau",
    "reservoir",
    "reset",
    "reset_button",
    "residence",
    "resistance_pyrometer",
    "resistor",
    "resonator",
    "resonator",
    "resort_hotel",
    "respirator",
    "restaurant",
    "rest_house",
    "restraint",
    "resuscitator",
    "retainer",
    "retaining_wall",
    "reticle",
    "reticulation",
    "reticule",
    "retort",
    "retractor",
    "return_key",
    "reverberatory_furnace",
    "revers",
    "reverse",
    "reversible",
    "revetment",
    "revetment",
    "revolver",
    "revolving_door",
    "rheometer",
    "rheostat",
    "rhinoscope",
    "rib",
    "riband",
    "ribbed_vault",
    "ribbing",
    "ribbon_development",
    "rib_joint_pliers",
    "ricer",
    "riddle",
    "ride",
    "ridge",
    "ridge_rope",
    "riding_boot",
    "riding_crop",
    "riding_mower",
    "rifle",
    "rifle_ball",
    "rifle_grenade",
    "rig",
    "rigger",
    "rigger",
    "rigging",
    "rigout",
    "ringlet",
    "rings",
    "rink",
    "riot_gun",
    "ripcord",
    "ripcord",
    "ripping_bar",
    "ripping_chisel",
    "ripsaw",
    "riser",
    "riser",
    "Ritz",
    "river_boat",
    "rivet",
    "riveting_machine",
    "roach_clip",
    "road",
    "roadbed",
    "roadblock",
    "roadhouse",
    "roadster",
    "roadway",
    "roaster",
    "robe",
    "robotics_equipment",
    "Rochon_prism",
    "rock_bit",
    "rocker",
    "rocker",
    "rocker_arm",
    "rocket",
    "rocket",
    "rocking_chair",
    "rod",
    "rodeo",
    "roll",
    "roller",
    "roller",
    "roller_bandage",
    "in-line_skate",
    "Rollerblade",
    "roller_blind",
    "roller_coaster",
    "roller_skate",
    "roller_towel",
    "roll_film",
    "rolling_hitch",
    "rolling_mill",
    "rolling_pin",
    "rolling_stock",
    "roll-on",
    "roll-on",
    "roll-on_roll-off",
    "Rolodex",
    "Roman_arch",
    "Roman_building",
    "romper",
    "rood_screen",
    "roof",
    "roof",
    "roofing",
    "room",
    "roomette",
    "room_light",
    "roost",
    "rope",
    "rope_bridge",
    "rope_tow",
    "rose_water",
    "rose_window",
    "rosin_bag",
    "rotary_actuator",
    "rotary_engine",
    "rotary_press",
    "rotating_mechanism",
    "rotating_shaft",
    "rotisserie",
    "rotisserie",
    "rotor",
    "rotor",
    "rotor",
    "rotor_blade",
    "rotor_head",
    "rotunda",
    "rotunda",
    "rouge",
    "roughcast",
    "rouleau",
    "roulette",
    "roulette_ball",
    "roulette_wheel",
    "round",
    "round_arch",
    "round-bottom_flask",
    "roundel",
    "round_file",
    "roundhouse",
    "router",
    "router",
    "router_plane",
    "rowel",
    "row_house",
    "rowing_boat",
    "rowlock_arch",
    "royal",
    "royal_mast",
    "rubber_band",
    "rubber_boot",
    "rubber_bullet",
    "rubber_eraser",
    "rudder",
    "rudder",
    "rudder_blade",
    "rug",
    "rugby_ball",
    "ruin",
    "rule",
    "rumble",
    "rumble_seat",
    "rummer",
    "rumpus_room",
    "runcible_spoon",
    "rundle",
    "running_shoe",
    "running_suit",
    "runway",
    "rushlight",
    "russet",
    "rya",
    "saber",
    "saber_saw",
    "sable",
    "sable",
    "sable_coat",
    "sabot",
    "sachet",
    "sack",
    "sack",
    "sackbut",
    "sackcloth",
    "sackcloth",
    "sack_coat",
    "sacking",
    "saddle",
    "saddlebag",
    "saddle_blanket",
    "saddle_oxford",
    "saddlery",
    "saddle_seat",
    "saddle_stitch",
    "safe",
    "safe",
    "safe-deposit",
    "safe_house",
    "safety_arch",
    "safety_belt",
    "safety_bicycle",
    "safety_bolt",
    "safety_curtain",
    "safety_fuse",
    "safety_lamp",
    "safety_match",
    "safety_net",
    "safety_pin",
    "safety_rail",
    "safety_razor",
    "safety_valve",
    "sail",
    "sail",
    "sailboat",
    "sailcloth",
    "sailing_vessel",
    "sailing_warship",
    "sailor_cap",
    "sailor_suit",
    "salad_bar",
    "salad_bowl",
    "salinometer",
    "sallet",
    "salon",
    "salon",
    "salon",
    "saltbox",
    "saltcellar",
    "saltshaker",
    "saltworks",
    "salver",
    "salwar",
    "Sam_Browne_belt",
    "samisen",
    "samite",
    "samovar",
    "sampan",
    "sandal",
    "sandbag",
    "sandblaster",
    "sandbox",
    "sandglass",
    "sand_wedge",
    "sandwich_board",
    "sanitary_napkin",
    "cling_film",
    "sarcenet",
    "sarcophagus",
    "sari",
    "sarong",
    "sash",
    "sash_fastener",
    "sash_window",
    "satchel",
    "sateen",
    "satellite",
    "satellite_receiver",
    "satellite_television",
    "satellite_transmitter",
    "satin",
    "Saturday_night_special",
    "saucepan",
    "saucepot",
    "sauna",
    "savings_bank",
    "saw",
    "sawed-off_shotgun",
    "sawhorse",
    "sawmill",
    "saw_set",
    "sax",
    "saxhorn",
    "scabbard",
    "scaffolding",
    "scale",
    "scale",
    "scaler",
    "scaling_ladder",
    "scalpel",
    "scanner",
    "scanner",
    "scanner",
    "scantling",
    "scarf",
    "scarf_joint",
    "scatter_rug",
    "scauper",
    "Schmidt_telescope",
    "school",
    "schoolbag",
    "school_bell",
    "school_bus",
    "school_ship",
    "school_system",
    "schooner",
    "schooner",
    "scientific_instrument",
    "scimitar",
    "scintillation_counter",
    "scissors",
    "sclerometer",
    "scoinson_arch",
    "sconce",
    "sconce",
    "scoop",
    "scooter",
    "scoreboard",
    "scouring_pad",
    "scow",
    "scow",
    "scraper",
    "scratcher",
    "screen",
    "screen",
    "screen",
    "screen",
    "screen_door",
    "screening",
    "screw",
    "screw",
    "screw",
    "screwdriver",
    "screw_eye",
    "screw_key",
    "screw_thread",
    "screwtop",
    "screw_wrench",
    "scriber",
    "scrim",
    "scrimshaw",
    "scriptorium",
    "scrubber",
    "scrub_brush",
    "scrub_plane",
    "scuffer",
    "scuffle",
    "scull",
    "scull",
    "scullery",
    "sculpture",
    "scuttle",
    "scyphus",
    "scythe",
    "seabag",
    "sea_boat",
    "sea_chest",
    "sealing_wax",
    "sealskin",
    "seam",
    "seaplane",
    "searchlight",
    "searing_iron",
    "seat",
    "seat",
    "seat",
    "seat_belt",
    "secateurs",
    "secondary_coil",
    "second_balcony",
    "second_base",
    "second_hand",
    "secretary",
    "sectional",
    "security_blanket",
    "security_system",
    "security_system",
    "sedan",
    "sedan",
    "seeder",
    "seeker",
    "seersucker",
    "segmental_arch",
    "Segway",
    "seidel",
    "seine",
    "seismograph",
    "selector",
    "selenium_cell",
    "self-propelled_vehicle",
    "self-registering_thermometer",
    "self-starter",
    "selsyn",
    "selvage",
    "semaphore",
    "semiautomatic_firearm",
    "semiautomatic_pistol",
    "semiconductor_device",
    "semi-detached_house",
    "semigloss",
    "semitrailer",
    "sennit",
    "sensitometer",
    "sentry_box",
    "separate",
    "septic_tank",
    "sequence",
    "sequencer",
    "serape",
    "serge",
    "serger",
    "serial_port",
    "serpent",
    "serration",
    "server",
    "server",
    "service_club",
    "serving_cart",
    "serving_dish",
    "servo",
    "set",
    "set_gun",
    "setscrew",
    "setscrew",
    "set_square",
    "settee",
    "settle",
    "settlement_house",
    "seventy-eight",
    "Seven_Wonders_of_the_Ancient_World",
    "sewage_disposal_plant",
    "sewer",
    "sewing_basket",
    "sewing_kit",
    "sewing_machine",
    "sewing_needle",
    "sewing_room",
    "sextant",
    "sgraffito",
    "shackle",
    "shackle",
    "shade",
    "shadow_box",
    "shaft",
    "shag_rug",
    "shaker",
    "shank",
    "shank",
    "shantung",
    "shaper",
    "shaping_tool",
    "sharkskin",
    "sharpener",
    "Sharpie",
    "shaver",
    "shaving_brush",
    "shaving_cream",
    "shaving_foam",
    "shawl",
    "shawm",
    "shears",
    "sheath",
    "sheathing",
    "shed",
    "sheep_bell",
    "sheepshank",
    "sheepskin_coat",
    "sheepwalk",
    "sheet",
    "sheet_bend",
    "sheeting",
    "sheet_pile",
    "Sheetrock",
    "shelf",
    "shelf_bracket",
    "shell",
    "shell",
    "shell",
    "shellac",
    "shelter",
    "shelter",
    "shelter",
    "sheltered_workshop",
    "Sheraton",
    "shield",
    "shield",
    "shielding",
    "shift_key",
    "shillelagh",
    "shim",
    "shingle",
    "shin_guard",
    "ship",
    "shipboard_system",
    "shipping",
    "shipping_room",
    "ship-towed_long-range_acoustic_detection_system",
    "shipwreck",
    "shirt",
    "shirt_button",
    "shirtdress",
    "shirtfront",
    "shirting",
    "shirtsleeve",
    "shirttail",
    "shirtwaist",
    "shiv",
    "shock_absorber",
    "shoe",
    "shoe",
    "shoebox",
    "shoehorn",
    "shoe_shop",
    "shoetree",
    "shofar",
    "shoji",
    "shooting_brake",
    "shooting_lodge",
    "shooting_stick",
    "shop",
    "shop_bell",
    "shopping_bag",
    "shopping_basket",
    "shopping_cart",
    "short_circuit",
    "short_iron",
    "short_pants",
    "short_sleeve",
    "shortwave_diathermy_machine",
    "shot",
    "shot_glass",
    "shotgun",
    "shotgun_shell",
    "shot_tower",
    "shoulder",
    "shoulder_bag",
    "shouldered_arch",
    "shoulder_holster",
    "shoulder_pad",
    "shoulder_patch",
    "shovel",
    "shovel",
    "shovel_hat",
    "showboat",
    "shower",
    "shower_cap",
    "shower_curtain",
    "shower_room",
    "shower_stall",
    "showroom",
    "shrapnel",
    "shredder",
    "shrimper",
    "shrine",
    "shrink-wrap",
    "shunt",
    "shunt",
    "shunter",
    "shutter",
    "shutter",
    "shuttle",
    "shuttle",
    "shuttle_bus",
    "shuttlecock",
    "shuttle_helicopter",
    "Sibley_tent",
    "sickbay",
    "sickbed",
    "sickle",
    "sickroom",
    "sideboard",
    "sidecar",
    "side_chapel",
    "sidelight",
    "sidesaddle",
    "sidewalk",
    "sidewall",
    "side-wheeler",
    "sidewinder",
    "sieve",
    "sifter",
    "sights",
    "sigmoidoscope",
    "signal_box",
    "signaling_device",
    "signboard",
    "silencer",
    "silent_butler",
    "Silex",
    "silk",
    "silks",
    "silo",
    "silver_plate",
    "silverpoint",
    "simple_pendulum",
    "simulator",
    "single_bed",
    "single-breasted_jacket",
    "single-breasted_suit",
    "single_prop",
    "single-reed_instrument",
    "single-rotor_helicopter",
    "singlestick",
    "singlet",
    "siren",
    "sister_ship",
    "sitar",
    "sitz_bath",
    "six-pack",
    "skate",
    "skateboard",
    "skeg",
    "skein",
    "skeleton",
    "skeleton_key",
    "skep",
    "skep",
    "sketch",
    "sketcher",
    "skew_arch",
    "skewer",
    "ski",
    "ski_binding",
    "skibob",
    "ski_boot",
    "ski_cap",
    "skidder",
    "skid_lid",
    "skiff",
    "ski_jump",
    "ski_lodge",
    "ski_mask",
    "skimmer",
    "ski_parka",
    "ski-plane",
    "ski_pole",
    "ski_rack",
    "skirt",
    "skirt",
    "ski_tow",
    "Skivvies",
    "skullcap",
    "skybox",
    "skyhook",
    "skylight",
    "skysail",
    "skyscraper",
    "skywalk",
    "slacks",
    "slack_suit",
    "slasher",
    "slash_pocket",
    "slat",
    "slate",
    "slate_pencil",
    "slate_roof",
    "sled",
    "sleeper",
    "sleeper",
    "sleeping_bag",
    "sleeping_car",
    "sleeve",
    "sleeve",
    "sleigh_bed",
    "sleigh_bell",
    "slice_bar",
    "slicer",
    "slicer",
    "slide",
    "slide_fastener",
    "slide_projector",
    "slide_rule",
    "slide_valve",
    "sliding_door",
    "sliding_seat",
    "sliding_window",
    "sling",
    "sling",
    "slingback",
    "slinger_ring",
    "slip_clutch",
    "slipcover",
    "slip-joint_pliers",
    "slipknot",
    "slip-on",
    "slipper",
    "slip_ring",
    "slit_lamp",
    "slit_trench",
    "sloop",
    "sloop_of_war",
    "slop_basin",
    "slop_pail",
    "slops",
    "slopshop",
    "slot",
    "slot_machine",
    "sluice",
    "smack",
    "small_boat",
    "small_computer_system_interface",
    "small_ship",
    "small_stores",
    "smart_bomb",
    "smelling_bottle",
    "smocking",
    "smoke_bomb",
    "smokehouse",
    "smoker",
    "smoke_screen",
    "smoking_room",
    "smoothbore",
    "smooth_plane",
    "snack_bar",
    "snaffle",
    "snap",
    "snap_brim",
    "snap-brim_hat",
    "snare",
    "snare_drum",
    "snatch_block",
    "snifter",
    "sniper_rifle",
    "snips",
    "Sno-cat",
    "snood",
    "snorkel",
    "snorkel",
    "snowbank",
    "snowboard",
    "snowmobile",
    "snowplow",
    "snowshoe",
    "snowsuit",
    "snow_thrower",
    "snuffbox",
    "snuffer",
    "snuffers",
    "soapbox",
    "soap_dish",
    "soap_dispenser",
    "soap_pad",
    "soccer_ball",
    "sock",
    "socket",
    "socket_wrench",
    "socle",
    "soda_can",
    "soda_fountain",
    "soda_fountain",
    "sod_house",
    "sodium-vapor_lamp",
    "sofa",
    "soffit",
    "softball",
    "soft_pedal",
    "soil_pipe",
    "solar_array",
    "solar_cell",
    "solar_dish",
    "solar_heater",
    "solar_house",
    "solar_telescope",
    "solar_thermal_system",
    "soldering_iron",
    "solenoid",
    "solleret",
    "sombrero",
    "sonic_depth_finder",
    "sonogram",
    "sonograph",
    "sorter",
    "souk",
    "sound_bow",
    "soundbox",
    "sound_camera",
    "sounder",
    "sound_film",
    "sounding_board",
    "sounding_rocket",
    "sound_recording",
    "sound_spectrograph",
    "soup_bowl",
    "soup_ladle",
    "soupspoon",
    "source_of_illumination",
    "sourdine",
    "soutache",
    "soutane",
    "sou'wester",
    "soybean_future",
    "space_bar",
    "space_capsule",
    "spacecraft",
    "space_heater",
    "space_helmet",
    "space_rocket",
    "space_shuttle",
    "space_station",
    "spacesuit",
    "spade",
    "spade_bit",
    "spaghetti_junction",
    "Spandau",
    "spandex",
    "spandrel",
    "spanker",
    "spar",
    "sparge_pipe",
    "spark_arrester",
    "spark_arrester",
    "spark_chamber",
    "spark_coil",
    "spark_gap",
    "spark_lever",
    "spark_plug",
    "sparkplug_wrench",
    "spark_transmitter",
    "spat",
    "spatula",
    "spatula",
    "speakerphone",
    "speaking_trumpet",
    "spear",
    "spear",
    "specialty_store",
    "specimen_bottle",
    "spectacle",
    "spectacles",
    "spectator_pump",
    "spectrograph",
    "spectrophotometer",
    "spectroscope",
    "speculum",
    "speedboat",
    "speed_bump",
    "speedometer",
    "speed_skate",
    "spherometer",
    "sphygmomanometer",
    "spicemill",
    "spice_rack",
    "spider",
    "spider_web",
    "spike",
    "spike",
    "spindle",
    "spindle",
    "spindle",
    "spin_dryer",
    "spinet",
    "spinet",
    "spinnaker",
    "spinner",
    "spinning_frame",
    "spinning_jenny",
    "spinning_machine",
    "spinning_rod",
    "spinning_wheel",
    "spiral_bandage",
    "spiral_ratchet_screwdriver",
    "spiral_spring",
    "spirit_lamp",
    "spirit_stove",
    "spirometer",
    "spit",
    "spittoon",
    "splashboard",
    "splasher",
    "splice",
    "splicer",
    "splint",
    "split_rail",
    "Spode",
    "spoiler",
    "spoiler",
    "spoke",
    "spokeshave",
    "sponge_cloth",
    "sponge_mop",
    "spoon",
    "spoon",
    "Spork",
    "sporran",
    "sport_kite",
    "sports_car",
    "sports_equipment",
    "sports_implement",
    "sportswear",
    "sport_utility",
    "spot",
    "spotlight",
    "spot_weld",
    "spouter",
    "sprag",
    "spray_gun",
    "spray_paint",
    "spreader",
    "sprig",
    "spring",
    "spring_balance",
    "springboard",
    "sprinkler",
    "sprinkler_system",
    "sprit",
    "spritsail",
    "sprocket",
    "sprocket",
    "spun_yarn",
    "spur",
    "spur_gear",
    "sputnik",
    "spy_satellite",
    "squad_room",
    "square",
    "square_knot",
    "square-rigger",
    "square_sail",
    "squash_ball",
    "squash_racket",
    "squawk_box",
    "squeegee",
    "squeezer",
    "squelch_circuit",
    "squinch",
    "stabilizer",
    "stabilizer",
    "stabilizer_bar",
    "stable",
    "stable_gear",
    "stabling",
    "stacks",
    "staddle",
    "stadium",
    "stage",
    "stagecoach",
    "stained-glass_window",
    "stair-carpet",
    "stair-rod",
    "stairwell",
    "stake",
    "stall",
    "stall",
    "stamp",
    "stamp_mill",
    "stamping_machine",
    "stanchion",
    "stand",
    "standard",
    "standard_cell",
    "standard_transmission",
    "standing_press",
    "stanhope",
    "Stanley_Steamer",
    "staple",
    "staple",
    "staple_gun",
    "stapler",
    "starship",
    "starter",
    "starting_gate",
    "Stassano_furnace",
    "Statehouse",
    "stately_home",
    "state_prison",
    "stateroom",
    "static_tube",
    "station",
    "stator",
    "statue",
    "stay",
    "staysail",
    "steakhouse",
    "steak_knife",
    "stealth_aircraft",
    "stealth_bomber",
    "stealth_fighter",
    "steam_bath",
    "steamboat",
    "steam_chest",
    "steam_engine",
    "steamer",
    "steamer",
    "steam_iron",
    "steam_locomotive",
    "steamroller",
    "steam_shovel",
    "steam_turbine",
    "steam_whistle",
    "steel",
    "steel_arch_bridge",
    "steel_drum",
    "steel_mill",
    "steel-wool_pad",
    "steelyard",
    "steeple",
    "steerage",
    "steering_gear",
    "steering_linkage",
    "steering_system",
    "steering_wheel",
    "stele",
    "stem-winder",
    "stencil",
    "Sten_gun",
    "stenograph",
    "step",
    "step-down_transformer",
    "step_stool",
    "step-up_transformer",
    "stereo",
    "stereoscope",
    "stern_chaser",
    "sternpost",
    "sternwheeler",
    "stethoscope",
    "stewing_pan",
    "stick",
    "stick",
    "stick",
    "stick",
    "stile",
    "stiletto",
    "still",
    "stillroom",
    "Stillson_wrench",
    "stilt",
    "Stinger",
    "stink_bomb",
    "stirrer",
    "stirrup",
    "stirrup_pump",
    "stob",
    "stock",
    "stockade",
    "stockcar",
    "stock_car",
    "stockinet",
    "stocking",
    "stock-in-trade",
    "stockpot",
    "stockroom",
    "stocks",
    "stock_saddle",
    "stockyard",
    "stole",
    "stomacher",
    "stomach_pump",
    "stone_wall",
    "stoneware",
    "stonework",
    "stool",
    "stoop",
    "stop_bath",
    "stopcock",
    "stopper_knot",
    "stopwatch",
    "storage_battery",
    "storage_cell",
    "storage_ring",
    "storage_space",
    "storeroom",
    "storm_cellar",
    "storm_door",
    "storm_window",
    "stoup",
    "stoup",
    "stove",
    "stove",
    "stove_bolt",
    "stovepipe",
    "stovepipe_iron",
    "Stradavarius",
    "straight_chair",
    "straightedge",
    "straightener",
    "straight_flute",
    "straight_pin",
    "straight_razor",
    "strainer",
    "straitjacket",
    "strap",
    "strap",
    "strap_hinge",
    "strapless",
    "streamer_fly",
    "streamliner",
    "street",
    "street",
    "streetcar",
    "street_clothes",
    "streetlight",
    "stretcher",
    "stretcher",
    "stretch_pants",
    "strickle",
    "strickle",
    "stringed_instrument",
    "stringer",
    "stringer",
    "string_tie",
    "strip",
    "strip_lighting",
    "strip_mall",
    "stroboscope",
    "strongbox",
    "stronghold",
    "strongroom",
    "strop",
    "structural_member",
    "structure",
    "student_center",
    "student_lamp",
    "student_union",
    "stud_finder",
    "studio_apartment",
    "studio_couch",
    "study",
    "study_hall",
    "stuffing_nut",
    "stump",
    "stun_gun",
    "stupa",
    "sty",
    "stylus",
    "stylus",
    "sub-assembly",
    "subcompact",
    "submachine_gun",
    "submarine",
    "submarine_torpedo",
    "submersible",
    "submersible",
    "subtracter",
    "subway_token",
    "subway_train",
    "subwoofer",
    "suction_cup",
    "suction_pump",
    "sudatorium",
    "suede_cloth",
    "sugar_bowl",
    "sugar_refinery",
    "sugar_spoon",
    "suit",
    "suite",
    "suiting",
    "sulky",
    "summer_house",
    "sumo_ring",
    "sump",
    "sump_pump",
    "sunbonnet",
    "Sunday_best",
    "sun_deck",
    "sundial",
    "sundress",
    "sundries",
    "sun_gear",
    "sunglass",
    "sunglasses",
    "sunhat",
    "sunlamp",
    "sun_parlor",
    "sunroof",
    "sunscreen",
    "sunsuit",
    "supercharger",
    "supercomputer",
    "superconducting_supercollider",
    "superhighway",
    "supermarket",
    "superstructure",
    "supertanker",
    "supper_club",
    "supplejack",
    "supply_chamber",
    "supply_closet",
    "support",
    "support",
    "support_column",
    "support_hose",
    "supporting_structure",
    "supporting_tower",
    "surcoat",
    "surface_gauge",
    "surface_lift",
    "surface_search_radar",
    "surface_ship",
    "surface-to-air_missile",
    "surface-to-air_missile_system",
    "surfboat",
    "surcoat",
    "surgeon's_knot",
    "surgery",
    "surge_suppressor",
    "surgical_dressing",
    "surgical_instrument",
    "surgical_knife",
    "surplice",
    "surrey",
    "surtout",
    "surveillance_system",
    "surveying_instrument",
    "surveyor's_level",
    "sushi_bar",
    "suspension",
    "suspension_bridge",
    "suspensory",
    "sustaining_pedal",
    "suture",
    "swab",
    "swab",
    "swaddling_clothes",
    "swag",
    "swage_block",
    "swagger_stick",
    "swallow-tailed_coat",
    "swamp_buggy",
    "swan's_down",
    "swathe",
    "swatter",
    "sweat_bag",
    "sweatband",
    "sweater",
    "sweat_pants",
    "sweatshirt",
    "sweatshop",
    "sweat_suit",
    "sweep",
    "sweep_hand",
    "swimming_trunks",
    "swimsuit",
    "swing",
    "swing_door",
    "switch",
    "switchblade",
    "switch_engine",
    "swivel",
    "swivel_chair",
    "swizzle_stick",
    "sword",
    "sword_cane",
    "S_wrench",
    "synagogue",
    "synchrocyclotron",
    "synchroflash",
    "synchromesh",
    "synchronous_converter",
    "synchronous_motor",
    "synchrotron",
    "synchroscope",
    "synthesizer",
    "syringe",
    "system",
    "tabard",
    "Tabernacle",
    "tabi",
    "tab_key",
    "table",
    "table",
    "tablefork",
    "table_knife",
    "table_lamp",
    "table_saw",
    "tablespoon",
    "tablet-armed_chair",
    "table-tennis_table",
    "table-tennis_racquet",
    "tabletop",
    "tableware",
    "tabor",
    "taboret",
    "tachistoscope",
    "tachograph",
    "tachometer",
    "tachymeter",
    "tack",
    "tack_hammer",
    "taffeta",
    "taffrail",
    "tailgate",
    "taillight",
    "tailor-made",
    "tailor's_chalk",
    "tailpipe",
    "tail_rotor",
    "tailstock",
    "take-up",
    "talaria",
    "talcum",
    "tam",
    "tambour",
    "tambour",
    "tambourine",
    "tammy",
    "tamp",
    "Tampax",
    "tampion",
    "tampon",
    "tandoor",
    "tangram",
    "tank",
    "tank",
    "tankard",
    "tank_car",
    "tank_destroyer",
    "tank_engine",
    "tanker_plane",
    "tank_shell",
    "tank_top",
    "tannoy",
    "tap",
    "tapa",
    "tape",
    "tape",
    "tape_deck",
    "tape_drive",
    "tape_player",
    "tape_recorder",
    "taper_file",
    "tapestry",
    "tappet",
    "tap_wrench",
    "tare",
    "target",
    "target_acquisition_system",
    "tarmacadam",
    "tarpaulin",
    "tartan",
    "tasset",
    "tattoo",
    "tavern",
    "tawse",
    "taximeter",
    "T-bar_lift",
    "tea_bag",
    "tea_ball",
    "tea_cart",
    "tea_chest",
    "teaching_aid",
    "teacup",
    "tea_gown",
    "teakettle",
    "tea_maker",
    "teapot",
    "teashop",
    "teaspoon",
    "tea-strainer",
    "tea_table",
    "tea_tray",
    "tea_urn",
    "tee",
    "tee_hinge",
    "telecom_hotel",
    "telecommunication_system",
    "telegraph",
    "telegraph_key",
    "telemeter",
    "telephone",
    "telephone_bell",
    "telephone_booth",
    "telephone_cord",
    "telephone_jack",
    "telephone_line",
    "telephone_plug",
    "telephone_pole",
    "telephone_receiver",
    "telephone_system",
    "telephone_wire",
    "telephoto_lens",
    "Teleprompter",
    "telescope",
    "telescopic_sight",
    "telethermometer",
    "teletypewriter",
    "television",
    "television_antenna",
    "television_camera",
    "television_equipment",
    "television_monitor",
    "television_receiver",
    "television_room",
    "television_transmitter",
    "telpher",
    "telpherage",
    "tempera",
    "temple",
    "temple",
    "temporary_hookup",
    "tender",
    "tender",
    "tender",
    "tenement",
    "tennis_ball",
    "tennis_camp",
    "tennis_racket",
    "tenon",
    "tenor_drum",
    "tenoroon",
    "tenpenny_nail",
    "tenpin",
    "tensimeter",
    "tensiometer",
    "tensiometer",
    "tensiometer",
    "tent",
    "tenter",
    "tenterhook",
    "tent-fly",
    "tent_peg",
    "tepee",
    "terminal",
    "terminal",
    "terraced_house",
    "terra_cotta",
    "terrarium",
    "terra_sigillata",
    "terry",
    "Tesla_coil",
    "tessera",
    "test_equipment",
    "test_rocket",
    "test_room",
    "testudo",
    "tetraskelion",
    "tetrode",
    "textile_machine",
    "textile_mill",
    "thatch",
    "theater",
    "theater_curtain",
    "theater_light",
    "theodolite",
    "theremin",
    "thermal_printer",
    "thermal_reactor",
    "thermocouple",
    "thermoelectric_thermometer",
    "thermograph",
    "thermograph",
    "thermohydrometer",
    "thermojunction",
    "thermometer",
    "thermonuclear_reactor",
    "thermopile",
    "thermos",
    "thermostat",
    "thigh_pad",
    "thill",
    "thimble",
    "thinning_shears",
    "third_base",
    "third_gear",
    "third_rail",
    "thong",
    "thong",
    "three-centered_arch",
    "three-decker",
    "three-dimensional_radar",
    "three-piece_suit",
    "three-quarter_binding",
    "three-way_switch",
    "thresher",
    "threshing_floor",
    "thriftshop",
    "throat_protector",
    "throne",
    "thrust_bearing",
    "thruster",
    "thumb",
    "thumbhole",
    "thumbscrew",
    "thumbstall",
    "thumbtack",
    "thunderer",
    "thwart",
    "tiara",
    "ticking",
    "tickler_coil",
    "tie",
    "tie",
    "tie_rack",
    "tie_rod",
    "tights",
    "tile",
    "tile_cutter",
    "tile_roof",
    "tiller",
    "tilter",
    "tilt-top_table",
    "timber",
    "timber",
    "timber_hitch",
    "timbrel",
    "time_bomb",
    "time_capsule",
    "time_clock",
    "time-delay_measuring_instrument",
    "time-fuse",
    "timepiece",
    "timer",
    "timer",
    "time-switch",
    "tin",
    "tinderbox",
    "tine",
    "tinfoil",
    "tippet",
    "tire_chain",
    "tire_iron",
    "titfer",
    "tithe_barn",
    "titrator",
    "toaster",
    "toaster_oven",
    "toasting_fork",
    "toastrack",
    "tobacco_pouch",
    "tobacco_shop",
    "toboggan",
    "toby",
    "tocsin",
    "toe",
    "toecap",
    "toehold",
    "toga",
    "toga_virilis",
    "toggle",
    "toggle_bolt",
    "toggle_joint",
    "toggle_switch",
    "togs",
    "toilet",
    "toilet_bag",
    "toilet_bowl",
    "toilet_kit",
    "toilet_powder",
    "toiletry",
    "toilet_seat",
    "toilet_water",
    "tokamak",
    "token",
    "tollbooth",
    "toll_bridge",
    "tollgate",
    "toll_line",
    "tomahawk",
    "Tommy_gun",
    "tomograph",
    "tone_arm",
    "toner",
    "tongs",
    "tongue",
    "tongue_and_groove_joint",
    "tongue_depressor",
    "tonometer",
    "tool",
    "tool_bag",
    "toolbox",
    "toolshed",
    "tooth",
    "tooth",
    "toothbrush",
    "toothpick",
    "top",
    "top",
    "topgallant",
    "topgallant",
    "topiary",
    "topknot",
    "topmast",
    "topper",
    "topsail",
    "toque",
    "torch",
    "torpedo",
    "torpedo",
    "torpedo",
    "torpedo_boat",
    "torpedo-boat_destroyer",
    "torpedo_tube",
    "torque_converter",
    "torque_wrench",
    "torture_chamber",
    "totem_pole",
    "touch_screen",
    "toupee",
    "touring_car",
    "tourist_class",
    "towel",
    "toweling",
    "towel_rack",
    "towel_rail",
    "tower",
    "town_hall",
    "towpath",
    "tow_truck",
    "toy",
    "toy_box",
    "toyshop",
    "trace_detector",
    "track",
    "track",
    "trackball",
    "tracked_vehicle",
    "tract_house",
    "tract_housing",
    "traction_engine",
    "tractor",
    "tractor",
    "trail_bike",
    "trailer",
    "trailer",
    "trailer_camp",
    "trailer_truck",
    "trailing_edge",
    "train",
    "tramline",
    "trammel",
    "trampoline",
    "tramp_steamer",
    "tramway",
    "transdermal_patch",
    "transept",
    "transformer",
    "transistor",
    "transit_instrument",
    "transmission",
    "transmission_shaft",
    "transmitter",
    "transom",
    "transom",
    "transponder",
    "transporter",
    "transporter",
    "transport_ship",
    "trap",
    "trap_door",
    "trapeze",
    "trave",
    "travel_iron",
    "trawl",
    "trawl",
    "trawler",
    "tray",
    "tray_cloth",
    "tread",
    "tread",
    "treadmill",
    "treadmill",
    "treasure_chest",
    "treasure_ship",
    "treenail",
    "trefoil_arch",
    "trellis",
    "trench",
    "trench_coat",
    "trench_knife",
    "trepan",
    "trepan",
    "trestle",
    "trestle",
    "trestle_bridge",
    "trestle_table",
    "trestlework",
    "trews",
    "trial_balloon",
    "triangle",
    "triangle",
    "triclinium",
    "triclinium",
    "tricorn",
    "tricot",
    "tricycle",
    "trident",
    "trigger",
    "trimaran",
    "trimmer",
    "trimmer_arch",
    "triode",
    "tripod",
    "triptych",
    "trip_wire",
    "trireme",
    "triskelion",
    "triumphal_arch",
    "trivet",
    "trivet",
    "troika",
    "troll",
    "trolleybus",
    "trombone",
    "troop_carrier",
    "troopship",
    "trophy_case",
    "trough",
    "trouser",
    "trouser_cuff",
    "trouser_press",
    "trouser",
    "trousseau",
    "trowel",
    "truck",
    "trumpet_arch",
    "truncheon",
    "trundle_bed",
    "trunk",
    "trunk_hose",
    "trunk_lid",
    "trunk_line",
    "truss",
    "truss_bridge",
    "try_square",
    "T-square",
    "tub",
    "tube",
    "tuck_box",
    "tucker",
    "tucker-bag",
    "tuck_shop",
    "Tudor_arch",
    "tudung",
    "tugboat",
    "tulle",
    "tumble-dryer",
    "tumbler",
    "tumbrel",
    "tun",
    "tunic",
    "tuning_fork",
    "tupik",
    "turban",
    "turbine",
    "turbogenerator",
    "tureen",
    "Turkish_bath",
    "Turkish_towel",
    "Turk's_head",
    "turnbuckle",
    "turner",
    "turnery",
    "turnpike",
    "turnspit",
    "turnstile",
    "turntable",
    "turntable",
    "turret",
    "turret_clock",
    "turtleneck",
    "tweed",
    "tweeter",
    "twenty-two",
    "twenty-two_pistol",
    "twenty-two_rifle",
    "twill",
    "twill",
    "twin_bed",
    "twinjet",
    "twist_bit",
    "two-by-four",
    "two-man_tent",
    "two-piece",
    "typesetting_machine",
    "typewriter",
    "typewriter_carriage",
    "typewriter_keyboard",
    "tyrolean",
    "uke",
    "ulster",
    "ultracentrifuge",
    "ultramicroscope",
    "Ultrasuede",
    "ultraviolet_lamp",
    "umbrella",
    "umbrella_tent",
    "undercarriage",
    "undercoat",
    "undergarment",
    "underpants",
    "underwear",
    "undies",
    "uneven_parallel_bars",
    "unicycle",
    "uniform",
    "universal_joint",
    "university",
    "upholstery",
    "upholstery_material",
    "upholstery_needle",
    "uplift",
    "upper_berth",
    "upright",
    "upset",
    "upstairs",
    "urceole",
    "urn",
    "urn",
    "used-car",
    "utensil",
    "Uzi",
    "vacation_home",
    "vacuum",
    "vacuum_chamber",
    "vacuum_flask",
    "vacuum_gauge",
    "Valenciennes",
    "valise",
    "valve",
    "valve",
    "valve-in-head_engine",
    "vambrace",
    "van",
    "van",
    "vane",
    "vaporizer",
    "variable-pitch_propeller",
    "variometer",
    "varnish",
    "vase",
    "vault",
    "vault",
    "vaulting_horse",
    "vehicle",
    "Velcro",
    "velocipede",
    "velour",
    "velvet",
    "velveteen",
    "vending_machine",
    "veneer",
    "Venetian_blind",
    "Venn_diagram",
    "ventilation",
    "ventilation_shaft",
    "ventilator",
    "veranda",
    "verdigris",
    "vernier_caliper",
    "vernier_scale",
    "vertical_file",
    "vertical_stabilizer",
    "vertical_tail",
    "Very_pistol",
    "vessel",
    "vessel",
    "vest",
    "vestiture",
    "vestment",
    "vest_pocket",
    "vestry",
    "viaduct",
    "vibraphone",
    "vibrator",
    "vibrator",
    "Victrola",
    "vicuna",
    "videocassette",
    "videocassette_recorder",
    "videodisk",
    "video_recording",
    "videotape",
    "videotape",
    "vigil_light",
    "villa",
    "villa",
    "villa",
    "viol",
    "viola",
    "viola_da_braccio",
    "viola_da_gamba",
    "viola_d'amore",
    "violin",
    "virginal",
    "viscometer",
    "viscose_rayon",
    "vise",
    "visor",
    "visual_display_unit",
    "vivarium",
    "Viyella",
    "voile",
    "volleyball",
    "volleyball_net",
    "voltage_regulator",
    "voltaic_cell",
    "voltaic_pile",
    "voltmeter",
    "vomitory",
    "von_Neumann_machine",
    "voting_booth",
    "voting_machine",
    "voussoir",
    "vox_angelica",
    "vox_humana",
    "waders",
    "wading_pool",
    "waffle_iron",
    "wagon",
    "wagon",
    "wagon_tire",
    "wagon_wheel",
    "wain",
    "wainscot",
    "wainscoting",
    "waist_pack",
    "walker",
    "walker",
    "walker",
    "walkie-talkie",
    "walk-in",
    "walking_shoe",
    "walking_stick",
    "Walkman",
    "walk-up_apartment",
    "wall",
    "wall",
    "wall_clock",
    "wallet",
    "wall_tent",
    "wall_unit",
    "wand",
    "Wankel_engine",
    "ward",
    "wardrobe",
    "wardroom",
    "warehouse",
    "warming_pan",
    "war_paint",
    "warplane",
    "war_room",
    "warship",
    "wash",
    "wash-and-wear",
    "washbasin",
    "washboard",
    "washboard",
    "washer",
    "washer",
    "washhouse",
    "washroom",
    "washstand",
    "washtub",
    "wastepaper_basket",
    "watch",
    "watch_cap",
    "watch_case",
    "watch_glass",
    "watchtower",
    "water-base_paint",
    "water_bed",
    "water_bottle",
    "water_butt",
    "water_cart",
    "water_chute",
    "water_closet",
    "watercolor",
    "water-cooled_reactor",
    "water_cooler",
    "water_faucet",
    "water_filter",
    "water_gauge",
    "water_glass",
    "water_hazard",
    "water_heater",
    "watering_can",
    "watering_cart",
    "water_jacket",
    "water_jug",
    "water_jump",
    "water_level",
    "water_meter",
    "water_mill",
    "waterproof",
    "waterproofing",
    "water_pump",
    "water_scooter",
    "water_ski",
    "waterspout",
    "water_tower",
    "water_wagon",
    "waterwheel",
    "waterwheel",
    "water_wings",
    "waterworks",
    "wattmeter",
    "waxwork",
    "ways",
    "weapon",
    "weaponry",
    "weapons_carrier",
    "weathercock",
    "weatherglass",
    "weather_satellite",
    "weather_ship",
    "weathervane",
    "web",
    "web",
    "webbing",
    "webcam",
    "wedge",
    "wedge",
    "wedgie",
    "Wedgwood",
    "weeder",
    "weeds",
    "weekender",
    "weighbridge",
    "weight",
    "weir",
    "weir",
    "welcome_wagon",
    "weld",
    "welder's_mask",
    "weldment",
    "well",
    "wellhead",
    "welt",
    "Weston_cell",
    "wet_bar",
    "wet-bulb_thermometer",
    "wet_cell",
    "wet_fly",
    "wet_suit",
    "whaleboat",
    "whaler",
    "whaling_gun",
    "wheel",
    "wheel",
    "wheel_and_axle",
    "wheelchair",
    "wheeled_vehicle",
    "wheelwork",
    "wherry",
    "wherry",
    "whetstone",
    "whiffletree",
    "whip",
    "whipcord",
    "whipping_post",
    "whipstitch",
    "whirler",
    "whisk",
    "whisk",
    "whiskey_bottle",
    "whiskey_jug",
    "whispering_gallery",
    "whistle",
    "whistle",
    "white",
    "white_goods",
    "whitewash",
    "whorehouse",
    "wick",
    "wicker",
    "wicker_basket",
    "wicket",
    "wicket",
    "wickiup",
    "wide-angle_lens",
    "widebody_aircraft",
    "wide_wale",
    "widow's_walk",
    "Wiffle",
    "wig",
    "wigwam",
    "Wilton",
    "wimple",
    "wincey",
    "winceyette",
    "winch",
    "Winchester",
    "windbreak",
    "winder",
    "wind_instrument",
    "windjammer",
    "windmill",
    "windmill",
    "window",
    "window",
    "window_blind",
    "window_box",
    "window_envelope",
    "window_frame",
    "window_screen",
    "window_seat",
    "window_shade",
    "windowsill",
    "windshield",
    "windshield_wiper",
    "Windsor_chair",
    "Windsor_knot",
    "Windsor_tie",
    "wind_tee",
    "wind_tunnel",
    "wind_turbine",
    "wine_bar",
    "wine_bottle",
    "wine_bucket",
    "wine_cask",
    "wineglass",
    "winepress",
    "winery",
    "wineskin",
    "wing",
    "wing_chair",
    "wing_nut",
    "wing_tip",
    "wing_tip",
    "winker",
    "wiper",
    "wiper_motor",
    "wire",
    "wire",
    "wire_cloth",
    "wire_cutter",
    "wire_gauge",
    "wireless_local_area_network",
    "wire_matrix_printer",
    "wire_recorder",
    "wire_stripper",
    "wirework",
    "wiring",
    "wishing_cap",
    "witness_box",
    "wok",
    "woman's_clothing",
    "wood",
    "woodcarving",
    "wood_chisel",
    "woodenware",
    "wooden_spoon",
    "woodscrew",
    "woodshed",
    "wood_vise",
    "woodwind",
    "woof",
    "woofer",
    "wool",
    "workbasket",
    "workbench",
    "work-clothing",
    "workhouse",
    "workhouse",
    "workpiece",
    "workroom",
    "works",
    "work-shirt",
    "workstation",
    "worktable",
    "workwear",
    "World_Wide_Web",
    "worm_fence",
    "worm_gear",
    "worm_wheel",
    "worsted",
    "worsted",
    "wrap",
    "wraparound",
    "wrapping",
    "wreck",
    "wrench",
    "wrestling_mat",
    "wringer",
    "wrist_pad",
    "wrist_pin",
    "wristwatch",
    "writing_arm",
    "writing_desk",
    "writing_desk",
    "writing_implement",
    "xerographic_printer",
    "Xerox",
    "X-ray_film",
    "X-ray_machine",
    "X-ray_tube",
    "yacht",
    "yacht_chair",
    "yagi",
    "yard",
    "yard",
    "yardarm",
    "yard_marker",
    "yardstick",
    "yarmulke",
    "yashmak",
    "yataghan",
    "yawl",
    "yawl",
    "yoke",
    "yoke",
    "yoke",
    "yurt",
    "Zamboni",
    "zero",
    "ziggurat",
    "zill",
    "zip_gun",
    "zither",
    "zoot_suit",
    "shading",
    "grain",
    "wood_grain",
    "graining",
    "marbleization",
    "light",
    "aura",
    "sunniness",
    "glint",
    "opalescence",
    "polish",
    "primary_color_for_pigments",
    "primary_color_for_light",
    "colorlessness",
    "mottle",
    "achromia",
    "shade",
    "chromatic_color",
    "black",
    "coal_black",
    "alabaster",
    "bone",
    "gray",
    "ash_grey",
    "charcoal",
    "sanguine",
    "Turkey_red",
    "crimson",
    "dark_red",
    "claret",
    "fuschia",
    "maroon",
    "orange",
    "reddish_orange",
    "yellow",
    "gamboge",
    "pale_yellow",
    "green",
    "greenishness",
    "sea_green",
    "sage_green",
    "bottle_green",
    "emerald",
    "olive_green",
    "jade_green",
    "blue",
    "azure",
    "steel_blue",
    "greenish_blue",
    "purplish_blue",
    "purple",
    "Tyrian_purple",
    "indigo",
    "lavender",
    "reddish_purple",
    "pink",
    "carnation",
    "rose",
    "chestnut",
    "chocolate",
    "light_brown",
    "tan",
    "beige",
    "reddish_brown",
    "brick_red",
    "copper",
    "Indian_red",
    "puce",
    "olive",
    "ultramarine",
    "complementary_color",
    "pigmentation",
    "complexion",
    "ruddiness",
    "nonsolid_color",
    "aposematic_coloration",
    "cryptic_coloration",
    "ring",
    "center_of_curvature",
    "cadaver",
    "mandibular_notch",
    "rib",
    "skin",
    "skin_graft",
    "epidermal_cell",
    "melanocyte",
    "prickle_cell",
    "columnar_cell",
    "spongioblast",
    "squamous_cell",
    "amyloid_plaque",
    "dental_plaque",
    "macule",
    "freckle",
    "bouffant",
    "sausage_curl",
    "forelock",
    "spit_curl",
    "pigtail",
    "pageboy",
    "pompadour",
    "thatch",
    "soup-strainer",
    "mustachio",
    "walrus_mustache",
    "stubble",
    "vandyke_beard",
    "soul_patch",
    "esophageal_smear",
    "paraduodenal_smear",
    "specimen",
    "punctum",
    "glenoid_fossa",
    "diastema",
    "marrow",
    "mouth",
    "canthus",
    "milk",
    "mother's_milk",
    "colostrum",
    "vein",
    "ganglion_cell",
    "X_chromosome",
    "embryonic_cell",
    "myeloblast",
    "sideroblast",
    "osteocyte",
    "megalocyte",
    "leukocyte",
    "histiocyte",
    "fixed_phagocyte",
    "lymphocyte",
    "monoblast",
    "neutrophil",
    "microphage",
    "sickle_cell",
    "siderocyte",
    "spherocyte",
    "ootid",
    "oocyte",
    "spermatid",
    "Leydig_cell",
    "striated_muscle_cell",
    "smooth_muscle_cell",
    "Ranvier's_nodes",
    "neuroglia",
    "astrocyte",
    "protoplasmic_astrocyte",
    "oligodendrocyte",
    "proprioceptor",
    "dendrite",
    "sensory_fiber",
    "subarachnoid_space",
    "cerebral_cortex",
    "renal_cortex",
    "prepuce",
    "head",
    "scalp",
    "frontal_eminence",
    "suture",
    "foramen_magnum",
    "esophagogastric_junction",
    "heel",
    "cuticle",
    "hangnail",
    "exoskeleton",
    "abdominal_wall",
    "lemon",
    "coordinate_axis",
    "landscape",
    "medium",
    "vehicle",
    "paper",
    "channel",
    "film",
    "silver_screen",
    "free_press",
    "press",
    "print_media",
    "storage_medium",
    "magnetic_storage_medium",
    "journalism",
    "Fleet_Street",
    "photojournalism",
    "news_photography",
    "rotogravure",
    "newspaper",
    "daily",
    "gazette",
    "school_newspaper",
    "tabloid",
    "yellow_journalism",
    "telecommunication",
    "telephone",
    "voice_mail",
    "call",
    "call-back",
    "collect_call",
    "call_forwarding",
    "call-in",
    "call_waiting",
    "crank_call",
    "local_call",
    "long_distance",
    "toll_call",
    "wake-up_call",
    "three-way_calling",
    "telegraphy",
    "cable",
    "wireless",
    "radiotelegraph",
    "radiotelephone",
    "broadcasting",
    "Rediffusion",
    "multiplex",
    "radio",
    "television",
    "cable_television",
    "high-definition_television",
    "reception",
    "signal_detection",
    "Hakham",
    "web_site",
    "chat_room",
    "portal_site",
    "jotter",
    "breviary",
    "wordbook",
    "desk_dictionary",
    "reckoner",
    "document",
    "album",
    "concept_album",
    "rock_opera",
    "tribute_album",
    "magazine",
    "colour_supplement",
    "comic_book",
    "news_magazine",
    "pulp",
    "slick",
    "trade_magazine",
    "movie",
    "outtake",
    "shoot-'em-up",
    "spaghetti_Western",
    "encyclical",
    "crossword_puzzle",
    "sign",
    "street_sign",
    "traffic_light",
    "swastika",
    "concert",
    "artwork",
    "lobe",
    "book_jacket",
    "cairn",
    "three-day_event",
    "comfort_food",
    "comestible",
    "tuck",
    "course",
    "dainty",
    "dish",
    "fast_food",
    "finger_food",
    "ingesta",
    "kosher",
    "fare",
    "diet",
    "diet",
    "dietary",
    "balanced_diet",
    "bland_diet",
    "clear_liquid_diet",
    "diabetic_diet",
    "dietary_supplement",
    "carbohydrate_loading",
    "fad_diet",
    "gluten-free_diet",
    "high-protein_diet",
    "high-vitamin_diet",
    "light_diet",
    "liquid_diet",
    "low-calorie_diet",
    "low-fat_diet",
    "low-sodium_diet",
    "macrobiotic_diet",
    "reducing_diet",
    "soft_diet",
    "vegetarianism",
    "menu",
    "chow",
    "board",
    "mess",
    "ration",
    "field_ration",
    "K_ration",
    "C-ration",
    "foodstuff",
    "starches",
    "breadstuff",
    "coloring",
    "concentrate",
    "tomato_concentrate",
    "meal",
    "kibble",
    "cornmeal",
    "farina",
    "matzo_meal",
    "oatmeal",
    "pea_flour",
    "roughage",
    "bran",
    "flour",
    "plain_flour",
    "wheat_flour",
    "whole_wheat_flour",
    "soybean_meal",
    "semolina",
    "corn_gluten_feed",
    "nutriment",
    "commissariat",
    "larder",
    "frozen_food",
    "canned_food",
    "canned_meat",
    "Spam",
    "dehydrated_food",
    "square_meal",
    "meal",
    "potluck",
    "refection",
    "refreshment",
    "breakfast",
    "continental_breakfast",
    "brunch",
    "lunch",
    "business_lunch",
    "high_tea",
    "tea",
    "dinner",
    "supper",
    "buffet",
    "picnic",
    "cookout",
    "barbecue",
    "clambake",
    "fish_fry",
    "bite",
    "nosh",
    "nosh-up",
    "ploughman's_lunch",
    "coffee_break",
    "banquet",
    "entree",
    "piece_de_resistance",
    "plate",
    "adobo",
    "side_dish",
    "special",
    "casserole",
    "chicken_casserole",
    "chicken_cacciatore",
    "antipasto",
    "appetizer",
    "canape",
    "cocktail",
    "fruit_cocktail",
    "crab_cocktail",
    "shrimp_cocktail",
    "hors_d'oeuvre",
    "relish",
    "dip",
    "bean_dip",
    "cheese_dip",
    "clam_dip",
    "guacamole",
    "soup",
    "soup_du_jour",
    "alphabet_soup",
    "consomme",
    "madrilene",
    "bisque",
    "borsch",
    "broth",
    "barley_water",
    "bouillon",
    "beef_broth",
    "chicken_broth",
    "broth",
    "stock_cube",
    "chicken_soup",
    "cock-a-leekie",
    "gazpacho",
    "gumbo",
    "julienne",
    "marmite",
    "mock_turtle_soup",
    "mulligatawny",
    "oxtail_soup",
    "pea_soup",
    "pepper_pot",
    "petite_marmite",
    "potage",
    "pottage",
    "turtle_soup",
    "eggdrop_soup",
    "chowder",
    "corn_chowder",
    "clam_chowder",
    "Manhattan_clam_chowder",
    "New_England_clam_chowder",
    "fish_chowder",
    "won_ton",
    "split-pea_soup",
    "green_pea_soup",
    "lentil_soup",
    "Scotch_broth",
    "vichyssoise",
    "stew",
    "bigos",
    "Brunswick_stew",
    "burgoo",
    "burgoo",
    "olla_podrida",
    "mulligan_stew",
    "purloo",
    "goulash",
    "hotchpotch",
    "hot_pot",
    "beef_goulash",
    "pork-and-veal_goulash",
    "porkholt",
    "Irish_stew",
    "oyster_stew",
    "lobster_stew",
    "lobscouse",
    "fish_stew",
    "bouillabaisse",
    "matelote",
    "paella",
    "fricassee",
    "chicken_stew",
    "turkey_stew",
    "beef_stew",
    "ragout",
    "ratatouille",
    "salmi",
    "pot-au-feu",
    "slumgullion",
    "smorgasbord",
    "viand",
    "ready-mix",
    "brownie_mix",
    "cake_mix",
    "lemonade_mix",
    "self-rising_flour",
    "choice_morsel",
    "savory",
    "calf's-foot_jelly",
    "caramel",
    "lump_sugar",
    "cane_sugar",
    "castor_sugar",
    "powdered_sugar",
    "granulated_sugar",
    "icing_sugar",
    "corn_sugar",
    "brown_sugar",
    "demerara",
    "sweet",
    "confectionery",
    "confiture",
    "sweetmeat",
    "candy",
    "candy_bar",
    "carob_bar",
    "hardbake",
    "hard_candy",
    "barley-sugar",
    "brandyball",
    "jawbreaker",
    "lemon_drop",
    "sourball",
    "patty",
    "peppermint_patty",
    "bonbon",
    "brittle",
    "peanut_brittle",
    "chewing_gum",
    "gum_ball",
    "bubble_gum",
    "butterscotch",
    "candied_fruit",
    "candied_apple",
    "crystallized_ginger",
    "grapefruit_peel",
    "lemon_peel",
    "orange_peel",
    "candied_citrus_peel",
    "candy_cane",
    "candy_corn",
    "caramel",
    "center",
    "comfit",
    "cotton_candy",
    "dragee",
    "dragee",
    "fondant",
    "fudge",
    "chocolate_fudge",
    "divinity",
    "penuche",
    "gumdrop",
    "jujube",
    "honey_crisp",
    "mint",
    "horehound",
    "peppermint",
    "jelly_bean",
    "kiss",
    "molasses_kiss",
    "meringue_kiss",
    "chocolate_kiss",
    "licorice",
    "Life_Saver",
    "lollipop",
    "lozenge",
    "cachou",
    "cough_drop",
    "marshmallow",
    "marzipan",
    "nougat",
    "nougat_bar",
    "nut_bar",
    "peanut_bar",
    "popcorn_ball",
    "praline",
    "rock_candy",
    "rock_candy",
    "sugar_candy",
    "sugarplum",
    "taffy",
    "molasses_taffy",
    "truffle",
    "Turkish_Delight",
    "dessert",
    "ambrosia",
    "ambrosia",
    "baked_Alaska",
    "blancmange",
    "charlotte",
    "compote",
    "dumpling",
    "flan",
    "frozen_dessert",
    "junket",
    "mousse",
    "mousse",
    "pavlova",
    "peach_melba",
    "whip",
    "prune_whip",
    "pudding",
    "pudding",
    "syllabub",
    "tiramisu",
    "trifle",
    "tipsy_cake",
    "jello",
    "apple_dumpling",
    "ice",
    "water_ice",
    "ice_cream",
    "ice-cream_cone",
    "chocolate_ice_cream",
    "Neapolitan_ice_cream",
    "peach_ice_cream",
    "sherbert",
    "strawberry_ice_cream",
    "tutti-frutti",
    "vanilla_ice_cream",
    "ice_lolly",
    "ice_milk",
    "frozen_yogurt",
    "snowball",
    "snowball",
    "parfait",
    "ice-cream_sundae",
    "split",
    "banana_split",
    "frozen_pudding",
    "frozen_custard",
    "pudding",
    "flummery",
    "fish_mousse",
    "chicken_mousse",
    "chocolate_mousse",
    "plum_pudding",
    "carrot_pudding",
    "corn_pudding",
    "steamed_pudding",
    "duff",
    "vanilla_pudding",
    "chocolate_pudding",
    "brown_Betty",
    "Nesselrode",
    "pease_pudding",
    "custard",
    "creme_caramel",
    "creme_anglais",
    "creme_brulee",
    "fruit_custard",
    "tapioca",
    "tapioca_pudding",
    "roly-poly",
    "suet_pudding",
    "Bavarian_cream",
    "maraschino",
    "nonpareil",
    "zabaglione",
    "garnish",
    "pastry",
    "turnover",
    "apple_turnover",
    "knish",
    "pirogi",
    "samosa",
    "timbale",
    "puff_paste",
    "phyllo",
    "puff_batter",
    "ice-cream_cake",
    "doughnut",
    "fish_cake",
    "fish_stick",
    "conserve",
    "apple_butter",
    "chowchow",
    "jam",
    "lemon_curd",
    "strawberry_jam",
    "jelly",
    "apple_jelly",
    "crabapple_jelly",
    "grape_jelly",
    "marmalade",
    "orange_marmalade",
    "gelatin",
    "gelatin_dessert",
    "buffalo_wing",
    "barbecued_wing",
    "mess",
    "mince",
    "puree",
    "barbecue",
    "biryani",
    "escalope_de_veau_Orloff",
    "saute",
    "patty",
    "veal_parmesan",
    "veal_cordon_bleu",
    "margarine",
    "mincemeat",
    "stuffing",
    "turkey_stuffing",
    "oyster_stuffing",
    "forcemeat",
    "bread",
    "anadama_bread",
    "bap",
    "barmbrack",
    "breadstick",
    "grissino",
    "brown_bread",
    "bun",
    "tea_bread",
    "caraway_seed_bread",
    "challah",
    "cinnamon_bread",
    "cracked-wheat_bread",
    "cracker",
    "crouton",
    "dark_bread",
    "English_muffin",
    "flatbread",
    "garlic_bread",
    "gluten_bread",
    "graham_bread",
    "Host",
    "flatbrod",
    "bannock",
    "chapatti",
    "pita",
    "loaf_of_bread",
    "French_loaf",
    "matzo",
    "nan",
    "onion_bread",
    "raisin_bread",
    "quick_bread",
    "banana_bread",
    "date_bread",
    "date-nut_bread",
    "nut_bread",
    "oatcake",
    "Irish_soda_bread",
    "skillet_bread",
    "rye_bread",
    "black_bread",
    "Jewish_rye_bread",
    "limpa",
    "Swedish_rye_bread",
    "salt-rising_bread",
    "simnel",
    "sour_bread",
    "toast",
    "wafer",
    "white_bread",
    "baguet",
    "French_bread",
    "Italian_bread",
    "cornbread",
    "corn_cake",
    "skillet_corn_bread",
    "ashcake",
    "hoecake",
    "cornpone",
    "corn_dab",
    "hush_puppy",
    "johnnycake",
    "Shawnee_cake",
    "spoon_bread",
    "cinnamon_toast",
    "orange_toast",
    "Melba_toast",
    "zwieback",
    "frankfurter_bun",
    "hamburger_bun",
    "muffin",
    "bran_muffin",
    "corn_muffin",
    "Yorkshire_pudding",
    "popover",
    "scone",
    "drop_scone",
    "cross_bun",
    "brioche",
    "crescent_roll",
    "hard_roll",
    "soft_roll",
    "kaiser_roll",
    "Parker_House_roll",
    "clover-leaf_roll",
    "onion_roll",
    "bialy",
    "sweet_roll",
    "bear_claw",
    "cinnamon_roll",
    "honey_bun",
    "pinwheel_roll",
    "danish",
    "bagel",
    "onion_bagel",
    "biscuit",
    "rolled_biscuit",
    "baking-powder_biscuit",
    "buttermilk_biscuit",
    "shortcake",
    "hardtack",
    "saltine",
    "soda_cracker",
    "oyster_cracker",
    "water_biscuit",
    "graham_cracker",
    "pretzel",
    "soft_pretzel",
    "sandwich",
    "sandwich_plate",
    "butty",
    "ham_sandwich",
    "chicken_sandwich",
    "club_sandwich",
    "open-face_sandwich",
    "hamburger",
    "cheeseburger",
    "tunaburger",
    "hotdog",
    "Sloppy_Joe",
    "bomber",
    "gyro",
    "bacon-lettuce-tomato_sandwich",
    "Reuben",
    "western",
    "wrap",
    "spaghetti",
    "hasty_pudding",
    "gruel",
    "congee",
    "skilly",
    "edible_fruit",
    "vegetable",
    "julienne",
    "raw_vegetable",
    "crudites",
    "celery_stick",
    "legume",
    "pulse",
    "potherb",
    "greens",
    "chop-suey_greens",
    "bean_curd",
    "solanaceous_vegetable",
    "root_vegetable",
    "potato",
    "baked_potato",
    "french_fries",
    "home_fries",
    "jacket_potato",
    "mashed_potato",
    "potato_skin",
    "Uruguay_potato",
    "yam",
    "sweet_potato",
    "yam",
    "snack_food",
    "chip",
    "corn_chip",
    "tortilla_chip",
    "nacho",
    "eggplant",
    "pieplant",
    "cruciferous_vegetable",
    "mustard",
    "cabbage",
    "kale",
    "collards",
    "Chinese_cabbage",
    "bok_choy",
    "head_cabbage",
    "red_cabbage",
    "savoy_cabbage",
    "broccoli",
    "cauliflower",
    "brussels_sprouts",
    "broccoli_rabe",
    "squash",
    "summer_squash",
    "yellow_squash",
    "crookneck",
    "zucchini",
    "marrow",
    "cocozelle",
    "pattypan_squash",
    "spaghetti_squash",
    "winter_squash",
    "acorn_squash",
    "butternut_squash",
    "hubbard_squash",
    "turban_squash",
    "buttercup_squash",
    "cushaw",
    "winter_crookneck_squash",
    "cucumber",
    "gherkin",
    "artichoke",
    "artichoke_heart",
    "Jerusalem_artichoke",
    "asparagus",
    "bamboo_shoot",
    "sprout",
    "bean_sprout",
    "alfalfa_sprout",
    "beet",
    "beet_green",
    "sugar_beet",
    "mangel-wurzel",
    "chard",
    "pepper",
    "sweet_pepper",
    "bell_pepper",
    "green_pepper",
    "globe_pepper",
    "pimento",
    "hot_pepper",
    "chili",
    "jalapeno",
    "chipotle",
    "cayenne",
    "tabasco",
    "onion",
    "Bermuda_onion",
    "green_onion",
    "Vidalia_onion",
    "Spanish_onion",
    "purple_onion",
    "leek",
    "shallot",
    "salad_green",
    "lettuce",
    "butterhead_lettuce",
    "buttercrunch",
    "Bibb_lettuce",
    "Boston_lettuce",
    "crisphead_lettuce",
    "cos",
    "leaf_lettuce",
    "celtuce",
    "bean",
    "goa_bean",
    "lentil",
    "pea",
    "green_pea",
    "marrowfat_pea",
    "snow_pea",
    "sugar_snap_pea",
    "split-pea",
    "chickpea",
    "cajan_pea",
    "field_pea",
    "mushy_peas",
    "black-eyed_pea",
    "common_bean",
    "kidney_bean",
    "navy_bean",
    "pinto_bean",
    "frijole",
    "black_bean",
    "fresh_bean",
    "flageolet",
    "green_bean",
    "snap_bean",
    "string_bean",
    "Kentucky_wonder",
    "scarlet_runner",
    "haricot_vert",
    "wax_bean",
    "shell_bean",
    "lima_bean",
    "Fordhooks",
    "sieva_bean",
    "fava_bean",
    "soy",
    "green_soybean",
    "field_soybean",
    "cardoon",
    "carrot",
    "carrot_stick",
    "celery",
    "pascal_celery",
    "celeriac",
    "chicory",
    "radicchio",
    "coffee_substitute",
    "chicory",
    "Postum",
    "chicory_escarole",
    "Belgian_endive",
    "corn",
    "sweet_corn",
    "hominy",
    "lye_hominy",
    "pearl_hominy",
    "popcorn",
    "cress",
    "watercress",
    "garden_cress",
    "winter_cress",
    "dandelion_green",
    "gumbo",
    "kohlrabi",
    "lamb's-quarter",
    "wild_spinach",
    "tomato",
    "beefsteak_tomato",
    "cherry_tomato",
    "plum_tomato",
    "tomatillo",
    "mushroom",
    "stuffed_mushroom",
    "salsify",
    "oyster_plant",
    "scorzonera",
    "parsnip",
    "pumpkin",
    "radish",
    "turnip",
    "white_turnip",
    "rutabaga",
    "turnip_greens",
    "sorrel",
    "French_sorrel",
    "spinach",
    "taro",
    "truffle",
    "edible_nut",
    "bunya_bunya",
    "peanut",
    "freestone",
    "cling",
    "windfall",
    "apple",
    "crab_apple",
    "eating_apple",
    "Baldwin",
    "Cortland",
    "Cox's_Orange_Pippin",
    "Delicious",
    "Golden_Delicious",
    "Red_Delicious",
    "Empire",
    "Grimes'_golden",
    "Jonathan",
    "McIntosh",
    "Macoun",
    "Northern_Spy",
    "Pearmain",
    "Pippin",
    "Prima",
    "Stayman",
    "Winesap",
    "Stayman_Winesap",
    "cooking_apple",
    "Bramley's_Seedling",
    "Granny_Smith",
    "Lane's_Prince_Albert",
    "Newtown_Wonder",
    "Rome_Beauty",
    "berry",
    "bilberry",
    "huckleberry",
    "blueberry",
    "wintergreen",
    "cranberry",
    "lingonberry",
    "currant",
    "gooseberry",
    "black_currant",
    "red_currant",
    "blackberry",
    "boysenberry",
    "dewberry",
    "loganberry",
    "raspberry",
    "saskatoon",
    "strawberry",
    "sugarberry",
    "persimmon",
    "acerola",
    "carambola",
    "ceriman",
    "carissa_plum",
    "citrus",
    "orange",
    "temple_orange",
    "mandarin",
    "clementine",
    "satsuma",
    "tangerine",
    "tangelo",
    "bitter_orange",
    "sweet_orange",
    "Jaffa_orange",
    "navel_orange",
    "Valencia_orange",
    "kumquat",
    "lemon",
    "lime",
    "key_lime",
    "grapefruit",
    "pomelo",
    "citrange",
    "citron",
    "almond",
    "Jordan_almond",
    "apricot",
    "peach",
    "nectarine",
    "pitahaya",
    "plum",
    "damson",
    "greengage",
    "beach_plum",
    "sloe",
    "Victoria_plum",
    "dried_fruit",
    "dried_apricot",
    "prune",
    "raisin",
    "seedless_raisin",
    "seeded_raisin",
    "currant",
    "fig",
    "pineapple",
    "anchovy_pear",
    "banana",
    "passion_fruit",
    "granadilla",
    "sweet_calabash",
    "bell_apple",
    "breadfruit",
    "jackfruit",
    "cacao_bean",
    "cocoa",
    "canistel",
    "melon",
    "melon_ball",
    "muskmelon",
    "cantaloup",
    "winter_melon",
    "honeydew",
    "Persian_melon",
    "net_melon",
    "casaba",
    "watermelon",
    "cherry",
    "sweet_cherry",
    "bing_cherry",
    "heart_cherry",
    "blackheart",
    "capulin",
    "sour_cherry",
    "amarelle",
    "morello",
    "cocoa_plum",
    "gherkin",
    "grape",
    "fox_grape",
    "Concord_grape",
    "Catawba",
    "muscadine",
    "scuppernong",
    "slipskin_grape",
    "vinifera_grape",
    "emperor",
    "muscat",
    "ribier",
    "sultana",
    "Tokay",
    "flame_tokay",
    "Thompson_Seedless",
    "custard_apple",
    "cherimoya",
    "soursop",
    "sweetsop",
    "ilama",
    "pond_apple",
    "papaw",
    "papaya",
    "kai_apple",
    "ketembilla",
    "ackee",
    "durian",
    "feijoa",
    "genip",
    "genipap",
    "kiwi",
    "loquat",
    "mangosteen",
    "mango",
    "sapodilla",
    "sapote",
    "tamarind",
    "avocado",
    "date",
    "elderberry",
    "guava",
    "mombin",
    "hog_plum",
    "hog_plum",
    "jaboticaba",
    "jujube",
    "litchi",
    "longanberry",
    "mamey",
    "marang",
    "medlar",
    "medlar",
    "mulberry",
    "olive",
    "black_olive",
    "green_olive",
    "pear",
    "bosc",
    "anjou",
    "bartlett",
    "seckel",
    "plantain",
    "plumcot",
    "pomegranate",
    "prickly_pear",
    "Barbados_gooseberry",
    "quandong",
    "quandong_nut",
    "quince",
    "rambutan",
    "pulasan",
    "rose_apple",
    "sorb",
    "sour_gourd",
    "edible_seed",
    "pumpkin_seed",
    "betel_nut",
    "beechnut",
    "walnut",
    "black_walnut",
    "English_walnut",
    "brazil_nut",
    "butternut",
    "souari_nut",
    "cashew",
    "chestnut",
    "chincapin",
    "hazelnut",
    "coconut",
    "coconut_milk",
    "grugru_nut",
    "hickory_nut",
    "cola_extract",
    "macadamia_nut",
    "pecan",
    "pine_nut",
    "pistachio",
    "sunflower_seed",
    "anchovy_paste",
    "rollmops",
    "feed",
    "cattle_cake",
    "creep_feed",
    "fodder",
    "feed_grain",
    "eatage",
    "silage",
    "oil_cake",
    "oil_meal",
    "alfalfa",
    "broad_bean",
    "hay",
    "timothy",
    "stover",
    "grain",
    "grist",
    "groats",
    "millet",
    "barley",
    "pearl_barley",
    "buckwheat",
    "bulgur",
    "wheat",
    "cracked_wheat",
    "stodge",
    "wheat_germ",
    "oat",
    "rice",
    "brown_rice",
    "white_rice",
    "wild_rice",
    "paddy",
    "slop",
    "mash",
    "chicken_feed",
    "cud",
    "bird_feed",
    "petfood",
    "dog_food",
    "cat_food",
    "canary_seed",
    "salad",
    "tossed_salad",
    "green_salad",
    "Caesar_salad",
    "salmagundi",
    "salad_nicoise",
    "combination_salad",
    "chef's_salad",
    "potato_salad",
    "pasta_salad",
    "macaroni_salad",
    "fruit_salad",
    "Waldorf_salad",
    "crab_Louis",
    "herring_salad",
    "tuna_fish_salad",
    "chicken_salad",
    "coleslaw",
    "aspic",
    "molded_salad",
    "tabbouleh",
    "ingredient",
    "flavorer",
    "bouillon_cube",
    "condiment",
    "herb",
    "fines_herbes",
    "spice",
    "spearmint_oil",
    "lemon_oil",
    "wintergreen_oil",
    "salt",
    "celery_salt",
    "onion_salt",
    "seasoned_salt",
    "sour_salt",
    "five_spice_powder",
    "allspice",
    "cinnamon",
    "stick_cinnamon",
    "clove",
    "cumin",
    "fennel",
    "ginger",
    "ginger",
    "mace",
    "nutmeg",
    "pepper",
    "black_pepper",
    "white_pepper",
    "sassafras",
    "basil",
    "bay_leaf",
    "borage",
    "hyssop",
    "caraway",
    "chervil",
    "chives",
    "comfrey",
    "coriander",
    "coriander",
    "costmary",
    "fennel",
    "fennel",
    "fennel_seed",
    "fenugreek",
    "garlic",
    "clove",
    "garlic_chive",
    "lemon_balm",
    "lovage",
    "marjoram",
    "mint",
    "mustard_seed",
    "mustard",
    "Chinese_mustard",
    "nasturtium",
    "parsley",
    "salad_burnet",
    "rosemary",
    "rue",
    "sage",
    "clary_sage",
    "savory",
    "summer_savory",
    "winter_savory",
    "sweet_woodruff",
    "sweet_cicely",
    "tarragon",
    "thyme",
    "turmeric",
    "caper",
    "catsup",
    "cardamom",
    "cayenne",
    "chili_powder",
    "chili_sauce",
    "chutney",
    "steak_sauce",
    "taco_sauce",
    "salsa",
    "mint_sauce",
    "cranberry_sauce",
    "curry_powder",
    "curry",
    "lamb_curry",
    "duck_sauce",
    "horseradish",
    "marinade",
    "paprika",
    "Spanish_paprika",
    "pickle",
    "dill_pickle",
    "bread_and_butter_pickle",
    "pickle_relish",
    "piccalilli",
    "sweet_pickle",
    "applesauce",
    "soy_sauce",
    "Tabasco",
    "tomato_paste",
    "angelica",
    "angelica",
    "almond_extract",
    "anise",
    "Chinese_anise",
    "juniper_berries",
    "saffron",
    "sesame_seed",
    "caraway_seed",
    "poppy_seed",
    "dill",
    "dill_seed",
    "celery_seed",
    "lemon_extract",
    "monosodium_glutamate",
    "vanilla_bean",
    "vinegar",
    "cider_vinegar",
    "wine_vinegar",
    "sauce",
    "anchovy_sauce",
    "hot_sauce",
    "hard_sauce",
    "horseradish_sauce",
    "bolognese_pasta_sauce",
    "carbonara",
    "tomato_sauce",
    "tartare_sauce",
    "wine_sauce",
    "marchand_de_vin",
    "bread_sauce",
    "plum_sauce",
    "peach_sauce",
    "apricot_sauce",
    "pesto",
    "ravigote",
    "remoulade_sauce",
    "dressing",
    "sauce_Louis",
    "bleu_cheese_dressing",
    "blue_cheese_dressing",
    "French_dressing",
    "Lorenzo_dressing",
    "anchovy_dressing",
    "Italian_dressing",
    "half-and-half_dressing",
    "mayonnaise",
    "green_mayonnaise",
    "aioli",
    "Russian_dressing",
    "salad_cream",
    "Thousand_Island_dressing",
    "barbecue_sauce",
    "hollandaise",
    "bearnaise",
    "Bercy",
    "bordelaise",
    "bourguignon",
    "brown_sauce",
    "Espagnole",
    "Chinese_brown_sauce",
    "blanc",
    "cheese_sauce",
    "chocolate_sauce",
    "hot-fudge_sauce",
    "cocktail_sauce",
    "Colbert",
    "white_sauce",
    "cream_sauce",
    "Mornay_sauce",
    "demiglace",
    "gravy",
    "gravy",
    "spaghetti_sauce",
    "marinara",
    "mole",
    "hunter's_sauce",
    "mushroom_sauce",
    "mustard_sauce",
    "Nantua",
    "Hungarian_sauce",
    "pepper_sauce",
    "roux",
    "Smitane",
    "Soubise",
    "Lyonnaise_sauce",
    "veloute",
    "allemande",
    "caper_sauce",
    "poulette",
    "curry_sauce",
    "Worcester_sauce",
    "coconut_milk",
    "egg",
    "egg_white",
    "egg_yolk",
    "boiled_egg",
    "hard-boiled_egg",
    "Easter_egg",
    "Easter_egg",
    "chocolate_egg",
    "candy_egg",
    "poached_egg",
    "scrambled_eggs",
    "deviled_egg",
    "shirred_egg",
    "omelet",
    "firm_omelet",
    "French_omelet",
    "fluffy_omelet",
    "western_omelet",
    "souffle",
    "fried_egg",
    "dairy_product",
    "milk",
    "milk",
    "sour_milk",
    "soya_milk",
    "formula",
    "pasteurized_milk",
    "cows'_milk",
    "yak's_milk",
    "goats'_milk",
    "acidophilus_milk",
    "raw_milk",
    "scalded_milk",
    "homogenized_milk",
    "certified_milk",
    "powdered_milk",
    "nonfat_dry_milk",
    "evaporated_milk",
    "condensed_milk",
    "skim_milk",
    "semi-skimmed_milk",
    "whole_milk",
    "low-fat_milk",
    "buttermilk",
    "cream",
    "clotted_cream",
    "double_creme",
    "half-and-half",
    "heavy_cream",
    "light_cream",
    "sour_cream",
    "whipping_cream",
    "butter",
    "clarified_butter",
    "ghee",
    "brown_butter",
    "Meuniere_butter",
    "yogurt",
    "blueberry_yogurt",
    "raita",
    "whey",
    "curd",
    "curd",
    "clabber",
    "cheese",
    "paring",
    "cream_cheese",
    "double_cream",
    "mascarpone",
    "triple_cream",
    "cottage_cheese",
    "process_cheese",
    "bleu",
    "Stilton",
    "Roquefort",
    "gorgonzola",
    "Danish_blue",
    "Bavarian_blue",
    "Brie",
    "brick_cheese",
    "Camembert",
    "cheddar",
    "rat_cheese",
    "Cheshire_cheese",
    "double_Gloucester",
    "Edam",
    "goat_cheese",
    "Gouda",
    "grated_cheese",
    "hand_cheese",
    "Liederkranz",
    "Limburger",
    "mozzarella",
    "Muenster",
    "Parmesan",
    "quark_cheese",
    "ricotta",
    "string_cheese",
    "Swiss_cheese",
    "Emmenthal",
    "Gruyere",
    "sapsago",
    "Velveeta",
    "nut_butter",
    "peanut_butter",
    "marshmallow_fluff",
    "onion_butter",
    "pimento_butter",
    "shrimp_butter",
    "lobster_butter",
    "yak_butter",
    "spread",
    "cheese_spread",
    "anchovy_butter",
    "fishpaste",
    "garlic_butter",
    "miso",
    "wasabi",
    "snail_butter",
    "hummus",
    "pate",
    "duck_pate",
    "foie_gras",
    "tapenade",
    "tahini",
    "sweetening",
    "aspartame",
    "honey",
    "saccharin",
    "sugar",
    "syrup",
    "sugar_syrup",
    "molasses",
    "sorghum",
    "treacle",
    "grenadine",
    "maple_syrup",
    "corn_syrup",
    "miraculous_food",
    "batter",
    "dough",
    "bread_dough",
    "pancake_batter",
    "fritter_batter",
    "coq_au_vin",
    "chicken_provencale",
    "chicken_and_rice",
    "moo_goo_gai_pan",
    "arroz_con_pollo",
    "bacon_and_eggs",
    "barbecued_spareribs",
    "beef_Bourguignonne",
    "beef_Wellington",
    "bitok",
    "boiled_dinner",
    "Boston_baked_beans",
    "bubble_and_squeak",
    "pasta",
    "cannelloni",
    "carbonnade_flamande",
    "cheese_souffle",
    "chicken_Marengo",
    "chicken_cordon_bleu",
    "Maryland_chicken",
    "chicken_paprika",
    "chicken_Tetrazzini",
    "Tetrazzini",
    "chicken_Kiev",
    "chili",
    "chili_dog",
    "chop_suey",
    "chow_mein",
    "codfish_ball",
    "coquille",
    "coquilles_Saint-Jacques",
    "croquette",
    "cottage_pie",
    "rissole",
    "dolmas",
    "egg_foo_yong",
    "egg_roll",
    "eggs_Benedict",
    "enchilada",
    "falafel",
    "fish_and_chips",
    "fondue",
    "cheese_fondue",
    "chocolate_fondue",
    "fondue",
    "beef_fondue",
    "French_toast",
    "fried_rice",
    "frittata",
    "frog_legs",
    "galantine",
    "gefilte_fish",
    "haggis",
    "ham_and_eggs",
    "hash",
    "corned_beef_hash",
    "jambalaya",
    "kabob",
    "kedgeree",
    "souvlaki",
    "lasagna",
    "seafood_Newburg",
    "lobster_Newburg",
    "shrimp_Newburg",
    "Newburg_sauce",
    "lobster_thermidor",
    "lutefisk",
    "macaroni_and_cheese",
    "macedoine",
    "meatball",
    "porcupine_ball",
    "Swedish_meatball",
    "meat_loaf",
    "moussaka",
    "osso_buco",
    "marrow",
    "pheasant_under_glass",
    "pigs_in_blankets",
    "pilaf",
    "bulgur_pilaf",
    "pizza",
    "sausage_pizza",
    "pepperoni_pizza",
    "cheese_pizza",
    "anchovy_pizza",
    "Sicilian_pizza",
    "poi",
    "pork_and_beans",
    "porridge",
    "oatmeal",
    "loblolly",
    "potpie",
    "rijsttaffel",
    "risotto",
    "roulade",
    "fish_loaf",
    "salmon_loaf",
    "Salisbury_steak",
    "sauerbraten",
    "sauerkraut",
    "scallopine",
    "veal_scallopini",
    "scampi",
    "Scotch_egg",
    "Scotch_woodcock",
    "scrapple",
    "spaghetti_and_meatballs",
    "Spanish_rice",
    "steak_tartare",
    "pepper_steak",
    "steak_au_poivre",
    "beef_Stroganoff",
    "stuffed_cabbage",
    "kishke",
    "stuffed_peppers",
    "stuffed_tomato",
    "stuffed_tomato",
    "succotash",
    "sukiyaki",
    "sashimi",
    "sushi",
    "Swiss_steak",
    "tamale",
    "tamale_pie",
    "tempura",
    "teriyaki",
    "terrine",
    "Welsh_rarebit",
    "schnitzel",
    "taco",
    "chicken_taco",
    "burrito",
    "beef_burrito",
    "quesadilla",
    "tostada",
    "bean_tostada",
    "refried_beans",
    "beverage",
    "wish-wash",
    "concoction",
    "mix",
    "filling",
    "lekvar",
    "potion",
    "elixir",
    "elixir_of_life",
    "philter",
    "alcohol",
    "proof_spirit",
    "home_brew",
    "hooch",
    "kava",
    "aperitif",
    "brew",
    "beer",
    "draft_beer",
    "suds",
    "Munich_beer",
    "bock",
    "lager",
    "light_beer",
    "Oktoberfest",
    "Pilsner",
    "shebeen",
    "Weissbier",
    "Weizenbock",
    "malt",
    "wort",
    "malt",
    "ale",
    "bitter",
    "Burton",
    "pale_ale",
    "porter",
    "stout",
    "Guinness",
    "kvass",
    "mead",
    "metheglin",
    "hydromel",
    "oenomel",
    "near_beer",
    "ginger_beer",
    "sake",
    "wine",
    "vintage",
    "red_wine",
    "white_wine",
    "blush_wine",
    "altar_wine",
    "sparkling_wine",
    "champagne",
    "cold_duck",
    "Burgundy",
    "Beaujolais",
    "Medoc",
    "Canary_wine",
    "Chablis",
    "Montrachet",
    "Chardonnay",
    "Pinot_noir",
    "Pinot_blanc",
    "Bordeaux",
    "claret",
    "Chianti",
    "Cabernet",
    "Merlot",
    "Sauvignon_blanc",
    "California_wine",
    "Cotes_de_Provence",
    "dessert_wine",
    "Dubonnet",
    "jug_wine",
    "macon",
    "Moselle",
    "Muscadet",
    "plonk",
    "retsina",
    "Rhine_wine",
    "Riesling",
    "liebfraumilch",
    "Rhone_wine",
    "Rioja",
    "sack",
    "Saint_Emilion",
    "Soave",
    "zinfandel",
    "Sauterne",
    "straw_wine",
    "table_wine",
    "Tokay",
    "vin_ordinaire",
    "vermouth",
    "sweet_vermouth",
    "dry_vermouth",
    "Chenin_blanc",
    "Verdicchio",
    "Vouvray",
    "Yquem",
    "generic",
    "varietal",
    "fortified_wine",
    "Madeira",
    "malmsey",
    "port",
    "sherry",
    "Marsala",
    "muscat",
    "liquor",
    "neutral_spirits",
    "aqua_vitae",
    "eau_de_vie",
    "moonshine",
    "bathtub_gin",
    "aquavit",
    "arrack",
    "bitters",
    "brandy",
    "applejack",
    "Calvados",
    "Armagnac",
    "Cognac",
    "grappa",
    "kirsch",
    "slivovitz",
    "gin",
    "sloe_gin",
    "geneva",
    "grog",
    "ouzo",
    "rum",
    "demerara",
    "Jamaica_rum",
    "schnapps",
    "pulque",
    "mescal",
    "tequila",
    "vodka",
    "whiskey",
    "blended_whiskey",
    "bourbon",
    "corn_whiskey",
    "firewater",
    "Irish",
    "poteen",
    "rye",
    "Scotch",
    "sour_mash",
    "liqueur",
    "absinth",
    "amaretto",
    "anisette",
    "benedictine",
    "Chartreuse",
    "coffee_liqueur",
    "creme_de_cacao",
    "creme_de_menthe",
    "creme_de_fraise",
    "Drambuie",
    "Galliano",
    "orange_liqueur",
    "curacao",
    "triple_sec",
    "Grand_Marnier",
    "kummel",
    "maraschino",
    "pastis",
    "Pernod",
    "pousse-cafe",
    "Kahlua",
    "ratafia",
    "sambuca",
    "mixed_drink",
    "cocktail",
    "Dom_Pedro",
    "highball",
    "mixer",
    "bishop",
    "Bloody_Mary",
    "Virgin_Mary",
    "bullshot",
    "cobbler",
    "collins",
    "cooler",
    "refresher",
    "smoothie",
    "daiquiri",
    "strawberry_daiquiri",
    "NADA_daiquiri",
    "spritzer",
    "flip",
    "gimlet",
    "gin_and_tonic",
    "grasshopper",
    "Harvey_Wallbanger",
    "julep",
    "manhattan",
    "Rob_Roy",
    "margarita",
    "martini",
    "gin_and_it",
    "vodka_martini",
    "old_fashioned",
    "pink_lady",
    "Sazerac",
    "screwdriver",
    "sidecar",
    "Scotch_and_soda",
    "sling",
    "brandy_sling",
    "gin_sling",
    "rum_sling",
    "sour",
    "whiskey_sour",
    "stinger",
    "swizzle",
    "hot_toddy",
    "zombie",
    "fizz",
    "Irish_coffee",
    "cafe_au_lait",
    "cafe_noir",
    "decaffeinated_coffee",
    "drip_coffee",
    "espresso",
    "caffe_latte",
    "cappuccino",
    "iced_coffee",
    "instant_coffee",
    "mocha",
    "mocha",
    "cassareep",
    "Turkish_coffee",
    "chocolate_milk",
    "cider",
    "hard_cider",
    "scrumpy",
    "sweet_cider",
    "mulled_cider",
    "perry",
    "rotgut",
    "slug",
    "cocoa",
    "criollo",
    "juice",
    "fruit_juice",
    "nectar",
    "apple_juice",
    "cranberry_juice",
    "grape_juice",
    "must",
    "grapefruit_juice",
    "orange_juice",
    "frozen_orange_juice",
    "pineapple_juice",
    "lemon_juice",
    "lime_juice",
    "papaya_juice",
    "tomato_juice",
    "carrot_juice",
    "V-8_juice",
    "koumiss",
    "fruit_drink",
    "lemonade",
    "limeade",
    "orangeade",
    "malted_milk",
    "mate",
    "mulled_wine",
    "negus",
    "soft_drink",
    "pop",
    "birch_beer",
    "bitter_lemon",
    "cola",
    "cream_soda",
    "egg_cream",
    "ginger_ale",
    "orange_soda",
    "phosphate",
    "Coca_Cola",
    "Pepsi",
    "root_beer",
    "sarsaparilla",
    "tonic",
    "coffee_bean",
    "coffee",
    "cafe_royale",
    "fruit_punch",
    "milk_punch",
    "mimosa",
    "pina_colada",
    "punch",
    "cup",
    "champagne_cup",
    "claret_cup",
    "wassail",
    "planter's_punch",
    "White_Russian",
    "fish_house_punch",
    "May_wine",
    "eggnog",
    "cassiri",
    "spruce_beer",
    "rickey",
    "gin_rickey",
    "tea",
    "tea_bag",
    "tea",
    "tea-like_drink",
    "cambric_tea",
    "cuppa",
    "herb_tea",
    "tisane",
    "camomile_tea",
    "ice_tea",
    "sun_tea",
    "black_tea",
    "congou",
    "Darjeeling",
    "orange_pekoe",
    "souchong",
    "green_tea",
    "hyson",
    "oolong",
    "water",
    "bottled_water",
    "branch_water",
    "spring_water",
    "sugar_water",
    "drinking_water",
    "ice_water",
    "soda_water",
    "mineral_water",
    "seltzer",
    "Vichy_water",
    "perishable",
    "couscous",
    "ramekin",
    "multivitamin",
    "vitamin_pill",
    "soul_food",
    "mold",
    "people",
    "collection",
    "book",
    "library",
    "baseball_club",
    "crowd",
    "class",
    "core",
    "concert_band",
    "dance",
    "wedding",
    "chain",
    "power_breakfast",
    "aerie",
    "agora",
    "amusement_park",
    "aphelion",
    "apron",
    "interplanetary_space",
    "interstellar_space",
    "intergalactic_space",
    "bush",
    "semidesert",
    "beam-ends",
    "bridgehead",
    "bus_stop",
    "campsite",
    "detention_basin",
    "cemetery",
    "trichion",
    "city",
    "business_district",
    "outskirts",
    "borough",
    "cow_pasture",
    "crest",
    "eparchy",
    "suburb",
    "stockbroker_belt",
    "crawlspace",
    "sheikdom",
    "residence",
    "domicile",
    "dude_ranch",
    "farmland",
    "midfield",
    "firebreak",
    "flea_market",
    "battlefront",
    "garbage_heap",
    "benthos",
    "goldfield",
    "grainfield",
    "half-mast",
    "hemline",
    "heronry",
    "hipline",
    "hipline",
    "hole-in-the-wall",
    "junkyard",
    "isoclinic_line",
    "littoral",
    "magnetic_pole",
    "grassland",
    "mecca",
    "observer's_meridian",
    "prime_meridian",
    "nombril",
    "no-parking_zone",
    "outdoors",
    "fairground",
    "pasture",
    "perihelion",
    "periselene",
    "locus_of_infection",
    "kasbah",
    "waterfront",
    "resort",
    "resort_area",
    "rough",
    "ashram",
    "harborage",
    "scrubland",
    "weald",
    "wold",
    "schoolyard",
    "showplace",
    "bedside",
    "sideline",
    "ski_resort",
    "soil_horizon",
    "geological_horizon",
    "coal_seam",
    "coalface",
    "field",
    "oilfield",
    "Temperate_Zone",
    "terreplein",
    "three-mile_limit",
    "desktop",
    "top",
    "kampong",
    "subtropics",
    "barrio",
    "veld",
    "vertex",
    "waterline",
    "high-water_mark",
    "low-water_mark",
    "continental_divide",
    "zodiac",
    "Aegean_island",
    "sultanate",
    "Swiss_canton",
    "abyssal_zone",
    "aerie",
    "air_bubble",
    "alluvial_flat",
    "alp",
    "Alpine_glacier",
    "anthill",
    "aquifer",
    "archipelago",
    "arete",
    "arroyo",
    "ascent",
    "asterism",
    "asthenosphere",
    "atoll",
    "bank",
    "bank",
    "bar",
    "barbecue_pit",
    "barrier_reef",
    "baryon",
    "basin",
    "beach",
    "honeycomb",
    "belay",
    "ben",
    "berm",
    "bladder_stone",
    "bluff",
    "borrow_pit",
    "brae",
    "bubble",
    "burrow",
    "butte",
    "caldera",
    "canyon",
    "canyonside",
    "cave",
    "cavern",
    "chasm",
    "cirque",
    "cliff",
    "cloud",
    "coast",
    "coastland",
    "col",
    "collector",
    "comet",
    "continental_glacier",
    "coral_reef",
    "cove",
    "crag",
    "crater",
    "cultivated_land",
    "dale",
    "defile",
    "delta",
    "descent",
    "diapir",
    "divot",
    "divot",
    "down",
    "downhill",
    "draw",
    "drey",
    "drumlin",
    "dune",
    "escarpment",
    "esker",
    "fireball",
    "flare_star",
    "floor",
    "fomite",
    "foothill",
    "footwall",
    "foreland",
    "foreshore",
    "gauge_boson",
    "geological_formation",
    "geyser",
    "glacier",
    "glen",
    "gopher_hole",
    "gorge",
    "grotto",
    "growler",
    "gulch",
    "gully",
    "hail",
    "highland",
    "hill",
    "hillside",
    "hole",
    "hollow",
    "hot_spring",
    "iceberg",
    "icecap",
    "ice_field",
    "ice_floe",
    "ice_mass",
    "inclined_fault",
    "ion",
    "isthmus",
    "kidney_stone",
    "knoll",
    "kopje",
    "Kuiper_belt",
    "lake_bed",
    "lakefront",
    "lakeside",
    "landfall",
    "landfill",
    "lather",
    "leak",
    "ledge",
    "lepton",
    "lithosphere",
    "lowland",
    "lunar_crater",
    "maar",
    "massif",
    "meander",
    "mesa",
    "meteorite",
    "microfossil",
    "midstream",
    "molehill",
    "monocline",
    "mountain",
    "mountainside",
    "mouth",
    "mull",
    "natural_depression",
    "natural_elevation",
    "nullah",
    "ocean",
    "ocean_floor",
    "oceanfront",
    "outcrop",
    "oxbow",
    "pallasite",
    "perforation",
    "photosphere",
    "piedmont",
    "Piedmont_glacier",
    "pinetum",
    "plage",
    "plain",
    "point",
    "polar_glacier",
    "pothole",
    "precipice",
    "promontory",
    "ptyalith",
    "pulsar",
    "quicksand",
    "rabbit_burrow",
    "radiator",
    "rainbow",
    "range",
    "rangeland",
    "ravine",
    "reef",
    "ridge",
    "ridge",
    "rift_valley",
    "riparian_forest",
    "ripple_mark",
    "riverbank",
    "riverbed",
    "rock",
    "roof",
    "saltpan",
    "sandbank",
    "sandbar",
    "sandpit",
    "sanitary_landfill",
    "sawpit",
    "scablands",
    "seashore",
    "seaside",
    "seif_dune",
    "shell",
    "shiner",
    "shoal",
    "shore",
    "shoreline",
    "sinkhole",
    "ski_slope",
    "sky",
    "slope",
    "snowcap",
    "snowdrift",
    "snowfield",
    "soapsuds",
    "spit",
    "spoor",
    "spume",
    "star",
    "steep",
    "steppe",
    "strand",
    "streambed",
    "sun",
    "supernova",
    "swale",
    "swamp",
    "swell",
    "tableland",
    "talus",
    "tangle",
    "tar_pit",
    "terrace",
    "tidal_basin",
    "tideland",
    "tor",
    "tor",
    "Trapezium",
    "troposphere",
    "tundra",
    "twinkler",
    "uphill",
    "urolith",
    "valley",
    "vehicle-borne_transmission",
    "vein",
    "volcanic_crater",
    "volcano",
    "wadi",
    "wall",
    "warren",
    "wasp's_nest",
    "watercourse",
    "waterside",
    "water_table",
    "whinstone",
    "wormcast",
    "xenolith",
    "Circe",
    "gryphon",
    "spiritual_leader",
    "messiah",
    "Rhea_Silvia",
    "number_one",
    "adventurer",
    "anomaly",
    "appointee",
    "argonaut",
    "Ashkenazi",
    "benefactor",
    "color-blind_person",
    "commoner",
    "conservator",
    "contrarian",
    "contadino",
    "contestant",
    "cosigner",
    "discussant",
    "enologist",
    "entertainer",
    "eulogist",
    "ex-gambler",
    "experimenter",
    "experimenter",
    "exponent",
    "ex-president",
    "face",
    "female",
    "finisher",
    "inhabitant",
    "native",
    "native",
    "juvenile",
    "lover",
    "male",
    "mediator",
    "mediatrix",
    "national",
    "peer",
    "prize_winner",
    "recipient",
    "religionist",
    "sensualist",
    "traveler",
    "unwelcome_person",
    "unskilled_person",
    "worker",
    "wrongdoer",
    "Black_African",
    "Afrikaner",
    "Aryan",
    "Black",
    "Black_woman",
    "mulatto",
    "White",
    "Circassian",
    "Semite",
    "Chaldean",
    "Elamite",
    "white_man",
    "WASP",
    "gook",
    "Mongol",
    "Tatar",
    "Nahuatl",
    "Aztec",
    "Olmec",
    "Biloxi",
    "Blackfoot",
    "Brule",
    "Caddo",
    "Cheyenne",
    "Chickasaw",
    "Cocopa",
    "Comanche",
    "Creek",
    "Delaware",
    "Diegueno",
    "Esselen",
    "Eyeish",
    "Havasupai",
    "Hunkpapa",
    "Iowa",
    "Kalapooia",
    "Kamia",
    "Kekchi",
    "Kichai",
    "Kickapoo",
    "Kiliwa",
    "Malecite",
    "Maricopa",
    "Mohican",
    "Muskhogean",
    "Navaho",
    "Nootka",
    "Oglala",
    "Osage",
    "Oneida",
    "Paiute",
    "Passamaquody",
    "Penobscot",
    "Penutian",
    "Potawatomi",
    "Powhatan",
    "kachina",
    "Salish",
    "Shahaptian",
    "Shasta",
    "Shawnee",
    "Sihasapa",
    "Teton",
    "Taracahitian",
    "Tarahumara",
    "Tuscarora",
    "Tutelo",
    "Yana",
    "Yavapai",
    "Yokuts",
    "Yuma",
    "Gadaba",
    "Kolam",
    "Kui",
    "Toda",
    "Tulu",
    "Gujarati",
    "Kashmiri",
    "Punjabi",
    "Slav",
    "Anabaptist",
    "Adventist",
    "gentile",
    "gentile",
    "Catholic",
    "Old_Catholic",
    "Uniat",
    "Copt",
    "Jewess",
    "Jihadist",
    "Buddhist",
    "Zen_Buddhist",
    "Mahayanist",
    "swami",
    "Hare_Krishna",
    "Shintoist",
    "Eurafrican",
    "Eurasian",
    "Gael",
    "Frank",
    "Afghan",
    "Albanian",
    "Algerian",
    "Altaic",
    "Andorran",
    "Angolan",
    "Anguillan",
    "Austrian",
    "Bahamian",
    "Bahraini",
    "Basotho",
    "Herero",
    "Luba",
    "Barbadian",
    "Bolivian",
    "Bornean",
    "Carioca",
    "Tupi",
    "Bruneian",
    "Bulgarian",
    "Byelorussian",
    "Cameroonian",
    "Canadian",
    "French_Canadian",
    "Central_American",
    "Chilean",
    "Congolese",
    "Cypriot",
    "Dane",
    "Djiboutian",
    "Britisher",
    "English_person",
    "Englishwoman",
    "Anglo-Saxon",
    "Angle",
    "West_Saxon",
    "Lombard",
    "limey",
    "Cantabrigian",
    "Cornishman",
    "Cornishwoman",
    "Lancastrian",
    "Lancastrian",
    "Geordie",
    "Oxonian",
    "Ethiopian",
    "Amhara",
    "Eritrean",
    "Finn",
    "Komi",
    "Livonian",
    "Lithuanian",
    "Selkup",
    "Parisian",
    "Parisienne",
    "Creole",
    "Creole",
    "Gabonese",
    "Greek",
    "Dorian",
    "Athenian",
    "Laconian",
    "Guyanese",
    "Haitian",
    "Malay",
    "Moro",
    "Netherlander",
    "Icelander",
    "Iraqi",
    "Irishman",
    "Irishwoman",
    "Dubliner",
    "Italian",
    "Roman",
    "Sabine",
    "Japanese",
    "Jordanian",
    "Korean",
    "Kenyan",
    "Lao",
    "Lapp",
    "Latin_American",
    "Lebanese",
    "Levantine",
    "Liberian",
    "Luxemburger",
    "Macedonian",
    "Sabahan",
    "Mexican",
    "Chicano",
    "Mexican-American",
    "Namibian",
    "Nauruan",
    "Gurkha",
    "New_Zealander",
    "Nicaraguan",
    "Nigerian",
    "Hausa",
    "North_American",
    "Nova_Scotian",
    "Omani",
    "Pakistani",
    "Brahui",
    "South_American_Indian",
    "Carib",
    "Filipino",
    "Polynesian",
    "Qatari",
    "Romanian",
    "Muscovite",
    "Georgian",
    "Sarawakian",
    "Scandinavian",
    "Senegalese",
    "Slovene",
    "South_African",
    "South_American",
    "Sudanese",
    "Syrian",
    "Tahitian",
    "Tanzanian",
    "Tibetan",
    "Togolese",
    "Tuareg",
    "Turki",
    "Chuvash",
    "Turkoman",
    "Uzbek",
    "Ugandan",
    "Ukranian",
    "Yakut",
    "Tungus",
    "Igbo",
    "American",
    "Anglo-American",
    "Alaska_Native",
    "Arkansan",
    "Carolinian",
    "Coloradan",
    "Connecticuter",
    "Delawarean",
    "Floridian",
    "German_American",
    "Illinoisan",
    "Mainer",
    "Marylander",
    "Minnesotan",
    "Nebraskan",
    "New_Hampshirite",
    "New_Jerseyan",
    "New_Yorker",
    "North_Carolinian",
    "Oregonian",
    "Pennsylvanian",
    "Texan",
    "Utahan",
    "Uruguayan",
    "Vietnamese",
    "Gambian",
    "East_German",
    "Berliner",
    "Prussian",
    "Ghanian",
    "Guinean",
    "Papuan",
    "Walloon",
    "Yemeni",
    "Yugoslav",
    "Serbian",
    "Xhosa",
    "Zairese",
    "Zimbabwean",
    "Zulu",
    "Gemini",
    "Sagittarius",
    "Pisces",
    "abbe",
    "abbess",
    "abnegator",
    "abridger",
    "abstractor",
    "absconder",
    "absolver",
    "abecedarian",
    "aberrant",
    "abettor",
    "abhorrer",
    "abomination",
    "abseiler",
    "abstainer",
    "academic_administrator",
    "academician",
    "accessory_before_the_fact",
    "companion",
    "accompanist",
    "accomplice",
    "account_executive",
    "accused",
    "accuser",
    "acid_head",
    "acquaintance",
    "acquirer",
    "aerialist",
    "action_officer",
    "active",
    "active_citizen",
    "actor",
    "actor",
    "addict",
    "adducer",
    "adjuster",
    "adjutant",
    "adjutant_general",
    "admirer",
    "adoptee",
    "adulterer",
    "adulteress",
    "advertiser",
    "advisee",
    "advocate",
    "aeronautical_engineer",
    "affiliate",
    "affluent",
    "aficionado",
    "buck_sergeant",
    "agent-in-place",
    "aggravator",
    "agitator",
    "agnostic",
    "agnostic",
    "agonist",
    "agony_aunt",
    "agriculturist",
    "air_attache",
    "air_force_officer",
    "airhead",
    "air_traveler",
    "alarmist",
    "albino",
    "alcoholic",
    "alderman",
    "alexic",
    "alienee",
    "alienor",
    "aliterate",
    "algebraist",
    "allegorizer",
    "alliterator",
    "almoner",
    "alpinist",
    "altar_boy",
    "alto",
    "ambassador",
    "ambassador",
    "ambusher",
    "amicus_curiae",
    "amoralist",
    "amputee",
    "analogist",
    "analphabet",
    "analyst",
    "industry_analyst",
    "market_strategist",
    "anarchist",
    "anathema",
    "ancestor",
    "anchor",
    "ancient",
    "anecdotist",
    "angler",
    "animator",
    "animist",
    "annotator",
    "announcer",
    "announcer",
    "anti",
    "anti-American",
    "anti-Semite",
    "Anzac",
    "ape-man",
    "aphakic",
    "appellant",
    "appointee",
    "apprehender",
    "April_fool",
    "aspirant",
    "appreciator",
    "appropriator",
    "Arabist",
    "archaist",
    "archbishop",
    "archer",
    "architect",
    "archivist",
    "archpriest",
    "Aristotelian",
    "armiger",
    "army_attache",
    "army_engineer",
    "army_officer",
    "arranger",
    "arrival",
    "arthritic",
    "articulator",
    "artilleryman",
    "artist's_model",
    "assayer",
    "assemblyman",
    "assemblywoman",
    "assenter",
    "asserter",
    "assignee",
    "assistant",
    "assistant_professor",
    "associate",
    "associate",
    "associate_professor",
    "astronaut",
    "cosmographer",
    "atheist",
    "athlete",
    "attendant",
    "attorney_general",
    "auditor",
    "augur",
    "aunt",
    "au_pair_girl",
    "authoritarian",
    "authority",
    "authorizer",
    "automobile_mechanic",
    "aviator",
    "aviatrix",
    "ayah",
    "babu",
    "baby",
    "baby",
    "baby_boomer",
    "baby_farmer",
    "back",
    "backbencher",
    "backpacker",
    "backroom_boy",
    "backscratcher",
    "bad_person",
    "baggage",
    "bag_lady",
    "bailee",
    "bailiff",
    "bailor",
    "bairn",
    "baker",
    "balancer",
    "balker",
    "ball-buster",
    "ball_carrier",
    "ballet_dancer",
    "ballet_master",
    "ballet_mistress",
    "balletomane",
    "ball_hawk",
    "balloonist",
    "ballplayer",
    "bullfighter",
    "banderillero",
    "matador",
    "picador",
    "bandsman",
    "banker",
    "bank_robber",
    "bankrupt",
    "bantamweight",
    "barmaid",
    "baron",
    "baron",
    "baron",
    "bartender",
    "baseball_coach",
    "base_runner",
    "basketball_player",
    "basketweaver",
    "Basket_Maker",
    "bass",
    "bastard",
    "bat_boy",
    "bather",
    "batman",
    "baton_twirler",
    "Bavarian",
    "beadsman",
    "beard",
    "beatnik",
    "beauty_consultant",
    "Bedouin",
    "bedwetter",
    "beekeeper",
    "beer_drinker",
    "beggarman",
    "beggarwoman",
    "beldam",
    "theist",
    "believer",
    "bell_founder",
    "benedick",
    "berserker",
    "besieger",
    "best",
    "betrothed",
    "Big_Brother",
    "bigot",
    "big_shot",
    "big_sister",
    "billiard_player",
    "biochemist",
    "biographer",
    "bird_fancier",
    "birth",
    "birth-control_campaigner",
    "bisexual",
    "black_belt",
    "blackmailer",
    "Black_Muslim",
    "blacksmith",
    "blade",
    "bleacher",
    "blind_date",
    "bluecoat",
    "bluestocking",
    "boatbuilder",
    "boatman",
    "boatswain",
    "bobby",
    "bodyguard",
    "boffin",
    "Bolshevik",
    "Bolshevik",
    "bombshell",
    "bondman",
    "bondwoman",
    "bondwoman",
    "bond_servant",
    "book_agent",
    "bookbinder",
    "bookkeeper",
    "bookmaker",
    "bookworm",
    "booster",
    "bootblack",
    "bootlegger",
    "bootmaker",
    "borderer",
    "border_patrolman",
    "botanist",
    "bottom_feeder",
    "boulevardier",
    "bounty_hunter",
    "bounty_hunter",
    "Bourbon",
    "bowler",
    "slugger",
    "cub",
    "Boy_Scout",
    "boy_scout",
    "boy_wonder",
    "bragger",
    "brahman",
    "brawler",
    "breadwinner",
    "breaststroker",
    "breeder",
    "brick",
    "bride",
    "bridesmaid",
    "bridge_agent",
    "broadcast_journalist",
    "Brother",
    "brother-in-law",
    "browser",
    "Brummie",
    "buddy",
    "bull",
    "bully",
    "bunny",
    "burglar",
    "bursar",
    "busboy",
    "business_editor",
    "business_traveler",
    "buster",
    "busybody",
    "buttinsky",
    "cabinetmaker",
    "caddie",
    "cadet",
    "caller",
    "call_girl",
    "calligrapher",
    "campaigner",
    "camper",
    "camp_follower",
    "candidate",
    "canonist",
    "capitalist",
    "captain",
    "captain",
    "captain",
    "captain",
    "captive",
    "captive",
    "cardinal",
    "cardiologist",
    "card_player",
    "cardsharp",
    "careerist",
    "career_man",
    "caregiver",
    "caretaker",
    "caretaker",
    "caricaturist",
    "carillonneur",
    "caroler",
    "carpenter",
    "carper",
    "Cartesian",
    "cashier",
    "casualty",
    "casualty",
    "casuist",
    "catechist",
    "catechumen",
    "caterer",
    "Catholicos",
    "cat_fancier",
    "Cavalier",
    "cavalryman",
    "caveman",
    "celebrant",
    "celebrant",
    "celebrity",
    "cellist",
    "censor",
    "censor",
    "centenarian",
    "centrist",
    "centurion",
    "certified_public_accountant",
    "chachka",
    "chambermaid",
    "chameleon",
    "champion",
    "chandler",
    "prison_chaplain",
    "charcoal_burner",
    "charge_d'affaires",
    "charioteer",
    "charmer",
    "chartered_accountant",
    "chartist",
    "charwoman",
    "male_chauvinist",
    "cheapskate",
    "Chechen",
    "checker",
    "cheerer",
    "cheerleader",
    "cheerleader",
    "Cheops",
    "chess_master",
    "chief_executive_officer",
    "chief_of_staff",
    "chief_petty_officer",
    "Chief_Secretary",
    "child",
    "child",
    "child",
    "child_prodigy",
    "chimneysweeper",
    "chiropractor",
    "chit",
    "choker",
    "choragus",
    "choreographer",
    "chorus_girl",
    "chosen",
    "cicerone",
    "cigar_smoker",
    "cipher",
    "circus_acrobat",
    "citizen",
    "city_editor",
    "city_father",
    "city_man",
    "city_slicker",
    "civic_leader",
    "civil_rights_leader",
    "cleaner",
    "clergyman",
    "cleric",
    "clerk",
    "clever_Dick",
    "climatologist",
    "climber",
    "clinician",
    "closer",
    "closet_queen",
    "clown",
    "clown",
    "coach",
    "coach",
    "pitching_coach",
    "coachman",
    "coal_miner",
    "coastguardsman",
    "cobber",
    "cobbler",
    "codger",
    "co-beneficiary",
    "cog",
    "cognitive_neuroscientist",
    "coiffeur",
    "coiner",
    "collaborator",
    "colleen",
    "college_student",
    "collegian",
    "colonial",
    "colonialist",
    "colonizer",
    "coloratura",
    "color_guard",
    "colossus",
    "comedian",
    "comedienne",
    "comer",
    "commander",
    "commander_in_chief",
    "commanding_officer",
    "commissar",
    "commissioned_officer",
    "commissioned_military_officer",
    "commissioner",
    "commissioner",
    "committee_member",
    "committeewoman",
    "commodore",
    "communicant",
    "communist",
    "Communist",
    "commuter",
    "compere",
    "complexifier",
    "compulsive",
    "computational_linguist",
    "computer_scientist",
    "computer_user",
    "Comrade",
    "concert-goer",
    "conciliator",
    "conductor",
    "confectioner",
    "Confederate",
    "confessor",
    "confidant",
    "Confucian",
    "rep",
    "conqueror",
    "Conservative",
    "Nonconformist",
    "Anglican",
    "consignee",
    "consigner",
    "constable",
    "constructivist",
    "contractor",
    "contralto",
    "contributor",
    "control_freak",
    "convalescent",
    "convener",
    "convict",
    "copilot",
    "copycat",
    "coreligionist",
    "cornerback",
    "corporatist",
    "correspondent",
    "cosmetician",
    "cosmopolitan",
    "Cossack",
    "cost_accountant",
    "co-star",
    "costumier",
    "cotter",
    "cotter",
    "counselor",
    "counterterrorist",
    "counterspy",
    "countess",
    "compromiser",
    "countrywoman",
    "county_agent",
    "courtier",
    "cousin",
    "cover_girl",
    "cow",
    "craftsman",
    "craftsman",
    "crapshooter",
    "crazy",
    "creature",
    "creditor",
    "creep",
    "criminologist",
    "critic",
    "Croesus",
    "cross-examiner",
    "crossover_voter",
    "croupier",
    "crown_prince",
    "crown_princess",
    "cryptanalyst",
    "Cub_Scout",
    "cuckold",
    "cultist",
    "curandera",
    "curate",
    "curator",
    "customer_agent",
    "cutter",
    "cyberpunk",
    "cyborg",
    "cymbalist",
    "Cynic",
    "cytogeneticist",
    "cytologist",
    "czar",
    "czar",
    "dad",
    "dairyman",
    "Dalai_Lama",
    "dallier",
    "dancer",
    "dancer",
    "clog_dancer",
    "dancing-master",
    "dark_horse",
    "darling",
    "date",
    "daughter",
    "dawdler",
    "day_boarder",
    "day_laborer",
    "deacon",
    "deaconess",
    "deadeye",
    "deipnosophist",
    "dropout",
    "deadhead",
    "deaf_person",
    "debtor",
    "deckhand",
    "defamer",
    "defense_contractor",
    "deist",
    "delegate",
    "deliveryman",
    "demagogue",
    "demigod",
    "demographer",
    "demonstrator",
    "den_mother",
    "department_head",
    "depositor",
    "deputy",
    "dermatologist",
    "descender",
    "designated_hitter",
    "designer",
    "desk_clerk",
    "desk_officer",
    "desk_sergeant",
    "detainee",
    "detective",
    "detective",
    "detractor",
    "developer",
    "deviationist",
    "devisee",
    "devisor",
    "devourer",
    "dialectician",
    "diarist",
    "dietician",
    "diocesan",
    "director",
    "director",
    "dirty_old_man",
    "disbeliever",
    "disk_jockey",
    "dispatcher",
    "distortionist",
    "distributor",
    "district_attorney",
    "district_manager",
    "diver",
    "divorcee",
    "ex-wife",
    "divorce_lawyer",
    "docent",
    "doctor",
    "dodo",
    "doge",
    "dog_in_the_manger",
    "dogmatist",
    "dolichocephalic",
    "domestic_partner",
    "Dominican",
    "dominus",
    "don",
    "Donatist",
    "donna",
    "dosser",
    "double",
    "double-crosser",
    "down-and-out",
    "doyenne",
    "draftsman",
    "dramatist",
    "dreamer",
    "dressmaker",
    "dressmaker's_model",
    "dribbler",
    "dribbler",
    "drinker",
    "drinker",
    "drug_addict",
    "drug_user",
    "Druid",
    "drum_majorette",
    "drummer",
    "drunk",
    "drunkard",
    "Druze",
    "dry",
    "dry_nurse",
    "duchess",
    "duke",
    "duffer",
    "dunker",
    "Dutch_uncle",
    "dyspeptic",
    "eager_beaver",
    "earl",
    "earner",
    "eavesdropper",
    "eccentric",
    "eclectic",
    "econometrician",
    "economist",
    "ectomorph",
    "editor",
    "egocentric",
    "egotist",
    "ejaculator",
    "elder",
    "elder_statesman",
    "elected_official",
    "electrician",
    "elegist",
    "elocutionist",
    "emancipator",
    "embryologist",
    "emeritus",
    "emigrant",
    "emissary",
    "empress",
    "employee",
    "employer",
    "enchantress",
    "enchantress",
    "encyclopedist",
    "endomorph",
    "enemy",
    "energizer",
    "end_man",
    "end_man",
    "endorser",
    "enjoyer",
    "enlisted_woman",
    "enophile",
    "entrant",
    "entrant",
    "entrepreneur",
    "envoy",
    "enzymologist",
    "eparch",
    "epidemiologist",
    "epigone",
    "epileptic",
    "Episcopalian",
    "equerry",
    "equerry",
    "erotic",
    "escapee",
    "escapist",
    "Eskimo",
    "espionage_agent",
    "esthetician",
    "etcher",
    "ethnologist",
    "Etonian",
    "etymologist",
    "evangelist",
    "Evangelist",
    "event_planner",
    "examiner",
    "examiner",
    "exarch",
    "executant",
    "executive_secretary",
    "executive_vice_president",
    "executrix",
    "exegete",
    "exhibitor",
    "exhibitionist",
    "exile",
    "existentialist",
    "exorcist",
    "ex-spouse",
    "extern",
    "extremist",
    "extrovert",
    "eyewitness",
    "facilitator",
    "fairy_godmother",
    "falangist",
    "falconer",
    "falsifier",
    "familiar",
    "fan",
    "fanatic",
    "fancier",
    "farm_boy",
    "farmer",
    "farmhand",
    "fascist",
    "fascista",
    "fatalist",
    "father",
    "Father",
    "father-figure",
    "father-in-law",
    "Fauntleroy",
    "Fauve",
    "favorite_son",
    "featherweight",
    "federalist",
    "fellow_traveler",
    "female_aristocrat",
    "female_offspring",
    "female_child",
    "fence",
    "fiance",
    "fielder",
    "field_judge",
    "fighter_pilot",
    "filer",
    "film_director",
    "finder",
    "fire_chief",
    "fire-eater",
    "fire-eater",
    "fireman",
    "fire_marshall",
    "fire_walker",
    "first_baseman",
    "firstborn",
    "first_lady",
    "first_lieutenant",
    "first_offender",
    "first_sergeant",
    "fishmonger",
    "flagellant",
    "flag_officer",
    "flak_catcher",
    "flanker_back",
    "flapper",
    "flatmate",
    "flatterer",
    "flibbertigibbet",
    "flight_surgeon",
    "floorwalker",
    "flop",
    "Florentine",
    "flower_girl",
    "flower_girl",
    "flutist",
    "fly-by-night",
    "flyweight",
    "flyweight",
    "foe",
    "folk_dancer",
    "folk_poet",
    "follower",
    "football_hero",
    "football_player",
    "footman",
    "forefather",
    "foremother",
    "foreign_agent",
    "foreigner",
    "boss",
    "foreman",
    "forester",
    "forewoman",
    "forger",
    "forward",
    "foster-brother",
    "foster-father",
    "foster-mother",
    "foster-sister",
    "foster-son",
    "founder",
    "foundress",
    "four-minute_man",
    "framer",
    "Francophobe",
    "freak",
    "free_agent",
    "free_agent",
    "freedom_rider",
    "free-liver",
    "freeloader",
    "free_trader",
    "Freudian",
    "friar",
    "monk",
    "frontierswoman",
    "front_man",
    "frotteur",
    "fucker",
    "fucker",
    "fuddy-duddy",
    "fullback",
    "funambulist",
    "fundamentalist",
    "fundraiser",
    "futurist",
    "gadgeteer",
    "gagman",
    "gagman",
    "gainer",
    "gal",
    "galoot",
    "gambist",
    "gambler",
    "gamine",
    "garbage_man",
    "gardener",
    "garment_cutter",
    "garroter",
    "gasman",
    "gastroenterologist",
    "gatherer",
    "gawker",
    "gendarme",
    "general",
    "generator",
    "geneticist",
    "genitor",
    "gent",
    "geologist",
    "geophysicist",
    "ghostwriter",
    "Gibson_girl",
    "girl",
    "girlfriend",
    "girlfriend",
    "girl_wonder",
    "Girondist",
    "gitano",
    "gladiator",
    "glassblower",
    "gleaner",
    "goat_herder",
    "godchild",
    "godfather",
    "godparent",
    "godson",
    "gofer",
    "goffer",
    "goldsmith",
    "golfer",
    "gondolier",
    "good_guy",
    "good_old_boy",
    "good_Samaritan",
    "gossip_columnist",
    "gouger",
    "governor_general",
    "grabber",
    "grader",
    "graduate_nurse",
    "grammarian",
    "granddaughter",
    "grande_dame",
    "grandfather",
    "Grand_Inquisitor",
    "grandma",
    "grandmaster",
    "grandparent",
    "grantee",
    "granter",
    "grass_widower",
    "great-aunt",
    "great_grandchild",
    "great_granddaughter",
    "great_grandmother",
    "great_grandparent",
    "great_grandson",
    "great-nephew",
    "great-niece",
    "Green_Beret",
    "grenadier",
    "greeter",
    "gringo",
    "grinner",
    "grocer",
    "groom",
    "groom",
    "grouch",
    "group_captain",
    "grunter",
    "prison_guard",
    "guard",
    "guesser",
    "guest",
    "guest",
    "guest_of_honor",
    "guest_worker",
    "guide",
    "guitarist",
    "gunnery_sergeant",
    "guru",
    "guru",
    "guvnor",
    "guy",
    "gymnast",
    "gym_rat",
    "gynecologist",
    "Gypsy",
    "hack",
    "hacker",
    "haggler",
    "hairdresser",
    "hakim",
    "Hakka",
    "halberdier",
    "halfback",
    "half_blood",
    "hand",
    "animal_trainer",
    "handyman",
    "hang_glider",
    "hardliner",
    "harlequin",
    "harmonizer",
    "hash_head",
    "hatchet_man",
    "hater",
    "hatmaker",
    "headman",
    "headmaster",
    "head_nurse",
    "hearer",
    "heartbreaker",
    "heathen",
    "heavyweight",
    "heavy",
    "heckler",
    "hedger",
    "hedger",
    "hedonist",
    "heir",
    "heir_apparent",
    "heiress",
    "heir_presumptive",
    "hellion",
    "helmsman",
    "hire",
    "hematologist",
    "hemiplegic",
    "herald",
    "herbalist",
    "herder",
    "hermaphrodite",
    "heroine",
    "heroin_addict",
    "hero_worshiper",
    "Herr",
    "highbinder",
    "highbrow",
    "high_commissioner",
    "highflier",
    "Highlander",
    "high-muck-a-muck",
    "high_priest",
    "highjacker",
    "hireling",
    "historian",
    "hitchhiker",
    "hitter",
    "hobbyist",
    "holdout",
    "holdover",
    "holdup_man",
    "homeboy",
    "homeboy",
    "home_buyer",
    "homegirl",
    "homeless",
    "homeopath",
    "honest_woman",
    "honor_guard",
    "hooker",
    "hoper",
    "hornist",
    "horseman",
    "horse_trader",
    "horsewoman",
    "horse_wrangler",
    "horticulturist",
    "hospital_chaplain",
    "host",
    "host",
    "hostess",
    "hotelier",
    "housekeeper",
    "housemaster",
    "housemate",
    "house_physician",
    "house_sitter",
    "housing_commissioner",
    "huckster",
    "hugger",
    "humanist",
    "humanitarian",
    "hunk",
    "huntress",
    "ex-husband",
    "hydrologist",
    "hyperope",
    "hypertensive",
    "hypnotist",
    "hypocrite",
    "iceman",
    "iconoclast",
    "ideologist",
    "idol",
    "idolizer",
    "imam",
    "imperialist",
    "important_person",
    "inamorato",
    "incumbent",
    "incurable",
    "inductee",
    "industrialist",
    "infanticide",
    "inferior",
    "infernal",
    "infielder",
    "infiltrator",
    "informer",
    "ingenue",
    "ingenue",
    "polymath",
    "in-law",
    "inquiry_agent",
    "inspector",
    "inspector_general",
    "instigator",
    "insurance_broker",
    "insurgent",
    "intelligence_analyst",
    "interior_designer",
    "interlocutor",
    "interlocutor",
    "International_Grandmaster",
    "internationalist",
    "internist",
    "interpreter",
    "interpreter",
    "intervenor",
    "introvert",
    "invader",
    "invalidator",
    "investigator",
    "investor",
    "invigilator",
    "irreligionist",
    "Ivy_Leaguer",
    "Jack_of_all_trades",
    "Jacksonian",
    "Jane_Doe",
    "janissary",
    "Jat",
    "Javanese",
    "Jekyll_and_Hyde",
    "jester",
    "Jesuit",
    "jezebel",
    "jilt",
    "jobber",
    "job_candidate",
    "Job's_comforter",
    "jockey",
    "John_Doe",
    "journalist",
    "judge",
    "judge_advocate",
    "juggler",
    "Jungian",
    "junior",
    "junior",
    "Junior",
    "junior_lightweight",
    "junior_middleweight",
    "jurist",
    "juror",
    "justice_of_the_peace",
    "justiciar",
    "kachina",
    "keyboardist",
    "Khedive",
    "kingmaker",
    "king",
    "King's_Counsel",
    "Counsel_to_the_Crown",
    "kin",
    "enate",
    "kink",
    "kinswoman",
    "kisser",
    "kitchen_help",
    "kitchen_police",
    "Klansman",
    "kleptomaniac",
    "kneeler",
    "knight",
    "knocker",
    "knower",
    "know-it-all",
    "kolkhoznik",
    "Kshatriya",
    "labor_coach",
    "laborer",
    "Labourite",
    "lady",
    "lady-in-waiting",
    "lady's_maid",
    "lama",
    "lamb",
    "lame_duck",
    "lamplighter",
    "land_agent",
    "landgrave",
    "landlubber",
    "landlubber",
    "landowner",
    "landscape_architect",
    "langlaufer",
    "languisher",
    "lapidary",
    "lass",
    "Latin",
    "Latin",
    "latitudinarian",
    "Jehovah's_Witness",
    "law_agent",
    "lawgiver",
    "lawman",
    "law_student",
    "lawyer",
    "lay_reader",
    "lazybones",
    "leaker",
    "leaseholder",
    "lector",
    "lector",
    "lecturer",
    "left-hander",
    "legal_representative",
    "legate",
    "legatee",
    "legionnaire",
    "letterman",
    "liberator",
    "licenser",
    "licentiate",
    "lieutenant",
    "lieutenant_colonel",
    "lieutenant_commander",
    "lieutenant_junior_grade",
    "life",
    "lifeguard",
    "life_tenant",
    "light_flyweight",
    "light_heavyweight",
    "light_heavyweight",
    "light-o'-love",
    "lightweight",
    "lightweight",
    "lightweight",
    "lilliputian",
    "limnologist",
    "lineman",
    "line_officer",
    "lion-hunter",
    "lisper",
    "lister",
    "literary_critic",
    "literate",
    "litigant",
    "litterer",
    "little_brother",
    "little_sister",
    "lobbyist",
    "locksmith",
    "locum_tenens",
    "Lord",
    "loser",
    "loser",
    "failure",
    "Lothario",
    "loudmouth",
    "lowerclassman",
    "Lowlander",
    "loyalist",
    "Luddite",
    "lumberman",
    "lumper",
    "bedlamite",
    "pyromaniac",
    "lutist",
    "Lutheran",
    "lyricist",
    "macebearer",
    "machinist",
    "madame",
    "maenad",
    "maestro",
    "magdalen",
    "magician",
    "magus",
    "maharani",
    "mahatma",
    "maid",
    "maid",
    "major",
    "major",
    "major-domo",
    "maker",
    "malahini",
    "malcontent",
    "malik",
    "malingerer",
    "Malthusian",
    "adonis",
    "man",
    "man",
    "manageress",
    "mandarin",
    "maneuverer",
    "maniac",
    "Manichaean",
    "manicurist",
    "manipulator",
    "man-at-arms",
    "man_of_action",
    "man_of_letters",
    "manufacturer",
    "marcher",
    "marchioness",
    "margrave",
    "margrave",
    "Marine",
    "marquess",
    "marquis",
    "marshal",
    "martinet",
    "mascot",
    "masochist",
    "mason",
    "masquerader",
    "masseur",
    "masseuse",
    "master",
    "master",
    "master-at-arms",
    "master_of_ceremonies",
    "masturbator",
    "matchmaker",
    "mate",
    "mate",
    "mate",
    "mater",
    "material",
    "materialist",
    "matriarch",
    "matriarch",
    "matriculate",
    "matron",
    "mayor",
    "mayoress",
    "mechanical_engineer",
    "medalist",
    "medical_officer",
    "medical_practitioner",
    "medical_scientist",
    "medium",
    "megalomaniac",
    "melancholic",
    "Melkite",
    "melter",
    "nonmember",
    "board_member",
    "clansman",
    "memorizer",
    "Mendelian",
    "mender",
    "Mesoamerican",
    "messmate",
    "mestiza",
    "meteorologist",
    "meter_maid",
    "Methodist",
    "Metis",
    "metropolitan",
    "mezzo-soprano",
    "microeconomist",
    "middle-aged_man",
    "middlebrow",
    "middleweight",
    "midwife",
    "mikado",
    "Milanese",
    "miler",
    "miles_gloriosus",
    "military_attache",
    "military_chaplain",
    "military_leader",
    "military_officer",
    "military_policeman",
    "mill_agent",
    "mill-hand",
    "millionairess",
    "millwright",
    "minder",
    "mining_engineer",
    "minister",
    "ministrant",
    "minor_leaguer",
    "Minuteman",
    "misanthrope",
    "misfit",
    "mistress",
    "mistress",
    "mixed-blood",
    "model",
    "class_act",
    "modeler",
    "modifier",
    "molecular_biologist",
    "Monegasque",
    "monetarist",
    "moneygrubber",
    "moneymaker",
    "Mongoloid",
    "monolingual",
    "monologist",
    "moonlighter",
    "moralist",
    "morosoph",
    "morris_dancer",
    "mortal_enemy",
    "mortgagee",
    "mortician",
    "moss-trooper",
    "mother",
    "mother",
    "mother",
    "mother_figure",
    "mother_hen",
    "mother-in-law",
    "mother's_boy",
    "mother's_daughter",
    "motorcycle_cop",
    "motorcyclist",
    "Mound_Builder",
    "mountebank",
    "mourner",
    "mouthpiece",
    "mover",
    "moviegoer",
    "muffin_man",
    "mugwump",
    "Mullah",
    "muncher",
    "murderess",
    "murder_suspect",
    "musher",
    "musician",
    "musicologist",
    "music_teacher",
    "musketeer",
    "Muslimah",
    "mutilator",
    "mutineer",
    "mute",
    "mutterer",
    "muzzler",
    "Mycenaen",
    "mycologist",
    "myope",
    "myrmidon",
    "mystic",
    "mythologist",
    "naif",
    "nailer",
    "namby-pamby",
    "name_dropper",
    "namer",
    "nan",
    "nanny",
    "narc",
    "narcissist",
    "nark",
    "nationalist",
    "nautch_girl",
    "naval_commander",
    "Navy_SEAL",
    "obstructionist",
    "Nazarene",
    "Nazarene",
    "Nazi",
    "nebbish",
    "necker",
    "neonate",
    "nephew",
    "neurobiologist",
    "neurologist",
    "neurosurgeon",
    "neutral",
    "neutralist",
    "newcomer",
    "newcomer",
    "New_Dealer",
    "newspaper_editor",
    "newsreader",
    "Newtonian",
    "niece",
    "niggard",
    "night_porter",
    "night_rider",
    "NIMBY",
    "niqaabi",
    "nitpicker",
    "Nobelist",
    "NOC",
    "noncandidate",
    "noncommissioned_officer",
    "nondescript",
    "nondriver",
    "nonparticipant",
    "nonperson",
    "nonresident",
    "nonsmoker",
    "Northern_Baptist",
    "noticer",
    "novelist",
    "novitiate",
    "nuclear_chemist",
    "nudger",
    "nullipara",
    "number_theorist",
    "nurse",
    "nursling",
    "nymph",
    "nymphet",
    "nympholept",
    "nymphomaniac",
    "oarswoman",
    "oboist",
    "obscurantist",
    "observer",
    "obstetrician",
    "occupier",
    "occultist",
    "wine_lover",
    "offerer",
    "office-bearer",
    "office_boy",
    "officeholder",
    "officiant",
    "Federal",
    "oilman",
    "oil_tycoon",
    "old-age_pensioner",
    "old_boy",
    "old_lady",
    "old_man",
    "oldster",
    "old-timer",
    "old_woman",
    "oligarch",
    "Olympian",
    "omnivore",
    "oncologist",
    "onlooker",
    "onomancer",
    "operator",
    "opportunist",
    "optimist",
    "Orangeman",
    "orator",
    "orderly",
    "orderly",
    "orderly_sergeant",
    "ordinand",
    "ordinary",
    "organ-grinder",
    "organist",
    "organization_man",
    "organizer",
    "organizer",
    "originator",
    "ornithologist",
    "orphan",
    "orphan",
    "osteopath",
    "out-and-outer",
    "outdoorswoman",
    "outfielder",
    "outfielder",
    "right_fielder",
    "right-handed_pitcher",
    "outlier",
    "owner-occupier",
    "oyabun",
    "packrat",
    "padrone",
    "padrone",
    "page",
    "painter",
    "Paleo-American",
    "paleontologist",
    "pallbearer",
    "palmist",
    "pamperer",
    "Panchen_Lama",
    "panelist",
    "panhandler",
    "paparazzo",
    "paperboy",
    "paperhanger",
    "paperhanger",
    "papoose",
    "pardoner",
    "paretic",
    "parishioner",
    "park_commissioner",
    "Parliamentarian",
    "parliamentary_agent",
    "parodist",
    "parricide",
    "parrot",
    "partaker",
    "part-timer",
    "party",
    "party_man",
    "passenger",
    "passer",
    "paster",
    "pater",
    "patient",
    "patriarch",
    "patriarch",
    "patriarch",
    "patriot",
    "patron",
    "patternmaker",
    "pawnbroker",
    "payer",
    "peacekeeper",
    "peasant",
    "pedant",
    "peddler",
    "pederast",
    "penologist",
    "pentathlete",
    "Pentecostal",
    "percussionist",
    "periodontist",
    "peshmerga",
    "personality",
    "personal_representative",
    "personage",
    "persona_grata",
    "persona_non_grata",
    "personification",
    "perspirer",
    "pervert",
    "pessimist",
    "pest",
    "Peter_Pan",
    "petitioner",
    "petit_juror",
    "pet_sitter",
    "petter",
    "Pharaoh",
    "pharmacist",
    "philanthropist",
    "philatelist",
    "philosopher",
    "phonetician",
    "phonologist",
    "photojournalist",
    "photometrist",
    "physical_therapist",
    "physicist",
    "piano_maker",
    "picker",
    "picnicker",
    "pilgrim",
    "pill",
    "pillar",
    "pill_head",
    "pilot",
    "Piltdown_man",
    "pimp",
    "pipe_smoker",
    "pip-squeak",
    "pisser",
    "pitcher",
    "pitchman",
    "placeman",
    "placer_miner",
    "plagiarist",
    "plainsman",
    "planner",
    "planter",
    "plasterer",
    "platinum_blond",
    "platitudinarian",
    "playboy",
    "player",
    "playmate",
    "pleaser",
    "pledger",
    "plenipotentiary",
    "plier",
    "plodder",
    "plodder",
    "plotter",
    "plumber",
    "pluralist",
    "pluralist",
    "poet",
    "pointsman",
    "point_woman",
    "policyholder",
    "political_prisoner",
    "political_scientist",
    "politician",
    "politician",
    "pollster",
    "polluter",
    "pool_player",
    "portraitist",
    "poseuse",
    "positivist",
    "postdoc",
    "poster_girl",
    "postulator",
    "private_citizen",
    "problem_solver",
    "pro-lifer",
    "prosthetist",
    "postulant",
    "potboy",
    "poultryman",
    "power_user",
    "power_worker",
    "practitioner",
    "prayer",
    "preceptor",
    "predecessor",
    "preemptor",
    "preemptor",
    "premature_baby",
    "presbyter",
    "presenter",
    "presentist",
    "preserver",
    "president",
    "President_of_the_United_States",
    "president",
    "press_agent",
    "press_photographer",
    "priest",
    "prima_ballerina",
    "prima_donna",
    "prima_donna",
    "primigravida",
    "primordial_dwarf",
    "prince_charming",
    "prince_consort",
    "princeling",
    "Prince_of_Wales",
    "princess",
    "princess_royal",
    "principal",
    "principal",
    "print_seller",
    "prior",
    "private",
    "probationer",
    "processor",
    "process-server",
    "proconsul",
    "proconsul",
    "proctologist",
    "proctor",
    "procurator",
    "procurer",
    "profit_taker",
    "programmer",
    "promiser",
    "promoter",
    "promulgator",
    "propagandist",
    "propagator",
    "property_man",
    "prophetess",
    "prophet",
    "prosecutor",
    "prospector",
    "protectionist",
    "protegee",
    "protozoologist",
    "provost_marshal",
    "pruner",
    "psalmist",
    "psephologist",
    "psychiatrist",
    "psychic",
    "psycholinguist",
    "psychophysicist",
    "publican",
    "pudge",
    "puerpera",
    "punching_bag",
    "punter",
    "punter",
    "puppeteer",
    "puppy",
    "purchasing_agent",
    "puritan",
    "Puritan",
    "pursuer",
    "pusher",
    "pusher",
    "pusher",
    "putz",
    "Pygmy",
    "qadi",
    "quadriplegic",
    "quadruplet",
    "quaker",
    "quarter",
    "quarterback",
    "quartermaster",
    "quartermaster_general",
    "Quebecois",
    "queen",
    "Queen_of_England",
    "queen",
    "queen",
    "queen_consort",
    "queen_mother",
    "Queen's_Counsel",
    "question_master",
    "quick_study",
    "quietist",
    "quitter",
    "rabbi",
    "racist",
    "radiobiologist",
    "radiologic_technologist",
    "radiologist",
    "rainmaker",
    "raiser",
    "raja",
    "rake",
    "ramrod",
    "ranch_hand",
    "ranker",
    "ranter",
    "rape_suspect",
    "rapper",
    "rapporteur",
    "rare_bird",
    "ratepayer",
    "raw_recruit",
    "reader",
    "reading_teacher",
    "realist",
    "real_estate_broker",
    "rear_admiral",
    "receiver",
    "reciter",
    "recruit",
    "recruit",
    "recruiter",
    "recruiting-sergeant",
    "redcap",
    "redhead",
    "redneck",
    "reeler",
    "reenactor",
    "referral",
    "referee",
    "refiner",
    "Reform_Jew",
    "registered_nurse",
    "registrar",
    "Regius_professor",
    "reliever",
    "anchorite",
    "religious_leader",
    "remover",
    "Renaissance_man",
    "renegade",
    "rentier",
    "repairman",
    "reporter",
    "newswoman",
    "representative",
    "reprobate",
    "rescuer",
    "reservist",
    "resident_commissioner",
    "respecter",
    "restaurateur",
    "restrainer",
    "retailer",
    "retiree",
    "returning_officer",
    "revenant",
    "revisionist",
    "revolutionist",
    "rheumatologist",
    "Rhodesian_man",
    "rhymer",
    "rich_person",
    "rider",
    "riding_master",
    "rifleman",
    "right-hander",
    "right-hand_man",
    "ringer",
    "ringleader",
    "roadman",
    "roarer",
    "rocket_engineer",
    "rocket_scientist",
    "rock_star",
    "Romanov",
    "romanticist",
    "ropemaker",
    "roper",
    "roper",
    "ropewalker",
    "rosebud",
    "Rosicrucian",
    "Mountie",
    "Rough_Rider",
    "roundhead",
    "civil_authority",
    "runner",
    "runner",
    "runner",
    "running_back",
    "rusher",
    "rustic",
    "saboteur",
    "sadist",
    "sailing_master",
    "sailor",
    "salesgirl",
    "salesman",
    "salesperson",
    "salvager",
    "sandwichman",
    "sangoma",
    "sannup",
    "sapper",
    "Sassenach",
    "satrap",
    "saunterer",
    "Savoyard",
    "sawyer",
    "scalper",
    "scandalmonger",
    "scapegrace",
    "scene_painter",
    "schemer",
    "schizophrenic",
    "schlemiel",
    "schlockmeister",
    "scholar",
    "scholiast",
    "schoolchild",
    "schoolfriend",
    "Schoolman",
    "schoolmaster",
    "schoolmate",
    "scientist",
    "scion",
    "scoffer",
    "scofflaw",
    "scorekeeper",
    "scorer",
    "scourer",
    "scout",
    "scoutmaster",
    "scrambler",
    "scratcher",
    "screen_actor",
    "scrutineer",
    "scuba_diver",
    "sculptor",
    "Sea_Scout",
    "seasonal_worker",
    "seasoner",
    "second_baseman",
    "second_cousin",
    "seconder",
    "second_fiddle",
    "second-in-command",
    "second_lieutenant",
    "second-rater",
    "secretary",
    "Secretary_of_Agriculture",
    "Secretary_of_Health_and_Human_Services",
    "Secretary_of_State",
    "Secretary_of_the_Interior",
    "sectarian",
    "section_hand",
    "secularist",
    "security_consultant",
    "seeded_player",
    "seeder",
    "seeker",
    "segregate",
    "segregator",
    "selectman",
    "selectwoman",
    "selfish_person",
    "self-starter",
    "seller",
    "selling_agent",
    "semanticist",
    "semifinalist",
    "seminarian",
    "senator",
    "sendee",
    "senior",
    "senior_vice_president",
    "separatist",
    "septuagenarian",
    "serf",
    "spree_killer",
    "serjeant-at-law",
    "server",
    "serviceman",
    "settler",
    "settler",
    "sex_symbol",
    "sexton",
    "shaheed",
    "Shakespearian",
    "shanghaier",
    "sharecropper",
    "shaver",
    "Shavian",
    "sheep",
    "sheik",
    "shelver",
    "shepherd",
    "ship-breaker",
    "shipmate",
    "shipowner",
    "shipping_agent",
    "shirtmaker",
    "shogun",
    "shopaholic",
    "shop_girl",
    "shop_steward",
    "shot_putter",
    "shrew",
    "shuffler",
    "shyster",
    "sibling",
    "sick_person",
    "sightreader",
    "signaler",
    "signer",
    "signor",
    "signora",
    "signore",
    "signorina",
    "silent_partner",
    "addle-head",
    "simperer",
    "singer",
    "Sinologist",
    "sipper",
    "sirrah",
    "Sister",
    "sister",
    "waverer",
    "sitar_player",
    "sixth-former",
    "skateboarder",
    "skeptic",
    "sketcher",
    "skidder",
    "skier",
    "skinny-dipper",
    "skin-diver",
    "skinhead",
    "slasher",
    "slattern",
    "sleeper",
    "sleeper",
    "sleeping_beauty",
    "sleuth",
    "slob",
    "sloganeer",
    "slopseller",
    "smasher",
    "smirker",
    "smith",
    "smoothie",
    "smuggler",
    "sneezer",
    "snob",
    "snoop",
    "snorer",
    "sob_sister",
    "soccer_player",
    "social_anthropologist",
    "social_climber",
    "socialist",
    "socializer",
    "social_scientist",
    "social_secretary",
    "Socinian",
    "sociolinguist",
    "sociologist",
    "soda_jerk",
    "sodalist",
    "sodomite",
    "soldier",
    "son",
    "songster",
    "songstress",
    "songwriter",
    "sorcerer",
    "sorehead",
    "soul_mate",
    "Southern_Baptist",
    "sovereign",
    "spacewalker",
    "Spanish_American",
    "sparring_partner",
    "spastic",
    "speaker",
    "native_speaker",
    "Speaker",
    "speechwriter",
    "specialist",
    "specifier",
    "spectator",
    "speech_therapist",
    "speedskater",
    "spellbinder",
    "sphinx",
    "spinster",
    "split_end",
    "sport",
    "sport",
    "sporting_man",
    "sports_announcer",
    "sports_editor",
    "sprog",
    "square_dancer",
    "square_shooter",
    "squatter",
    "squire",
    "squire",
    "staff_member",
    "staff_sergeant",
    "stage_director",
    "stainer",
    "stakeholder",
    "stalker",
    "stalking-horse",
    "stammerer",
    "stamper",
    "standee",
    "stand-in",
    "star",
    "starlet",
    "starter",
    "statesman",
    "state_treasurer",
    "stationer",
    "stenographer",
    "stentor",
    "stepbrother",
    "stepmother",
    "stepparent",
    "stevedore",
    "steward",
    "steward",
    "steward",
    "stickler",
    "stiff",
    "stifler",
    "stipendiary",
    "stitcher",
    "stockjobber",
    "stock_trader",
    "stockist",
    "stoker",
    "stooper",
    "store_detective",
    "strafer",
    "straight_man",
    "stranger",
    "stranger",
    "strategist",
    "straw_boss",
    "streetwalker",
    "stretcher-bearer",
    "struggler",
    "stud",
    "student",
    "stumblebum",
    "stylist",
    "subaltern",
    "subcontractor",
    "subduer",
    "subject",
    "subordinate",
    "substitute",
    "successor",
    "successor",
    "succorer",
    "Sufi",
    "suffragan",
    "suffragette",
    "sugar_daddy",
    "suicide_bomber",
    "suitor",
    "sumo_wrestler",
    "sunbather",
    "sundowner",
    "super_heavyweight",
    "superior",
    "supermom",
    "supernumerary",
    "supremo",
    "surgeon",
    "Surgeon_General",
    "Surgeon_General",
    "surpriser",
    "surveyor",
    "surveyor",
    "survivor",
    "sutler",
    "sweeper",
    "sweetheart",
    "swinger",
    "switcher",
    "swot",
    "sycophant",
    "sylph",
    "sympathizer",
    "symphonist",
    "syncopator",
    "syndic",
    "tactician",
    "tagger",
    "tailback",
    "tallyman",
    "tallyman",
    "tanker",
    "tapper",
    "Tartuffe",
    "Tarzan",
    "taster",
    "tax_assessor",
    "taxer",
    "taxi_dancer",
    "taxonomist",
    "teacher",
    "teaching_fellow",
    "tearaway",
    "technical_sergeant",
    "technician",
    "Ted",
    "teetotaler",
    "television_reporter",
    "temporizer",
    "tempter",
    "term_infant",
    "toiler",
    "tenant",
    "tenant",
    "tenderfoot",
    "tennis_player",
    "tennis_pro",
    "tenor_saxophonist",
    "termer",
    "terror",
    "tertigravida",
    "testator",
    "testatrix",
    "testee",
    "test-tube_baby",
    "Texas_Ranger",
    "thane",
    "theatrical_producer",
    "theologian",
    "theorist",
    "theosophist",
    "therapist",
    "Thessalonian",
    "thinker",
    "thinker",
    "thrower",
    "thurifer",
    "ticket_collector",
    "tight_end",
    "tiler",
    "timekeeper",
    "Timorese",
    "tinkerer",
    "tinsmith",
    "tinter",
    "tippler",
    "tipster",
    "T-man",
    "toastmaster",
    "toast_mistress",
    "tobogganist",
    "tomboy",
    "toolmaker",
    "torchbearer",
    "Tory",
    "Tory",
    "tosser",
    "tosser",
    "totalitarian",
    "tourist",
    "tout",
    "tout",
    "tovarich",
    "towhead",
    "town_clerk",
    "town_crier",
    "townsman",
    "toxicologist",
    "track_star",
    "trader",
    "trade_unionist",
    "traditionalist",
    "traffic_cop",
    "tragedian",
    "tragedian",
    "tragedienne",
    "trail_boss",
    "trainer",
    "traitor",
    "traitress",
    "transactor",
    "transcriber",
    "transfer",
    "transferee",
    "translator",
    "transvestite",
    "traveling_salesman",
    "traverser",
    "trawler",
    "Treasury",
    "trencher",
    "trend-setter",
    "tribesman",
    "trier",
    "trifler",
    "trooper",
    "trooper",
    "Trotskyite",
    "truant",
    "trumpeter",
    "trusty",
    "Tudor",
    "tumbler",
    "tutee",
    "twin",
    "two-timer",
    "Tyke",
    "tympanist",
    "typist",
    "tyrant",
    "umpire",
    "understudy",
    "undesirable",
    "unicyclist",
    "unilateralist",
    "Unitarian",
    "Arminian",
    "universal_donor",
    "UNIX_guru",
    "Unknown_Soldier",
    "upsetter",
    "upstager",
    "upstart",
    "upstart",
    "urchin",
    "urologist",
    "usherette",
    "usher",
    "usurper",
    "utility_man",
    "utilizer",
    "Utopian",
    "uxoricide",
    "vacationer",
    "valedictorian",
    "valley_girl",
    "vaulter",
    "vegetarian",
    "vegan",
    "venerator",
    "venture_capitalist",
    "venturer",
    "vermin",
    "very_important_person",
    "vibist",
    "vicar",
    "vicar",
    "vicar-general",
    "vice_chancellor",
    "vicegerent",
    "vice_president",
    "vice-regent",
    "victim",
    "Victorian",
    "victualer",
    "vigilante",
    "villager",
    "vintager",
    "vintner",
    "violator",
    "violator",
    "violist",
    "virago",
    "virologist",
    "Visayan",
    "viscountess",
    "viscount",
    "Visigoth",
    "visionary",
    "visiting_fireman",
    "visiting_professor",
    "visualizer",
    "vixen",
    "vizier",
    "voicer",
    "volunteer",
    "volunteer",
    "votary",
    "votary",
    "vouchee",
    "vower",
    "voyager",
    "voyeur",
    "vulcanizer",
    "waffler",
    "Wagnerian",
    "waif",
    "wailer",
    "waiter",
    "waitress",
    "walking_delegate",
    "walk-on",
    "wallah",
    "wally",
    "waltzer",
    "wanderer",
    "Wandering_Jew",
    "wanton",
    "warrantee",
    "warrantee",
    "washer",
    "washerman",
    "washwoman",
    "wassailer",
    "wastrel",
    "Wave",
    "weatherman",
    "weekend_warrior",
    "weeder",
    "welder",
    "welfare_case",
    "westerner",
    "West-sider",
    "wetter",
    "whaler",
    "Whig",
    "whiner",
    "whipper-in",
    "whisperer",
    "whiteface",
    "Carmelite",
    "Augustinian",
    "white_hope",
    "white_supremacist",
    "whoremaster",
    "whoremaster",
    "widow",
    "wife",
    "wiggler",
    "wimp",
    "wing_commander",
    "winger",
    "winner",
    "winner",
    "window_dresser",
    "winker",
    "wiper",
    "wireman",
    "wise_guy",
    "witch_doctor",
    "withdrawer",
    "withdrawer",
    "woman",
    "woman",
    "wonder_boy",
    "wonderer",
    "working_girl",
    "workman",
    "workmate",
    "worldling",
    "worshiper",
    "worthy",
    "wrecker",
    "wright",
    "write-in_candidate",
    "writer",
    "Wykehamist",
    "yakuza",
    "yard_bird",
    "yardie",
    "yardman",
    "yardmaster",
    "yenta",
    "yogi",
    "young_buck",
    "young_Turk",
    "Young_Turk",
    "Zionist",
    "zoo_keeper",
    "Genet",
    "Kennan",
    "Munro",
    "Popper",
    "Stoker",
    "Townes",
    "dust_storm",
    "parhelion",
    "snow",
    "facula",
    "wave",
    "microflora",
    "wilding",
    "semi-climber",
    "volva",
    "basidiocarp",
    "domatium",
    "apomict",
    "aquatic",
    "bryophyte",
    "acrocarp",
    "sphagnum",
    "liverwort",
    "hepatica",
    "pecopteris",
    "pteridophyte",
    "fern",
    "fern_ally",
    "spore",
    "carpospore",
    "chlamydospore",
    "conidium",
    "oospore",
    "tetraspore",
    "zoospore",
    "cryptogam",
    "spermatophyte",
    "seedling",
    "annual",
    "biennial",
    "perennial",
    "hygrophyte",
    "gymnosperm",
    "gnetum",
    "Catha_edulis",
    "ephedra",
    "mahuang",
    "welwitschia",
    "cycad",
    "sago_palm",
    "false_sago",
    "zamia",
    "coontie",
    "ceratozamia",
    "dioon",
    "encephalartos",
    "kaffir_bread",
    "macrozamia",
    "burrawong",
    "pine",
    "pinon",
    "nut_pine",
    "pinon_pine",
    "Rocky_mountain_pinon",
    "single-leaf",
    "bishop_pine",
    "California_single-leaf_pinyon",
    "Parry's_pinyon",
    "spruce_pine",
    "black_pine",
    "pitch_pine",
    "pond_pine",
    "stone_pine",
    "Swiss_pine",
    "cembra_nut",
    "Swiss_mountain_pine",
    "ancient_pine",
    "white_pine",
    "American_white_pine",
    "western_white_pine",
    "southwestern_white_pine",
    "limber_pine",
    "whitebark_pine",
    "yellow_pine",
    "ponderosa",
    "Jeffrey_pine",
    "shore_pine",
    "Sierra_lodgepole_pine",
    "loblolly_pine",
    "jack_pine",
    "swamp_pine",
    "longleaf_pine",
    "shortleaf_pine",
    "red_pine",
    "Scotch_pine",
    "scrub_pine",
    "Monterey_pine",
    "bristlecone_pine",
    "table-mountain_pine",
    "knobcone_pine",
    "Japanese_red_pine",
    "Japanese_black_pine",
    "Torrey_pine",
    "larch",
    "American_larch",
    "western_larch",
    "subalpine_larch",
    "European_larch",
    "Siberian_larch",
    "golden_larch",
    "fir",
    "silver_fir",
    "amabilis_fir",
    "European_silver_fir",
    "white_fir",
    "balsam_fir",
    "Fraser_fir",
    "lowland_fir",
    "Alpine_fir",
    "Santa_Lucia_fir",
    "cedar",
    "cedar_of_Lebanon",
    "deodar",
    "Atlas_cedar",
    "spruce",
    "Norway_spruce",
    "weeping_spruce",
    "Engelmann_spruce",
    "white_spruce",
    "black_spruce",
    "Siberian_spruce",
    "Sitka_spruce",
    "oriental_spruce",
    "Colorado_spruce",
    "red_spruce",
    "hemlock",
    "eastern_hemlock",
    "Carolina_hemlock",
    "mountain_hemlock",
    "western_hemlock",
    "douglas_fir",
    "green_douglas_fir",
    "big-cone_spruce",
    "Cathaya",
    "cedar",
    "cypress",
    "gowen_cypress",
    "pygmy_cypress",
    "Santa_Cruz_cypress",
    "Arizona_cypress",
    "Guadalupe_cypress",
    "Monterey_cypress",
    "Mexican_cypress",
    "Italian_cypress",
    "King_William_pine",
    "Chilean_cedar",
    "incense_cedar",
    "southern_white_cedar",
    "Oregon_cedar",
    "yellow_cypress",
    "Japanese_cedar",
    "juniper_berry",
    "incense_cedar",
    "kawaka",
    "pahautea",
    "metasequoia",
    "arborvitae",
    "western_red_cedar",
    "American_arborvitae",
    "Oriental_arborvitae",
    "hiba_arborvitae",
    "keteleeria",
    "Wollemi_pine",
    "araucaria",
    "monkey_puzzle",
    "norfolk_island_pine",
    "new_caledonian_pine",
    "bunya_bunya",
    "hoop_pine",
    "kauri_pine",
    "kauri",
    "amboina_pine",
    "dundathu_pine",
    "red_kauri",
    "plum-yew",
    "California_nutmeg",
    "stinking_cedar",
    "celery_pine",
    "celery_top_pine",
    "tanekaha",
    "Alpine_celery_pine",
    "yellowwood",
    "gymnospermous_yellowwood",
    "podocarp",
    "yacca",
    "brown_pine",
    "cape_yellowwood",
    "South-African_yellowwood",
    "alpine_totara",
    "totara",
    "common_yellowwood",
    "kahikatea",
    "rimu",
    "tarwood",
    "common_sickle_pine",
    "yellow-leaf_sickle_pine",
    "tarwood",
    "westland_pine",
    "huon_pine",
    "Chilean_rimu",
    "mountain_rimu",
    "nagi",
    "miro",
    "matai",
    "plum-fruited_yew",
    "Prince_Albert_yew",
    "Sundacarpus_amara",
    "Japanese_umbrella_pine",
    "yew",
    "Old_World_yew",
    "Pacific_yew",
    "Japanese_yew",
    "Florida_yew",
    "New_Caledonian_yew",
    "white-berry_yew",
    "ginkgo",
    "angiosperm",
    "dicot",
    "monocot",
    "floret",
    "flower",
    "bloomer",
    "wildflower",
    "apetalous_flower",
    "inflorescence",
    "rosebud",
    "gynostegium",
    "pollinium",
    "pistil",
    "gynobase",
    "gynophore",
    "stylopodium",
    "carpophore",
    "cornstalk",
    "petiolule",
    "mericarp",
    "micropyle",
    "germ_tube",
    "pollen_tube",
    "gemma",
    "galbulus",
    "nectary",
    "pericarp",
    "epicarp",
    "mesocarp",
    "pip",
    "silique",
    "cataphyll",
    "perisperm",
    "monocarp",
    "sporophyte",
    "gametophyte",
    "megasporangium",
    "microspore",
    "microsporangium",
    "microsporophyll",
    "archespore",
    "bonduc_nut",
    "Job's_tears",
    "oilseed",
    "castor_bean",
    "cottonseed",
    "candlenut",
    "peach_pit",
    "hypanthium",
    "petal",
    "corolla",
    "lip",
    "perianth",
    "thistledown",
    "custard_apple",
    "cherimoya",
    "ilama",
    "soursop",
    "bullock's_heart",
    "sweetsop",
    "pond_apple",
    "pawpaw",
    "ilang-ilang",
    "lancewood",
    "Guinea_pepper",
    "barberry",
    "American_barberry",
    "common_barberry",
    "Japanese_barberry",
    "Oregon_grape",
    "Oregon_grape",
    "mayapple",
    "May_apple",
    "allspice",
    "Carolina_allspice",
    "spicebush",
    "katsura_tree",
    "laurel",
    "true_laurel",
    "camphor_tree",
    "cinnamon",
    "cassia",
    "cassia_bark",
    "Saigon_cinnamon",
    "cinnamon_bark",
    "spicebush",
    "avocado",
    "laurel-tree",
    "sassafras",
    "California_laurel",
    "anise_tree",
    "purple_anise",
    "star_anise",
    "star_anise",
    "magnolia",
    "southern_magnolia",
    "umbrella_tree",
    "earleaved_umbrella_tree",
    "cucumber_tree",
    "large-leaved_magnolia",
    "saucer_magnolia",
    "star_magnolia",
    "sweet_bay",
    "manglietia",
    "tulip_tree",
    "moonseed",
    "common_moonseed",
    "Carolina_moonseed",
    "nutmeg",
    "water_nymph",
    "European_white_lily",
    "southern_spatterdock",
    "lotus",
    "water_chinquapin",
    "water-shield",
    "water-shield",
    "peony",
    "buttercup",
    "meadow_buttercup",
    "water_crowfoot",
    "lesser_celandine",
    "lesser_spearwort",
    "greater_spearwort",
    "western_buttercup",
    "creeping_buttercup",
    "cursed_crowfoot",
    "aconite",
    "monkshood",
    "wolfsbane",
    "baneberry",
    "baneberry",
    "red_baneberry",
    "pheasant's-eye",
    "anemone",
    "Alpine_anemone",
    "Canada_anemone",
    "thimbleweed",
    "wood_anemone",
    "wood_anemone",
    "longheaded_thimbleweed",
    "snowdrop_anemone",
    "Virginia_thimbleweed",
    "rue_anemone",
    "columbine",
    "meeting_house",
    "blue_columbine",
    "granny's_bonnets",
    "marsh_marigold",
    "American_bugbane",
    "black_cohosh",
    "fetid_bugbane",
    "clematis",
    "pine_hyacinth",
    "blue_jasmine",
    "golden_clematis",
    "scarlet_clematis",
    "leather_flower",
    "leather_flower",
    "virgin's_bower",
    "purple_clematis",
    "goldthread",
    "rocket_larkspur",
    "delphinium",
    "larkspur",
    "winter_aconite",
    "lenten_rose",
    "green_hellebore",
    "hepatica",
    "goldenseal",
    "false_rue_anemone",
    "giant_buttercup",
    "nigella",
    "love-in-a-mist",
    "fennel_flower",
    "black_caraway",
    "pasqueflower",
    "meadow_rue",
    "false_bugbane",
    "globeflower",
    "winter's_bark",
    "pepper_shrub",
    "sweet_gale",
    "wax_myrtle",
    "bay_myrtle",
    "bayberry",
    "sweet_fern",
    "corkwood",
    "jointed_rush",
    "toad_rush",
    "slender_rush",
    "zebrawood",
    "Connarus_guianensis",
    "legume",
    "legume",
    "peanut",
    "granadilla_tree",
    "arariba",
    "tonka_bean",
    "courbaril",
    "melilotus",
    "darling_pea",
    "smooth_darling_pea",
    "clover",
    "alpine_clover",
    "hop_clover",
    "crimson_clover",
    "red_clover",
    "buffalo_clover",
    "white_clover",
    "mimosa",
    "acacia",
    "shittah",
    "wattle",
    "black_wattle",
    "gidgee",
    "catechu",
    "silver_wattle",
    "huisache",
    "lightwood",
    "golden_wattle",
    "fever_tree",
    "coralwood",
    "albizzia",
    "silk_tree",
    "siris",
    "rain_tree",
    "calliandra",
    "conacaste",
    "inga",
    "ice-cream_bean",
    "guama",
    "lead_tree",
    "wild_tamarind",
    "sabicu",
    "nitta_tree",
    "Parkia_javanica",
    "manila_tamarind",
    "cat's-claw",
    "honey_mesquite",
    "algarroba",
    "screw_bean",
    "screw_bean",
    "dogbane",
    "Indian_hemp",
    "bushman's_poison",
    "impala_lily",
    "allamanda",
    "common_allamanda",
    "dita",
    "Nepal_trumpet_flower",
    "carissa",
    "hedge_thorn",
    "natal_plum",
    "periwinkle",
    "ivory_tree",
    "white_dipladenia",
    "Chilean_jasmine",
    "oleander",
    "frangipani",
    "West_Indian_jasmine",
    "rauwolfia",
    "snakewood",
    "Strophanthus_kombe",
    "yellow_oleander",
    "myrtle",
    "large_periwinkle",
    "arum",
    "cuckoopint",
    "black_calla",
    "calamus",
    "alocasia",
    "giant_taro",
    "amorphophallus",
    "pungapung",
    "devil's_tongue",
    "anthurium",
    "flamingo_flower",
    "jack-in-the-pulpit",
    "friar's-cowl",
    "caladium",
    "Caladium_bicolor",
    "wild_calla",
    "taro",
    "taro",
    "cryptocoryne",
    "dracontium",
    "golden_pothos",
    "skunk_cabbage",
    "monstera",
    "ceriman",
    "nephthytis",
    "Nephthytis_afzelii",
    "arrow_arum",
    "green_arrow_arum",
    "philodendron",
    "pistia",
    "pothos",
    "spathiphyllum",
    "skunk_cabbage",
    "yautia",
    "calla_lily",
    "pink_calla",
    "golden_calla",
    "duckweed",
    "common_duckweed",
    "star-duckweed",
    "great_duckweed",
    "watermeal",
    "common_wolffia",
    "aralia",
    "American_angelica_tree",
    "American_spikenard",
    "bristly_sarsaparilla",
    "Japanese_angelica_tree",
    "Chinese_angelica",
    "ivy",
    "puka",
    "ginseng",
    "ginseng",
    "umbrella_tree",
    "birthwort",
    "Dutchman's-pipe",
    "Virginia_snakeroot",
    "Canada_ginger",
    "heartleaf",
    "heartleaf",
    "asarabacca",
    "caryophyllaceous_plant",
    "corn_cockle",
    "sandwort",
    "mountain_sandwort",
    "pine-barren_sandwort",
    "seabeach_sandwort",
    "rock_sandwort",
    "thyme-leaved_sandwort",
    "mouse-ear_chickweed",
    "snow-in-summer",
    "Alpine_mouse-ear",
    "pink",
    "sweet_William",
    "carnation",
    "china_pink",
    "Japanese_pink",
    "maiden_pink",
    "cheddar_pink",
    "button_pink",
    "cottage_pink",
    "fringed_pink",
    "drypis",
    "baby's_breath",
    "coral_necklace",
    "lychnis",
    "ragged_robin",
    "scarlet_lychnis",
    "mullein_pink",
    "sandwort",
    "sandwort",
    "soapwort",
    "knawel",
    "silene",
    "moss_campion",
    "wild_pink",
    "red_campion",
    "white_campion",
    "fire_pink",
    "bladder_campion",
    "corn_spurry",
    "sand_spurry",
    "chickweed",
    "common_chickweed",
    "cowherb",
    "Hottentot_fig",
    "livingstone_daisy",
    "fig_marigold",
    "ice_plant",
    "New_Zealand_spinach",
    "amaranth",
    "amaranth",
    "tumbleweed",
    "prince's-feather",
    "pigweed",
    "thorny_amaranth",
    "alligator_weed",
    "cockscomb",
    "cottonweed",
    "globe_amaranth",
    "bloodleaf",
    "saltwort",
    "lamb's-quarters",
    "good-king-henry",
    "Jerusalem_oak",
    "oak-leaved_goosefoot",
    "sowbane",
    "nettle-leaved_goosefoot",
    "red_goosefoot",
    "stinking_goosefoot",
    "orach",
    "saltbush",
    "garden_orache",
    "desert_holly",
    "quail_bush",
    "beet",
    "beetroot",
    "chard",
    "mangel-wurzel",
    "winged_pigweed",
    "halogeton",
    "glasswort",
    "saltwort",
    "Russian_thistle",
    "greasewood",
    "scarlet_musk_flower",
    "sand_verbena",
    "sweet_sand_verbena",
    "yellow_sand_verbena",
    "beach_pancake",
    "beach_sand_verbena",
    "desert_sand_verbena",
    "trailing_four_o'clock",
    "bougainvillea",
    "umbrellawort",
    "four_o'clock",
    "common_four-o'clock",
    "California_four_o'clock",
    "sweet_four_o'clock",
    "desert_four_o'clock",
    "mountain_four_o'clock",
    "cockspur",
    "rattail_cactus",
    "saguaro",
    "night-blooming_cereus",
    "echinocactus",
    "hedgehog_cactus",
    "golden_barrel_cactus",
    "hedgehog_cereus",
    "rainbow_cactus",
    "epiphyllum",
    "barrel_cactus",
    "night-blooming_cereus",
    "chichipe",
    "mescal",
    "mescal_button",
    "mammillaria",
    "feather_ball",
    "garambulla",
    "Knowlton's_cactus",
    "nopal",
    "prickly_pear",
    "cholla",
    "nopal",
    "tuna",
    "Barbados_gooseberry",
    "mistletoe_cactus",
    "Christmas_cactus",
    "night-blooming_cereus",
    "crab_cactus",
    "pokeweed",
    "Indian_poke",
    "poke",
    "ombu",
    "bloodberry",
    "portulaca",
    "rose_moss",
    "common_purslane",
    "rock_purslane",
    "red_maids",
    "Carolina_spring_beauty",
    "spring_beauty",
    "Virginia_spring_beauty",
    "siskiyou_lewisia",
    "bitterroot",
    "broad-leaved_montia",
    "blinks",
    "toad_lily",
    "winter_purslane",
    "flame_flower",
    "pigmy_talinum",
    "jewels-of-opar",
    "caper",
    "native_pomegranate",
    "caper_tree",
    "caper_tree",
    "common_caper",
    "spiderflower",
    "Rocky_Mountain_bee_plant",
    "clammyweed",
    "crucifer",
    "cress",
    "watercress",
    "stonecress",
    "garlic_mustard",
    "alyssum",
    "rose_of_Jericho",
    "Arabidopsis_thaliana",
    "Arabidopsis_lyrata",
    "rock_cress",
    "sicklepod",
    "tower_mustard",
    "horseradish",
    "winter_cress",
    "yellow_rocket",
    "hoary_alison",
    "buckler_mustard",
    "wild_cabbage",
    "cabbage",
    "head_cabbage",
    "savoy_cabbage",
    "brussels_sprout",
    "cauliflower",
    "broccoli",
    "collard",
    "kohlrabi",
    "turnip_plant",
    "turnip",
    "rutabaga",
    "broccoli_raab",
    "mustard",
    "chinese_mustard",
    "bok_choy",
    "rape",
    "rapeseed",
    "shepherd's_purse",
    "lady's_smock",
    "coral-root_bittercress",
    "crinkleroot",
    "American_watercress",
    "spring_cress",
    "purple_cress",
    "wallflower",
    "prairie_rocket",
    "scurvy_grass",
    "sea_kale",
    "tansy_mustard",
    "draba",
    "wallflower",
    "prairie_rocket",
    "Siberian_wall_flower",
    "western_wall_flower",
    "wormseed_mustard",
    "heliophila",
    "damask_violet",
    "tansy-leaved_rocket",
    "candytuft",
    "woad",
    "dyer's_woad",
    "bladderpod",
    "sweet_alyssum",
    "Malcolm_stock",
    "Virginian_stock",
    "stock",
    "brompton_stock",
    "bladderpod",
    "chamois_cress",
    "radish_plant",
    "jointed_charlock",
    "radish",
    "radish",
    "marsh_cress",
    "great_yellowcress",
    "schizopetalon",
    "field_mustard",
    "hedge_mustard",
    "desert_plume",
    "pennycress",
    "field_pennycress",
    "fringepod",
    "bladderpod",
    "wasabi",
    "poppy",
    "Iceland_poppy",
    "western_poppy",
    "prickly_poppy",
    "Iceland_poppy",
    "oriental_poppy",
    "corn_poppy",
    "opium_poppy",
    "prickly_poppy",
    "Mexican_poppy",
    "bocconia",
    "celandine",
    "corydalis",
    "climbing_corydalis",
    "California_poppy",
    "horn_poppy",
    "golden_cup",
    "plume_poppy",
    "blue_poppy",
    "Welsh_poppy",
    "creamcups",
    "matilija_poppy",
    "wind_poppy",
    "celandine_poppy",
    "climbing_fumitory",
    "bleeding_heart",
    "Dutchman's_breeches",
    "squirrel_corn",
    "composite",
    "compass_plant",
    "everlasting",
    "achillea",
    "yarrow",
    "pink-and-white_everlasting",
    "white_snakeroot",
    "ageratum",
    "common_ageratum",
    "sweet_sultan",
    "ragweed",
    "common_ragweed",
    "great_ragweed",
    "western_ragweed",
    "ammobium",
    "winged_everlasting",
    "pellitory",
    "pearly_everlasting",
    "andryala",
    "plantain-leaved_pussytoes",
    "field_pussytoes",
    "solitary_pussytoes",
    "mountain_everlasting",
    "mayweed",
    "yellow_chamomile",
    "corn_chamomile",
    "woolly_daisy",
    "burdock",
    "great_burdock",
    "African_daisy",
    "blue-eyed_African_daisy",
    "marguerite",
    "silversword",
    "arnica",
    "heartleaf_arnica",
    "Arnica_montana",
    "lamb_succory",
    "artemisia",
    "mugwort",
    "sweet_wormwood",
    "field_wormwood",
    "tarragon",
    "sand_sage",
    "wormwood_sage",
    "western_mugwort",
    "Roman_wormwood",
    "bud_brush",
    "common_mugwort",
    "aster",
    "wood_aster",
    "whorled_aster",
    "heath_aster",
    "heart-leaved_aster",
    "white_wood_aster",
    "bushy_aster",
    "heath_aster",
    "white_prairie_aster",
    "stiff_aster",
    "goldilocks",
    "large-leaved_aster",
    "New_England_aster",
    "Michaelmas_daisy",
    "upland_white_aster",
    "Short's_aster",
    "sea_aster",
    "prairie_aster",
    "annual_salt-marsh_aster",
    "aromatic_aster",
    "arrow_leaved_aster",
    "azure_aster",
    "bog_aster",
    "crooked-stemmed_aster",
    "Eastern_silvery_aster",
    "flat-topped_white_aster",
    "late_purple_aster",
    "panicled_aster",
    "perennial_salt_marsh_aster",
    "purple-stemmed_aster",
    "rough-leaved_aster",
    "rush_aster",
    "Schreiber's_aster",
    "small_white_aster",
    "smooth_aster",
    "southern_aster",
    "starved_aster",
    "tradescant's_aster",
    "wavy-leaved_aster",
    "Western_silvery_aster",
    "willow_aster",
    "ayapana",
    "mule_fat",
    "balsamroot",
    "daisy",
    "common_daisy",
    "bur_marigold",
    "Spanish_needles",
    "tickseed_sunflower",
    "European_beggar-ticks",
    "slender_knapweed",
    "false_chamomile",
    "Swan_River_daisy",
    "woodland_oxeye",
    "Indian_plantain",
    "calendula",
    "common_marigold",
    "China_aster",
    "thistle",
    "welted_thistle",
    "musk_thistle",
    "carline_thistle",
    "stemless_carline_thistle",
    "common_carline_thistle",
    "safflower",
    "safflower_seed",
    "catananche",
    "blue_succory",
    "centaury",
    "dusty_miller",
    "cornflower",
    "star-thistle",
    "knapweed",
    "sweet_sultan",
    "great_knapweed",
    "Barnaby's_thistle",
    "chamomile",
    "chaenactis",
    "chrysanthemum",
    "corn_marigold",
    "crown_daisy",
    "chop-suey_greens",
    "golden_aster",
    "Maryland_golden_aster",
    "goldenbush",
    "rabbit_brush",
    "chicory",
    "endive",
    "chicory",
    "plume_thistle",
    "Canada_thistle",
    "field_thistle",
    "woolly_thistle",
    "European_woolly_thistle",
    "melancholy_thistle",
    "brook_thistle",
    "bull_thistle",
    "blessed_thistle",
    "mistflower",
    "horseweed",
    "coreopsis",
    "giant_coreopsis",
    "sea_dahlia",
    "calliopsis",
    "cosmos",
    "brass_buttons",
    "billy_buttons",
    "hawk's-beard",
    "artichoke",
    "cardoon",
    "dahlia",
    "German_ivy",
    "florist's_chrysanthemum",
    "cape_marigold",
    "leopard's-bane",
    "coneflower",
    "globe_thistle",
    "elephant's-foot",
    "tassel_flower",
    "brittlebush",
    "sunray",
    "engelmannia",
    "fireweed",
    "fleabane",
    "blue_fleabane",
    "daisy_fleabane",
    "orange_daisy",
    "spreading_fleabane",
    "seaside_daisy",
    "Philadelphia_fleabane",
    "robin's_plantain",
    "showy_daisy",
    "woolly_sunflower",
    "golden_yarrow",
    "dog_fennel",
    "Joe-Pye_weed",
    "boneset",
    "Joe-Pye_weed",
    "blue_daisy",
    "kingfisher_daisy",
    "cotton_rose",
    "herba_impia",
    "gaillardia",
    "gazania",
    "treasure_flower",
    "African_daisy",
    "Barberton_daisy",
    "desert_sunflower",
    "cudweed",
    "chafeweed",
    "gumweed",
    "Grindelia_robusta",
    "curlycup_gumweed",
    "little-head_snakeweed",
    "rabbitweed",
    "broomweed",
    "velvet_plant",
    "goldenbush",
    "camphor_daisy",
    "yellow_spiny_daisy",
    "hoary_golden_bush",
    "sneezeweed",
    "orange_sneezeweed",
    "rosilla",
    "sunflower",
    "swamp_sunflower",
    "common_sunflower",
    "giant_sunflower",
    "showy_sunflower",
    "Maximilian's_sunflower",
    "prairie_sunflower",
    "Jerusalem_artichoke",
    "Jerusalem_artichoke",
    "strawflower",
    "heliopsis",
    "strawflower",
    "hairy_golden_aster",
    "hawkweed",
    "rattlesnake_weed",
    "alpine_coltsfoot",
    "alpine_gold",
    "dwarf_hulsea",
    "cat's-ear",
    "inula",
    "marsh_elder",
    "burweed_marsh_elder",
    "krigia",
    "dwarf_dandelion",
    "garden_lettuce",
    "cos_lettuce",
    "leaf_lettuce",
    "celtuce",
    "prickly_lettuce",
    "goldfields",
    "tidytips",
    "hawkbit",
    "fall_dandelion",
    "edelweiss",
    "oxeye_daisy",
    "oxeye_daisy",
    "shasta_daisy",
    "Pyrenees_daisy",
    "north_island_edelweiss",
    "blazing_star",
    "dotted_gayfeather",
    "dense_blazing_star",
    "Texas_star",
    "African_daisy",
    "tahoka_daisy",
    "sticky_aster",
    "Mojave_aster",
    "tarweed",
    "sweet_false_chamomile",
    "pineapple_weed",
    "climbing_hempweed",
    "mutisia",
    "rattlesnake_root",
    "white_lettuce",
    "daisybush",
    "New_Zealand_daisybush",
    "cotton_thistle",
    "othonna",
    "cascade_everlasting",
    "butterweed",
    "American_feverfew",
    "cineraria",
    "florest's_cineraria",
    "butterbur",
    "winter_heliotrope",
    "sweet_coltsfoot",
    "oxtongue",
    "hawkweed",
    "mouse-ear_hawkweed",
    "stevia",
    "rattlesnake_root",
    "fleabane",
    "sheep_plant",
    "coneflower",
    "Mexican_hat",
    "long-head_coneflower",
    "prairie_coneflower",
    "Swan_River_everlasting",
    "coneflower",
    "black-eyed_Susan",
    "cutleaved_coneflower",
    "golden_glow",
    "lavender_cotton",
    "creeping_zinnia",
    "golden_thistle",
    "Spanish_oyster_plant",
    "nodding_groundsel",
    "dusty_miller",
    "butterweed",
    "ragwort",
    "arrowleaf_groundsel",
    "black_salsify",
    "white-topped_aster",
    "narrow-leaved_white-topped_aster",
    "silver_sage",
    "sea_wormwood",
    "sawwort",
    "rosinweed",
    "milk_thistle",
    "goldenrod",
    "silverrod",
    "meadow_goldenrod",
    "Missouri_goldenrod",
    "alpine_goldenrod",
    "grey_goldenrod",
    "Blue_Mountain_tea",
    "dyer's_weed",
    "seaside_goldenrod",
    "narrow_goldenrod",
    "Boott's_goldenrod",
    "Elliott's_goldenrod",
    "Ohio_goldenrod",
    "rough-stemmed_goldenrod",
    "showy_goldenrod",
    "tall_goldenrod",
    "zigzag_goldenrod",
    "sow_thistle",
    "milkweed",
    "stevia",
    "stokes'_aster",
    "marigold",
    "African_marigold",
    "French_marigold",
    "painted_daisy",
    "pyrethrum",
    "northern_dune_tansy",
    "feverfew",
    "dusty_miller",
    "tansy",
    "dandelion",
    "common_dandelion",
    "dandelion_green",
    "Russian_dandelion",
    "stemless_hymenoxys",
    "Mexican_sunflower",
    "Easter_daisy",
    "yellow_salsify",
    "salsify",
    "meadow_salsify",
    "scentless_camomile",
    "turfing_daisy",
    "coltsfoot",
    "ursinia",
    "crownbeard",
    "wingstem",
    "cowpen_daisy",
    "gravelweed",
    "Virginia_crownbeard",
    "ironweed",
    "mule's_ears",
    "white-rayed_mule's_ears",
    "cocklebur",
    "xeranthemum",
    "immortelle",
    "zinnia",
    "white_zinnia",
    "little_golden_zinnia",
    "blazing_star",
    "bartonia",
    "achene",
    "samara",
    "campanula",
    "creeping_bellflower",
    "Canterbury_bell",
    "tall_bellflower",
    "marsh_bellflower",
    "clustered_bellflower",
    "peach_bells",
    "chimney_plant",
    "rampion",
    "tussock_bellflower",
    "orchid",
    "orchis",
    "male_orchis",
    "butterfly_orchid",
    "showy_orchis",
    "aerides",
    "angrecum",
    "jewel_orchid",
    "puttyroot",
    "arethusa",
    "bog_rose",
    "bletia",
    "Bletilla_striata",
    "brassavola",
    "spider_orchid",
    "spider_orchid",
    "caladenia",
    "calanthe",
    "grass_pink",
    "calypso",
    "cattleya",
    "helleborine",
    "red_helleborine",
    "spreading_pogonia",
    "rosebud_orchid",
    "satyr_orchid",
    "frog_orchid",
    "coelogyne",
    "coral_root",
    "spotted_coral_root",
    "striped_coral_root",
    "early_coral_root",
    "swan_orchid",
    "cymbid",
    "cypripedia",
    "lady's_slipper",
    "moccasin_flower",
    "common_lady's-slipper",
    "ram's-head",
    "yellow_lady's_slipper",
    "large_yellow_lady's_slipper",
    "California_lady's_slipper",
    "clustered_lady's_slipper",
    "mountain_lady's_slipper",
    "marsh_orchid",
    "common_spotted_orchid",
    "dendrobium",
    "disa",
    "phantom_orchid",
    "tulip_orchid",
    "butterfly_orchid",
    "butterfly_orchid",
    "epidendron",
    "helleborine",
    "Epipactis_helleborine",
    "stream_orchid",
    "tongueflower",
    "rattlesnake_plantain",
    "fragrant_orchid",
    "short-spurred_fragrant_orchid",
    "fringed_orchis",
    "frog_orchid",
    "rein_orchid",
    "bog_rein_orchid",
    "white_fringed_orchis",
    "elegant_Habenaria",
    "purple-fringed_orchid",
    "coastal_rein_orchid",
    "Hooker's_orchid",
    "ragged_orchid",
    "prairie_orchid",
    "snowy_orchid",
    "round-leaved_rein_orchid",
    "purple_fringeless_orchid",
    "purple-fringed_orchid",
    "Alaska_rein_orchid",
    "crested_coral_root",
    "Texas_purple_spike",
    "lizard_orchid",
    "laelia",
    "liparis",
    "twayblade",
    "fen_orchid",
    "broad-leaved_twayblade",
    "lesser_twayblade",
    "twayblade",
    "green_adder's_mouth",
    "masdevallia",
    "maxillaria",
    "pansy_orchid",
    "odontoglossum",
    "oncidium",
    "bee_orchid",
    "fly_orchid",
    "spider_orchid",
    "early_spider_orchid",
    "Venus'_slipper",
    "phaius",
    "moth_orchid",
    "butterfly_plant",
    "rattlesnake_orchid",
    "lesser_butterfly_orchid",
    "greater_butterfly_orchid",
    "prairie_white-fringed_orchid",
    "tangle_orchid",
    "Indian_crocus",
    "pleurothallis",
    "pogonia",
    "butterfly_orchid",
    "Psychopsis_krameriana",
    "Psychopsis_papilio",
    "helmet_orchid",
    "foxtail_orchid",
    "orange-blossom_orchid",
    "sobralia",
    "ladies'_tresses",
    "screw_augur",
    "hooded_ladies'_tresses",
    "western_ladies'_tresses",
    "European_ladies'_tresses",
    "stanhopea",
    "stelis",
    "fly_orchid",
    "vanda",
    "blue_orchid",
    "vanilla",
    "vanilla_orchid",
    "yam",
    "yam",
    "white_yam",
    "cinnamon_vine",
    "elephant's-foot",
    "wild_yam",
    "cush-cush",
    "black_bryony",
    "primrose",
    "English_primrose",
    "cowslip",
    "oxlip",
    "Chinese_primrose",
    "polyanthus",
    "pimpernel",
    "scarlet_pimpernel",
    "bog_pimpernel",
    "chaffweed",
    "cyclamen",
    "sowbread",
    "sea_milkwort",
    "featherfoil",
    "water_gillyflower",
    "water_violet",
    "loosestrife",
    "gooseneck_loosestrife",
    "yellow_pimpernel",
    "fringed_loosestrife",
    "moneywort",
    "swamp_candles",
    "whorled_loosestrife",
    "water_pimpernel",
    "brookweed",
    "brookweed",
    "coralberry",
    "marlberry",
    "plumbago",
    "leadwort",
    "thrift",
    "sea_lavender",
    "barbasco",
    "gramineous_plant",
    "grass",
    "midgrass",
    "shortgrass",
    "sword_grass",
    "tallgrass",
    "herbage",
    "goat_grass",
    "wheatgrass",
    "crested_wheatgrass",
    "bearded_wheatgrass",
    "western_wheatgrass",
    "intermediate_wheatgrass",
    "slender_wheatgrass",
    "velvet_bent",
    "cloud_grass",
    "meadow_foxtail",
    "foxtail",
    "broom_grass",
    "broom_sedge",
    "tall_oat_grass",
    "toetoe",
    "oat",
    "cereal_oat",
    "wild_oat",
    "slender_wild_oat",
    "wild_red_oat",
    "brome",
    "chess",
    "field_brome",
    "grama",
    "black_grama",
    "buffalo_grass",
    "reed_grass",
    "feather_reed_grass",
    "Australian_reed_grass",
    "burgrass",
    "buffel_grass",
    "Rhodes_grass",
    "pampas_grass",
    "giant_star_grass",
    "orchard_grass",
    "Egyptian_grass",
    "crabgrass",
    "smooth_crabgrass",
    "large_crabgrass",
    "barnyard_grass",
    "Japanese_millet",
    "yardgrass",
    "finger_millet",
    "lyme_grass",
    "wild_rye",
    "giant_ryegrass",
    "sea_lyme_grass",
    "Canada_wild_rye",
    "teff",
    "weeping_love_grass",
    "plume_grass",
    "Ravenna_grass",
    "fescue",
    "reed_meadow_grass",
    "velvet_grass",
    "creeping_soft_grass",
    "barleycorn",
    "barley_grass",
    "little_barley",
    "rye_grass",
    "perennial_ryegrass",
    "Italian_ryegrass",
    "darnel",
    "nimblewill",
    "cultivated_rice",
    "ricegrass",
    "smilo",
    "switch_grass",
    "broomcorn_millet",
    "goose_grass",
    "dallisgrass",
    "Bahia_grass",
    "knotgrass",
    "fountain_grass",
    "reed_canary_grass",
    "canary_grass",
    "timothy",
    "bluegrass",
    "meadowgrass",
    "wood_meadowgrass",
    "noble_cane",
    "munj",
    "broom_beard_grass",
    "bluestem",
    "rye",
    "bristlegrass",
    "giant_foxtail",
    "yellow_bristlegrass",
    "green_bristlegrass",
    "Siberian_millet",
    "German_millet",
    "millet",
    "rattan",
    "malacca",
    "reed",
    "sorghum",
    "grain_sorghum",
    "durra",
    "feterita",
    "hegari",
    "kaoliang",
    "milo",
    "shallu",
    "broomcorn",
    "cordgrass",
    "salt_reed_grass",
    "prairie_cordgrass",
    "smut_grass",
    "sand_dropseed",
    "rush_grass",
    "St._Augustine_grass",
    "grain",
    "cereal",
    "wheat",
    "wheat_berry",
    "durum",
    "spelt",
    "emmer",
    "wild_wheat",
    "corn",
    "mealie",
    "corn",
    "dent_corn",
    "flint_corn",
    "popcorn",
    "zoysia",
    "Manila_grass",
    "Korean_lawn_grass",
    "bamboo",
    "common_bamboo",
    "giant_bamboo",
    "umbrella_plant",
    "chufa",
    "galingale",
    "nutgrass",
    "sand_sedge",
    "cypress_sedge",
    "cotton_grass",
    "common_cotton_grass",
    "hardstem_bulrush",
    "wool_grass",
    "spike_rush",
    "water_chestnut",
    "needle_spike_rush",
    "creeping_spike_rush",
    "pandanus",
    "textile_screw_pine",
    "cattail",
    "cat's-tail",
    "bur_reed",
    "grain",
    "kernel",
    "rye",
    "gourd",
    "gourd",
    "pumpkin",
    "squash",
    "summer_squash",
    "yellow_squash",
    "marrow",
    "zucchini",
    "cocozelle",
    "cymling",
    "spaghetti_squash",
    "winter_squash",
    "acorn_squash",
    "hubbard_squash",
    "turban_squash",
    "buttercup_squash",
    "butternut_squash",
    "winter_crookneck",
    "cushaw",
    "prairie_gourd",
    "prairie_gourd",
    "bryony",
    "white_bryony",
    "sweet_melon",
    "cantaloupe",
    "winter_melon",
    "net_melon",
    "cucumber",
    "squirting_cucumber",
    "bottle_gourd",
    "luffa",
    "loofah",
    "angled_loofah",
    "loofa",
    "balsam_apple",
    "balsam_pear",
    "lobelia",
    "water_lobelia",
    "mallow",
    "musk_mallow",
    "common_mallow",
    "okra",
    "okra",
    "abelmosk",
    "flowering_maple",
    "velvetleaf",
    "hollyhock",
    "rose_mallow",
    "althea",
    "marsh_mallow",
    "poppy_mallow",
    "fringed_poppy_mallow",
    "purple_poppy_mallow",
    "clustered_poppy_mallow",
    "sea_island_cotton",
    "Levant_cotton",
    "upland_cotton",
    "Peruvian_cotton",
    "wild_cotton",
    "kenaf",
    "sorrel_tree",
    "rose_mallow",
    "cotton_rose",
    "roselle",
    "mahoe",
    "flower-of-an-hour",
    "lacebark",
    "wild_hollyhock",
    "mountain_hollyhock",
    "seashore_mallow",
    "salt_marsh_mallow",
    "chaparral_mallow",
    "malope",
    "false_mallow",
    "waxmallow",
    "glade_mallow",
    "pavonia",
    "ribbon_tree",
    "bush_hibiscus",
    "Virginia_mallow",
    "Queensland_hemp",
    "Indian_mallow",
    "checkerbloom",
    "globe_mallow",
    "prairie_mallow",
    "tulipwood_tree",
    "portia_tree",
    "red_silk-cotton_tree",
    "cream-of-tartar_tree",
    "baobab",
    "kapok",
    "durian",
    "Montezuma",
    "shaving-brush_tree",
    "quandong",
    "quandong",
    "makomako",
    "Jamaican_cherry",
    "breakax",
    "sterculia",
    "Panama_tree",
    "kalumpang",
    "bottle-tree",
    "flame_tree",
    "flame_tree",
    "kurrajong",
    "Queensland_bottletree",
    "kola",
    "kola_nut",
    "Chinese_parasol_tree",
    "flannelbush",
    "screw_tree",
    "nut-leaved_screw_tree",
    "red_beech",
    "looking_glass_tree",
    "looking-glass_plant",
    "honey_bell",
    "mayeng",
    "silver_tree",
    "cacao",
    "obeche",
    "linden",
    "American_basswood",
    "small-leaved_linden",
    "white_basswood",
    "Japanese_linden",
    "silver_lime",
    "corchorus",
    "African_hemp",
    "herb",
    "protea",
    "honeypot",
    "honeyflower",
    "banksia",
    "honeysuckle",
    "smoke_bush",
    "Chilean_firebush",
    "Chilean_nut",
    "grevillea",
    "red-flowered_silky_oak",
    "silky_oak",
    "beefwood",
    "cushion_flower",
    "rewa-rewa",
    "honeyflower",
    "silver_tree",
    "lomatia",
    "macadamia",
    "Macadamia_integrifolia",
    "macadamia_nut",
    "Queensland_nut",
    "prickly_ash",
    "geebung",
    "wheel_tree",
    "scrub_beefwood",
    "waratah",
    "waratah",
    "casuarina",
    "she-oak",
    "beefwood",
    "Australian_pine",
    "heath",
    "tree_heath",
    "briarroot",
    "winter_heath",
    "bell_heather",
    "Cornish_heath",
    "Spanish_heath",
    "Prince-of-Wales'-heath",
    "bog_rosemary",
    "marsh_andromeda",
    "madrona",
    "strawberry_tree",
    "bearberry",
    "alpine_bearberry",
    "heartleaf_manzanita",
    "Parry_manzanita",
    "spike_heath",
    "bryanthus",
    "leatherleaf",
    "Connemara_heath",
    "trailing_arbutus",
    "creeping_snowberry",
    "salal",
    "huckleberry",
    "black_huckleberry",
    "dangleberry",
    "box_huckleberry",
    "kalmia",
    "mountain_laurel",
    "swamp_laurel",
    "trapper's_tea",
    "wild_rosemary",
    "sand_myrtle",
    "leucothoe",
    "dog_laurel",
    "sweet_bells",
    "alpine_azalea",
    "staggerbush",
    "maleberry",
    "fetterbush",
    "false_azalea",
    "minniebush",
    "sorrel_tree",
    "mountain_heath",
    "purple_heather",
    "fetterbush",
    "rhododendron",
    "coast_rhododendron",
    "rosebay",
    "swamp_azalea",
    "azalea",
    "cranberry",
    "American_cranberry",
    "European_cranberry",
    "blueberry",
    "farkleberry",
    "low-bush_blueberry",
    "rabbiteye_blueberry",
    "dwarf_bilberry",
    "evergreen_blueberry",
    "evergreen_huckleberry",
    "bilberry",
    "bilberry",
    "bog_bilberry",
    "dryland_blueberry",
    "grouseberry",
    "deerberry",
    "cowberry",
    "diapensia",
    "galax",
    "pyxie",
    "shortia",
    "oconee_bells",
    "Australian_heath",
    "epacris",
    "common_heath",
    "common_heath",
    "Port_Jackson_heath",
    "native_cranberry",
    "pink_fivecorner",
    "wintergreen",
    "false_wintergreen",
    "lesser_wintergreen",
    "wild_lily_of_the_valley",
    "wild_lily_of_the_valley",
    "pipsissewa",
    "love-in-winter",
    "one-flowered_wintergreen",
    "Indian_pipe",
    "pinesap",
    "beech",
    "common_beech",
    "copper_beech",
    "American_beech",
    "weeping_beech",
    "Japanese_beech",
    "chestnut",
    "American_chestnut",
    "European_chestnut",
    "Chinese_chestnut",
    "Japanese_chestnut",
    "Allegheny_chinkapin",
    "Ozark_chinkapin",
    "oak_chestnut",
    "giant_chinkapin",
    "dwarf_golden_chinkapin",
    "tanbark_oak",
    "Japanese_oak",
    "southern_beech",
    "myrtle_beech",
    "Coigue",
    "New_Zealand_beech",
    "silver_beech",
    "roble_beech",
    "rauli_beech",
    "black_beech",
    "hard_beech",
    "acorn",
    "cupule",
    "oak",
    "live_oak",
    "coast_live_oak",
    "white_oak",
    "American_white_oak",
    "Arizona_white_oak",
    "swamp_white_oak",
    "European_turkey_oak",
    "canyon_oak",
    "scarlet_oak",
    "jack_oak",
    "red_oak",
    "southern_red_oak",
    "Oregon_white_oak",
    "holm_oak",
    "bear_oak",
    "shingle_oak",
    "bluejack_oak",
    "California_black_oak",
    "American_turkey_oak",
    "laurel_oak",
    "California_white_oak",
    "overcup_oak",
    "bur_oak",
    "scrub_oak",
    "blackjack_oak",
    "swamp_chestnut_oak",
    "Japanese_oak",
    "chestnut_oak",
    "chinquapin_oak",
    "myrtle_oak",
    "water_oak",
    "Nuttall_oak",
    "durmast",
    "basket_oak",
    "pin_oak",
    "willow_oak",
    "dwarf_chinkapin_oak",
    "common_oak",
    "northern_red_oak",
    "Shumard_oak",
    "post_oak",
    "cork_oak",
    "Spanish_oak",
    "huckleberry_oak",
    "Chinese_cork_oak",
    "black_oak",
    "southern_live_oak",
    "interior_live_oak",
    "mast",
    "birch",
    "yellow_birch",
    "American_white_birch",
    "grey_birch",
    "silver_birch",
    "downy_birch",
    "black_birch",
    "sweet_birch",
    "Yukon_white_birch",
    "swamp_birch",
    "Newfoundland_dwarf_birch",
    "alder",
    "common_alder",
    "grey_alder",
    "seaside_alder",
    "white_alder",
    "red_alder",
    "speckled_alder",
    "smooth_alder",
    "green_alder",
    "green_alder",
    "hornbeam",
    "European_hornbeam",
    "American_hornbeam",
    "hop_hornbeam",
    "Old_World_hop_hornbeam",
    "Eastern_hop_hornbeam",
    "hazelnut",
    "American_hazel",
    "cobnut",
    "beaked_hazelnut",
    "centaury",
    "rosita",
    "lesser_centaury",
    "seaside_centaury",
    "slender_centaury",
    "prairie_gentian",
    "Persian_violet",
    "columbo",
    "gentian",
    "gentianella",
    "closed_gentian",
    "explorer's_gentian",
    "closed_gentian",
    "great_yellow_gentian",
    "marsh_gentian",
    "soapwort_gentian",
    "striped_gentian",
    "agueweed",
    "felwort",
    "fringed_gentian",
    "Gentianopsis_crinita",
    "Gentianopsis_detonsa",
    "Gentianopsid_procera",
    "Gentianopsis_thermalis",
    "tufted_gentian",
    "spurred_gentian",
    "sabbatia",
    "toothbrush_tree",
    "olive_tree",
    "olive",
    "olive",
    "black_maire",
    "white_maire",
    "fringe_tree",
    "fringe_bush",
    "forestiera",
    "forsythia",
    "ash",
    "white_ash",
    "swamp_ash",
    "flowering_ash",
    "European_ash",
    "Oregon_ash",
    "black_ash",
    "manna_ash",
    "red_ash",
    "green_ash",
    "blue_ash",
    "mountain_ash",
    "pumpkin_ash",
    "Arizona_ash",
    "jasmine",
    "primrose_jasmine",
    "winter_jasmine",
    "common_jasmine",
    "privet",
    "Amur_privet",
    "Japanese_privet",
    "Ligustrum_obtusifolium",
    "common_privet",
    "devilwood",
    "mock_privet",
    "lilac",
    "Himalayan_lilac",
    "Persian_lilac",
    "Japanese_tree_lilac",
    "Japanese_lilac",
    "common_lilac",
    "bloodwort",
    "kangaroo_paw",
    "Virginian_witch_hazel",
    "vernal_witch_hazel",
    "winter_hazel",
    "fothergilla",
    "liquidambar",
    "sweet_gum",
    "iron_tree",
    "walnut",
    "California_black_walnut",
    "butternut",
    "black_walnut",
    "English_walnut",
    "hickory",
    "water_hickory",
    "pignut",
    "bitternut",
    "pecan",
    "big_shellbark",
    "nutmeg_hickory",
    "shagbark",
    "mockernut",
    "wing_nut",
    "Caucasian_walnut",
    "dhawa",
    "combretum",
    "hiccup_nut",
    "bush_willow",
    "bush_willow",
    "button_tree",
    "white_mangrove",
    "oleaster",
    "water_milfoil",
    "anchovy_pear",
    "brazil_nut",
    "loosestrife",
    "purple_loosestrife",
    "grass_poly",
    "crape_myrtle",
    "Queen's_crape_myrtle",
    "myrtaceous_tree",
    "myrtle",
    "common_myrtle",
    "bayberry",
    "allspice",
    "allspice_tree",
    "sour_cherry",
    "nakedwood",
    "Surinam_cherry",
    "rose_apple",
    "feijoa",
    "jaboticaba",
    "guava",
    "guava",
    "cattley_guava",
    "Brazilian_guava",
    "gum_tree",
    "eucalyptus",
    "flooded_gum",
    "mallee",
    "stringybark",
    "smoothbark",
    "red_gum",
    "red_gum",
    "river_red_gum",
    "mountain_swamp_gum",
    "snow_gum",
    "alpine_ash",
    "white_mallee",
    "white_stringybark",
    "white_mountain_ash",
    "blue_gum",
    "rose_gum",
    "cider_gum",
    "swamp_gum",
    "spotted_gum",
    "lemon-scented_gum",
    "black_mallee",
    "forest_red_gum",
    "mountain_ash",
    "manna_gum",
    "clove",
    "clove",
    "tupelo",
    "water_gum",
    "sour_gum",
    "enchanter's_nightshade",
    "Circaea_lutetiana",
    "willowherb",
    "fireweed",
    "California_fuchsia",
    "fuchsia",
    "lady's-eardrop",
    "evening_primrose",
    "common_evening_primrose",
    "sundrops",
    "Missouri_primrose",
    "pomegranate",
    "mangrove",
    "daphne",
    "garland_flower",
    "spurge_laurel",
    "mezereon",
    "Indian_rhododendron",
    "Medinilla_magnifica",
    "deer_grass",
    "canna",
    "achira",
    "arrowroot",
    "banana",
    "dwarf_banana",
    "Japanese_banana",
    "plantain",
    "edible_banana",
    "abaca",
    "Abyssinian_banana",
    "ginger",
    "common_ginger",
    "turmeric",
    "galangal",
    "shellflower",
    "grains_of_paradise",
    "cardamom",
    "begonia",
    "fibrous-rooted_begonia",
    "tuberous_begonia",
    "rhizomatous_begonia",
    "Christmas_begonia",
    "angel-wing_begonia",
    "beefsteak_begonia",
    "star_begonia",
    "rex_begonia",
    "wax_begonia",
    "Socotra_begonia",
    "hybrid_tuberous_begonia",
    "dillenia",
    "guinea_gold_vine",
    "poon",
    "calaba",
    "Maria",
    "laurelwood",
    "Alexandrian_laurel",
    "clusia",
    "wild_fig",
    "waxflower",
    "pitch_apple",
    "mangosteen",
    "gamboge_tree",
    "St_John's_wort",
    "common_St_John's_wort",
    "great_St_John's_wort",
    "creeping_St_John's_wort",
    "low_St_Andrew's_cross",
    "klammath_weed",
    "shrubby_St_John's_wort",
    "St_Peter's_wort",
    "marsh_St-John's_wort",
    "mammee_apple",
    "rose_chestnut",
    "bower_actinidia",
    "Chinese_gooseberry",
    "silvervine",
    "wild_cinnamon",
    "papaya",
    "souari",
    "rockrose",
    "white-leaved_rockrose",
    "common_gum_cistus",
    "frostweed",
    "dipterocarp",
    "red_lauan",
    "governor's_plum",
    "kei_apple",
    "ketembilla",
    "chaulmoogra",
    "wild_peach",
    "candlewood",
    "boojum_tree",
    "bird's-eye_bush",
    "granadilla",
    "granadilla",
    "granadilla",
    "maypop",
    "Jamaica_honeysuckle",
    "banana_passion_fruit",
    "sweet_calabash",
    "love-in-a-mist",
    "reseda",
    "mignonette",
    "dyer's_rocket",
    "false_tamarisk",
    "halophyte",
    "viola",
    "violet",
    "field_pansy",
    "American_dog_violet",
    "dog_violet",
    "horned_violet",
    "two-eyed_violet",
    "bird's-foot_violet",
    "downy_yellow_violet",
    "long-spurred_violet",
    "pale_violet",
    "hedge_violet",
    "nettle",
    "stinging_nettle",
    "Roman_nettle",
    "ramie",
    "wood_nettle",
    "Australian_nettle",
    "pellitory-of-the-wall",
    "richweed",
    "artillery_plant",
    "friendship_plant",
    "Queensland_grass-cloth_plant",
    "Pipturus_albidus",
    "cannabis",
    "Indian_hemp",
    "mulberry",
    "white_mulberry",
    "black_mulberry",
    "red_mulberry",
    "osage_orange",
    "breadfruit",
    "jackfruit",
    "marang",
    "fig_tree",
    "fig",
    "caprifig",
    "golden_fig",
    "banyan",
    "pipal",
    "India-rubber_tree",
    "mistletoe_fig",
    "Port_Jackson_fig",
    "sycamore",
    "paper_mulberry",
    "trumpetwood",
    "elm",
    "winged_elm",
    "American_elm",
    "smooth-leaved_elm",
    "cedar_elm",
    "witch_elm",
    "Dutch_elm",
    "Huntingdon_elm",
    "water_elm",
    "Chinese_elm",
    "English_elm",
    "Siberian_elm",
    "slippery_elm",
    "Jersey_elm",
    "September_elm",
    "rock_elm",
    "hackberry",
    "European_hackberry",
    "American_hackberry",
    "sugarberry",
    "iridaceous_plant",
    "bearded_iris",
    "beardless_iris",
    "orrisroot",
    "dwarf_iris",
    "Dutch_iris",
    "Florentine_iris",
    "stinking_iris",
    "German_iris",
    "Japanese_iris",
    "German_iris",
    "Dalmatian_iris",
    "Persian_iris",
    "Dutch_iris",
    "dwarf_iris",
    "Spanish_iris",
    "blackberry-lily",
    "crocus",
    "saffron",
    "corn_lily",
    "blue-eyed_grass",
    "wandflower",
    "amaryllis",
    "salsilla",
    "salsilla",
    "blood_lily",
    "Cape_tulip",
    "hippeastrum",
    "narcissus",
    "daffodil",
    "jonquil",
    "jonquil",
    "Jacobean_lily",
    "liliaceous_plant",
    "mountain_lily",
    "Canada_lily",
    "tiger_lily",
    "Columbia_tiger_lily",
    "tiger_lily",
    "Easter_lily",
    "coast_lily",
    "Turk's-cap",
    "Michigan_lily",
    "leopard_lily",
    "Turk's-cap",
    "African_lily",
    "colicroot",
    "ague_root",
    "yellow_colicroot",
    "alliaceous_plant",
    "Hooker's_onion",
    "wild_leek",
    "Canada_garlic",
    "keeled_garlic",
    "onion",
    "shallot",
    "nodding_onion",
    "Welsh_onion",
    "red-skinned_onion",
    "daffodil_garlic",
    "few-flowered_leek",
    "garlic",
    "sand_leek",
    "chives",
    "crow_garlic",
    "wild_garlic",
    "garlic_chive",
    "round-headed_leek",
    "three-cornered_leek",
    "cape_aloe",
    "kniphofia",
    "poker_plant",
    "red-hot_poker",
    "fly_poison",
    "amber_lily",
    "asparagus",
    "asparagus_fern",
    "smilax",
    "asphodel",
    "Jacob's_rod",
    "aspidistra",
    "coral_drops",
    "Christmas_bells",
    "climbing_onion",
    "mariposa",
    "globe_lily",
    "cat's-ear",
    "white_globe_lily",
    "yellow_globe_lily",
    "rose_globe_lily",
    "star_tulip",
    "desert_mariposa_tulip",
    "yellow_mariposa_tulip",
    "sagebrush_mariposa_tulip",
    "sego_lily",
    "camas",
    "common_camas",
    "Leichtlin's_camas",
    "wild_hyacinth",
    "dogtooth_violet",
    "white_dogtooth_violet",
    "yellow_adder's_tongue",
    "European_dogtooth",
    "fawn_lily",
    "glacier_lily",
    "avalanche_lily",
    "fritillary",
    "mission_bells",
    "mission_bells",
    "stink_bell",
    "crown_imperial",
    "white_fritillary",
    "snake's_head_fritillary",
    "adobe_lily",
    "scarlet_fritillary",
    "tulip",
    "dwarf_tulip",
    "lady_tulip",
    "Tulipa_gesneriana",
    "cottage_tulip",
    "Darwin_tulip",
    "gloriosa",
    "lemon_lily",
    "common_hyacinth",
    "Roman_hyacinth",
    "summer_hyacinth",
    "star-of-Bethlehem",
    "bath_asparagus",
    "grape_hyacinth",
    "common_grape_hyacinth",
    "tassel_hyacinth",
    "scilla",
    "spring_squill",
    "false_asphodel",
    "Scotch_asphodel",
    "sea_squill",
    "squill",
    "butcher's_broom",
    "bog_asphodel",
    "European_bog_asphodel",
    "American_bog_asphodel",
    "hellebore",
    "white_hellebore",
    "squaw_grass",
    "death_camas",
    "alkali_grass",
    "white_camas",
    "poison_camas",
    "grassy_death_camas",
    "prairie_wake-robin",
    "dwarf-white_trillium",
    "herb_Paris",
    "sarsaparilla",
    "bullbrier",
    "rough_bindweed",
    "clintonia",
    "false_lily_of_the_valley",
    "false_lily_of_the_valley",
    "Solomon's-seal",
    "great_Solomon's-seal",
    "bellwort",
    "strawflower",
    "pia",
    "agave",
    "American_agave",
    "sisal",
    "maguey",
    "maguey",
    "Agave_tequilana",
    "cabbage_tree",
    "dracaena",
    "tuberose",
    "sansevieria",
    "African_bowstring_hemp",
    "Ceylon_bowstring_hemp",
    "mother-in-law's_tongue",
    "Spanish_bayonet",
    "Spanish_bayonet",
    "Joshua_tree",
    "soapweed",
    "Adam's_needle",
    "bear_grass",
    "Spanish_dagger",
    "Our_Lord's_candle",
    "water_shamrock",
    "butterfly_bush",
    "yellow_jasmine",
    "flax",
    "calabar_bean",
    "bonduc",
    "divi-divi",
    "Mysore_thorn",
    "brazilian_ironwood",
    "bird_of_paradise",
    "shingle_tree",
    "mountain_ebony",
    "msasa",
    "cassia",
    "golden_shower_tree",
    "pink_shower",
    "rainbow_shower",
    "horse_cassia",
    "carob",
    "carob",
    "paloverde",
    "royal_poinciana",
    "locust_tree",
    "water_locust",
    "honey_locust",
    "Kentucky_coffee_tree",
    "logwood",
    "Jerusalem_thorn",
    "palo_verde",
    "Dalmatian_laburnum",
    "senna",
    "avaram",
    "Alexandria_senna",
    "wild_senna",
    "sicklepod",
    "coffee_senna",
    "tamarind",
    "false_indigo",
    "false_indigo",
    "hog_peanut",
    "angelim",
    "cabbage_bark",
    "kidney_vetch",
    "groundnut",
    "rooibos",
    "milk_vetch",
    "alpine_milk_vetch",
    "purple_milk_vetch",
    "camwood",
    "wild_indigo",
    "blue_false_indigo",
    "white_false_indigo",
    "indigo_broom",
    "dhak",
    "pigeon_pea",
    "sword_bean",
    "pea_tree",
    "Siberian_pea_tree",
    "Chinese_pea_tree",
    "Moreton_Bay_chestnut",
    "butterfly_pea",
    "Judas_tree",
    "redbud",
    "western_redbud",
    "tagasaste",
    "weeping_tree_broom",
    "flame_pea",
    "chickpea",
    "chickpea",
    "Kentucky_yellowwood",
    "glory_pea",
    "desert_pea",
    "parrot's_beak",
    "butterfly_pea",
    "blue_pea",
    "telegraph_plant",
    "bladder_senna",
    "axseed",
    "crotalaria",
    "guar",
    "white_broom",
    "common_broom",
    "rosewood",
    "Indian_blackwood",
    "sissoo",
    "kingwood",
    "Brazilian_rosewood",
    "cocobolo",
    "blackwood",
    "bitter_pea",
    "derris",
    "derris_root",
    "prairie_mimosa",
    "tick_trefoil",
    "beggarweed",
    "Australian_pea",
    "coral_tree",
    "kaffir_boom",
    "coral_bean_tree",
    "ceibo",
    "kaffir_boom",
    "Indian_coral_tree",
    "cork_tree",
    "goat's_rue",
    "poison_bush",
    "Spanish_broom",
    "woodwaxen",
    "chanar",
    "gliricidia",
    "soy",
    "licorice",
    "wild_licorice",
    "licorice_root",
    "Western_Australia_coral_pea",
    "sweet_vetch",
    "French_honeysuckle",
    "anil",
    "scarlet_runner",
    "hyacinth_bean",
    "Scotch_laburnum",
    "vetchling",
    "wild_pea",
    "everlasting_pea",
    "beach_pea",
    "grass_vetch",
    "marsh_pea",
    "common_vetchling",
    "grass_pea",
    "Tangier_pea",
    "heath_pea",
    "bicolor_lespediza",
    "japanese_clover",
    "Korean_lespedeza",
    "sericea_lespedeza",
    "lentil",
    "lentil",
    "prairie_bird's-foot_trefoil",
    "bird's_foot_trefoil",
    "winged_pea",
    "lupine",
    "white_lupine",
    "tree_lupine",
    "wild_lupine",
    "bluebonnet",
    "Texas_bluebonnet",
    "medic",
    "moon_trefoil",
    "sickle_alfalfa",
    "Calvary_clover",
    "black_medick",
    "alfalfa",
    "millettia",
    "mucuna",
    "cowage",
    "tolu_tree",
    "Peruvian_balsam",
    "sainfoin",
    "restharrow",
    "bead_tree",
    "jumby_bead",
    "locoweed",
    "purple_locoweed",
    "tumbleweed",
    "yam_bean",
    "shamrock_pea",
    "pole_bean",
    "kidney_bean",
    "haricot",
    "wax_bean",
    "scarlet_runner",
    "lima_bean",
    "sieva_bean",
    "tepary_bean",
    "chaparral_pea",
    "Jamaica_dogwood",
    "pea",
    "garden_pea",
    "edible-pod_pea",
    "sugar_snap_pea",
    "field_pea",
    "field_pea",
    "common_flat_pea",
    "quira",
    "roble",
    "Panama_redwood_tree",
    "Indian_beech",
    "winged_bean",
    "breadroot",
    "bloodwood_tree",
    "kino",
    "red_sandalwood",
    "kudzu",
    "bristly_locust",
    "black_locust",
    "clammy_locust",
    "carib_wood",
    "Colorado_River_hemp",
    "scarlet_wisteria_tree",
    "Japanese_pagoda_tree",
    "mescal_bean",
    "kowhai",
    "jade_vine",
    "hoary_pea",
    "bastard_indigo",
    "catgut",
    "bush_pea",
    "false_lupine",
    "Carolina_lupine",
    "tipu",
    "bird's_foot_trefoil",
    "fenugreek",
    "gorse",
    "vetch",
    "tufted_vetch",
    "broad_bean",
    "bitter_betch",
    "bush_vetch",
    "moth_bean",
    "snailflower",
    "mung",
    "cowpea",
    "cowpea",
    "asparagus_bean",
    "swamp_oak",
    "keurboom",
    "keurboom",
    "Japanese_wistaria",
    "Chinese_wistaria",
    "American_wistaria",
    "silky_wisteria",
    "palm",
    "sago_palm",
    "feather_palm",
    "fan_palm",
    "palmetto",
    "coyol",
    "grugru",
    "areca",
    "betel_palm",
    "sugar_palm",
    "piassava_palm",
    "coquilla_nut",
    "palmyra",
    "calamus",
    "rattan",
    "lawyer_cane",
    "fishtail_palm",
    "wine_palm",
    "wax_palm",
    "coconut",
    "carnauba",
    "caranday",
    "corozo",
    "gebang_palm",
    "latanier",
    "talipot",
    "oil_palm",
    "African_oil_palm",
    "American_oil_palm",
    "palm_nut",
    "cabbage_palm",
    "cabbage_palm",
    "true_sago_palm",
    "nipa_palm",
    "babassu",
    "babassu_nut",
    "cohune_palm",
    "cohune_nut",
    "date_palm",
    "ivory_palm",
    "raffia_palm",
    "bamboo_palm",
    "lady_palm",
    "miniature_fan_palm",
    "reed_rhapis",
    "royal_palm",
    "cabbage_palm",
    "cabbage_palmetto",
    "saw_palmetto",
    "thatch_palm",
    "key_palm",
    "English_plantain",
    "broad-leaved_plantain",
    "hoary_plantain",
    "fleawort",
    "rugel's_plantain",
    "hoary_plantain",
    "buckwheat",
    "prince's-feather",
    "eriogonum",
    "umbrella_plant",
    "wild_buckwheat",
    "rhubarb",
    "Himalayan_rhubarb",
    "pie_plant",
    "Chinese_rhubarb",
    "sour_dock",
    "sheep_sorrel",
    "bitter_dock",
    "French_sorrel",
    "yellow-eyed_grass",
    "commelina",
    "spiderwort",
    "pineapple",
    "pipewort",
    "water_hyacinth",
    "water_star_grass",
    "naiad",
    "water_plantain",
    "narrow-leaved_water_plantain",
    "hydrilla",
    "American_frogbit",
    "waterweed",
    "Canadian_pondweed",
    "tape_grass",
    "pondweed",
    "curled_leaf_pondweed",
    "loddon_pondweed",
    "frog's_lettuce",
    "arrow_grass",
    "horned_pondweed",
    "eelgrass",
    "rose",
    "hip",
    "banksia_rose",
    "damask_rose",
    "sweetbrier",
    "Cherokee_rose",
    "musk_rose",
    "agrimonia",
    "harvest-lice",
    "fragrant_agrimony",
    "alderleaf_Juneberry",
    "flowering_quince",
    "japonica",
    "coco_plum",
    "cotoneaster",
    "Cotoneaster_dammeri",
    "Cotoneaster_horizontalis",
    "parsley_haw",
    "scarlet_haw",
    "blackthorn",
    "cockspur_thorn",
    "mayhaw",
    "red_haw",
    "red_haw",
    "quince",
    "mountain_avens",
    "loquat",
    "beach_strawberry",
    "Virginia_strawberry",
    "avens",
    "yellow_avens",
    "yellow_avens",
    "prairie_smoke",
    "bennet",
    "toyon",
    "apple_tree",
    "apple",
    "wild_apple",
    "crab_apple",
    "Siberian_crab",
    "wild_crab",
    "American_crab_apple",
    "Oregon_crab_apple",
    "Southern_crab_apple",
    "Iowa_crab",
    "Bechtel_crab",
    "medlar",
    "cinquefoil",
    "silverweed",
    "salad_burnet",
    "plum",
    "wild_plum",
    "Allegheny_plum",
    "American_red_plum",
    "chickasaw_plum",
    "beach_plum",
    "common_plum",
    "bullace",
    "damson_plum",
    "big-tree_plum",
    "Canada_plum",
    "plumcot",
    "apricot",
    "Japanese_apricot",
    "common_apricot",
    "purple_apricot",
    "cherry",
    "wild_cherry",
    "wild_cherry",
    "sweet_cherry",
    "heart_cherry",
    "gean",
    "capulin",
    "cherry_laurel",
    "cherry_plum",
    "sour_cherry",
    "amarelle",
    "morello",
    "marasca",
    "almond_tree",
    "almond",
    "bitter_almond",
    "jordan_almond",
    "dwarf_flowering_almond",
    "holly-leaved_cherry",
    "fuji",
    "flowering_almond",
    "cherry_laurel",
    "Catalina_cherry",
    "bird_cherry",
    "hagberry_tree",
    "hagberry",
    "pin_cherry",
    "peach",
    "nectarine",
    "sand_cherry",
    "Japanese_plum",
    "black_cherry",
    "flowering_cherry",
    "oriental_cherry",
    "Japanese_flowering_cherry",
    "Sierra_plum",
    "rosebud_cherry",
    "Russian_almond",
    "flowering_almond",
    "chokecherry",
    "chokecherry",
    "western_chokecherry",
    "Pyracantha",
    "pear",
    "fruit_tree",
    "bramble_bush",
    "lawyerbush",
    "stone_bramble",
    "sand_blackberry",
    "boysenberry",
    "loganberry",
    "American_dewberry",
    "Northern_dewberry",
    "Southern_dewberry",
    "swamp_dewberry",
    "European_dewberry",
    "raspberry",
    "wild_raspberry",
    "American_raspberry",
    "black_raspberry",
    "salmonberry",
    "salmonberry",
    "wineberry",
    "mountain_ash",
    "rowan",
    "rowanberry",
    "American_mountain_ash",
    "Western_mountain_ash",
    "service_tree",
    "wild_service_tree",
    "spirea",
    "bridal_wreath",
    "madderwort",
    "Indian_madder",
    "madder",
    "woodruff",
    "dagame",
    "blolly",
    "coffee",
    "Arabian_coffee",
    "Liberian_coffee",
    "robusta_coffee",
    "cinchona",
    "Cartagena_bark",
    "calisaya",
    "cinchona_tree",
    "cinchona",
    "bedstraw",
    "sweet_woodruff",
    "Northern_bedstraw",
    "yellow_bedstraw",
    "wild_licorice",
    "cleavers",
    "wild_madder",
    "cape_jasmine",
    "genipa",
    "genipap_fruit",
    "hamelia",
    "scarlet_bush",
    "lemonwood",
    "negro_peach",
    "wild_medlar",
    "Spanish_tamarind",
    "abelia",
    "bush_honeysuckle",
    "American_twinflower",
    "honeysuckle",
    "American_fly_honeysuckle",
    "Italian_honeysuckle",
    "yellow_honeysuckle",
    "hairy_honeysuckle",
    "Japanese_honeysuckle",
    "Hall's_honeysuckle",
    "Morrow's_honeysuckle",
    "woodbine",
    "trumpet_honeysuckle",
    "European_fly_honeysuckle",
    "swamp_fly_honeysuckle",
    "snowberry",
    "coralberry",
    "blue_elder",
    "dwarf_elder",
    "American_red_elder",
    "European_red_elder",
    "feverroot",
    "cranberry_bush",
    "wayfaring_tree",
    "guelder_rose",
    "arrow_wood",
    "black_haw",
    "weigela",
    "teasel",
    "common_teasel",
    "fuller's_teasel",
    "wild_teasel",
    "scabious",
    "sweet_scabious",
    "field_scabious",
    "jewelweed",
    "geranium",
    "cranesbill",
    "wild_geranium",
    "meadow_cranesbill",
    "Richardson's_geranium",
    "herb_robert",
    "sticky_geranium",
    "dove's_foot_geranium",
    "rose_geranium",
    "fish_geranium",
    "ivy_geranium",
    "apple_geranium",
    "lemon_geranium",
    "storksbill",
    "musk_clover",
    "incense_tree",
    "elephant_tree",
    "gumbo-limbo",
    "Boswellia_carteri",
    "salai",
    "balm_of_gilead",
    "myrrh_tree",
    "Protium_heptaphyllum",
    "Protium_guianense",
    "water_starwort",
    "barbados_cherry",
    "mahogany",
    "chinaberry",
    "neem",
    "neem_seed",
    "Spanish_cedar",
    "satinwood",
    "African_scented_mahogany",
    "silver_ash",
    "native_beech",
    "bunji-bunji",
    "African_mahogany",
    "lanseh_tree",
    "true_mahogany",
    "Honduras_mahogany",
    "Philippine_mahogany",
    "caracolito",
    "common_wood_sorrel",
    "Bermuda_buttercup",
    "creeping_oxalis",
    "goatsfoot",
    "violet_wood_sorrel",
    "oca",
    "carambola",
    "bilimbi",
    "milkwort",
    "senega",
    "orange_milkwort",
    "flowering_wintergreen",
    "Seneca_snakeroot",
    "common_milkwort",
    "rue",
    "citrus",
    "orange",
    "sour_orange",
    "bergamot",
    "pomelo",
    "citron",
    "grapefruit",
    "mandarin",
    "tangerine",
    "clementine",
    "satsuma",
    "sweet_orange",
    "temple_orange",
    "tangelo",
    "rangpur",
    "lemon",
    "sweet_lemon",
    "lime",
    "citrange",
    "fraxinella",
    "kumquat",
    "marumi",
    "nagami",
    "cork_tree",
    "trifoliate_orange",
    "prickly_ash",
    "toothache_tree",
    "Hercules'-club",
    "bitterwood_tree",
    "marupa",
    "paradise_tree",
    "ailanthus",
    "tree_of_heaven",
    "wild_mango",
    "pepper_tree",
    "Jamaica_quassia",
    "quassia",
    "nasturtium",
    "garden_nasturtium",
    "bush_nasturtium",
    "canarybird_flower",
    "bean_caper",
    "palo_santo",
    "lignum_vitae",
    "creosote_bush",
    "caltrop",
    "willow",
    "osier",
    "white_willow",
    "silver_willow",
    "golden_willow",
    "cricket-bat_willow",
    "arctic_willow",
    "weeping_willow",
    "Wisconsin_weeping_willow",
    "pussy_willow",
    "sallow",
    "goat_willow",
    "peachleaf_willow",
    "almond_willow",
    "hoary_willow",
    "crack_willow",
    "prairie_willow",
    "dwarf_willow",
    "grey_willow",
    "arroyo_willow",
    "shining_willow",
    "swamp_willow",
    "bay_willow",
    "purple_willow",
    "balsam_willow",
    "creeping_willow",
    "Sitka_willow",
    "dwarf_grey_willow",
    "bearberry_willow",
    "common_osier",
    "poplar",
    "balsam_poplar",
    "white_poplar",
    "grey_poplar",
    "black_poplar",
    "Lombardy_poplar",
    "cottonwood",
    "Eastern_cottonwood",
    "black_cottonwood",
    "swamp_cottonwood",
    "aspen",
    "quaking_aspen",
    "American_quaking_aspen",
    "Canadian_aspen",
    "sandalwood_tree",
    "quandong",
    "rabbitwood",
    "Loranthaceae",
    "mistletoe",
    "American_mistletoe",
    "mistletoe",
    "American_mistletoe",
    "aalii",
    "soapberry",
    "wild_China_tree",
    "China_tree",
    "akee",
    "soapberry_vine",
    "heartseed",
    "balloon_vine",
    "longan",
    "harpullia",
    "harpulla",
    "Moreton_Bay_tulipwood",
    "litchi",
    "Spanish_lime",
    "rambutan",
    "pulasan",
    "pachysandra",
    "Allegheny_spurge",
    "bittersweet",
    "spindle_tree",
    "winged_spindle_tree",
    "wahoo",
    "strawberry_bush",
    "evergreen_bittersweet",
    "cyrilla",
    "titi",
    "crowberry",
    "maple",
    "silver_maple",
    "sugar_maple",
    "red_maple",
    "moosewood",
    "Oregon_maple",
    "dwarf_maple",
    "mountain_maple",
    "vine_maple",
    "hedge_maple",
    "Norway_maple",
    "sycamore",
    "box_elder",
    "California_box_elder",
    "pointed-leaf_maple",
    "Japanese_maple",
    "Japanese_maple",
    "holly",
    "Chinese_holly",
    "bearberry",
    "inkberry",
    "mate",
    "American_holly",
    "low_gallberry_holly",
    "tall_gallberry_holly",
    "yaupon_holly",
    "deciduous_holly",
    "juneberry_holly",
    "largeleaf_holly",
    "Geogia_holly",
    "common_winterberry_holly",
    "smooth_winterberry_holly",
    "cashew",
    "goncalo_alves",
    "Venetian_sumac",
    "laurel_sumac",
    "mango",
    "pistachio",
    "terebinth",
    "mastic",
    "Australian_sumac",
    "sumac",
    "smooth_sumac",
    "sugar-bush",
    "staghorn_sumac",
    "squawbush",
    "aroeira_blanca",
    "pepper_tree",
    "Brazilian_pepper_tree",
    "hog_plum",
    "mombin",
    "poison_ash",
    "poison_ivy",
    "western_poison_oak",
    "eastern_poison_oak",
    "varnish_tree",
    "horse_chestnut",
    "buckeye",
    "sweet_buckeye",
    "Ohio_buckeye",
    "dwarf_buckeye",
    "red_buckeye",
    "particolored_buckeye",
    "ebony",
    "marblewood",
    "marblewood",
    "persimmon",
    "Japanese_persimmon",
    "American_persimmon",
    "date_plum",
    "buckthorn",
    "southern_buckthorn",
    "false_buckthorn",
    "star_apple",
    "satinleaf",
    "balata",
    "sapodilla",
    "gutta-percha_tree",
    "gutta-percha_tree",
    "canistel",
    "marmalade_tree",
    "sweetleaf",
    "Asiatic_sweetleaf",
    "styrax",
    "snowbell",
    "Japanese_snowbell",
    "Texas_snowbell",
    "silver-bell_tree",
    "carnivorous_plant",
    "pitcher_plant",
    "common_pitcher_plant",
    "hooded_pitcher_plant",
    "huntsman's_horn",
    "tropical_pitcher_plant",
    "sundew",
    "Venus's_flytrap",
    "waterwheel_plant",
    "Drosophyllum_lusitanicum",
    "roridula",
    "Australian_pitcher_plant",
    "sedum",
    "stonecrop",
    "rose-root",
    "orpine",
    "pinwheel",
    "Christmas_bush",
    "hortensia",
    "fall-blooming_hydrangea",
    "carpenteria",
    "decumary",
    "deutzia",
    "philadelphus",
    "mock_orange",
    "saxifrage",
    "yellow_mountain_saxifrage",
    "meadow_saxifrage",
    "mossy_saxifrage",
    "western_saxifrage",
    "purple_saxifrage",
    "star_saxifrage",
    "strawberry_geranium",
    "astilbe",
    "false_goatsbeard",
    "dwarf_astilbe",
    "spirea",
    "bergenia",
    "coast_boykinia",
    "golden_saxifrage",
    "umbrella_plant",
    "bridal_wreath",
    "alumroot",
    "coralbells",
    "leatherleaf_saxifrage",
    "woodland_star",
    "prairie_star",
    "miterwort",
    "five-point_bishop's_cap",
    "parnassia",
    "bog_star",
    "fringed_grass_of_Parnassus",
    "false_alumroot",
    "foamflower",
    "false_miterwort",
    "pickaback_plant",
    "currant",
    "black_currant",
    "white_currant",
    "gooseberry",
    "plane_tree",
    "London_plane",
    "American_sycamore",
    "oriental_plane",
    "California_sycamore",
    "Arizona_sycamore",
    "Greek_valerian",
    "northern_Jacob's_ladder",
    "skunkweed",
    "phlox",
    "moss_pink",
    "evening-snow",
    "acanthus",
    "bear's_breech",
    "caricature_plant",
    "black-eyed_Susan",
    "catalpa",
    "Catalpa_bignioides",
    "Catalpa_speciosa",
    "desert_willow",
    "calabash",
    "calabash",
    "borage",
    "common_amsinckia",
    "anchusa",
    "bugloss",
    "cape_forget-me-not",
    "cape_forget-me-not",
    "Spanish_elm",
    "princewood",
    "Chinese_forget-me-not",
    "hound's-tongue",
    "hound's-tongue",
    "blueweed",
    "beggar's_lice",
    "gromwell",
    "puccoon",
    "Virginia_bluebell",
    "garden_forget-me-not",
    "forget-me-not",
    "false_gromwell",
    "comfrey",
    "common_comfrey",
    "convolvulus",
    "bindweed",
    "field_bindweed",
    "scammony",
    "silverweed",
    "dodder",
    "dichondra",
    "cypress_vine",
    "moonflower",
    "wild_potato_vine",
    "red_morning-glory",
    "man-of-the-earth",
    "scammony",
    "Japanese_morning_glory",
    "imperial_Japanese_morning_glory",
    "gesneriad",
    "gesneria",
    "achimenes",
    "aeschynanthus",
    "lace-flower_vine",
    "columnea",
    "episcia",
    "gloxinia",
    "Canterbury_bell",
    "kohleria",
    "African_violet",
    "streptocarpus",
    "Cape_primrose",
    "waterleaf",
    "Virginia_waterleaf",
    "yellow_bells",
    "yerba_santa",
    "nemophila",
    "baby_blue-eyes",
    "five-spot",
    "scorpionweed",
    "California_bluebell",
    "California_bluebell",
    "fiddleneck",
    "fiesta_flower",
    "basil_thyme",
    "giant_hyssop",
    "yellow_giant_hyssop",
    "anise_hyssop",
    "Mexican_hyssop",
    "bugle",
    "creeping_bugle",
    "erect_bugle",
    "pyramid_bugle",
    "wood_mint",
    "hairy_wood_mint",
    "downy_wood_mint",
    "calamint",
    "common_calamint",
    "large-flowered_calamint",
    "lesser_calamint",
    "wild_basil",
    "horse_balm",
    "coleus",
    "country_borage",
    "painted_nettle",
    "Apalachicola_rosemary",
    "dragonhead",
    "elsholtzia",
    "hemp_nettle",
    "ground_ivy",
    "pennyroyal",
    "hyssop",
    "dead_nettle",
    "white_dead_nettle",
    "henbit",
    "English_lavender",
    "French_lavender",
    "spike_lavender",
    "dagga",
    "lion's-ear",
    "motherwort",
    "pitcher_sage",
    "bugleweed",
    "water_horehound",
    "gipsywort",
    "origanum",
    "oregano",
    "sweet_marjoram",
    "horehound",
    "common_horehound",
    "lemon_balm",
    "corn_mint",
    "water-mint",
    "bergamot_mint",
    "horsemint",
    "peppermint",
    "spearmint",
    "apple_mint",
    "pennyroyal",
    "yerba_buena",
    "molucca_balm",
    "monarda",
    "bee_balm",
    "horsemint",
    "bee_balm",
    "lemon_mint",
    "plains_lemon_monarda",
    "basil_balm",
    "mustang_mint",
    "catmint",
    "basil",
    "beefsteak_plant",
    "phlomis",
    "Jerusalem_sage",
    "physostegia",
    "plectranthus",
    "patchouli",
    "self-heal",
    "mountain_mint",
    "rosemary",
    "clary_sage",
    "purple_sage",
    "cancerweed",
    "common_sage",
    "meadow_clary",
    "clary",
    "pitcher_sage",
    "Mexican_mint",
    "wild_sage",
    "savory",
    "summer_savory",
    "winter_savory",
    "skullcap",
    "blue_pimpernel",
    "hedge_nettle",
    "hedge_nettle",
    "germander",
    "American_germander",
    "cat_thyme",
    "wood_sage",
    "thyme",
    "common_thyme",
    "wild_thyme",
    "blue_curls",
    "turpentine_camphor_weed",
    "bastard_pennyroyal",
    "bladderwort",
    "butterwort",
    "genlisea",
    "martynia",
    "common_unicorn_plant",
    "sand_devil's_claw",
    "sweet_unicorn_plant",
    "figwort",
    "snapdragon",
    "white_snapdragon",
    "yellow_twining_snapdragon",
    "Mediterranean_snapdragon",
    "kitten-tails",
    "Alpine_besseya",
    "false_foxglove",
    "false_foxglove",
    "calceolaria",
    "Indian_paintbrush",
    "desert_paintbrush",
    "giant_red_paintbrush",
    "great_plains_paintbrush",
    "sulfur_paintbrush",
    "shellflower",
    "maiden_blue-eyed_Mary",
    "blue-eyed_Mary",
    "foxglove",
    "common_foxglove",
    "yellow_foxglove",
    "gerardia",
    "blue_toadflax",
    "toadflax",
    "golden-beard_penstemon",
    "scarlet_bugler",
    "red_shrubby_penstemon",
    "Platte_River_penstemon",
    "hot-rock_penstemon",
    "Jones'_penstemon",
    "shrubby_penstemon",
    "narrow-leaf_penstemon",
    "balloon_flower",
    "Parry's_penstemon",
    "rock_penstemon",
    "Rydberg's_penstemon",
    "cascade_penstemon",
    "Whipple's_penstemon",
    "moth_mullein",
    "white_mullein",
    "purple_mullein",
    "common_mullein",
    "veronica",
    "field_speedwell",
    "brooklime",
    "corn_speedwell",
    "brooklime",
    "germander_speedwell",
    "water_speedwell",
    "common_speedwell",
    "purslane_speedwell",
    "thyme-leaved_speedwell",
    "nightshade",
    "horse_nettle",
    "African_holly",
    "potato_vine",
    "garden_huckleberry",
    "naranjilla",
    "potato_vine",
    "potato_tree",
    "belladonna",
    "bush_violet",
    "lady-of-the-night",
    "angel's_trumpet",
    "angel's_trumpet",
    "red_angel's_trumpet",
    "cone_pepper",
    "bird_pepper",
    "day_jessamine",
    "night_jasmine",
    "tree_tomato",
    "thorn_apple",
    "jimsonweed",
    "pichi",
    "henbane",
    "Egyptian_henbane",
    "matrimony_vine",
    "common_matrimony_vine",
    "Christmasberry",
    "plum_tomato",
    "mandrake",
    "mandrake_root",
    "apple_of_Peru",
    "flowering_tobacco",
    "common_tobacco",
    "wild_tobacco",
    "cupflower",
    "whitecup",
    "petunia",
    "large_white_petunia",
    "violet-flowered_petunia",
    "hybrid_petunia",
    "cape_gooseberry",
    "strawberry_tomato",
    "tomatillo",
    "tomatillo",
    "yellow_henbane",
    "cock's_eggs",
    "salpiglossis",
    "painted_tongue",
    "butterfly_flower",
    "Scopolia_carniolica",
    "chalice_vine",
    "verbena",
    "lantana",
    "black_mangrove",
    "white_mangrove",
    "black_mangrove",
    "teak",
    "spurge",
    "sun_spurge",
    "petty_spurge",
    "medusa's_head",
    "wild_spurge",
    "snow-on-the-mountain",
    "cypress_spurge",
    "leafy_spurge",
    "hairy_spurge",
    "poinsettia",
    "Japanese_poinsettia",
    "fire-on-the-mountain",
    "wood_spurge",
    "dwarf_spurge",
    "scarlet_plume",
    "naboom",
    "crown_of_thorns",
    "toothed_spurge",
    "three-seeded_mercury",
    "croton",
    "cascarilla",
    "cascarilla_bark",
    "castor-oil_plant",
    "spurge_nettle",
    "physic_nut",
    "Para_rubber_tree",
    "cassava",
    "bitter_cassava",
    "cassava",
    "sweet_cassava",
    "candlenut",
    "tung_tree",
    "slipper_spurge",
    "candelilla",
    "Jewbush",
    "jumping_bean",
    "camellia",
    "japonica",
    "umbellifer",
    "wild_parsley",
    "fool's_parsley",
    "dill",
    "angelica",
    "garden_angelica",
    "wild_angelica",
    "chervil",
    "cow_parsley",
    "wild_celery",
    "astrantia",
    "greater_masterwort",
    "caraway",
    "whorled_caraway",
    "water_hemlock",
    "spotted_cowbane",
    "hemlock",
    "earthnut",
    "cumin",
    "wild_carrot",
    "eryngo",
    "sea_holly",
    "button_snakeroot",
    "rattlesnake_master",
    "fennel",
    "common_fennel",
    "Florence_fennel",
    "cow_parsnip",
    "lovage",
    "sweet_cicely",
    "water_fennel",
    "parsnip",
    "cultivated_parsnip",
    "wild_parsnip",
    "parsley",
    "Italian_parsley",
    "Hamburg_parsley",
    "anise",
    "sanicle",
    "purple_sanicle",
    "European_sanicle",
    "water_parsnip",
    "greater_water_parsnip",
    "skirret",
    "dogwood",
    "common_white_dogwood",
    "red_osier",
    "silky_dogwood",
    "silky_cornel",
    "common_European_dogwood",
    "bunchberry",
    "cornelian_cherry",
    "puka",
    "kapuka",
    "valerian",
    "common_valerian",
    "common_corn_salad",
    "red_valerian",
    "filmy_fern",
    "bristle_fern",
    "hare's-foot_bristle_fern",
    "Killarney_fern",
    "kidney_fern",
    "flowering_fern",
    "royal_fern",
    "interrupted_fern",
    "crape_fern",
    "crepe_fern",
    "curly_grass",
    "pine_fern",
    "climbing_fern",
    "creeping_fern",
    "climbing_maidenhair",
    "scented_fern",
    "clover_fern",
    "nardoo",
    "water_clover",
    "pillwort",
    "regnellidium",
    "floating-moss",
    "mosquito_fern",
    "adder's_tongue",
    "ribbon_fern",
    "grape_fern",
    "daisyleaf_grape_fern",
    "leathery_grape_fern",
    "rattlesnake_fern",
    "flowering_fern",
    "powdery_mildew",
    "Dutch_elm_fungus",
    "ergot",
    "rye_ergot",
    "black_root_rot_fungus",
    "dead-man's-fingers",
    "sclerotinia",
    "brown_cup",
    "earthball",
    "Scleroderma_citrinum",
    "Scleroderma_flavidium",
    "Scleroderma_bovista",
    "Podaxaceae",
    "stalked_puffball",
    "stalked_puffball",
    "false_truffle",
    "Rhizopogon_idahoensis",
    "Truncocolumella_citrina",
    "mucor",
    "rhizopus",
    "bread_mold",
    "slime_mold",
    "true_slime_mold",
    "cellular_slime_mold",
    "dictostylium",
    "pond-scum_parasite",
    "potato_wart_fungus",
    "white_fungus",
    "water_mold",
    "downy_mildew",
    "blue_mold_fungus",
    "onion_mildew",
    "tobacco_mildew",
    "white_rust",
    "pythium",
    "damping_off_fungus",
    "Phytophthora_citrophthora",
    "Phytophthora_infestans",
    "clubroot_fungus",
    "Geglossaceae",
    "Sarcosomataceae",
    "Rufous_rubber_cup",
    "devil's_cigar",
    "devil's_urn",
    "truffle",
    "club_fungus",
    "coral_fungus",
    "tooth_fungus",
    "lichen",
    "ascolichen",
    "basidiolichen",
    "lecanora",
    "manna_lichen",
    "archil",
    "roccella",
    "beard_lichen",
    "horsehair_lichen",
    "reindeer_moss",
    "crottle",
    "Iceland_moss",
    "fungus",
    "promycelium",
    "true_fungus",
    "basidiomycete",
    "mushroom",
    "agaric",
    "mushroom",
    "mushroom",
    "toadstool",
    "horse_mushroom",
    "meadow_mushroom",
    "shiitake",
    "scaly_lentinus",
    "royal_agaric",
    "false_deathcap",
    "fly_agaric",
    "death_cap",
    "blushing_mushroom",
    "destroying_angel",
    "chanterelle",
    "floccose_chanterelle",
    "pig's_ears",
    "cinnabar_chanterelle",
    "jack-o-lantern_fungus",
    "inky_cap",
    "shaggymane",
    "milkcap",
    "fairy-ring_mushroom",
    "fairy_ring",
    "oyster_mushroom",
    "olive-tree_agaric",
    "Pholiota_astragalina",
    "Pholiota_aurea",
    "Pholiota_destruens",
    "Pholiota_flammans",
    "Pholiota_flavida",
    "nameko",
    "Pholiota_squarrosa-adiposa",
    "Pholiota_squarrosa",
    "Pholiota_squarrosoides",
    "Stropharia_ambigua",
    "Stropharia_hornemannii",
    "Stropharia_rugoso-annulata",
    "gill_fungus",
    "Entoloma_lividum",
    "Entoloma_aprile",
    "Chlorophyllum_molybdites",
    "lepiota",
    "parasol_mushroom",
    "poisonous_parasol",
    "Lepiota_naucina",
    "Lepiota_rhacodes",
    "American_parasol",
    "Lepiota_rubrotincta",
    "Lepiota_clypeolaria",
    "onion_stem",
    "pink_disease_fungus",
    "bottom_rot_fungus",
    "potato_fungus",
    "coffee_fungus",
    "blewits",
    "sandy_mushroom",
    "Tricholoma_pessundatum",
    "Tricholoma_sejunctum",
    "man-on-a-horse",
    "Tricholoma_venenata",
    "Tricholoma_pardinum",
    "Tricholoma_vaccinum",
    "Tricholoma_aurantium",
    "Volvaria_bombycina",
    "Pluteus_aurantiorugosus",
    "Pluteus_magnus",
    "deer_mushroom",
    "straw_mushroom",
    "Volvariella_bombycina",
    "Clitocybe_clavipes",
    "Clitocybe_dealbata",
    "Clitocybe_inornata",
    "Clitocybe_robusta",
    "Clitocybe_irina",
    "Clitocybe_subconnexa",
    "winter_mushroom",
    "mycelium",
    "sclerotium",
    "sac_fungus",
    "ascomycete",
    "Clavicipitaceae",
    "grainy_club",
    "yeast",
    "baker's_yeast",
    "wine-maker's_yeast",
    "Aspergillus_fumigatus",
    "brown_root_rot_fungus",
    "discomycete",
    "Leotia_lubrica",
    "Mitrula_elegans",
    "Sarcoscypha_coccinea",
    "Caloscypha_fulgens",
    "Aleuria_aurantia",
    "elf_cup",
    "Peziza_domicilina",
    "blood_cup",
    "Urnula_craterium",
    "Galiella_rufa",
    "Jafnea_semitosta",
    "morel",
    "common_morel",
    "Disciotis_venosa",
    "Verpa",
    "Verpa_bohemica",
    "Verpa_conica",
    "black_morel",
    "Morchella_crassipes",
    "Morchella_semilibera",
    "Wynnea_americana",
    "Wynnea_sparassoides",
    "false_morel",
    "lorchel",
    "helvella",
    "Helvella_crispa",
    "Helvella_acetabulum",
    "Helvella_sulcata",
    "discina",
    "gyromitra",
    "Gyromitra_californica",
    "Gyromitra_sphaerospora",
    "Gyromitra_esculenta",
    "Gyromitra_infula",
    "Gyromitra_fastigiata",
    "Gyromitra_gigas",
    "gasteromycete",
    "stinkhorn",
    "common_stinkhorn",
    "Phallus_ravenelii",
    "dog_stinkhorn",
    "Calostoma_lutescens",
    "Calostoma_cinnabarina",
    "Calostoma_ravenelii",
    "stinky_squid",
    "puffball",
    "giant_puffball",
    "earthstar",
    "Geastrum_coronatum",
    "Radiigera_fuscogleba",
    "Astreus_pteridis",
    "Astreus_hygrometricus",
    "bird's-nest_fungus",
    "Gastrocybe_lateritia",
    "Macowanites_americanus",
    "polypore",
    "bracket_fungus",
    "Albatrellus_dispansus",
    "Albatrellus_ovinus",
    "Neolentinus_ponderosus",
    "Oligoporus_leucospongia",
    "Polyporus_tenuiculus",
    "hen-of-the-woods",
    "Polyporus_squamosus",
    "beefsteak_fungus",
    "agaric",
    "bolete",
    "Boletus_chrysenteron",
    "Boletus_edulis",
    "Frost's_bolete",
    "Boletus_luridus",
    "Boletus_mirabilis",
    "Boletus_pallidus",
    "Boletus_pulcherrimus",
    "Boletus_pulverulentus",
    "Boletus_roxanae",
    "Boletus_subvelutipes",
    "Boletus_variipes",
    "Boletus_zelleri",
    "Fuscoboletinus_paluster",
    "Fuscoboletinus_serotinus",
    "Leccinum_fibrillosum",
    "Suillus_albivelatus",
    "old-man-of-the-woods",
    "Boletellus_russellii",
    "jelly_fungus",
    "snow_mushroom",
    "witches'_butter",
    "Tremella_foliacea",
    "Tremella_reticulata",
    "Jew's-ear",
    "rust",
    "aecium",
    "flax_rust",
    "blister_rust",
    "wheat_rust",
    "apple_rust",
    "smut",
    "covered_smut",
    "loose_smut",
    "cornsmut",
    "boil_smut",
    "Sphacelotheca",
    "head_smut",
    "bunt",
    "bunt",
    "onion_smut",
    "flag_smut_fungus",
    "wheat_flag_smut",
    "felt_fungus",
    "waxycap",
    "Hygrocybe_acutoconica",
    "Hygrophorus_borealis",
    "Hygrophorus_caeruleus",
    "Hygrophorus_inocybiformis",
    "Hygrophorus_kauffmanii",
    "Hygrophorus_marzuolus",
    "Hygrophorus_purpurascens",
    "Hygrophorus_russula",
    "Hygrophorus_sordidus",
    "Hygrophorus_tennesseensis",
    "Hygrophorus_turundus",
    "Neohygrophorus_angelesianus",
    "Cortinarius_armillatus",
    "Cortinarius_atkinsonianus",
    "Cortinarius_corrugatus",
    "Cortinarius_gentilis",
    "Cortinarius_mutabilis",
    "Cortinarius_semisanguineus",
    "Cortinarius_subfoetidus",
    "Cortinarius_violaceus",
    "Gymnopilus_spectabilis",
    "Gymnopilus_validipes",
    "Gymnopilus_ventricosus",
    "mold",
    "mildew",
    "verticillium",
    "monilia",
    "candida",
    "Candida_albicans",
    "blastomycete",
    "yellow_spot_fungus",
    "green_smut_fungus",
    "dry_rot",
    "rhizoctinia",
    "houseplant",
    "bedder",
    "succulent",
    "cultivar",
    "weed",
    "wort",
    "brier",
    "aril",
    "sporophyll",
    "sporangium",
    "sporangiophore",
    "ascus",
    "ascospore",
    "arthrospore",
    "eusporangium",
    "tetrasporangium",
    "gametangium",
    "sorus",
    "sorus",
    "partial_veil",
    "lignum",
    "vascular_ray",
    "phloem",
    "evergreen",
    "deciduous_plant",
    "poisonous_plant",
    "vine",
    "creeper",
    "tendril",
    "root_climber",
    "lignosae",
    "arborescent_plant",
    "snag",
    "tree",
    "timber_tree",
    "treelet",
    "arbor",
    "bean_tree",
    "pollard",
    "sapling",
    "shade_tree",
    "gymnospermous_tree",
    "conifer",
    "angiospermous_tree",
    "nut_tree",
    "spice_tree",
    "fever_tree",
    "stump",
    "bonsai",
    "ming_tree",
    "ming_tree",
    "undershrub",
    "subshrub",
    "bramble",
    "liana",
    "geophyte",
    "desert_plant",
    "mesophyte",
    "marsh_plant",
    "hemiepiphyte",
    "strangler",
    "lithophyte",
    "saprobe",
    "autophyte",
    "root",
    "taproot",
    "prop_root",
    "prophyll",
    "rootstock",
    "quickset",
    "stolon",
    "tuberous_plant",
    "rhizome",
    "rachis",
    "caudex",
    "cladode",
    "receptacle",
    "scape",
    "umbel",
    "petiole",
    "peduncle",
    "pedicel",
    "flower_cluster",
    "raceme",
    "panicle",
    "thyrse",
    "cyme",
    "cymule",
    "glomerule",
    "scorpioid_cyme",
    "ear",
    "spadix",
    "bulbous_plant",
    "bulbil",
    "cormous_plant",
    "fruit",
    "fruitlet",
    "seed",
    "bean",
    "nut",
    "nutlet",
    "kernel",
    "syconium",
    "berry",
    "aggregate_fruit",
    "simple_fruit",
    "acinus",
    "drupe",
    "drupelet",
    "pome",
    "pod",
    "loment",
    "pyxidium",
    "husk",
    "cornhusk",
    "pod",
    "accessory_fruit",
    "buckthorn",
    "buckthorn_berry",
    "cascara_buckthorn",
    "cascara",
    "Carolina_buckthorn",
    "coffeeberry",
    "redberry",
    "nakedwood",
    "jujube",
    "Christ's-thorn",
    "hazel",
    "fox_grape",
    "muscadine",
    "vinifera",
    "Pinot_blanc",
    "Sauvignon_grape",
    "Sauvignon_blanc",
    "Muscadet",
    "Riesling",
    "Zinfandel",
    "Chenin_blanc",
    "malvasia",
    "Verdicchio",
    "Boston_ivy",
    "Virginia_creeper",
    "true_pepper",
    "betel",
    "cubeb",
    "schizocarp",
    "peperomia",
    "watermelon_begonia",
    "yerba_mansa",
    "pinna",
    "frond",
    "bract",
    "bracteole",
    "involucre",
    "glume",
    "palmate_leaf",
    "pinnate_leaf",
    "bijugate_leaf",
    "decompound_leaf",
    "acuminate_leaf",
    "deltoid_leaf",
    "ensiform_leaf",
    "linear_leaf",
    "lyrate_leaf",
    "obtuse_leaf",
    "oblanceolate_leaf",
    "pandurate_leaf",
    "reniform_leaf",
    "spatulate_leaf",
    "even-pinnate_leaf",
    "odd-pinnate_leaf",
    "pedate_leaf",
    "crenate_leaf",
    "dentate_leaf",
    "denticulate_leaf",
    "erose_leaf",
    "runcinate_leaf",
    "prickly-edged_leaf",
    "deadwood",
    "haulm",
    "branchlet",
    "osier",
    "giant_scrambling_fern",
    "umbrella_fern",
    "floating_fern",
    "polypody",
    "licorice_fern",
    "grey_polypody",
    "leatherleaf",
    "rock_polypody",
    "common_polypody",
    "bear's-paw_fern",
    "strap_fern",
    "Florida_strap_fern",
    "basket_fern",
    "snake_polypody",
    "climbing_bird's_nest_fern",
    "golden_polypody",
    "staghorn_fern",
    "South_American_staghorn",
    "common_staghorn_fern",
    "felt_fern",
    "potato_fern",
    "myrmecophyte",
    "grass_fern",
    "spleenwort",
    "black_spleenwort",
    "bird's_nest_fern",
    "ebony_spleenwort",
    "black-stem_spleenwort",
    "walking_fern",
    "green_spleenwort",
    "mountain_spleenwort",
    "lobed_spleenwort",
    "lanceolate_spleenwort",
    "hart's-tongue",
    "scale_fern",
    "scolopendrium",
    "deer_fern",
    "doodia",
    "chain_fern",
    "Virginia_chain_fern",
    "silver_tree_fern",
    "davallia",
    "hare's-foot_fern",
    "Canary_Island_hare's_foot_fern",
    "squirrel's-foot_fern",
    "bracken",
    "soft_tree_fern",
    "Scythian_lamb",
    "false_bracken",
    "thyrsopteris",
    "shield_fern",
    "broad_buckler-fern",
    "fragrant_cliff_fern",
    "Goldie's_fern",
    "wood_fern",
    "male_fern",
    "marginal_wood_fern",
    "mountain_male_fern",
    "lady_fern",
    "Alpine_lady_fern",
    "silvery_spleenwort",
    "holly_fern",
    "bladder_fern",
    "brittle_bladder_fern",
    "mountain_bladder_fern",
    "bulblet_fern",
    "silvery_spleenwort",
    "oak_fern",
    "limestone_fern",
    "ostrich_fern",
    "hart's-tongue",
    "sensitive_fern",
    "Christmas_fern",
    "holly_fern",
    "Braun's_holly_fern",
    "western_holly_fern",
    "soft_shield_fern",
    "leather_fern",
    "button_fern",
    "Indian_button_fern",
    "woodsia",
    "rusty_woodsia",
    "Alpine_woodsia",
    "smooth_woodsia",
    "Boston_fern",
    "basket_fern",
    "golden_fern",
    "maidenhair",
    "common_maidenhair",
    "American_maidenhair_fern",
    "Bermuda_maidenhair",
    "brittle_maidenhair",
    "Farley_maidenhair",
    "annual_fern",
    "lip_fern",
    "smooth_lip_fern",
    "lace_fern",
    "wooly_lip_fern",
    "southwestern_lip_fern",
    "bamboo_fern",
    "American_rock_brake",
    "European_parsley_fern",
    "hand_fern",
    "cliff_brake",
    "coffee_fern",
    "purple_rock_brake",
    "bird's-foot_fern",
    "button_fern",
    "silver_fern",
    "golden_fern",
    "gold_fern",
    "Pteris_cretica",
    "spider_brake",
    "ribbon_fern",
    "potato_fern",
    "angiopteris",
    "skeleton_fork_fern",
    "horsetail",
    "common_horsetail",
    "swamp_horsetail",
    "scouring_rush",
    "marsh_horsetail",
    "wood_horsetail",
    "variegated_horsetail",
    "club_moss",
    "shining_clubmoss",
    "alpine_clubmoss",
    "fir_clubmoss",
    "ground_cedar",
    "ground_fir",
    "foxtail_grass",
    "spikemoss",
    "meadow_spikemoss",
    "desert_selaginella",
    "resurrection_plant",
    "florida_selaginella",
    "quillwort",
    "earthtongue",
    "snuffbox_fern",
    "christella",
    "mountain_fern",
    "New_York_fern",
    "Massachusetts_fern",
    "beech_fern",
    "broad_beech_fern",
    "long_beech_fern",
    "shoestring_fungus",
    "Armillaria_caligata",
    "Armillaria_ponderosa",
    "Armillaria_zelleri",
    "honey_mushroom",
    "milkweed",
    "white_milkweed",
    "poke_milkweed",
    "swamp_milkweed",
    "Mead's_milkweed",
    "purple_silkweed",
    "showy_milkweed",
    "poison_milkweed",
    "butterfly_weed",
    "whorled_milkweed",
    "cruel_plant",
    "wax_plant",
    "silk_vine",
    "stapelia",
    "Stapelias_asterias",
    "stephanotis",
    "Madagascar_jasmine",
    "negro_vine",
    "zygospore",
    "tree_of_knowledge",
    "orangery",
    "pocketbook",
    "shit",
    "cordage",
    "yard",
    "extremum",
    "leaf_shape",
    "equilateral",
    "figure",
    "pencil",
    "plane_figure",
    "solid_figure",
    "line",
    "bulb",
    "convex_shape",
    "concave_shape",
    "cylinder",
    "round_shape",
    "heart",
    "polygon",
    "convex_polygon",
    "concave_polygon",
    "reentrant_polygon",
    "amorphous_shape",
    "closed_curve",
    "simple_closed_curve",
    "S-shape",
    "wave",
    "extrados",
    "hook",
    "envelope",
    "bight",
    "diameter",
    "cone",
    "funnel",
    "oblong",
    "circle",
    "circle",
    "equator",
    "scallop",
    "ring",
    "loop",
    "bight",
    "helix",
    "element_of_a_cone",
    "element_of_a_cylinder",
    "ellipse",
    "quadrate",
    "triangle",
    "acute_triangle",
    "isosceles_triangle",
    "obtuse_triangle",
    "right_triangle",
    "scalene_triangle",
    "parallel",
    "trapezoid",
    "star",
    "pentagon",
    "hexagon",
    "heptagon",
    "octagon",
    "nonagon",
    "decagon",
    "rhombus",
    "spherical_polygon",
    "spherical_triangle",
    "convex_polyhedron",
    "concave_polyhedron",
    "cuboid",
    "quadrangular_prism",
    "bell",
    "angular_distance",
    "true_anomaly",
    "spherical_angle",
    "angle_of_refraction",
    "acute_angle",
    "groove",
    "rut",
    "bulge",
    "belly",
    "bow",
    "crescent",
    "ellipsoid",
    "hypotenuse",
    "balance",
    "conformation",
    "symmetry",
    "spheroid",
    "spherule",
    "toroid",
    "column",
    "barrel",
    "pipe",
    "pellet",
    "bolus",
    "dewdrop",
    "ridge",
    "rim",
    "taper",
    "boundary",
    "incisure",
    "notch",
    "wrinkle",
    "dermatoglyphic",
    "frown_line",
    "line_of_life",
    "line_of_heart",
    "crevice",
    "cleft",
    "roulette",
    "node",
    "tree",
    "stemma",
    "brachium",
    "fork",
    "block",
    "ovoid",
    "tetrahedron",
    "pentahedron",
    "hexahedron",
    "regular_polyhedron",
    "polyhedral_angle",
    "cube",
    "truncated_pyramid",
    "truncated_cone",
    "tail",
    "tongue",
    "trapezohedron",
    "wedge",
    "keel",
    "place",
    "herpes",
    "chlamydia",
    "wall",
    "micronutrient",
    "chyme",
    "ragweed_pollen",
    "pina_cloth",
    "chlorobenzylidenemalononitrile",
    "carbon",
    "charcoal",
    "rock",
    "gravel",
    "aflatoxin",
    "alpha-tocopheral",
    "leopard",
    "bricks_and_mortar",
    "lagging",
    "hydraulic_cement",
    "choline",
    "concrete",
    "glass_wool",
    "soil",
    "high_explosive",
    "litter",
    "fish_meal",
    "Greek_fire",
    "culture_medium",
    "agar",
    "blood_agar",
    "hip_tile",
    "hyacinth",
    "hydroxide_ion",
    "ice",
    "inositol",
    "linoleum",
    "lithia_water",
    "lodestone",
    "pantothenic_acid",
    "paper",
    "papyrus",
    "pantile",
    "blacktop",
    "tarmacadam",
    "paving",
    "plaster",
    "poison_gas",
    "ridge_tile",
    "roughcast",
    "sand",
    "spackle",
    "render",
    "wattle_and_daub",
    "stucco",
    "tear_gas",
    "toilet_tissue",
    "linseed",
    "vitamin",
    "fat-soluble_vitamin",
    "water-soluble_vitamin",
    "vitamin_A",
    "vitamin_A1",
    "vitamin_A2",
    "B-complex_vitamin",
    "vitamin_B1",
    "vitamin_B12",
    "vitamin_B2",
    "vitamin_B6",
    "vitamin_Bc",
    "niacin",
    "vitamin_D",
    "vitamin_E",
    "biotin",
    "vitamin_K",
    "vitamin_K1",
    "vitamin_K3",
    "vitamin_P",
    "vitamin_C",
    "planking",
    "chipboard",
    "knothole",
]  # noqa
